Package: rgdal
Title: Bindings for the 'Geospatial' Data Abstraction Library
Version: 1.5-29
Date: 2022-03-07
Depends: R (>= 3.5.0), methods, sp (>= 1.1-0)
Imports: grDevices, graphics, stats, utils
LinkingTo: sp
Suggests: knitr, DBI, RSQLite, maptools, mapview, rmarkdown, curl,
        rgeos
NeedsCompilation: yes
Description: Provides bindings to the 'Geospatial' Data Abstraction Library ('GDAL') (>= 1.11.4) and access to projection/transformation operations from the 'PROJ' library. Please note that 'rgdal' will be retired by the end of 2023, plan transition to sf/stars/'terra' functions using 'GDAL' and 'PROJ' at your earliest convenience. Use is made of classes defined in the 'sp' package. Raster and vector map data can be imported into R, and raster and vector 'sp' objects exported. The 'GDAL' and 'PROJ' libraries are external to the package, and, when installing the package from source, must be correctly installed first; it is important that 'GDAL' < 3 be matched with 'PROJ' < 6. From 'rgdal' 1.5-8, installed with to 'GDAL' >=3, 'PROJ' >=6 and 'sp' >= 1.4, coordinate reference systems use 'WKT2_2019' strings, not 'PROJ' strings. 'Windows' and 'macOS' binaries (including 'GDAL', 'PROJ' and their dependencies) are provided on 'CRAN'. 
Authors@R: c(
	person("Roger", "Bivand", role = c("cre", "aut"), email = "Roger.Bivand@nhh.no", comment=c(ORCID="0000-0003-2392-6140")),
	person("Tim", "Keitt", role = "aut"),
	person("Barry", "Rowlingson", role = c("aut", "ctb")),
	person("Edzer", "Pebesma", role = "ctb"),
	person("Michael", "Sumner", role = "ctb"),
	person("Robert", "Hijmans", role = "ctb"),
        person("Daniel", "Baston", role = "ctb"),
	person("Even", "Rouault", role = c("cph", "ctb")),
	person("Frank", "Warmerdam", role = c("cph", "ctb")),
	person("Jeroen", "Ooms", role = "ctb"),
	person("Colin", "Rundel", role = "ctb"))
License: GPL (>= 2)
URL: http://rgdal.r-forge.r-project.org, https://gdal.org,
        https://proj.org, https://r-forge.r-project.org/projects/rgdal/
SystemRequirements: PROJ (>= 4.8.0, https://proj.org/download.html) and
        GDAL (>= 1.11.4, https://gdal.org/download.html), with versions
        either (A) PROJ < 6 and GDAL < 3 or (B) PROJ >= 6 and GDAL >=
        3. For degraded PROJ >= 6 and GDAL < 3, use the configure
        argument '--with-proj_api="proj_api.h"'.
VignetteBuilder: knitr
RoxygenNote: 6.1.1
Packaged: 2022-03-15 08:11:38 UTC; rsb
Author: Roger Bivand [cre, aut] (<https://orcid.org/0000-0003-2392-6140>),
  Tim Keitt [aut],
  Barry Rowlingson [aut, ctb],
  Edzer Pebesma [ctb],
  Michael Sumner [ctb],
  Robert Hijmans [ctb],
  Daniel Baston [ctb],
  Even Rouault [cph, ctb],
  Frank Warmerdam [cph, ctb],
  Jeroen Ooms [ctb],
  Colin Rundel [ctb]
Maintainer: Roger Bivand <Roger.Bivand@nhh.no>
Repository: CRAN
Date/Publication: 2022-03-16 22:10:02 UTC
Built: R 4.0.5; x86_64-apple-darwin17.0; 2022-03-17 10:25:19 UTC; unix
Archs: rgdal.so.dSYM

# useDynLib(rgdal, .registration = TRUE, .fixes = "GDAL_")
useDynLib(rgdal, .registration = TRUE)

import(methods)
import(sp)
importFrom(grDevices, gray, rgb, hsv, col2rgb)
importFrom(graphics, par, image.default, text)
importFrom(stats, sd, offset, na.omit)
importFrom(utils, packageVersion)
importFrom("utils", "read.csv")

#exportMethods(spTransform)
exportMethods(closeDataset)
exportMethods(initialize)
exportMethods(summary)
exportMethods(coerce, "[", "[[")

exportClass(GDALDataset)
exportClass(GDALDriver)
exportClass(GDALMajorObject)
exportClass(GDALRasterBand)
exportClass(GDALReadOnlyDataset)
exportClass(GDALTransientDataset)
exportClass(SpatialGDAL, SpatialGDALWrite)

export(rgdal_extSoftVersion, PROJis6ormore, GDALis3ormore, GDAL_OSR_PROJ, 
 new_proj_and_gdal)

export(project, make_EPSG, projInfo, getPROJ4VersionInfo, getPROJ4libPath,
  projNAD, EPSG_version)
export(CRSargs, checkCRSArgs, checkCRSArgs_ng, compare_CRS)
export(get_P6_datum_hard_fail, set_P6_datum_hard_fail, get_transform_wkt_comment, set_transform_wkt_comment, get_last_coordOp, get_thin_PROJ6_warnings, set_thin_PROJ6_warnings, get_enforce_xy, set_enforce_xy, get_prefer_proj, set_prefer_proj, get_rgdal_show_exportToProj4_warnings, set_rgdal_show_exportToProj4_warnings, get_PROJ6_warnings_count)
export(putRasterData, saveDataset, copyDataset, deleteDataset, saveDatasetAs)
export(getGDALDriverNames, getDriverName, getDriverLongName, gdalDrivers)
export(getDescription)
export(getRasterData, getRasterTable, getProjectionRef)
export(getRasterBand, getRasterBlockSize)
export(toSigned, toUnSigned)
export(getDriver, GDAL.close, GDAL.open, getColorTable,
  getGeoTransFunc, sub.GDROD, SGDF2PCT, vec2RGB, asSGDF_GROD, RGB2PCT)
export(readGDAL, writeGDAL, create2GDAL)
export(showWKT, showEPSG, showP4, showSRID, list_coordOps, best_instantiable_coordOp)
export(readOGR, ogrInfo, ogrFIDs, ogrDrivers, writeOGR, ogrListLayers)
export(GDALinfo, print.GDALobj, getGDALVersionInfo, getGDALCheckVersion, getGDAL_DATA_Path, getGDALwithGEOS)
export(displayDataset)
export(open.SpatialGDAL, copy.SpatialGDAL)
export(llgridlines)
export(GDALcall)
export(rawTransform)
export(is_proj_CDN_enabled, enable_proj_CDN, disable_proj_CDN)
export(get_proj_search_paths, set_proj_search_paths, proj_CDN_user_writable_dir)

#export(spTransform, spTransform.SpatialPoints, 
#  spTransform.SpatialPointsDataFrame, spTransform.SpatialLines, 
#  spTransform.SpatialLinesDataFrame, spTransform.SpatialPolygons, 
#  spTransform.SpatialPolygonsDataFrame)

export(OGRSpatialRef, GDALSpatialRef, OSRIsProjected)

export(get_OVERRIDE_PROJ_DATUM_WITH_TOWGS84,
 set_OVERRIDE_PROJ_DATUM_WITH_TOWGS84)

export(getCPLConfigOption, setCPLConfigOption)

export(get_cached_orig_PROJ_LIB, get_cached_orig_GDAL_DATA,
 get_cached_set_PROJ_LIB, get_cached_set_GDAL_DATA)

#S3method(spTransform, SpatialPoints)
#S3method(spTransform, SpatialPointsDataFrame)
#S3method(spTransform, SpatialLines)
#S3method(spTransform, SpatialLinesDataFrame)
#S3method(spTransform, SpatialPolygons)
#S3method(spTransform, SpatialPolygonsDataFrame)
S3method(print, GDALobj)
S3method(print, ogrinfo)
S3method(open, SpatialGDAL)
S3method(close, SpatialGDAL)
S3method(close, SpatialGDALWrite)
S3method(print, summary.SpatialGDAL)
S3method(print, coordOps)

# Copyright (c) 2003-20 by Barry Rowlingson, Roger Bivand, and Edzer Pebesma

getPROJ4VersionInfo <- function() {
    res0 <- .Call("PROJ4VersionInfo", PACKAGE="rgdal")
    res <- paste(res0[[1]], ", [PJ_VERSION: ", res0[[2]], "]", sep="")
    attr(res, "short") <- res0[[2]]
    res
}

PROJis6ormore <- function() {
    verno <- .Call("PROJ4VersionInfo", PACKAGE="rgdal")[[2]]
    verno >= 600
}

GDALis3ormore <- function() {
    substring(getGDALVersionInfo(), 6, 6) >= "3"
}

new_proj_and_gdal <- function() {
    PROJis6ormore() && GDALis3ormore()
}

is_proj_CDN_enabled <- function() {
    .Call("proj_network_enabled", PACKAGE="rgdal")
}

enable_proj_CDN <- function() {
    .Call("enable_proj_network", PACKAGE="rgdal")
    paste0("Using: ", proj_CDN_user_writable_dir(), sep="")
}

disable_proj_CDN <- function() {
    invisible(.Call("disable_proj_network", PACKAGE="rgdal"))
}

proj_CDN_user_writable_dir <- function() {
    .Call("get_proj_user_writable_dir", PACKAGE="rgdal")
}


GDAL_OSR_PROJ <- function() {
    res <- .Call("R_GDAL_OSR_PROJ", PACKAGE="rgdal")
    if (is.null(res)) return(res)
    res1 <- (res[1]*100)+(res[2]*10)+res[3]
    verno <- .Call("PROJ4VersionInfo", PACKAGE="rgdal")[[2]]
    if (res1 != verno) warning("GDAL built with PROJ ", res1,
        " not running ", verno)
    res1
}

getPROJ4libPath <- function() {
    res <- Sys.getenv("PROJ_LIB")
    if (PROJis6ormore()) {
        attr(res, "search_path") <- .Call("get_proj_search_path",
            PACKAGE="rgdal")
    }
    res
}

get_proj_search_paths <- function() {
    if (PROJis6ormore()) {
        res <- .Call("get_proj_search_path", PACKAGE="rgdal")
        res <- strsplit(res, .Platform$path.sep)[[1]]
    } else {
        res <- NULL
    }
    res
}

set_proj_search_paths <- function(paths) {
    if (PROJis6ormore()) {
        stopifnot(!missing(paths))
        stopifnot(is.character(paths))
        stopifnot(length(paths) > 0)
        n <- length(paths)
        for (i in 1:n) stopifnot(dir.exists(paths[i]))
        res <- .Call("set_proj_paths", paths, PACKAGE="rgdal")
    } else {
        res <- NULL
    }
    res
}

projNAD <- function() {
    .Call("PROJ4NADsInstalled", PACKAGE="rgdal")
}

get_P6_datum_hard_fail <- function() {
    get("P6_datum_hard_fail", envir=.RGDAL_CACHE)
}

set_P6_datum_hard_fail <- function(value) {
    stopifnot(is.logical(value))
    stopifnot(length(value) == 1L)
    stopifnot(!is.na(value))
    assign("P6_datum_hard_fail", value, envir=.RGDAL_CACHE)
}

get_transform_wkt_comment <- function() {
    get("transform_wkt_comment", envir=.RGDAL_CACHE)
}

set_transform_wkt_comment <- function(value) {
    stopifnot(is.logical(value))
    stopifnot(length(value) == 1L)
    stopifnot(!is.na(value))
    assign("transform_wkt_comment", value, envir=.RGDAL_CACHE)
}

get_last_coordOp <- function() {
    get(".last_coordOp", envir=.RGDAL_CACHE)
}

get_aoi <- function(obj, xy, inv, proj) {
    if (!new_proj_and_gdal()) return(NULL)
    if (missing(obj)) { # used in project
        bb <- cbind(range(xy[,1], na.rm=TRUE), range(xy[,2], na.rm=TRUE))
        if (inv) {
            o <- try(project(bb, proj, inv=inv, use_aoi=FALSE), silent=TRUE)
            if (inherits(o, "try-error")) return(NULL)
        } else {
            o <- bb
        }
    } else { # used in spTransform
        if (is.projected(obj)) { 
            tg <- wkt(obj)
            if (is.null(tg)) {
                tg <- proj4string(obj)
                if (length(grep("+init", tg)) > 0L) {
                    tg <- wkt(CRS(tg))
                }
            }
            o <- try(project(t(bbox(obj))[,1:2], tg, inv=TRUE, use_aoi=FALSE),
                silent=TRUE)
            if (inherits(o, "try-error")) return(NULL)
        } else {
            o <- t(bbox(obj))[,1:2]
        }
    }
    if (any(!is.finite(c(o)))) return(NULL)
    aoi <- o + c(-0.1, +0.1) # stretch envelope
    aoi[,1] <- sapply(aoi[,1], function(x) { if (x > 180) 180 - x else if (x < -180) 360 + x else x})
    aoi[,2] <- sapply(aoi[,2], function(y) { if (y > 90) 90 else if (y < -90) -90 else y}) # constrain to -180/180, -90/90
    c(t(aoi))
}


OSRIsProjected <- function(obj) {
    stopifnot(inherits(obj, "CRS"))
    wkt2 <- wkt(obj)
    if (!is.null(wkt2) && new_proj_and_gdal())
        return(.Call("OSR_is_projected", wkt2, PACKAGE="rgdal"))
    p4str <- slot(obj, "projargs")
    if (is.na(p4str) || !nzchar(p4str)) 
	return(as.logical(NA))    
    .Call("OSR_is_projected", p4str, PACKAGE="rgdal")
}


"project" <- function(xy, proj, inv=FALSE, use_ob_tran=FALSE, legacy=TRUE, allowNAs_if_not_legacy=FALSE, coordOp=NULL, verbose=FALSE, use_aoi=TRUE) {

#    if (new_proj_and_gdal()) 
#        warning("project() will not be adapted for PROJ 6 and is deprecated")

    if (!is.numeric(xy)) stop("xy not numeric")
    if (is.matrix(xy)) nc <- dim(xy)[1]
    else if (length(xy) == 2L) nc <- 1
    else stop("xy malformed")
# 111216 RSB
    stopifnot(is.character(proj))
    stopifnot(length(proj) == 1)
    stopifnot(is.logical(inv))
# 120816 RSB
    stopifnot(is.logical(use_ob_tran))
    if (use_ob_tran) {
        gp <- grep("proj=ob_tran", proj)
        if (length(gp) == 0) {
            use_ob_tran <- FALSE
            warning("project: use_ob_tran set FALSE")
# 120820 RSB
        } else inv <- !inv
    }
    if (.Platform$OS.type == "windows" && .Platform$r_arch == "i386" &&
        !new_proj_and_gdal()) legacy <- FALSE
    if (!legacy) {
      if (allowNAs_if_not_legacy) {
       nas <- is.na(xy[,1]) | is.na(xy[,2])
      }
      if (!inv) {
        attr(nc, "ob_tran") <- as.integer(use_ob_tran)
        if (attr(nc, "ob_tran") == 0L) {
          res <- .Call("transform", "+proj=longlat", proj, nc,
	    as.double(xy[,1]), as.double(xy[,2]), NULL, PACKAGE="rgdal")
        } else {
          res <- .Call("transform", proj, "+proj=longlat", nc,
	    as.double(xy[,1]), as.double(xy[,2]), NULL, PACKAGE="rgdal")
        }
	if (!allowNAs_if_not_legacy) {
           if (any(!is.finite(res[[1]])) || any(!is.finite(res[[2]]))) {
	      k <- which(!is.finite(res[[1]]) | !is.finite(res[[2]]))
	      cat("non finite transformation detected:\n")
	      print(cbind(xy, res[[1]], res[[2]])[k,])
	      stop(paste("failure in points", paste(k, collapse=":")))
	   }
        } else {
           if (any(nas)) {
             res[[1]][nas] <- as.double(NA)
             res[[2]][nas] <- as.double(NA)
           }
        }
     } else {
        attr(nc, "ob_tran") <- -as.integer(use_ob_tran)
        if (attr(nc, "ob_tran") == 0L) {
          res <- .Call("transform", proj, "+proj=longlat", nc,
	    as.double(xy[,1]), as.double(xy[,2]), NULL, PACKAGE="rgdal")
        } else {
          res <- .Call("transform", "+proj=longlat", proj, nc,
	    as.double(xy[,1]), as.double(xy[,2]), NULL, PACKAGE="rgdal")
        }
	if (!allowNAs_if_not_legacy) {
	   if (any(!is.finite(res[[1]])) || any(!is.finite(res[[2]]))) {
	     k <- which(!is.finite(res[[1]]) | !is.finite(res[[2]]))
	     cat("non finite transformation detected:\n")
	     print(cbind(xy, res[[1]], res[[2]])[k,])
	     stop(paste("failure in points", paste(k, collapse=":")))
	   }
        } else {
           if (any(nas)) {
             res[[1]][nas] <- as.double(NA)
             res[[2]][nas] <- as.double(NA)   
           }        
        }
     }
    } else {
        if (new_proj_and_gdal()) {
            if (is.null(coordOp)) {
                if (substring(proj, 1, 1) == "+") {
                    if (length(grep("\\+init", proj)) > 0L)
                        proj <- comment(CRS(proj))
                    if (length(grep(" \\+type=crs", proj)) == 0L)
                        proj <- paste0(proj, " +type=crs")
                }
                aoi <- NULL
                if (!use_ob_tran && use_aoi) {
                    aoi <- get_aoi(xy=xy, inv=inv, proj=proj)
                    if (!is.null(aoi)) {
                        stopifnot(length(aoi) == 4)
                        aoi <- as.numeric(aoi)
                    }
                }
                coordOp <- .Call("project_ng_coordOp", proj,
                    as.logical(inv), aoi, as.logical(use_ob_tran), PACKAGE="rgdal")
                s2t <- strsplit(coordOp, " ")[[1]]
                if (length(s2t) == 0L) stop("malformed pipeline found")
                coordOp <- paste("+", s2t, sep="", collapse=" ")
            }
            if (verbose) cat(strwrap(coordOp), sep="\n")
            res <- .Call("project_ng",
                as.integer(nc),
                as.double(xy[,1]),
                as.double(xy[,2]),
                NULL,
                coordOp,
                PACKAGE="rgdal")
        } else {
            if(!inv) {
# 160404 RSB convert to .Call()
            res <- .Call("RGDAL_project",
                as.integer(nc),
                as.double(xy[,1]),
                as.double(xy[,2]),
                proj,
                as.logical(use_ob_tran),
                PACKAGE="rgdal")
            } else {
                res <- .Call("project_inv",
                as.integer(nc),
                as.double(xy[,1]),
                as.double(xy[,2]),
                proj,
                as.logical(use_ob_tran),
                PACKAGE="rgdal")
            }
        }
    }
    out <- cbind(res[[1]], res[[2]])
    if (!is.null(colnames(xy))) colnames(out) <- colnames(xy)
    out
}

## Fri, 28 Apr 2017 10:40:18
## From: Martin Aleksandrov Ivanov <Martin.Ivanov@geogr.uni-giessen.de>
# ob_tran when isTRUE(is.projected())

#if (packageVersion("sp") < "1.4.2") {

#if (!isGeneric("is.projected"))
#	setGeneric("is.projected", function(obj)
#		standardGeneric("is.projected"))
#
#is_projected_crs <- function (obj) 
#{
#    p4str <- CRSargs(obj)
#    if (is.na(p4str) || !nzchar(p4str)) 
#        return(as.logical(NA))
#    else {
#        res <- grep("longlat", p4str, fixed = TRUE)
#        if (length(res) == 0) 
#            return(TRUE)
#        else return(FALSE)
#    }
#}

#setMethod("is.projected", signature("CRS"), is_projected_crs)

#}

if (!isGeneric("spTransform"))
	setGeneric("spTransform", function(x, CRSobj, ...)
		standardGeneric("spTransform"))

"spTransform.SpatialPoints" <-  function(x, CRSobj, ...) {
        if (new_proj_and_gdal()) {
          if (get_transform_wkt_comment()) {
            if (is.null(comment(slot(x, "proj4string")))) {
                from_args <- paste0(slot(slot(x, "proj4string"),
                  "projargs"), " +type=crs")
                warning("NULL source CRS comment, falling back to PROJ string")
	        if (is.na(from_args)) 
		    stop("No transformation possible from NA source CRS")
                if (length(grep("\\+init\\=", from_args)) > 0) {
                    warning("+init dropped in PROJ string")
                    strres <- unlist(strsplit(from_args, " "))
                    from_args <- paste(strres[-grep("\\+init\\=", strres)],
                        collapse=" ")
                }
            } else {
                from_args <- comment(slot(x, "proj4string"))
            }
            if (is.null(comment(CRSobj))) {
                warning("NULL target CRS comment, falling back to PROJ string")
	        to_args <- paste0(slot(CRSobj, "projargs"), " +type=crs")
	        if (is.na(to_args)) 
		    stop("No transformation possible to NA target CRS")
                if (length(grep("\\+init\\=", to_args)) > 0) {
                    warning("+init dropped in PROJ string")
                    strres <- unlist(strsplit(to_args, " "))
                    to_args <- paste(strres[-grep("\\+init\\=", strres)],
                        collapse=" ")
                }
            } else {
                to_args <- comment(CRSobj)
            }
          } else {
            from_args <- paste0(slot(slot(x, "proj4string"), "projargs"),
              " +type=crs")
            to_args <- paste0(slot(CRSobj, "projargs"), " +type=crs")
          }
        } else {
	    from_args <- proj4string(x)
	    if (is.na(from_args)) 
		stop("No transformation possible from NA from reference system")
	    to_args <- slot(CRSobj, "projargs")
	    if (is.na(to_args)) 
		stop("No transformation possible to NA to reference system")
        }
	dots = list(...)
        if (!is.null(dots$use_ob_tran)) {
          stopifnot(is.logical(dots$use_ob_tran))
          if (dots$use_ob_tran) {
            gpf <- grep("proj=ob_tran", slot(CRSobj, "projargs"))
            gpi <- grep("proj=ob_tran", proj4string(x))
            if (length(gpf) == 0 && length(gpi) == 0) {
              use_ob_tran <- 0L
              warning("project: use_ob_tran set FALSE")
            } else {
              if (length(gpf) > 0) use_ob_tran <- -1L
              else use_ob_tran <- 1L
            }
            if (use_ob_tran != 0L) {
              if (is.projected(x) || length(grep("longlat", slot(CRSobj, "projargs"))) == 0)
                stop("Use ob_tran to or from unprojected objects only")
            }
          } else {
            use_ob_tran <- 0L
          }
        } else {
          use_ob_tran <- 0L
        }
        if (!is.null(dots$enforce_xy)) {
          stopifnot(is.logical(dots$enforce_xy))
          stopifnot(length(dots$enforce_xy) == 1L)
          stopifnot(!is.na(dots$enforce_xy))
          enforce_xy <- dots$enforce_xy
        } else {
            enforce_xy <- get_enforce_xy()
        }
        coordOp <- NULL
        if (!is.null(dots$coordOp)) {
            coordOp <- dots$coordOp
        }
        
        use_aoi <- TRUE
        if (!is.null(dots$use_aoi)) {
            use_aoi <- dots$use_aoi
            if (use_ob_tran != 0L) use_aoi <- FALSE
            if (!is.null(coordOp)) use_aoi <- FALSE
            if (!new_proj_and_gdal()) use_aoi <- FALSE
        }
        
        aoi <- NULL
        if (use_aoi && new_proj_and_gdal()) {
            aoi <- get_aoi(x)
            if (!is.null(aoi)) {
                stopifnot(length(aoi) == 4)
                aoi <- as.numeric(aoi)
            }
        }

	crds <- coordinates(x)
	crds.names <- dimnames(crds)[[2]] # crds is matrix
	n <- nrow(crds)
        attr(n, "ob_tran") <- use_ob_tran
        if (ncol(crds) == 2) {
            if (new_proj_and_gdal()) {
                if (use_ob_tran == 0L) {
                    attr(n, "enforce_xy") <- enforce_xy
                    res <- .Call("transform_ng", from_args, to_args, coordOp, n,
		        as.double(crds[,1]), as.double(crds[,2]), NULL, aoi,
		        PACKAGE="rgdal")
                    out_coordOp <- res[[5]]
                } else {
                    if (use_ob_tran == -1L) {
                        inv <- TRUE
                        proj <- paste0(slot(CRSobj, "projargs"), " +type=crs")
                    }
                    if (use_ob_tran == 1L) {
                        inv <- FALSE
                        proj <- paste0(slot(slot(x, "proj4string"),
                            "projargs"), " +type=crs")
                    }
                    if (use_ob_tran == 0L) use_ob_tran1 <- FALSE
                    else use_ob_tran1 <- TRUE
                    if (is.null(coordOp)) {
                        out_coordOp <- .Call("project_ng_coordOp", proj,
                            as.logical(inv), NULL, as.logical(use_ob_tran1),
                            PACKAGE="rgdal")
                        s2t <- strsplit(out_coordOp, " ")[[1]]
                        if (length(s2t) == 0L) stop("malformed pipeline found")
                        out_coordOp <- paste("+", s2t, sep="", collapse=" ")
                    }
                    res <- .Call("project_ng", as.integer(n), 
                        as.double(crds[,1]), as.double(crds[,2]), NULL,
                        out_coordOp, PACKAGE="rgdal")
                }
            } else {
	        res <- .Call("transform", from_args, to_args, n,
                    as.double(crds[,1]), as.double(crds[,2]), NULL,
                    PACKAGE="rgdal")
            }
	    if (any(!is.finite(res[[1]])) || any(!is.finite(res[[2]]))) {
		k <- which(!is.finite(res[[1]]) | !is.finite(res[[2]]))
		cat("non finite transformation detected:\n")
		print(cbind(crds, res[[1]], res[[2]])[k,])
		stop(paste("failure in points", paste(k, collapse=":")))
	    }
	    crds[,1:2] <- cbind(res[[1]], res[[2]])
        } else {
            if (new_proj_and_gdal()) {
                if (use_ob_tran == 0L) {
                    attr(n, "enforce_xy") <- enforce_xy
                    res <- .Call("transform_ng", from_args, to_args, coordOp, n,
		        as.double(crds[,1]), as.double(crds[,2]),
                        as.double(crds[,3]), aoi, PACKAGE="rgdal")
                    out_coordOp <- res[[6]]
                } else {
                    if (use_ob_tran == -1L) {
                        inv <- TRUE
                        proj <- paste0(slot(CRSobj, "projargs"), " +type=crs")
                    }
                    if (use_ob_tran == 1L) {
                        inv <- FALSE
                        proj <- paste0(slot(slot(x, "proj4string"),
                            "projargs"), " +type=crs")
                    }
                    if (use_ob_tran == 0L) use_ob_tran1 <- FALSE
                    else use_ob_tran1 <- TRUE
                    if (is.null(coordOp)) {
                        out_coordOp <- .Call("project_ng_coordOp", proj,
                            as.logical(inv), NULL, as.logical(use_ob_tran1),
                            PACKAGE="rgdal")
                        s2t <- strsplit(out_coordOp, " ")[[1]]
                        if (length(s2t) == 0L) stop("malformed pipeline found")
                        out_coordOp <- paste("+", s2t, sep="", collapse=" ")
                    }
                    res <- .Call("project_ng", as.integer(n), 
                        as.double(crds[,1]), as.double(crds[,2]),
                        as.double(crds[,3]), out_coordOp, PACKAGE="rgdal")
                }
            } else {
	    res <- .Call("transform", from_args, to_args, n,
		as.double(crds[,1]), as.double(crds[,2]), as.double(crds[,3]),
                PACKAGE="rgdal")
            }
	    if (any(!is.finite(res[[1]])) || any(!is.finite(res[[2]]))
                || any(!is.finite(res[[3]]))) {
		k <- which(!is.finite(res[[1]]) | !is.finite(res[[2]])
                    | !is.finite(res[[3]]))
		cat("non finite transformation detected:\n")
		print(cbind(crds, res[[1]], res[[2]], res[[3]])[k,])
		stop(paste("failure in points", paste(k, collapse=":")))
	    }
	    crds[,1:3] <- cbind(res[[1]], res[[2]], res[[3]])
        }
	# make sure coordinate names are set back:
        if (new_proj_and_gdal()) {
            if (substring(out_coordOp, 1, 1) != "+") {
                out_coordOp <-  gsub(" ", " +", paste0("+", out_coordOp))
            }
            assign(".last_coordOp", out_coordOp, envir=.RGDAL_CACHE)
        }
	dimnames(crds)[[2]] <- crds.names
	x <- SpatialPoints(coords=crds, proj4string=CRSobj)
	x
}
setMethod("spTransform", signature("SpatialPoints", "CRS"), spTransform.SpatialPoints)



"spTransform.SpatialPointsDataFrame" <- function(x, CRSobj, ...) {
	xSP <- as(x, "SpatialPoints")
	resSP <- spTransform(xSP, CRSobj, ...)
	# xDF <- as(x, "data.frame")
	xDF <- x@data # little need to add unique row.names here!
        if (packageVersion("sp") > "1.4.1") {
            crs <- rebuild_CRS(slot(resSP, "proj4string"))
        } else {
            rebuild_CRS <- NULL
            crs <- CRS(proj4string(resSP))
        }
        res <- SpatialPointsDataFrame(coords=coordinates(resSP), data=xDF,
		coords.nrs = numeric(0), proj4string = crs)
	res
}
setMethod("spTransform", signature("SpatialPointsDataFrame", "CRS"), 
	spTransform.SpatialPointsDataFrame)




setMethod("spTransform", signature("SpatialPixelsDataFrame", "CRS"), 
	function(x, CRSobj, ...) {
                warning("Grid warping not available, coercing to points")
		spTransform(as(x, "SpatialPointsDataFrame"), CRSobj, ...)})

setMethod("spTransform", signature("SpatialGridDataFrame", "CRS"), 
	function(x, CRSobj, ...) {
                warning("Grid warping not available, coercing to points")
		spTransform(as(x, "SpatialPixelsDataFrame"), CRSobj, ...)})


".spTransform_Line" <- function(x, to_args, from_args, ii, jj,
                use_ob_tran, coordOp, enforce_xy, aoi) {
	crds <- slot(x, "coords")
	n <- nrow(crds)
        attr(n, "ob_tran") <- use_ob_tran
        if (new_proj_and_gdal()) {
            if (use_ob_tran == 0L) {
                attr(n, "enforce_xy") <- enforce_xy
                res <- .Call("transform_ng", from_args, to_args, coordOp, n,
		    as.double(crds[,1]), as.double(crds[,2]), NULL, aoi,
		    PACKAGE="rgdal")
                out_coordOp <- res[[5]]
            } else {
                if (use_ob_tran == -1L) {
                    inv <- TRUE
                    if (is.null(attr(to_args, "p4s"))) proj <- to_args
                    else proj <- paste0(attr(to_args, "p4s"), " +type=crs")
                }
                if (use_ob_tran == 1L) {
                    inv <- FALSE
                    if (is.null(attr(from_args, "p4s"))) proj <- from_args
                    else proj <- paste0(attr(from_args, "p4s"), " +type=crs")
                }
                if (use_ob_tran == 0L) use_ob_tran1 <- FALSE
                else use_ob_tran1 <- TRUE
                if (is.null(coordOp)) {
                    coordOp <- .Call("project_ng_coordOp", proj,
                        as.logical(inv), NULL, as.logical(use_ob_tran1),
                        PACKAGE="rgdal")
                    s2t <- strsplit(coordOp, " ")[[1]]
                    if (length(s2t) == 0L) stop("malformed pipeline found")
                    coordOp <- paste("+", s2t, sep="", collapse=" ")
                }
                res <- .Call("project_ng", as.integer(n), 
                    as.double(crds[,1]), as.double(crds[,2]), NULL,
                    coordOp, PACKAGE="rgdal")
                out_coordOp <- coordOp
            }
        } else {
	    res <- .Call("transform", from_args, to_args, n,
		as.double(crds[,1]), as.double(crds[,2]), NULL,
		PACKAGE="rgdal")
        }
	if (any(!is.finite(res[[1]])) || any(!is.finite(res[[2]]))) {
		k <- which(!is.finite(res[[1]]) | !is.finite(res[[2]]))
		cat("non finite transformation detected:\n")
		print(cbind(crds, res[[1]], res[[2]])[k,])
		stop(paste("failure in Lines", ii, "Line", jj, 
			"points", paste(k, collapse=":")))
	}
	crds <- cbind(res[[1]], res[[2]])
	x <- Line(coords=crds)
        if (new_proj_and_gdal()) {
            if (substring(out_coordOp, 1, 1) == "+") {
                attr(x, "coordOp") <- out_coordOp
            } else {
                attr(x, "coordOp") <-  gsub(" ", " +", paste0("+", out_coordOp))
            }
        }
	x
}

#setMethod("spTransform", signature("Sline", "CRS"), spTransform.Sline)

".spTransform_Lines" <- function(x, to_args, from_args, ii,
                use_ob_tran, coordOp, enforce_xy, aoi) {
	ID <- slot(x, "ID")
	input <- slot(x, "Lines")
	n <- length(input)
	output <- vector(mode="list", length=n)
        out_coordOp <- NULL
	for (i in 1:n) {
            output[[i]] <- .spTransform_Line(input[[i]], 
		to_args=to_args, from_args=from_args, ii=ii, jj=i,
                use_ob_tran=use_ob_tran, coordOp=coordOp,
                enforce_xy=enforce_xy, aoi=aoi)
            if (is.null(coordOp) && new_proj_and_gdal() && i == 1) {
                coordOp <- attr(output[[i]], "coordOp")
#                coordOp <-  gsub(" ", " +", paste0("+", out_coordOp))
            }
            attr(output[[i]], "coordOp") <- NULL
        }
	x <- Lines(output, ID)
        if (new_proj_and_gdal()) attr(x, "coordOp") <- out_coordOp
	x
}

#setMethod("spTransform", signature("Slines", "CRS"), spTransform.Slines)

"spTransform.SpatialLines" <- function(x, CRSobj, ...) {
        if (new_proj_and_gdal()) {
          if (get_transform_wkt_comment()) {
            if (is.null(comment(slot(x, "proj4string")))) {
                from_args <- paste0(slot(slot(x, "proj4string"),
                  "projargs"), " +type=crs")
                warning("NULL source CRS comment, falling back to PROJ string")
	        if (is.na(from_args)) 
		    stop("No transformation possible from NA source CRS")
                if (length(grep("\\+init\\=", from_args)) > 0) {
                    warning("+init dropped in PROJ string")
                    strres <- unlist(strsplit(from_args, " "))
                    from_args <- paste(strres[-grep("\\+init\\=", strres)],
                        collapse=" ")
                }
            } else {
                from_args <- comment(slot(x, "proj4string"))
                attr(from_args, "p4s") <- slot(slot(x, "proj4string"),
                    "projargs")
            }
            if (is.null(comment(CRSobj))) {
                warning("NULL target CRS comment, falling back to PROJ string")
	        to_args <- paste0(slot(CRSobj, "projargs"), " +type=crs")
	        if (is.na(to_args)) 
		    stop("No transformation possible to NA target CRS")
                if (length(grep("\\+init\\=", to_args)) > 0) {
                    warning("+init dropped in PROJ string")
                    strres <- unlist(strsplit(to_args, " "))
                    to_args <- paste(strres[-grep("\\+init\\=", strres)],
                        collapse=" ")
                }
            } else {
                to_args <- comment(CRSobj)
                attr(to_args, "p4s") <- slot(CRSobj, "projargs")
            }
          } else {
            from_args <- paste0(slot(slot(x, "proj4string"), "projargs"),
              " +type=crs")
            to_args <- paste0(slot(CRSobj, "projargs"), " +type=crs")
          }
        } else {
	    from_args <- proj4string(x)
	    if (is.na(from_args)) 
		stop("No transformation possible from NA from reference system")
	    to_args <- slot(CRSobj, "projargs")
	    if (is.na(to_args)) 
		stop("No transformation possible to NA to reference system")
        }
	dots = list(...)
        if (!is.null(dots$use_ob_tran)) {
          stopifnot(is.logical(dots$use_ob_tran))
          if (dots$use_ob_tran) {
            gpf <- grep("proj=ob_tran", slot(CRSobj, "projargs"))
            gpi <- grep("proj=ob_tran", proj4string(x))
            if (length(gpf) == 0 && length(gpi) == 0) {
              use_ob_tran <- 0L
              warning("project: use_ob_tran set FALSE")
            } else {
              if (length(gpf) > 0) use_ob_tran <- -1L
              else use_ob_tran <- 1L
            }
            if (use_ob_tran != 0L) {
              if (is.projected(x) || length(grep("longlat", slot(CRSobj, "projargs"))) == 0)
                stop("Use ob_tran to or from unprojected objects only")
            }
          } else {
            use_ob_tran <- 0L
          }
        } else {
          use_ob_tran <- 0L
        }
        if (!is.null(dots$enforce_xy)) {
          stopifnot(is.logical(dots$enforce_xy))
          stopifnot(length(dots$enforce_xy) == 1L)
          stopifnot(!is.na(dots$enforce_xy))
          enforce_xy <- dots$enforce_xy
        } else {
            enforce_xy <- get_enforce_xy()
        }
        coordOp <- NULL
        if (!is.null(dots$coordOp)) {
            coordOp <- dots$coordOp
        }
        use_aoi <- TRUE
        if (!is.null(dots$use_aoi)) {
            use_aoi <- dots$use_aoi
            if (use_ob_tran != 0L) use_aoi <- FALSE
            if (!is.null(coordOp)) use_aoi <- FALSE
            if (!new_proj_and_gdal()) use_aoi <- FALSE
        }
        
        aoi <- NULL
        if (use_aoi && new_proj_and_gdal()) {
            aoi <- get_aoi(x)
            if (!is.null(aoi)) {
                stopifnot(length(aoi) == 4)
                aoi <- as.numeric(aoi)
            }
        }

	input <- slot(x, "lines")
	n <- length(input)
	output <- vector(mode="list", length=n)
	for (i in 1:n) {
            output[[i]] <- .spTransform_Lines(input[[i]], 
		to_args=to_args, from_args=from_args, ii=i,
                use_ob_tran=use_ob_tran, coordOp=coordOp,
                enforce_xy=enforce_xy, aoi=aoi)
            if (is.null(coordOp) && new_proj_and_gdal() && i == 1) {
                coordOp <- attr(output[[i]], "coordOp")
#                coordOp <-  gsub(" ", " +", paste0("+", out_coordOp))
            }
            attr(output[[i]], "coordOp") <- NULL
        }
        if (new_proj_and_gdal()) assign(".last_coordOp", coordOp,
            envir=.RGDAL_CACHE)
	res <- SpatialLines(output, proj4string=CRSobj)
	res
}
setMethod("spTransform", signature("SpatialLines", "CRS"), spTransform.SpatialLines)
"spTransform.SpatialLinesDataFrame" <- function(x, CRSobj, ...) {
	xSP <- as(x, "SpatialLines")
	resSP <- spTransform(xSP, CRSobj, ...)
	xDF <- as(x, "data.frame")
	res <- SpatialLinesDataFrame(sl=resSP, data=xDF, match.ID = FALSE)
	res
}
setMethod("spTransform", signature("SpatialLinesDataFrame", "CRS"), spTransform.SpatialLinesDataFrame)




".spTransform_Polygon" <- function(x, to_args, from_args, ii, jj,
                use_ob_tran, coordOp, enforce_xy, aoi) {
	crds <- slot(x, "coords")
	n <- nrow(crds)
        attr(n, "ob_tran") <- use_ob_tran
        if (new_proj_and_gdal()) {
            if (use_ob_tran == 0L) {
                attr(n, "enforce_xy") <- enforce_xy
                res <- .Call("transform_ng", from_args, to_args, coordOp, n,
		    as.double(crds[,1]), as.double(crds[,2]), NULL, aoi,
		    PACKAGE="rgdal")
                out_coordOp <- res[[5]]
            } else {
                if (use_ob_tran == -1L) {
                    inv <- TRUE
                    if (is.null(attr(to_args, "p4s"))) proj <- to_args
                    else proj <- paste0(attr(to_args, "p4s"), " +type=crs")
                }
                if (use_ob_tran == 1L) {
                    inv <- FALSE
                    if (is.null(attr(from_args, "p4s"))) proj <- from_args
                    else proj <- paste0(attr(from_args, "p4s"), " +type=crs")
                }
                if (use_ob_tran == 0L) use_ob_tran1 <- FALSE
                else use_ob_tran1 <- TRUE
                if (is.null(coordOp)) {
                    coordOp <- .Call("project_ng_coordOp", proj,
                        as.logical(inv), NULL, as.logical(use_ob_tran1),
                        PACKAGE="rgdal")
                    s2t <- strsplit(coordOp, " ")[[1]]
                    if (length(s2t) == 0L) stop("malformed pipeline found")
                    coordOp <- paste("+", s2t, sep="", collapse=" ")
                }
                res <- .Call("project_ng", as.integer(n), 
                    as.double(crds[,1]), as.double(crds[,2]), NULL,
                    coordOp, PACKAGE="rgdal")
                out_coordOp <- coordOp
            }
        } else {
	    res <- .Call("transform", from_args, to_args, n,
		as.double(crds[,1]), as.double(crds[,2]), NULL,
		PACKAGE="rgdal")
        }
	if (any(!is.finite(res[[1]])) || any(!is.finite(res[[2]]))) {
		k <- which(!is.finite(res[[1]]) | !is.finite(res[[2]]))
		cat("non finite transformation detected:\n")
		print(cbind(crds, res[[1]], res[[2]])[k,])
		stop(paste("failure in Polygons", ii, "Polygon", jj, 
			"points", paste(k, collapse=":")))
	}
	crds <- cbind(res[[1]], res[[2]])
	x <- Polygon(coords=crds)
        if (new_proj_and_gdal()) {
            if (substring(out_coordOp, 1, 1) == "+") {
                attr(x, "coordOp") <- out_coordOp
            } else {
                attr(x, "coordOp") <-  gsub(" ", " +", paste0("+", out_coordOp))
            }
        }
	x
}


".spTransform_Polygons" <- function(x, to_args, from_args, ii,
                use_ob_tran, coordOp, enforce_xy, aoi) {
	ID <- slot(x, "ID")
	input <- slot(x, "Polygons")
	n <- length(input)
	output <- vector(mode="list", length=n)
        out_coordOp <- NULL
	for (i in 1:n) {
            output[[i]] <- .spTransform_Polygon(input[[i]], 
		to_args=to_args, from_args=from_args, ii=ii, jj=i,
                use_ob_tran=use_ob_tran, coordOp=coordOp,
                enforce_xy=enforce_xy, aoi=aoi)
            if (is.null(coordOp) && new_proj_and_gdal() && i == 1) {
                coordOp <- attr(output[[i]], "coordOp")
#                coordOp <-  gsub(" ", " +", paste0("+", out_coordOp))
            }
            attr(output[[i]], "coordOp") <- NULL
        }
	res <- Polygons(output, ID)
        if (!is.null(comment(x))) comment(res) <- comment(x)
        if (new_proj_and_gdal()) attr(res, "coordOp") <- coordOp
	res
}


"spTransform.SpatialPolygons" <- function(x, CRSobj, ...) {
        if (new_proj_and_gdal()) {
          if (get_transform_wkt_comment()) {
            if (is.null(comment(slot(x, "proj4string")))) {
                from_args <- paste0(slot(slot(x, "proj4string"), "projargs"),
                  " +type=crs")
                warning("NULL source CRS comment, falling back to PROJ string")
	        if (is.na(from_args)) 
		    stop("No transformation possible from NA source CRS")
                if (length(grep("\\+init\\=", from_args)) > 0) {
                    warning("+init dropped in PROJ string")
                    strres <- unlist(strsplit(from_args, " "))
                    from_args <- paste(strres[-grep("\\+init\\=", strres)],
                        collapse=" ")
                }
            } else {
                from_args <- comment(slot(x, "proj4string"))
                attr(from_args, "p4s") <- slot(slot(x, "proj4string"),
                    "projargs")
            }
            if (is.null(comment(CRSobj))) {
                warning("NULL target CRS comment, falling back to PROJ string")
	        to_args <- paste0(slot(CRSobj, "projargs"), " +type=crs")
	        if (is.na(to_args)) 
		    stop("No transformation possible to NA target CRS")
                if (length(grep("\\+init\\=", to_args)) > 0) {
                    warning("+init dropped in PROJ string")
                    strres <- unlist(strsplit(to_args, " "))
                    to_args <- paste(strres[-grep("\\+init\\=", strres)],
                        collapse=" ")
                }
            } else {
                to_args <- comment(CRSobj)
                attr(to_args, "p4s") <- slot(CRSobj, "projargs")
            }
          } else {
            from_args <- paste0(slot(slot(x, "proj4string"), "projargs"),
              " +type=crs")
            to_args <- paste0(slot(CRSobj, "projargs"), " +type=crs")
          }
        } else {
	    from_args <- proj4string(x)
	    if (is.na(from_args)) 
		stop("No transformation possible from NA from reference system")
	    to_args <- slot(CRSobj, "projargs")
	    if (is.na(to_args)) 
		stop("No transformation possible to NA to reference system")
        }
	dots = list(...)
        if (!is.null(dots$use_ob_tran)) {
          stopifnot(is.logical(dots$use_ob_tran))
          if (dots$use_ob_tran) {
            gpf <- grep("proj=ob_tran", slot(CRSobj, "projargs"))
            gpi <- grep("proj=ob_tran", proj4string(x))
            if (length(gpf) == 0 && length(gpi) == 0) {
              use_ob_tran <- 0L
              warning("project: use_ob_tran set FALSE")
            } else {
              if (length(gpf) > 0) use_ob_tran <- -1L
              else use_ob_tran <- 1L
            }
            if (use_ob_tran != 0L) {
              if (is.projected(x) || length(grep("longlat", slot(CRSobj, "projargs"))) == 0)
                stop("Use ob_tran to or from unprojected objects only")
            }
          } else {
            use_ob_tran <- 0L
          }
        } else {
          use_ob_tran <- 0L
        }
        if (!is.null(dots$enforce_xy)) {
          stopifnot(is.logical(dots$enforce_xy))
          stopifnot(length(dots$enforce_xy) == 1L)
          stopifnot(!is.na(dots$enforce_xy))
          enforce_xy <- dots$enforce_xy
        } else {
            enforce_xy <- get_enforce_xy()
        }
        coordOp <- NULL
        if (!is.null(dots$coordOp)) {
            coordOp <- dots$coordOp
        }
        use_aoi <- TRUE
        if (!is.null(dots$use_aoi)) {
            use_aoi <- dots$use_aoi
            if (use_ob_tran != 0L) use_aoi <- FALSE
            if (!is.null(coordOp)) use_aoi <- FALSE
            if (!new_proj_and_gdal()) use_aoi <- FALSE
        }
        
        aoi <- NULL
        if (use_aoi && new_proj_and_gdal()) {
            aoi <- get_aoi(x)
            if (!is.null(aoi)) {
                stopifnot(length(aoi) == 4)
                aoi <- as.numeric(aoi)
            }
        }

	input <- slot(x, "polygons")
	n <- length(input)
	output <- vector(mode="list", length=n)
	for (i in 1:n) {
            output[[i]] <- .spTransform_Polygons(input[[i]], 
		to_args=to_args, from_args=from_args, ii=i,
                use_ob_tran=use_ob_tran, coordOp=coordOp,
                enforce_xy=enforce_xy, aoi=aoi)
            if (is.null(coordOp) && new_proj_and_gdal() && i == 1) {
                coordOp <- attr(output[[i]], "coordOp")
#                coordOp <-  gsub(" ", " +", paste0("+", out_coordOp))
            }
            attr(output[[i]], "coordOp") <- NULL
        }
        if (new_proj_and_gdal()) assign(".last_coordOp", coordOp, 
            envir=.RGDAL_CACHE)
	res <- SpatialPolygons(output, pO=slot(x, "plotOrder"), 
		proj4string=CRSobj)
	res
}
setMethod("spTransform", signature("SpatialPolygons", "CRS"), spTransform.SpatialPolygons)

"spTransform.SpatialPolygonsDataFrame" <- function(x, CRSobj, ...) {
	xSP <- as(x, "SpatialPolygons")
	resSP <- spTransform(xSP, CRSobj, ...)
	xDF <- as(x, "data.frame")
	res <- SpatialPolygonsDataFrame(Sr=resSP, data=xDF, match.ID = FALSE)
	res
}
setMethod("spTransform", signature("SpatialPolygonsDataFrame", "CRS"), spTransform.SpatialPolygonsDataFrame)

projInfo <- function(type="proj") {
    opts <- c("proj", "ellps", "datum", "units")
    if (!(type %in% opts)) stop("unknown type")
    t <- as.integer(match(type[1], opts) - 1)
    if (is.na(t)) stop("unknown type")
    res <- .Call("RGDAL_projInfo", t, PACKAGE="rgdal")
    if (type == "proj") res$description <- sapply(strsplit(as.character(
        res$description), "\n"), function(x) x[1])
    if (type == "datum" && is.null(res)) warning("datum list not available for PROJ > 5")
    res <- data.frame(res)
    res
}

setClass("SpatialGDAL",
    representation("Spatial", grid = "GridTopology", grod = "GDALReadOnlyDataset", name = "character"),
    validity = function(object) {
        if (is.null(object@grod))
            stop("grod is NULL; this should not happen")
        return(TRUE)
    }
)
setClass("SpatialGDALWrite", "SpatialGDAL")

open.SpatialGDAL = function(con, ..., silent = FALSE,
        allowedDrivers=NULL, options=NULL) {
	if (nchar(con) == 0) stop("empty file name")

	grod = GDAL.open(con, read.only = TRUE, silent=silent,
              allowedDrivers=allowedDrivers, options=options)

	d = dim(grod)
	if (!silent) {
		cat(paste(con, "has GDAL driver", getDriverName(getDriver(grod)),"\n"))
		cat(paste("and has", d[1], "rows and", d[2], "columns\n"))
	}
#	p4s <- .Call("RGDAL_GetProjectionRef", grod, PACKAGE="rgdal")
	p4s <- getProjectionRef(grod, OVERRIDE_PROJ_DATUM_WITH_TOWGS84=NULL)
	if (!is.na(p4s) && nchar(p4s) == 0) p4s <- as.character(NA)
        if (new_proj_and_gdal()) wkt2 <- comment(p4s)
        oCRS <- CRS(c(p4s))
        if (new_proj_and_gdal()) comment(oCRS) <- wkt2
	gt = .Call('RGDAL_GetGeoTransform', grod, PACKAGE="rgdal")
        if (attr(gt, "CE_Failure")) warning("GeoTransform values not available")
	# [1] 178400     40      0 334000      0    -40
	if (any(gt[c(3,5)] != 0.0)) 
		stop("rotated grid cannot be read; try readGDAL to read as points")
	cellsize = abs(c(gt[2],gt[6]))
	ysign <- sign(gt[6])
	output.dim <- dim(grod)[1:2] # rows=nx, cols=ny
	half.cell <- c(0.5, 0.5)
	co.x <- gt[1] + (half.cell[2]) * cellsize[1]
	co.y <- ifelse(ysign < 0, gt[4] + (ysign*((output.dim[1]) + (ysign*half.cell[1]))) * abs(cellsize[2]),
		gt[4] + (ysign*((offset[1]) + (ysign*half.cell[1]))) * 
		abs(cellsize[2]))
	cellcentre.offset <- c(x=co.x, y=co.y)
	grid = GridTopology(cellcentre.offset, cellsize, rev(output.dim))
	#bbox = bbox(SpatialGrid(grid)),

	x = SpatialPoints(rbind(grid@cellcentre.offset, 
			grid@cellcentre.offset + (grid@cells.dim - 1) * grid@cellsize))
	x@bbox[,1] = x@bbox[,1] - 0.5 * grid@cellsize
	x@bbox[,2] = x@bbox[,2] + 0.5 * grid@cellsize

	data = new("SpatialGDAL", 
		bbox = x@bbox,
		proj4string = oCRS, 
		grid = grid,
		# data = data.frame(), 
		grod = grod,
		name = con)
	return(data)
}

close.SpatialGDAL = function(con, ...) {
	GDAL.close(con@grod)
	invisible(NULL)
}
close.SpatialGDALWrite = close.SpatialGDAL 

copy.SpatialGDAL = function(dataset, fname, driver = 
		getDriver(dataset@grod), strict = FALSE, options = NULL, silent = FALSE)
{
	if (nchar(fname) == 0) 
		stop("empty file name")
	assertClass(dataset, 'SpatialGDAL')
	# grod = GDAL.open(con, read.only = FALSE)

	if (is.character(driver)) 
		driver <- new("GDALDriver", driver)
	if (nchar(fname) == 0) 
		stop("empty file name")

	if (!is.null(options) && !is.character(options))
		stop("options not character")

	grod <- new('GDALTransientDataset',
		handle = .Call('RGDAL_CopyDataset',
			dataset@grod, driver, as.integer(strict),
			as.character(options), fname, PACKAGE="rgdal"))

	data = new("SpatialGDALWrite", 
		bbox = bbox(dataset),
		proj4string = slot(dataset, "proj4string"), 
		grid = dataset@grid, grod = grod, name = fname)
	return(data)
}

setAs("SpatialGDAL", "SpatialGridDataFrame",
	function(from) { 
#		print("doing the coerce...")
		from[]
	}
)

setAs("SpatialGDAL", "SpatialPixelsDataFrame",
	function(from) as(from[], "SpatialPixelsDataFrame")
)

setMethod("[", "SpatialGDAL",
	function(x, i, j, ... , drop = FALSE)
		x@grod[i = i, j = j, ...]
)

setMethod("[[", c("SpatialGDAL", "ANY", "missing"),
    function(x, i, j, ...)
        x[,,i][[1]] # efficient: reads only band i
)

# avoid inheritance:
setMethod("$", "SpatialGDAL",
    function(x, name)
        stop("use [[ with numeric index to select single band")
)

# avoid inheritance:
setReplaceMethod("$", "SpatialGDAL",
    function(x, name, value)
		stop("no replacement method available")
)

# avoid inheritance:
setReplaceMethod("[[", c("SpatialGDAL", "ANY", "missing", "ANY"),
    function(x, i, j, value)
		stop("no replacement method available")
)

setMethod("summary", "SpatialGDAL",
	function(object, ...) {
		obj = list()
		obj$class = class(object)
		obj$grid = object@grid
		obj$name = object@name
		class(obj) = "summary.SpatialGDAL"
		obj
	}
)

print.summary.SpatialGDAL = function(x, ...) {
	cat(paste("object of class", x$class, "\n"))
	cat(paste("file name:", x$name, "\n"))
	print(x$grid)
}

setReplaceMethod("[", "SpatialGDALWrite", function(x, i, j, ..., value) {
	ncol = gridparameters(x@grid)[1,"cells.dim"]
	nrow = gridparameters(x@grid)[2,"cells.dim"]
	if (!is.numeric(value)) stop("Numeric bands required")
	if (missing(i)) i = 1:nrow
	if (missing(j)) j = 1:ncol
	i = i - 1 # y/row index -- zero offset
	j = j - 1 # x/cols -- zero offset
	if (length(i) * length(j) != length(value))
		stop("lengths do not match")
	dots = list(...)
	if (length(dots) > 0L)
		band = dots[[1]]
	else
		band = 1
#	mvFlag = .Call("RGDAL_GetNoDataValue", x@grod)
#	if (!is.na(mvFlag))
#		data[is.na(data)] = mvFlag
	if (!is.matrix(value))
		value = matrix(value, length(i), length(j))
	putRasterData(x@grod, value, band, c(i[1], j[1])) # offset y, x
	x
})

# Note that reg.finalizer does not finalize objects
# at the end of an R session. This could be a problem.
.setCollectorFun <- function(object, fun) {

  if (is.null(fun)) fun <- function(obj) obj
  reg.finalizer(object, fun, onexit=TRUE)

}

assertClass <- function(object, class) {
  
  if (class %in% is(object))
    invisible(object)
  else
    stop(paste('Object is not a member of class', class))

}

.GDALDataTypes <- c('Unknown', 'Byte', 'UInt16', 'Int16', 'UInt32',
                    'Int32', 'Float32', 'Float64', 'CInt16', 'CInt32',
                    'CFloat32', 'CFloat64')

.normalize_if_path <- function(file, mustWork = NA) {
  if (file.exists(file)) {
    file <- normalizePath(file, mustWork = mustWork)
  }
  file
}


setClass('GDALMajorObject',
         representation(handle = 'externalptr'))
 
getDescription <- function(object) {

  assertClass(object, 'GDALMajorObject')

  .Call('RGDAL_GetDescription', object, PACKAGE="rgdal")

}

getGDALVersionInfo <- function(str="--version") {
    stopifnot(is.character(str))
    .Call("RGDAL_GDALVersionInfo", str, PACKAGE="rgdal")
}

getGDALCheckVersion <- function() {
    .Call("RGDAL_GDALCheckVersion", PACKAGE="rgdal")
}

getGDALwithGEOS <- function() {
    res <- .Call("RGDAL_GDALwithGEOS", PACKAGE="rgdal")
    if (is.character(res)) {
        oo <- strsplit(res, "\n")[[1]]
        res <- "GEOS_ENABLED=YES" %in% oo
        attr(res, "GEOS_VERSION") <- substring(oo[grep("GEOS_VERSION", oo)], 14)
    }
    res
}

getGDAL_DATA_Path <- function() {
    res <- .Call("RGDAL_GDAL_DATA_Info", PACKAGE="rgdal")
    res <- sub("stateplane.csv", "", res)
    n <- nchar(res)
    res <- substring(res, 1, n-1)
    res
}

version_sp_linkingTo <- function() {
    .Call("rgdal_sp_linkingTo_version")
}


get_cached_orig_PROJ_LIB <- function() {
  get(".rgdal_old.PROJ_LIB", envir=.RGDAL_CACHE)
}

get_cached_orig_GDAL_DATA <- function() {
  get(".rgdal_old.GDAL_DATA", envir=.RGDAL_CACHE)
}

get_cached_set_PROJ_LIB <- function() {
  get(".rgdal_set.PROJ_LIB", envir=.RGDAL_CACHE)
}

get_cached_set_GDAL_DATA <- function() {
  get(".rgdal_set.GDAL_DATA", envir=.RGDAL_CACHE)
}




setClass('GDALDriver', 'GDALMajorObject')

setClass('GDALReadOnlyDataset', 'GDALMajorObject')

setClass('GDALDataset', 'GDALReadOnlyDataset')

setClass('GDALTransientDataset', 'GDALDataset')
         
setClass('GDALRasterBand', 'GDALMajorObject')

getGDALDriverNames <- function() {
  res <- .Call('RGDAL_GetDriverNames', PACKAGE="rgdal")
  has_isRaster <- 0L
  if (!is.null(attr(res, "isRaster"))) has_isRaster <- 1L
  if (has_isRaster) res$isRaster <- attr(res, "isRaster")
  res <- as.data.frame(res, stringsAsFactors=FALSE)
  if (has_isRaster) res <- res[res$isRaster,]
  res <- res[order(res$name),]
  row.names(res) <- NULL
  res
}

setMethod('initialize', 'GDALDriver',
          def = function(.Object, name, handle = NULL) {
            if (is.null(handle)) {
              slot(.Object, 'handle') <- {
                .Call('RGDAL_GetDriver', as.character(name), PACKAGE="rgdal")
              }
            } else {
              slot(.Object, 'handle') <- handle
            }
            .Object
          })

getDriverName <- function(driver) {

  assertClass(driver, 'GDALDriver')

  .Call('RGDAL_GetDriverShortName', driver, PACKAGE="rgdal")

}

getDriverLongName <- function(driver) {

  assertClass(driver, 'GDALDriver')

  .Call('RGDAL_GetDriverLongName', driver, PACKAGE="rgdal")

}

setMethod('initialize', 'GDALReadOnlyDataset',
          def = function(.Object, filename, silent=FALSE, handle = NULL,
            allowedDrivers=NULL, options=NULL) {
            if (is.null(handle)) {
              filename <- as.character(filename)
	      if (nchar(filename) == 0) stop("empty file name")
              silent <- as.logical(silent)
              if (length(silent) != 1L || is.na(silent) || !is.logical(silent))
                  stop("options(warn) not set")
	      if (!is.null(options)) options <- as.character(options)
	      if (!is.null(allowedDrivers))
                  allowedDrivers <- as.character(allowedDrivers)
              slot(.Object, 'handle') <- {
                .Call('RGDAL_OpenDataset',
                        .normalize_if_path(filename, mustWork=NA), 
			TRUE, silent, allowedDrivers, options, PACKAGE="rgdal")
              }
            } else {
              slot(.Object, 'handle') <- handle
            }
            cfn <- function(handle) .Call('RGDAL_CloseHandle', 
		handle, PACKAGE="rgdal")
            .setCollectorFun(slot(.Object, 'handle'), cfn)
            .Object
          })

setMethod('initialize', 'GDALDataset',
          def = function(.Object, filename, silent=FALSE, handle = NULL,
            allowedDrivers=NULL, options=NULL) {
            if (is.null(handle)) {
              filename <- as.character(filename)
	      if (nchar(filename) == 0) stop("empty file name")
              silent <- as.logical(silent)
              if (length(silent) != 1L || is.na(silent) || !is.logical(silent))
                  stop("options(warn) not set")
	      if (!is.null(options)) options <- as.character(options)
	      if (!is.null(allowedDrivers))
                  allowedDrivers <- as.character(allowedDrivers)
              slot(.Object, 'handle') <- {
                .Call('RGDAL_OpenDataset', 
                        .normalize_if_path(filename, mustWork=NA), 
			FALSE, silent, allowedDrivers, options, PACKAGE="rgdal")
              }
            } else {
              slot(.Object, 'handle') <- handle
            }
            cfn <- function(handle) .Call('RGDAL_CloseHandle', 
		handle, PACKAGE="rgdal")
            .setCollectorFun(slot(.Object, 'handle'), cfn)
            .Object
          })

setMethod('initialize', 'GDALTransientDataset',
          def = function(.Object, driver, rows, cols, bands = 1,
            type = 'Byte', options = NULL, fname = NULL, handle = NULL) {
            if (is.null(handle)) {
              typeNum <- match(type, .GDALDataTypes, 1) - 1
	      if (is.null(fname)) {
                  my_tempfile <- tempfile()
              } else {
                  my_tempfile <- paste(tempdir(), "/",
                      paste(sample(letters, 3), collapse=""),
                      basename(fname[1]), sep="")
              }
	      if (nchar(my_tempfile) == 0) stop("empty file name")
	      if (!is.null(options)) options <- as.character(options)
              slot(.Object, 'handle') <- .Call('RGDAL_CreateDataset', driver,
                                              as.integer(c(cols, rows, bands)),
                                              as.integer(typeNum),
                                              options,
                                              my_tempfile, PACKAGE="rgdal")
            } else {
              slot(.Object, 'handle') <- handle
            }
            cfn <- function(handle) .Call('RGDAL_CloseHandle',
#            cfn <- function(handle) .Call('RGDAL_CloseDataset', RSB 081030
		handle, PACKAGE="rgdal")
            .setCollectorFun(slot(.Object, 'handle'), cfn)
            .Object
          })

getDriver <- function(dataset) {

  assertClass(dataset, 'GDALReadOnlyDataset')

  new('GDALDriver',
      handle = .Call('RGDAL_GetDatasetDriver', dataset, PACKAGE="rgdal"))

}

copyDataset <- function(dataset, driver, strict = FALSE, options = NULL, fname = NULL) {

  assertClass(dataset, 'GDALReadOnlyDataset')
  
  if (missing(driver)) driver <- getDriver(dataset)
  else if (is.character(driver)) driver <- new("GDALDriver", driver)

  if (is.null(fname)) {
     my_tempfile <- tempfile()
  } else {
     my_tempfile <- paste(tempdir(), "/",
       paste(sample(letters, 3), collapse=""),
       basename(fname[1]), sep="")
  }
  if (nchar(my_tempfile) == 0) stop("empty file name")
  if (!is.null(options) && !is.character(options))
    stop("options not character")
  
  new.obj <- new('GDALTransientDataset',
                 handle = .Call('RGDAL_CopyDataset',
                   dataset, driver,
                   as.integer(strict),
                   as.character(options),
                   my_tempfile, PACKAGE="rgdal"))

  new.obj
  
}

saveDataset <- function(dataset, filename, options=NULL, returnNewObj=FALSE) {

  assertClass(dataset, 'GDALReadOnlyDataset')
  
  new.class <- ifelse(class(dataset) == 'GDALTransientDataset',
                      'GDALDataset', class(dataset))
  if (!is.null(options) && !is.character(options))
    stop("options not character")
  
  filename <- as.character(filename)
  if (nchar(filename) == 0) stop("empty file name")
  new.obj <- new(new.class,
                 handle = .Call('RGDAL_CopyDataset',
                   dataset, getDriver(dataset),
                   FALSE, options, normalizePath(filename, mustWork=FALSE),
                   PACKAGE="rgdal"))

  if (returnNewObj) return(new.obj)
  invisible(GDAL.close(new.obj))
}

setGeneric('closeDataset', function(dataset) standardGeneric('closeDataset'))

"closeDataset.default" <- function(dataset) 
	stop("No default method for closeDataset")

setMethod("closeDataset", signature("ANY"), closeDataset.default)

setMethod('closeDataset', 'GDALReadOnlyDataset',
          def = function(dataset) {
            .setCollectorFun(slot(dataset, 'handle'), NULL)
            .Call('RGDAL_CloseDataset', dataset, PACKAGE="rgdal")
            invisible(gc())
          })

setMethod('closeDataset', 'GDALTransientDataset',
          def = function(dataset) {
            driver <- getDriver(dataset)
#            filename <- getDescription(dataset)
#            .Call('RGDAL_CloseDataset', driver, filename, PACKAGE="rgdal")
            .Call('RGDAL_CloseDataset', driver, PACKAGE="rgdal")
            invisible(gc())
            callNextMethod()
          })


saveDatasetAs <- function(dataset, filename, driver = NULL, options=NULL) {

  .Deprecated("saveDataset")

  assertClass(dataset, 'GDALReadOnlyDataset')
  
  filename <- as.character(filename)
  if (nchar(filename) == 0) stop("empty file name")
  if (is.null(driver)) driver <- getDriver(dataset)
  if (!is.null(options) && !is.character(options))
    stop("options not character")
  
  new.obj <- new('GDALReadOnlyDataset',
                 handle = .Call('RGDAL_CopyDataset',
                   dataset, driver, FALSE, options,
                   normalizePath(filename, mustWork=FALSE), PACKAGE="rgdal"))
  
  closeDataset(new.obj)
  
  err.opt <- getOption('show.error.messages')

  options(show.error.messages = FALSE)

  new.obj <- try(new('GDALDataset', filename))

  options(show.error.messages = err.opt)

  if (inherits(new.obj, 'try-error'))
    new.obj <- new('GDALReadOnlyDataset', filename)

  closeDataset(dataset)

  eval.parent(dataset <- new.obj)

  invisible(new.obj)
  
}


deleteDataset <- function(dataset) {

  assertClass(dataset, 'GDALDataset')
  
  driver <- getDriver(dataset)
  
  filename <- getDescription(dataset)
  
  .Call('RGDAL_DeleteFile', driver, filename, PACKAGE="rgdal")
  
  closeDataset(dataset)

}

isObjPtrNULL <- function(ptr) {

  stopifnot(is(ptr, "GDALMajorObject"))

  .Call("isGDALObjPtrNULL", ptr, PACKAGE="rgdal")

}

GDAL.open <- function(filename, read.only = TRUE, silent = FALSE,
        allowedDrivers=NULL, options=NULL) {

        if (!is.null(options) && !is.character(options))
            stop("options not character")
  
  
	res <- if(read.only)
          new("GDALReadOnlyDataset", filename, silent=silent,
              allowedDrivers=allowedDrivers, options=options)
        else
          new("GDALDataset", filename, silent=silent,
              allowedDrivers=allowedDrivers, options=options)
        
	res
        
}

GDAL.close <- function(dataset) {
            isTrans <- is(dataset, "GDALTransientDataset")
            if (isTrans) {
                if (isObjPtrNULL(dataset)) stop("object already closed")
#                filename <- getDescription(dataset)
            }
            .setCollectorFun(slot(dataset, 'handle'), NULL)
            .Call('RGDAL_CloseDataset', dataset, PACKAGE="rgdal")
#            .Call("RGDAL_CloseHandle", slot(dataset, 'handle'),
#                PACKAGE="rgdal")
             invisible(NULL)
#            if (isTrans) {
#                basen <- basename(filename)
#                dirn <- dirname(filename)
#                lf <- list.files(path=dirn, pattern=basen)
#                flf <- paste(dirn, lf, sep="/")
#                unlink(flf)
#            }
#            invisible(gc())
}

setMethod('dim', 'GDALReadOnlyDataset',
          def = function(x) {
            nrows <- .Call('RGDAL_GetRasterYSize', x, PACKAGE="rgdal")
            ncols <- .Call('RGDAL_GetRasterXSize', x, PACKAGE="rgdal")
            nbands <- .Call('RGDAL_GetRasterCount', x, PACKAGE="rgdal")
            if (nbands < 1) warning("no bands in dataset")
            if (nbands > 1)
              c(nrows, ncols, nbands)
            else
              c(nrows, ncols)
          })

getProjectionRef <- function(dataset, OVERRIDE_PROJ_DATUM_WITH_TOWGS84=NULL, enforce_xy=NULL) {

  assertClass(dataset, 'GDALReadOnlyDataset')

  vs <- strsplit(strsplit(getGDALVersionInfo(), ",")[[1]][1], " ")[[1]][2]
  env_absent <- is.null(getCPLConfigOption("OVERRIDE_PROJ_DATUM_WITH_TOWGS84"))
  wkt2 <- NULL
  if (!is.null(enforce_xy)) {
    stopifnot(is.logical(enforce_xy))
    stopifnot(length(enforce_xy) == 1L)
    stopifnot(!is.na(enforce_xy))
  } else {
      enforce_xy <- get_enforce_xy()
  }

  if ((vs > "1.8.0") && env_absent) {
    if (is.null(OVERRIDE_PROJ_DATUM_WITH_TOWGS84))
      OVERRIDE_PROJ_DATUM_WITH_TOWGS84 <- get_OVERRIDE_PROJ_DATUM_WITH_TOWGS84()
    stopifnot(is.logical(OVERRIDE_PROJ_DATUM_WITH_TOWGS84))
    stopifnot(length(OVERRIDE_PROJ_DATUM_WITH_TOWGS84) == 1)
    if (!OVERRIDE_PROJ_DATUM_WITH_TOWGS84) {
      setCPLConfigOption("OVERRIDE_PROJ_DATUM_WITH_TOWGS84", "NO")
      res <- .Call('RGDAL_GetProjectionRef', dataset, enforce_xy, PACKAGE="rgdal")
      setCPLConfigOption("OVERRIDE_PROJ_DATUM_WITH_TOWGS84", NULL)
    } else {
      res <- .Call('RGDAL_GetProjectionRef', dataset, enforce_xy, PACKAGE="rgdal")
    }
  } else {
    res <- .Call('RGDAL_GetProjectionRef', dataset, enforce_xy, PACKAGE="rgdal")
  }
  no_ellps <- FALSE
  if (!(is.na(res)) && new_proj_and_gdal()) {
    no_towgs84 <- all(nchar(attr(res, "towgs84")) == 0)
    if ((length(grep("towgs84", c(res))) == 0L) && !no_towgs84)
      warning("TOWGS84 discarded")
    no_ellps <- (!is.null(attr(res, "ellps"))) &&
        (nchar(attr(res, "ellps")) > 0L) &&
        (length(grep("ellps", c(res))) == 0L)
    no_ellps <- no_ellps && length(grep("datum", c(res))) == 0L
    if (no_ellps) {
      msg <- paste0("Discarded ellps ", attr(res, "ellps"),
            " in Proj4 definition: ", c(res))
      if (get_rgdal_show_exportToProj4_warnings()) {
       if (!get_thin_PROJ6_warnings()) {
        warning(msg)
        } else {
          if (get("PROJ6_warnings_count",
            envir=.RGDAL_CACHE) == 0L) {
            warning(paste0("PROJ/GDAL PROJ string degradation in workflow\n repeated warnings suppressed\n ", msg))
          }
          assign("PROJ6_warnings_count",
            get("PROJ6_warnings_count",
            envir=.RGDAL_CACHE) + 1L, envir=.RGDAL_CACHE)
          }
         }
    }
# warning("Discarded ellps ", attr(res, "ellps"),
#            " in Proj4 definition: ", c(res))
    if ((!is.null(attr(res, "datum"))) && (nchar(attr(res, "datum")) > 0L)
      && (length(grep("datum", c(res))) == 0L)) {
      msg <- paste0("Discarded datum ", attr(res, "datum"),
          " in Proj4 definition: ", c(res))
      if (!no_towgs84 && (length(grep("towgs84", c(res))) > 0L))
        msg <- paste0(msg, ",\n but +towgs84= values preserved")
      if (get_P6_datum_hard_fail()) stop(msg)
      else {
       if (get_rgdal_show_exportToProj4_warnings()) {
        if (!get_thin_PROJ6_warnings()) {
          warning(msg)
        } else {
          if (get("PROJ6_warnings_count",
            envir=.RGDAL_CACHE) == 0L) {
            warning(paste0("PROJ/GDAL PROJ string degradation in workflow\n repeated warnings suppressed\n ", msg))
          }
          assign("PROJ6_warnings_count",
              get("PROJ6_warnings_count",
              envir=.RGDAL_CACHE) + 1L, envir=.RGDAL_CACHE)
          }
         }
        }
#warning(msg)
    }
    if (new_proj_and_gdal()) wkt2 <- attr(res, "WKT2_2018")
  }
  res <- c(res)
  if (new_proj_and_gdal()) {
    if (no_ellps) res <- showSRID(wkt2, "PROJ")
    comment(res) <- wkt2
  }
  res
}

putRasterData <- function(dataset,
                          rasterData,
                          band = 1,
                          offset = c(0, 0)) {

  assertClass(dataset, 'GDALDataset')

  offset <- rep(offset, length.out = 2)
  
  raster <- getRasterBand(dataset, band)
  
  .Call('RGDAL_PutRasterData', raster, rasterData, 
	as.integer(offset), PACKAGE="rgdal")

}

getRasterTable <- function(dataset,
                           band = NULL,
                           offset = c(0, 0),
                           region.dim = dim(dataset)) {

  assertClass(dataset, 'GDALReadOnlyDataset')

  offset <- rep(offset, length.out = 2)
  region.dim <- rep(region.dim, length.out = 2)

  rasterData <- getRasterData(dataset, band,
                              offset = offset,
                              region.dim = region.dim, list_out=TRUE)

  if (is.null(band)) {

    nbands <- .Call('RGDAL_GetRasterCount', dataset, PACKAGE="rgdal")
    if (nbands < 1) stop("no bands in dataset")
    band <- 1:nbands

  } else {

    nbands <- length(band)

  }

#  dim(rasterData) <- c(region.dim, nbands)

  geoTrans <- getGeoTransFunc(dataset)

  y.i <- 1:region.dim[1] - 0.5 + offset[1]
  x.i <- 1:region.dim[2] - 0.5 + offset[2]

  y.i <- rep(y.i, each = length(x.i))
  x.i <- rep(x.i, len = prod(region.dim))

  out <- geoTrans(x.i, y.i)

#  out <- cbind(out$x, out$y)
  out <- data.frame(x=out$x, y=out$y)
  rasterData <- as.data.frame(rasterData)

#  for (b in band) { 
#    vec <- as.numeric(rasterData[, , b])
#    out <- cbind(out, vec)
#  }

#  out <- as.data.frame(out)
    
#  names(out) <- c('x', 'y', paste('band', 1:nbands, sep = ''))
  out <- cbind(out, rasterData)

  out

}
                           
getRasterData <- function(dataset,
                          band = NULL,
                          offset = c(0, 0),
                          region.dim = dim(dataset),
                          output.dim = region.dim,
                          interleave = c(0, 0),
                          as.is = FALSE, list_out=FALSE) {

    assertClass(dataset, 'GDALReadOnlyDataset')

    offset <- rep(offset, length.out = 2)
    region.dim <- rep(region.dim, length.out = 2)
    output.dim <- rep(output.dim, length.out = 2)
    interleave <- rep(interleave, length.out = 2)

    nbands <- .Call('RGDAL_GetRasterCount', dataset, PACKAGE="rgdal")
    if (nbands < 1) stop("no bands in dataset")

    if (is.null(band)) band <- 1:nbands
  
    x <- array(dim = as.integer(c(rev(output.dim), length(band))))
    for (i in seq(along = band)) {

        raster <- getRasterBand(dataset, band[i])

        y <- .Call('RGDAL_GetRasterData', raster,
                      as.integer(c(offset, region.dim)),
                      as.integer(output.dim),
                      as.integer(interleave),
                      PACKAGE="rgdal")
        
        if (length(band) == 1) {
          # avoid surprisingly expensive slice assignment for
          # common case of a single band
          attributes(y) <- attributes(x)
          x <- y
        } else {
          x[,,i] <- y
        }
  
    }
    if (!as.is) {
        for (i in seq(along = band)) {

            raster <- getRasterBand(dataset, band[i])
            scale <- .Call('RGDAL_GetScale', raster, PACKAGE="rgdal")
            offset <- .Call('RGDAL_GetOffset', raster, PACKAGE="rgdal")

            if (scale != 1) x[,,i] <- x[,,i] * scale
            if (offset != 0) x[,,i] <- x[,,i] + offset
        }
    }
    if (!list_out) {
        if (length(band) == 1L) x <- drop(x)
        return(x)
    } else {
        X <- vector(mode="list", length=length(band))
        names(X) <- paste("band", 1:length(band), sep="")

        for (i in seq(along = band)) {

            X[[i]] <- as.vector(x[,,i])

            if (!as.is) {
  
                raster <- getRasterBand(dataset, band[i])
    
                catNames <- .Call('RGDAL_GetCategoryNames', raster,
                    PACKAGE="rgdal")
  
                if (!is.null(catNames)) {
                    ux <- sort(unique(na.omit(X[[i]])))
                    lCN <- length(catNames)
                    levels <- ((1:lCN)-1)
                    back_incls <- ux %in% levels
                    if (all(back_incls)) {
                        X[[i]] <- factor(X[[i]], levels=levels, labels=catNames)
                        if (!get("silent", envir=.RGDAL_CACHE)) {
                            cat("Input level values and names\n")
                            cat(paste(levels, " ", catNames, "\n", sep=""),
                                sep="")
                        }
                    } else {
                        warning("Assign CategoryNames manually, level/label mismatch")
                    }
                }
            }
        }
        return(X)
    }
}

getCategoryNames <- function(dataset, band = 1) {

  assertClass(dataset, 'GDALReadOnlyDataset')

  raster <- getRasterBand(dataset, band)
  
  catNames <- .Call('RGDAL_GetCategoryNames', raster, PACKAGE="rgdal")

  catNames
}

getBandColorTable <- function(raster) {

  ctab <- .Call('RGDAL_GetColorTable', raster, PACKAGE="rgdal") / 255

  if (length(ctab) == 0L) return(NULL)

  if (.Call('RGDAL_GetColorInterp', raster, PACKAGE="rgdal") == 'Palette')
    switch(.Call('RGDAL_GetPaletteInterp', raster, PACKAGE="rgdal"),  
           RGB = rgb(ctab[,1], ctab[,2], ctab[,3]),
           HSV = hsv(ctab[,1], ctab[,2], ctab[,3]), # Doesn't actually exist
           Gray = gray(ctab[,1]),
           gray(apply(ctab, 2, mean)))
  else
    gray(ctab[,1])

}

getColorTable <- function(dataset, band = 1) {

  assertClass(dataset, 'GDALReadOnlyDataset')

  raster <- getRasterBand(dataset, band)
  
  getBandColorTable(raster)
}

RGB2PCT <- function(x, band, driver.name = 'MEM',
                    ncolors = 256, set.ctab = TRUE) {
  
  assertClass(x, 'GDALReadOnlyDataset')

  if (ncolors < 2 || ncolors > 256)
    stop('Number of colors must be between 2 and 256')

  band <- rep(band, length.out = 3)

  dithered <- new('GDALTransientDataset',
                  new('GDALDriver', driver.name),
                  nrow(x), ncol(x))

  ctab <- .Call('RGDAL_GenCMap',
                getRasterBand(x, band[1]),
                getRasterBand(x, band[2]),
                getRasterBand(x, band[3]),
                getRasterBand(dithered),
                as.integer(ncolors),
                as.logical(set.ctab),
                PACKAGE = "rgdal") / 255

  if (set.ctab)
    dithered
  else
    list(dataset = dithered,
         pct = rgb(ctab[,1], ctab[,2], ctab[,3]))

}  

displayDataset <- function(x, offset = c(0, 0), region.dim = dim(x),
                           reduction = 1, band = 1, col = NULL,
                           reset.par = TRUE, max.dim = 500, ...) {

  assertClass(x, 'GDALReadOnlyDataset')

  offset <- rep(offset, length.out = 2)
  region.dim <- rep(region.dim, length.out = 2)
  reduction <- rep(reduction, length.out = 2)

  offset <- offset %% dim(x)[1:2]
  
  oob <- (region.dim + offset) > dim(x)[1:2]
  
  if (any(oob)) region.dim[oob]  <-  dim(x)[oob] - offset[oob]

  reduction[reduction < 1] <- 1

  plot.dim <- region.dim / reduction
            
  if (any(plot.dim > max.dim))
    plot.dim <- max.dim * plot.dim / max(plot.dim)

  image.data <- getRasterData(x, band[1], offset,
                              region.dim, plot.dim,
                              as.is = TRUE)
#  image.data <- array(image.data[[1]], t(plot.dim))

  if (is.null(col)) {
    
    max.val <- max(image.data, na.rm = TRUE)

    if (!is.finite(max.val)) {
      image.data[] <- 2
      max.val <- 2
    }

    col <- getColorTable(x, band)[1:(max.val + 1)]
      
  }

  if (is.null(col)) col <- gray(seq(0, 1, len = 256))
  
  par.in <- par(no.readonly = TRUE)

  if (reset.par) on.exit(par(par.in))

  par(pin = max(par.in$pin)
      * par.in$fin / max(par.in$fin)
      * rev(plot.dim) / max(plot.dim))
  
  image.data <- image.data[, ncol(image.data):1]

  image.default(image.data + 1, col = col, axes = FALSE, ...)
            
  invisible(list(image.data = image.data, col = col, par.in = par.in))

}

setMethod('initialize', 'GDALRasterBand',
          def =  function(.Object, dataset, band = 1) {
            slot(.Object, 'handle') <- .Call('RGDAL_GetRasterBand',
                                            dataset, as.integer(band), 
					    PACKAGE="rgdal")
            .Object
          })

setMethod('dim', 'GDALRasterBand',
          def = function(x) {
            c(.Call('RGDAL_GetYSize', x, PACKAGE="rgdal"),
              .Call('RGDAL_GetXSize', x, PACKAGE="rgdal"))
          })

getGeoTransFunc <- function(dataset) {

  assertClass(dataset, 'GDALReadOnlyDataset')

  geoTrans <- .Call('RGDAL_GetGeoTransform', dataset, PACKAGE="rgdal")
  if (attr(geoTrans, "CE_Failure")) warning("GeoTransform values not available")
  rotMat <- matrix(geoTrans[c(2, 3, 5, 6)], 2)

  offset <- geoTrans[c(1, 4)]

  function(x, y) {

    x <- cbind(x, y)

    x <- x %*% rotMat

    list(x = x[,1] + offset[1],
         y = x[,2] + offset[2])

  }

}

getRasterBand <- function(dataset, band = 1) {

  assertClass(dataset, 'GDALReadOnlyDataset')

  new('GDALRasterBand', dataset, band)

}

getRasterBlockSize <- function(raster) {

  assertClass(raster, 'GDALRasterBand')
  
  .Call('RGDAL_GetRasterBlockSize', raster, PACKAGE="rgdal")
  
}

get_OVERRIDE_PROJ_DATUM_WITH_TOWGS84 <- function() {
  get("OVERRIDE_PROJ_DATUM_WITH_TOWGS84", envir=.RGDAL_CACHE)
}

set_OVERRIDE_PROJ_DATUM_WITH_TOWGS84 <- function(value) {
        stopifnot(is.logical(value))
        stopifnot(length(value) == 1)
        stopifnot(!is.na(value))
        assign("OVERRIDE_PROJ_DATUM_WITH_TOWGS84", value, envir = .RGDAL_CACHE)
        get_OVERRIDE_PROJ_DATUM_WITH_TOWGS84()
}

getCPLConfigOption <- function(ConfigOption) {
    stopifnot(is.character(ConfigOption))
    stopifnot(length(ConfigOption) == 1)
    .Call("RGDAL_CPLGetConfigOption", ConfigOption, PACKAGE="rgdal")
}

setCPLConfigOption <- function(ConfigOption, value) {
    stopifnot(is.character(ConfigOption))
    stopifnot(length(ConfigOption) == 1)
    if (!is.null(value)) {
        stopifnot(is.character(value))
        stopifnot(length(value) == 1)
    }
    .Call("RGDAL_CPLSetConfigOption", ConfigOption, value, PACKAGE="rgdal")
    .Call("RGDAL_CPLGetConfigOption", ConfigOption, PACKAGE="rgdal")
}

GDAL_iconv <- function() {
    .Call("RGDAL_CPL_RECODE_ICONV", PACKAGE="rgdal")
}

#   if(!exists("Sys.setenv", envir = baseenv())
#     Sys.setenv <- Sys.putenv
#.First.lib <- function(lib, pkg) {
##.onLoad <- function(lib, pkg) {
##  require(methods, quietly = TRUE, warn.conflicts = FALSE)
##  require(sp)
#  .rgdal_old.PROJ_LIB <- Sys.getenv("PROJ_LIB")
#  Sys.setenv("PROJ_LIB"=system.file("proj", package = "rgdal")[1])
#  .rgdal_old.GDAL_DATA <- Sys.getenv("GDAL_DATA")
#  Sys.setenv("GDAL_DATA"=system.file("gdal", package = "rgdal")[1])
#
#  library.dynam('rgdal', pkg, lib)
#
#  .Call('RGDAL_Init', PACKAGE="rgdal")
##
##
#  cat('Geospatial Data Abstraction Library ')
#  cat('extensions to R successfully loaded\n')
#  
#}
#
#.Last.lib <- function(lib, pkg) {
##.onUnload <- function(libpath) {
#    Sys.setenv("PROJ_LIB"=.rgdal_old.PROJ_LIB)
#    Sys.setenv("GDAL_DATA"=.rgdal_old.GDAL_DATA)
#}
#

# Copyright 2003 (c) Barry Rowlingson
# Modified 2006-12 Roger Bivand
###
###
###  Routines for ogr layer data source 
###
###

#
ogrInfo <- function(dsn, layer, encoding=NULL,
  use_iconv=FALSE, swapAxisOrder=FALSE, require_geomType=NULL,
  morphFromESRI=NULL, dumpSRS=FALSE, enforce_xy=NULL, D3_if_2D3D_points=FALSE) {
  if (missing(dsn)) stop("missing dsn")
  stopifnot(is.character(dsn))
  stopifnot(length(dsn) == 1L)
# copy sf::st_read.default usage
  if (length(dsn) == 1 && file.exists(dsn))
    dsn <- enc2utf8(normalizePath(dsn))
  if (nchar(dsn) == 0) stop("empty dsn")
  if (missing(layer)) {
    layers <- ogrListLayers(dsn=dsn)
    if (length(layers) == 0L) stop("missing layer")
    if (length(layers) > 0L) layer <- enc2utf8(c(layers[1]))
    if (length(layers) > 1L)
      warning("First layer ", layer,
        " read; multiple layers present in\n", dsn,
        ", check layers with ogrListLayers()")
    
  } else layer <- enc2utf8(layer)
  if (nchar(layer) == 0) stop("empty name")
  WKB <- c("wkbPoint", "wkbLineString", "wkbPolygon", "wkbMultiPoint",
    "wkbMultiLineString", "wkbMultiPolygon", "wkbGeometryCollection")
  if (!is.null(require_geomType)) {
    stopifnot(is.character(require_geomType) && length(require_geomType)==1)
    m_require_geomType <- match(require_geomType, WKB)
    stopifnot(!is.na(m_require_geomType) || m_require_geomType <= 3)
  }
# a list with various ogr data source information
  
  stopifnot(is.logical(use_iconv))
  stopifnot(length(use_iconv) == 1)
  if (!is.null(encoding)) {
    stopifnot(is.character(encoding))
    stopifnot(length(encoding) == 1)
  }
  if (!use_iconv && !is.null(encoding)) {
    oSE <- getCPLConfigOption("SHAPE_ENCODING")
    tull <- setCPLConfigOption("SHAPE_ENCODING", encoding)
  }
  ogrinfo <- .Call("RGDAL_ogrInfo",as.character(dsn), as.character(layer),
    PACKAGE = "rgdal")
  if (!use_iconv && !is.null(encoding)) {
    tull <- setCPLConfigOption("SHAPE_ENCODING", oSE)
  }

  if (swapAxisOrder) ogrinfo[[5]] <- ogrinfo[[5]][c(2,1,4,3)]

  u_eType <- u_with_z <- null_geometries <- NULL
  deleted_geometries <- NULL
  retain <- NULL
  have_features <- NULL
  all_NULL <- FALSE
  keepGeoms <- NULL

  if (!is.na(ogrinfo[[1]])) {

  fids <- ogrFIDs(dsn=dsn, layer=layer)
  nrows_i <- attr(fids, "i")
  have_features <- nrows_i > 0
  if (have_features && (attr(fids, "i") != attr(fids, "nf"))) {
     retain <- 1:attr(fids, "i")
     afids <- 0:(attr(fids, "nf")-1)
     deleted <- afids[!(afids %in% fids[retain])]
     deleted_geometries <- paste("Deleted feature IDs:", paste(deleted,
        collapse=", "))
     fids <- fids[retain]
  } else {
     deleted_geometries <- NULL
     retain <- NULL
  }
#  attributes(fids) <- NULL
  if (have_features) {
    eTypes <- .Call("R_OGR_types",as.character(dsn), as.character(layer),
      PACKAGE = "rgdal")
    if (is.null(retain)) {
      eType <- eTypes[[4]]
      with_z <- eTypes[[5]]
      isNULL <- as.logical(eTypes[[6]])
   } else {
      eType <- eTypes[[4]][retain]
      with_z <- eTypes[[5]][retain]
      isNULL <- as.logical(eTypes[[6]])[retain]
    }
    null_geometries <- NULL
    if (any(isNULL)) {
      all_NULL <- (sum(isNULL) == length(eType))
      eType <- eType[!isNULL]
      with_z <- with_z[!isNULL]
      null_geometries <- paste("Null geometry IDs:", 
        paste(which(isNULL), collapse=", "))
    }        
    if (!all_NULL) {
      eType[eType == 5L] <- 2L
      eType[eType == 6L] <- 3L

      u_eType <- unique(sort(eType))
      u_with_z <- unique(sort(with_z))
      attr(u_with_z, "elevated") <- FALSE
      if (length(u_with_z) != 1L) {
        if (D3_if_2D3D_points) {
          warning("Accepting mixed 2D and 3D as 3D")
          u_with_z <- 1L
          attr(u_with_z, "elevated") <- TRUE
        } else {
          stop(paste("Multiple # dimensions:", paste((u_with_z + 2),
          collapse=":")))
        }
      }
      if (u_with_z < 0 || u_with_z > 1) stop(
        paste("Invalid # dimensions:", (u_with_z + 2)))

      t_eType <- table(eType)
      if (is.null(require_geomType)) {
        keepGeoms <- NULL
        if (length(u_eType) > 1L) stop(
          paste("Multiple incompatible geometries:", 
            paste(paste(WKB[as.integer(names(t_eType))], t_eType, sep=": "),
            collapse="; ")))
#  if (length(u_eType) == 2L) {
#    if (u_eType[1] == 2 && u_eType[2] == 5) u_eType = 2
#    else if (u_eType[1] == 3 && u_eType[2] == 6) u_eType = 3
#    else stop(paste("Multiple incompatible geometries:", 
#      paste(paste(WKB[as.integer(names(t_eType))], t_eType, sep=": "),
#        collapse="; ")))
#   }
      } else {
        if (!require_geomType %in% WKB[as.integer(names(t_eType))])
          stop(require_geomType, "not in", WKB[as.integer(names(t_eType))])
        u_eType <- match(require_geomType, WKB)
        keepGeoms <- WKB[eType] == require_geomType
        message("NOTE: keeping only ", sum(keepGeoms), " ", require_geomType,
          " of ", length(keepGeoms), " features\n",
          "    note that extent applies to all features")
      }
    } else {
      have_features <- FALSE
      warning("ogrInfo: all features NULL")
    }
  }
  } else {
    ogrinfo[[1]] <- attr(ogrinfo[[1]], "dFIDs")
    warning("ogrInfo: feature count overflow")
  }
  names(ogrinfo) <- c("nrows", "nitems", "iteminfo", "driver", "extent",
    "nListFields")
  if (ogrinfo$driver == "ESRI Shapefile") {
      DSN <- dsn
      if (!file.info(DSN)$isdir) DSN <- dirname(enc2utf8(normalizePath(dsn)))
      DBF_fn <- paste(DSN, .Platform$file.sep, layer, ".dbf", sep = "")
      if (file.exists(DBF_fn)) {
        con <- file(DBF_fn, "rb")
        vr <- readBin(con, "raw", n=32L)
        ldid <- as.integer(vr[30])
        attr(ogrinfo, "LDID") <- ldid
        close(con)
      } else {
        warning("ogrInfo: ", DBF_fn, " not found", sep="")
      }
  }
  names(ogrinfo$iteminfo) <- c("name","type","length","typeName","maxListCount")
  if (use_iconv && !is.null(encoding))
    ogrinfo$iteminfo$name <- iconv(ogrinfo$iteminfo$name, from=encoding)
  ogrinfo$have_features <- have_features
  ogrinfo$eType <- u_eType
  ogrinfo$with_z <- u_with_z
  ogrinfo$null_geometries <- null_geometries
  ogrinfo$deleted_geometries <- deleted_geometries
  ogrinfo$dsn <- dsn
  ogrinfo$layer <- layer
  if (is.null(morphFromESRI)) {
    if (ogrinfo$driver == "ESRI Shapefile") morphFromESRI <- TRUE
    else morphFromESRI <- FALSE
  }
  if (ogrinfo$driver != "ESRI Shapefile" && morphFromESRI)
    morphFromESRI <- FALSE
  p4s0 <- OGRSpatialRef(dsn, layer, morphFromESRI=morphFromESRI,
    dumpSRS=dumpSRS, enforce_xy=enforce_xy)
  if (new_proj_and_gdal()) wkt2 <- comment(p4s0)
  ogrinfo$p4s <- c(p4s0)
  if (new_proj_and_gdal()) ogrinfo$wkt2 <- wkt2
  if (!is.null(require_geomType))
    attr(ogrinfo, "require_geomType") <- require_geomType
  if (!is.null(keepGeoms)) attr(ogrinfo, "keepGeoms") <- keepGeoms
  class(ogrinfo) <- "ogrinfo"
  ogrinfo
}

print.ogrinfo <- function(x, ...) {
  cat("Source: \"", x$dsn, '\", layer: \"', x$layer, "\"", '\n', sep='')
  cat("Driver:", x$driver)
  if (x$have_features) cat("; number of rows:", x$nrows, "\n")
  WKB <- c("wkbPoint", "wkbLineString", "wkbPolygon", "wkbMultiPoint",
    "wkbMultiLineString", "wkbMultiPolygon", "wkbGeometryCollection")
  if (!is.null(attr(x, "require_geomType"))) {
    cat("  selected geometry type:", attr(x, "require_geomType"), "with",
      sum(attr(x, "keepGeoms")), "rows\n")
  }
  if (!x$have_features) cat(", no features found\n")
  if (x$have_features) cat("Feature type:", paste(WKB[x$eType],
    collapse=", "), "with", x$with_z+2, "dimensions\n")
  if (!is.null(x$extent)) cat("Extent: (", x$extent[1], " ",
    x$extent[2], ") - (", x$extent[3], " ", x$extent[4], ")\n", sep="")
  if (!is.null(x$null_geometries)) cat(x$null_geometries, "\n")
  if (!is.null(x$deleted_geometries)) cat(x$deleted_geometries, "\n")
  if ((nchar(x$p4s) > 1) && !is.na(x$p4s)) cat("CRS:", x$p4s, "\n")
  if (!is.null(attr(x, "LDID"))) cat("LDID:", attr(x, "LDID"), "\n")
  cat("Number of fields:", x$nitems, "\n")
  if (is.null(x$nListFields)) x$nListFields <- 0
  if (x$nListFields > 0) cat("Number of list fields:", x$nListFields, "\n")
  if (x$nitems > 0 && x$nListFields == 0) print(as.data.frame(x$iteminfo)[,1:4])
  if (x$nitems > 0 && x$nListFields > 0 && x$have_features)
    print(as.data.frame(x$iteminfo))
  if (x$nitems > 0 && x$nListFields > 0 && !x$have_features)
    print(as.data.frame(x$iteminfo)[,1:4])
  invisible(x)
}


ogrFIDs <- function(dsn, layer){
  if (missing(dsn)) stop("missing dsn")
  stopifnot(is.character(dsn))
  stopifnot(length(dsn) == 1L)
# copy sf::st_read.default usage
  if (length(dsn) == 1 && file.exists(dsn))
    dsn <- enc2utf8(normalizePath(dsn))
  if (nchar(dsn) == 0) stop("empty name")
  if (missing(layer)) stop("missing layer")
  layer <- enc2utf8(layer)
  if (nchar(layer) == 0) stop("empty name")
  fids <- .Call("RGDAL_ogrFIDs",as.character(dsn),as.character(layer), PACKAGE = "rgdal")
  if (attr(fids, "i") == 0L) warning("no features found")
  fids
}

ogrDrivers <- function() {
  if (strsplit(getGDALVersionInfo(), " ")[[1]][2] < "2") {
    res <- .Call("ogr_GetDriverNames", PACKAGE="rgdal")
    res <- as.data.frame(res, stringsAsFactors=FALSE)
  } else {
      res <- .Call('RGDAL_GetDriverNames', PACKAGE="rgdal")
      if (!is.null(attr(res, "isVector"))) res$isVector <- attr(res, "isVector")
      res <- as.data.frame(res, stringsAsFactors=FALSE)
      res <- res[res$isVector,]
      names(res)[3] <- "write"
  }
  res <- res[order(res$name),]
  row.names(res) <- NULL
  res
}

OGRSpatialRef <- function(dsn, layer, morphFromESRI=NULL, dumpSRS=FALSE,
  driver=NULL, enforce_xy=NULL) {
  stopifnot(is.character(dsn))
  stopifnot(length(dsn) == 1L)
  if (is.null(driver)) driver <- attr(ogrListLayers(dsn), "driver")
  if (is.null(morphFromESRI)) {
    if (driver == "ESRI Shapefile") morphFromESRI <- TRUE
    else morphFromESRI <- FALSE
  }
  if (driver != "ESRI Shapefile" && morphFromESRI)
    morphFromESRI <- FALSE
  stopifnot(is.logical(morphFromESRI))
  stopifnot(length(morphFromESRI) == 1L)
# copy sf::st_read.default usage
  if (length(dsn) == 1 && file.exists(dsn))
    dsn <- enc2utf8(normalizePath(dsn))
  layer <- enc2utf8(layer)
  stopifnot(is.logical(dumpSRS))
  stopifnot(length(dumpSRS) == 1)
  stopifnot(!is.na(dumpSRS))
  if (!is.null(enforce_xy)) {
    stopifnot(is.logical(enforce_xy))
    stopifnot(length(enforce_xy) == 1L)
    stopifnot(!is.na(enforce_xy))
  } else {
      enforce_xy <- get_enforce_xy()
  }
  attr(dumpSRS, "enforce_xy") <- enforce_xy
  wkt2 <- NULL
  no_ellps <- FALSE
  res <- .Call("ogrP4S", as.character(dsn), as.character(layer),
        as.logical(morphFromESRI), dumpSRS, PACKAGE="rgdal")
  if (!is.na(res) && new_proj_and_gdal()) {
    no_towgs84 <- ((is.null(attr(res, "towgs84"))) || 
      (all(nchar(attr(res, "towgs84")) == 0)))
    if ((length(grep("towgs84", c(res))) == 0L) && !no_towgs84)
      warning("TOWGS84 discarded")
    no_ellps <- (!is.null(attr(res, "ellps"))) &&
        (nchar(attr(res, "ellps")) > 0L) &&
        (length(grep("ellps", c(res))) == 0L) 
    no_ellps <- no_ellps && length(grep("datum", c(res))) == 0L
    if (no_ellps) {
      msg <- paste0("Discarded ellps ", attr(res, "ellps"),
            " in Proj4 definition: ", c(res))
      if (get_rgdal_show_exportToProj4_warnings()) {
       if (!get_thin_PROJ6_warnings()) {
        warning(msg)
        } else {
          if (get("PROJ6_warnings_count",
            envir=.RGDAL_CACHE) == 0L) {
            warning(paste0("PROJ/GDAL PROJ string degradation in workflow\n repeated warnings suppressed\n ", msg))
          }
          assign("PROJ6_warnings_count",
              get("PROJ6_warnings_count",
              envir=.RGDAL_CACHE) + 1L, envir=.RGDAL_CACHE)
      }
     }
    }
    if ((!is.null(attr(res, "datum"))) && (nchar(attr(res, "datum")) > 0L)
      && (length(grep("datum", c(res))) == 0L)) {
      msg <- paste0("Discarded datum ", attr(res, "datum"),
          " in Proj4 definition: ", c(res))
      if (!no_towgs84 && (length(grep("towgs84", c(res))) > 0L))
        msg <- paste0(msg, ",\n but +towgs84= values preserved")
      if (get_P6_datum_hard_fail()) stop(msg)
      else {
      if (get_rgdal_show_exportToProj4_warnings()) {
        if (!get_thin_PROJ6_warnings()) {
          warning(msg)
        } else {
          if (get("PROJ6_warnings_count",
            envir=.RGDAL_CACHE) == 0L) {
            warning(paste0("PROJ/GDAL PROJ string degradation in workflow\n repeated warnings suppressed\n ", msg))
          }
          assign("PROJ6_warnings_count",
              get("PROJ6_warnings_count",
              envir=.RGDAL_CACHE) + 1L, envir=.RGDAL_CACHE)
          }
         }
        }
#warning(msg)
    }
    if (new_proj_and_gdal()) wkt2 <- attr(res, "WKT2_2018")
  }
  res <- c(res)
  if (new_proj_and_gdal()) {
    if (no_ellps) res <- showSRID(wkt2, "PROJ")
    comment(res) <- wkt2
  }
  res
}

ogrListLayers <- function(dsn) {
  if (missing(dsn)) stop("missing dsn")
  stopifnot(is.character(dsn))
  stopifnot(length(dsn) == 1)
# copy sf::st_read.default usage
  if (length(dsn) == 1 && file.exists(dsn))
    dsn <- enc2utf8(normalizePath(dsn))
  if (nchar(dsn) == 0) stop("empty name")
  if (!is.null(attr(dsn, "debug"))) {
    stopifnot(is.logical(attr(dsn, "debug")))
    stopifnot(length(attr(dsn, "debug")) == 1)
  } else {
    attr(dsn, "debug") <- FALSE
  }
  layers <- .Call("RGDAL_ogrListLayers", dsn, PACKAGE = "rgdal")
  n <- length(layers)
  tmp <- layers[n]
  layers <- layers[-n]
  Encoding(layers) <- "UTF-8"
  layers <- enc2native(layers)
  attr(layers, "driver") <- tmp
  attr(layers, "nlayers") <- (n-1)
  layers
}

# Copyright (c) 2003-19 by Barry Rowlingson and Roger Bivand

#.valid.CRSobj <- function(object) {
#	if (exists("is.R") && is.function(is.R) && is.R()) {
#		projargNA <- is.na(object@projargs)
#	} else {
#		projargNA <- is.na(as.numeric(object@projargs))
#	}
#	if (!projargNA) {
#		res <- .Call("checkCRSArgs", object@projargs, 
#				PACKAGE="rgdal")
#	} else res <- list(TRUE, as.character(NA))
#	if (!res[[1]]) {
#	    	return(res[[2]])
#	} else {
#		return(res[[1]])
#	}
#
#}
#
#setValidity("CRS", .valid.CRSobj)


"CRSargs" <- function(object) {
	if (!is(object, "CRS")) stop("not a CRS object")

	if (!is.na(object@projargs)) {
            if (new_proj_and_gdal()) 
                res <- (checkCRSArgs_ng(object@projargs)[[2]])
            else res <- (checkCRSArgs(object@projargs)[[2]])
            res <- paste(unique(unlist(strsplit(res, " "))), 
		collapse=" ")
            return(sub("^\\s+", "", res))
	} else return(as.character(NA))
}

checkCRSArgs <- function(uprojargs=NA_character_) {
  if (packageVersion("rgdal") >= "1.5.1" && length(grep("ob_tran", uprojargs)) > 0L) {
    return(list(TRUE, uprojargs))
  }
# pkgdown work-around
  if (is.na(get("has_proj_def.dat", envir=.RGDAL_CACHE))) {
    assign("has_proj_def.dat", .Call("PROJ4_proj_def_dat_Installed",
      PACKAGE="rgdal"), envir=.RGDAL_CACHE)
  }
# RSB Web Mercator bug 180313 (for 5.0.0)
  drop_nadgrids <- FALSE
  if (strsplit(strsplit(getPROJ4VersionInfo(), ",")[[1]][1], " ")[[1]][2]
    == "5.0.0") {
    if (length(grep("+init=epsg:3857", uprojargs)) > 0L) drop_nadgrids <- TRUE
  }
# RSB 2015-05-21
# fix for omission of proj_defs.dat in PROJ.4 4.9.1
  if ((!get("has_proj_def.dat", envir=.RGDAL_CACHE)) && (!PROJis6ormore())) {
      message("NOTE: rgdal::checkCRSArgs: no proj_defs.dat in PROJ.4 shared files")
      uprojargs <- proj_def_bug_fix(uprojargs)
  }
  res <- .Call("RGDAL_checkCRSArgs", uprojargs, PACKAGE="rgdal")
  if (drop_nadgrids) {
    uuproj <- strsplit(res[[2]], " ")[[1]]
    hit_nad <- grep("nadgrids", uuproj)
    hit_init <- grep("init", uuproj)
    if (length(c(hit_nad, hit_init) > 0))
      res[[2]] <- paste(uuproj[-c(hit_nad, hit_init)], collapse=" ")
  }
  res[[2]] <- sub("^\\s+", "", res[[2]])
# fix for pj_get_def() +no_uoff/+no_off bug
  no_uoff <- length(grep("+no_uoff", uprojargs, fixed=TRUE) > 0)
  no_off <- length(grep("+no_off", uprojargs, fixed=TRUE) > 0)
  if (no_uoff) {
      if( length(grep("+no_uoff", res[[2]], fixed=TRUE)) == 0) 
          res[[2]] <- sub("+no_defs", "+no_uoff +no_defs", res[[2]], fixed=TRUE)
  }
  if (no_off) {
      if (length(grep("+no_off", res[[2]], fixed=TRUE)) == 0)
          res[[2]] <- sub("+no_defs", "+no_off +no_defs", res[[2]], fixed=TRUE)
  }
  res
}

checkCRSArgs_ng <- function(uprojargs=NA_character_, SRS_string=NULL,
  get_source_if_boundcrs=TRUE) {
  no_SRS <- is.null(SRS_string)
  no_PROJ <- is.na(uprojargs)
  prefer_proj <- get_prefer_proj()
  res <- vector(mode="list", length=3L)
  res[[1]] <- FALSE
  res[[2]] <- NA_character_
  res[[3]] <- NA_character_
  if (!no_SRS) {
    stopifnot(is.character(SRS_string))
    stopifnot(length(SRS_string) == 1L)
  }
  if (!no_PROJ) {
    stopifnot(is.character(uprojargs))
    stopifnot(length(uprojargs) == 1L)
    if (grepl("\\+init\\=", uprojargs)) prefer_proj <- FALSE
  }
  if (!no_SRS) {
    uprojargs1 <- try(showSRID(SRS_string, format="PROJ", multiline="NO",
        prefer_proj=prefer_proj), silent=TRUE)
    if (inherits(uprojargs1, "try-error")) {
      res[[1]] <- FALSE
      res[[2]] <- NA_character_
    } else {
      res[[1]] <- TRUE
      res[[2]] <- gsub(" \\+type\\=crs", "", uprojargs1)
    }
    wkt2 <- try(showSRID(SRS_string, format="WKT2", multiline="YES",
        prefer_proj=prefer_proj), silent=TRUE)
    if (!inherits(wkt2, "try-error")) {
      if (get_enforce_xy()) wkt2 <- try(.Call("proj_vis_order", wkt2,
        PACKAGE="rgdal"), silent=TRUE)
      if (!inherits(wkt2, "try-error")) {
        res[[3]] <- wkt2
      }
    }
  } else if (!no_PROJ) {
    uprojargs <- sub("^\\s+", "", uprojargs)
    if (prefer_proj && !grepl("\\+type\\=crs", uprojargs))
      uprojargs <- paste0(uprojargs, " +type=crs") 
    uprojargs1 <- try(showSRID(uprojargs, format="PROJ", multiline="NO",
        prefer_proj=prefer_proj), silent=TRUE)
    if (inherits(uprojargs1, "try-error")) {
      res[[1]] <- FALSE
      res[[2]] <- NA_character_
    } else {
      res[[1]] <- TRUE
      res[[2]] <- gsub(" \\+type\\=crs", "", uprojargs1)
    }
    wkt2 <- try(showSRID(uprojargs, format="WKT2", multiline="YES",
        prefer_proj=prefer_proj), silent=TRUE)
    if (!inherits(wkt2, "try-error")) {
      if (get_source_if_boundcrs) {
        if (length(grep("^BOUNDCRS", wkt2)) > 0L) {
          wkt2a <- try(.Call("get_source_crs", wkt2, PACKAGE="rgdal"),
            silent=TRUE)
          if (!inherits(wkt2a, "try-error")) wkt2 <- wkt2a
        }
      }
      if (get_enforce_xy()) {
        wkt2a <- try(.Call("proj_vis_order", wkt2, PACKAGE="rgdal"),
          silent=TRUE)
          if (!inherits(wkt2a, "try-error")) {
            wkt2 <- wkt2a
          }
      }
      res[[3]] <- wkt2
    }
  }
  res
}

compare_CRS <- function(CRS1, CRS2) {
    stopifnot(new_proj_and_gdal())
    stopifnot(inherits(CRS1, "CRS"))
    stopifnot(inherits(CRS2, "CRS"))
    type1 <- FALSE
    if (is.null(comment(CRS1))) {
        from_args <- paste0(slot(CRS1, "projargs"), " +type=crs")
        warning("NULL source CRS comment, falling back to PROJ string")
        if (is.na(from_args)) 
	    stop("No transformation possible from NA source CRS")
        if (length(grep("\\+init\\=", from_args)) > 0) {
            warning("+init dropped in PROJ string")
            strres <- unlist(strsplit(from_args, " "))
            from_args <- paste(strres[-grep("\\+init\\=", strres)],
                collapse=" ")
        }
    } else {
       type1 <- TRUE
       from_args <- comment(CRS1)
    }
    type2 <- FALSE
    if (is.null(comment(CRS2))) {
        warning("NULL target CRS comment, falling back to PROJ string")
        to_args <- paste0(slot(CRS2, "projargs"), " +type=crs")
        if (is.na(to_args)) 
	    stop("No transformation possible to NA target CRS")
        if (length(grep("\\+init\\=", to_args)) > 0) {
            warning("+init dropped in PROJ string")
            strres <- unlist(strsplit(to_args, " "))
            to_args <- paste(strres[-grep("\\+init\\=", strres)], collapse=" ")
        }
    } else {
        type2 <- TRUE
        to_args <- comment(CRS2)
    }
    res0 <- .Call("CRS_compare", as.character(from_args), as.character(to_args),
        as.logical(type1), as.logical(type2), PACKAGE="rgdal")
    res <- as.logical(res0)
    names(res) <- c("strict", "equivalent", "equivalent_except_axis_order")
    res
}

proj_def_bug_fix <- function(uprojargs) {
    if (length(grep("no_defs", uprojargs)) == 0L && 
# corrected 20150904
        length(grep("init", uprojargs)) == 0L) {
        if (length(grep("ellps", uprojargs)) == 0L && 
# corrected 20150905
            length(grep("datum", uprojargs)) == 0L) {
            tags <- sapply(strsplit(strsplit("+proj=longlat +no_defs",
                "\\+")[[1]], "="), "[", 1)
# based on proj/src/pj_init.c lines 191-197
            if (!any(c("datum", "ellps", "a", "b", "rf", "f") %in% tags)) {
                uprojargs <- paste(uprojargs, "+ellps=WGS84", sep=" ")
           }
         }
    }
    uprojargs
}

llgridlines = function(obj, easts, norths, ndiscr = 20, 
		lty = 2, offset=0.5, side="WS", 
                llcrs = "+proj=longlat +datum=WGS84",
		plotLines = TRUE, plotLabels = TRUE, ...) {
	obj_ll <- spTransform(obj, CRS(llcrs))
	if (missing(easts))
		easts = pretty(bbox(obj_ll)[1,])
	if (missing(norths))
		norths = pretty(bbox(obj_ll)[2,])
	grd <- gridlines(obj_ll, easts = easts, norths = norths, ndiscr = ndiscr)
	grd_x <- spTransform(grd, CRS(proj4string(obj)))
	if (plotLines)
		plot(grd_x, add = TRUE, lty = lty, ...)
        if (packageVersion("sp") >= "0.9.84") {
	    grdat_ll <- gridat(obj_ll, easts = easts, norths = norths,
                side=side, offset = offset)
        } else {
	    grdat_ll <- gridat(obj_ll, easts = easts, norths = norths, offset = offset)
        }
	grdat_x <- spTransform(grdat_ll, CRS(proj4string(obj)))
	if (plotLabels)
		text(coordinates(grdat_x), labels=parse(text=grdat_x$labels),
  			pos=grdat_x$pos, offset=grdat_x$offset, ...)
}

# Copyright 2006-2021 Roger Bivand

readOGR <- function(dsn, layer, verbose=TRUE, p4s=NULL, 
        stringsAsFactors=as.logical(NA),
        drop_unsupported_fields=FALSE,
	pointDropZ=FALSE, dropNULLGeometries=TRUE, useC=TRUE,
        disambiguateFIDs=FALSE, addCommentsToPolygons=TRUE, encoding=NULL,
        use_iconv=FALSE, swapAxisOrder=FALSE, require_geomType=NULL,
        integer64="no.loss", GDAL1_integer64_policy=FALSE,
        morphFromESRI=NULL, dumpSRS=FALSE, enforce_xy=NULL,
        D3_if_2D3D_points=FALSE, missing_3D=0) {
	if (missing(dsn)) stop("missing dsn")
        stopifnot(is.character(dsn))
        stopifnot(length(dsn) == 1L)
# copy sf::st_read.default usage
	if (length(dsn) == 1 && file.exists(dsn))
          dsn <- enc2utf8(normalizePath(dsn))
	if (nchar(dsn) == 0) stop("empty name")
	if (missing(layer)){
          layers <- ogrListLayers(dsn=dsn)
          if (length(layers) == 0L) stop("missing layer")
          if (length(layers) > 0L) layer <- c(layers[1])
          if (length(layers) > 1L)
            warning("First layer ", layer,
              " read; multiple layers present in\n", dsn,
              ", check layers with ogrListLayers()")
        }  else layer <- enc2utf8(layer)
# stop("missing dsn")
#	if (missing(layer)) stop("missing layer")
	if (nchar(layer) == 0) stop("empty name")
        integer64 <- match.arg(integer64,
          c("allow.loss", "warn.loss", "no.loss"))
        
        int64 <- switch(integer64,
          "allow.loss"=1L,
          "warn.loss"=2L, 
          "no.loss"=3L)
        if (GDAL1_integer64_policy) int64 <- 4L
# adding argument for SHAPE_ENCODING environment variable 121124
        stopifnot(is.logical(use_iconv))
        stopifnot(length(use_iconv) == 1)
        if (!is.null(encoding)) {
            stopifnot(is.character(encoding))
            stopifnot(length(encoding) == 1)
        }
        WKB <- c("wkbPoint", "wkbLineString", "wkbPolygon", "wkbMultiPoint",
          "wkbMultiLineString", "wkbMultiPolygon", "wkbGeometryCollection")
        if (!is.null(require_geomType)) {
          stopifnot(is.character(require_geomType) &&
            length(require_geomType)==1)
          m_require_geomType <- match(require_geomType, WKB)
          stopifnot(!is.na(m_require_geomType) || m_require_geomType <= 3)
        }
        if(is.na(stringsAsFactors)) {
            stringsAsFactors <-
                if(getRversion() < "4.1.0")
                    default.stringsAsFactors()
                else
                    FALSE
        }
        
	suppressMessages(ogr_info <- ogrInfo(dsn=dsn, layer=layer,
            encoding=encoding, use_iconv=use_iconv,
            swapAxisOrder=swapAxisOrder, require_geomType=require_geomType,
            morphFromESRI=morphFromESRI, dumpSRS=dumpSRS,
            D3_if_2D3D_points=D3_if_2D3D_points))
        HAS_FEATURES <- TRUE
        if (!ogr_info$have_features) {
            if (dropNULLGeometries) {
                stop("no features found")
            } else {
                warning("no features found; proceeding to atttributes only")
                HAS_FEATURES <- FALSE
            }
        }
        if (is.null(ogr_info$nListFields)) nListFields <- 0
        else nListFields <- ogr_info$nListFields
# 121130 RSB trap no field case (from PostGIS, Mathieu Basille)
        if (ogr_info$nitems > 0) {
          nodata_flag <- FALSE
          if (strsplit(getGDALVersionInfo(), " ")[[1]][2] < "2") {
            keep <- ogr_info$iteminfo$typeName %in% c("Integer", "Real",
              "String", "Date", "Time", "DateTime", "IntegerList",
              "RealList", "StringList")
          } else {
            keep <- ogr_info$iteminfo$typeName %in% c("Integer", "Real",
              "String", "Date", "Time", "DateTime", "IntegerList",
              "RealList", "StringList", "Integer64", "Integer64List")
          }
          if (nListFields > 0)
              ListFields <- as.integer(ogr_info$iteminfo$maxListCount)
          if (drop_unsupported_fields) {
             iflds <- as.integer((1:ogr_info$nitems)-1)
             iflds <- iflds[keep]
             fldnms <- ogr_info$iteminfo$name[keep]
             if (nListFields > 0) ListFields <- ListFields[keep]
             if (any(!keep)) warning(paste("Fields dropped:", 
                 paste(ogr_info$iteminfo$name[!keep], collapse=" ")))
          } else {
             if (any(!keep)) stop(paste("Unsupported field type:", 
                 paste(ogr_info$iteminfo$typeName[!keep], collapse=" ")))
             iflds <- as.integer((1:ogr_info$nitems)-1)
             fldnms <- ogr_info$iteminfo$name
          }
          int64_found <- grep("Integer64", ogr_info$iteminfo$typeName)
        } else {
          int64_found <- integer(0L)
          nodata_flag <- TRUE
          iflds <- integer(0)
        }
	fids <- ogrFIDs(dsn=dsn, layer=layer)
        if (attr(fids, "i") != attr(fids, "nf")) {
            retain <- 1:attr(fids, "i")
            afids <- 0:(attr(fids, "nf")-1)
            deleted <- afids[!(afids %in% fids[retain])]
            warning(paste("Deleted feature IDs: ", paste(deleted,
                    collapse=", ")))
            fids <- fids[retain]
        } else {
            retain <- NULL
        }
        attributes(fids) <- NULL
# suggestion by Dylan Beaudette 110620
        non_unique_fids <- max(table(fids)) > 1
        if (non_unique_fids) {
            if (disambiguateFIDs) {
                fids <- seq_along(fids) # if present, make new FIDs
            } else {
                stop("FIDs not unique")
            }
        }
	if (verbose) {
		cat("OGR data source with driver:", ogr_info$driver, "\n")
		cat("Source: \"", dsn, '\", layer: \"', layer, "\"", '\n',
			sep='')
		cat("with", length(fids), "features")
                if (!is.null(attr(ogr_info, "require_geomType")))
                    cat(";\nSelected", attr(ogr_info, "require_geomType"),
                      "feature type, with", sum(attr(ogr_info, "keepGeoms")),
                      "rows")
                cat("\n")
		cat("It has", length(iflds), "fields")
                if (nListFields > 0)
                  cat(", of which", nListFields, "list fields")
                cat("\n")
                if (length(int64_found > 0L)) {
                    if (GDAL1_integer64_policy) {
                        cat("Integer64 fields read as doubles: ",
                            paste(fldnms[int64_found], collapse=" "), "\n")
                    } else {
                        if (integer64 == "no.loss") {
                            cat("Integer64 fields read as strings: ",
                                paste(fldnms[int64_found], collapse=" "), "\n")
                        } else {
                            cat("Integer64 fields read as signed 32-bit integers: ",
                                paste(fldnms[int64_found], collapse=" "), "\n")
                        }
                    }
            }

	}

# adding argument for SHAPE_ENCODING environment variable 121124
        if (!use_iconv && !is.null(encoding) && 
            ogr_info$driver == "ESRI Shapefile") {
            oSE <- getCPLConfigOption("SHAPE_ENCODING")
            tull <- setCPLConfigOption("SHAPE_ENCODING", encoding)
        }
	if (nodata_flag) {
            dlist <- list(FID=as.integer(fids))
        } else {
            attr(iflds, "nListFields") <- as.integer(nListFields)
            nflds <- length(iflds)
            if (nListFields > 0) {
                attr(iflds, "ListFields") <- ListFields
                nflds <- nflds + sum(ListFields) - nListFields
                fldnms1 <- NULL
                for (i in seq(along=ListFields)) {
                    if (ListFields[i] == 0) fldnms1 <- c(fldnms1, fldnms[i])
                    else fldnms1 <- c(fldnms1,
                        paste(fldnms[i], 1:ListFields[i], sep=""))
                }
                stopifnot(length(fldnms1) == nflds)
                fldnms <- fldnms1
            }
            attr(iflds, "nflds") <- as.integer(nflds)
            attr(iflds, "int64") <- as.integer(int64)
            dlist <- .Call("ogrDataFrame", as.character(dsn),
                enc2utf8(as.character(layer)), as.integer(fids), iflds, PACKAGE="rgdal")
	    names(dlist) <- make.names(fldnms ,unique=TRUE)

            if (use_iconv && !is.null(encoding)) {
                for (i in seq(along=dlist)) {
                    if (is.character(dlist[[i]])) {
                       dlist[[i]] <- iconv(dlist[[i]], from=encoding)
                    }
                }
            }
        }
        if (!use_iconv && !is.null(encoding) && 
            ogr_info$driver == "ESRI Shapefile") {
            tull <- setCPLConfigOption("SHAPE_ENCODING", oSE)
        }
	data <- data.frame(dlist, row.names=fids,
            stringsAsFactors=stringsAsFactors)
        rm(dlist)
        gc(verbose = FALSE)

        if (!HAS_FEATURES) {
            return(data)
        }

# suggestion by Paul Hiemstra 070817
#        if (is.null(morphFromESRI)) {
#            if (ogr_info$driver == "ESRI Shapefile") morphFromESRI <- TRUE
#            else morphFromESRI <- FALSE
#        }
#        if (ogr_info$driver != "ESRI Shapefile" && morphFromESRI)
#            morphFromESRI <- FALSE
#        stopifnot(is.logical(morphFromESRI))
#        stopifnot(length(morphFromESRI) == 1)
#        stopifnot(is.logical(dumpSRS))
#        stopifnot(length(dumpSRS) == 1)
#	prj <- .Call("ogrP4S", as.character(dsn), enc2utf8(as.character(layer)),		as.logical(morphFromESRI), as.logical(dumpSRS), PACKAGE="rgdal")

#        prj <- OGRSpatialRef(dsn=dsn, layer=layer, morphFromESRI=morphFromESRI,
#          dumpSRS=dumpSRS, driver=ogr_info$driver, enforce_xy=enforce_xy)

        prj <- ogr_info$p4s

	if (!is.null(p4s)) {
          if (!is.na(prj)) {
            warning("p4s= argument given as: ", p4s, "\n and read as: ",
              c(prj), "\n read string overridden by given p4s= argument value")
          }
        } else {
          p4s <- prj
        }

	if (!is.na(p4s) && nchar(p4s) == 0) p4s <- as.character(NA)
#        if (new_proj_and_gdal()) wkt2 <- ogr_info$wkt2
        oCRS <- new("CRS", projargs=p4s)
        if (new_proj_and_gdal()) comment(oCRS) <- ogr_info$wkt2

	geometry <- .Call("R_OGR_CAPI_features", as.character(dsn), 
		enc2utf8(as.character(layer)), comments=addCommentsToPolygons,
                PACKAGE="rgdal")
	if (is.null(retain)) {
	    eType <- geometry[[4]]
	    with_z <- geometry[[6]]
            isNULL <- as.logical(geometry[[7]])
	    gFeatures <- geometry[[5]]
            gComments <- geometry[[8]]
        } else {
	    eType <- geometry[[4]][retain]
	    with_z <- geometry[[6]][retain]
            isNULL <- as.logical(geometry[[7]])[retain]
	    gFeatures <- geometry[[5]][retain]
            gComments <- geometry[[8]][retain]
        }
        rm(geometry);
        gc(verbose = FALSE)
        if (any(isNULL)) {
            eType <- eType[!isNULL]
            with_z <- with_z[!isNULL]
        }   
        elevated <- attr(ogr_info$with_z, "elevated")
	u_with_z <- unique(sort(with_z))
	if (length(u_with_z) != 1L) {
          if (elevated) {
            if (pointDropZ) {
              warning("Dropping mixed third dimension")
              u_with_z <- 0L
            } else {
              warning("Setting any missing third dimension to ", missing_3D)
              u_with_z <- 1L
            }
          } else {
            stop(paste("Multiple # dimensions:", paste((u_with_z + 2), 
              collapse=":")))
          }
        }
	if (u_with_z < 0 || u_with_z > 1) stop(
		paste("Invalid # dimensions:", (u_with_z + 2)))

        eType[eType == 5L] <- 2L
        eType[eType == 6L] <- 3L

	u_eType <- unique(sort(eType))

    t_eType <- table(eType)
    if (is.null(require_geomType)) {
      keepGeoms <- NULL
      if (length(u_eType) > 1L) stop(
        paste("Multiple incompatible geometries:", 
          paste(paste(WKB[as.integer(names(t_eType))], t_eType, sep=": "),
          collapse="; ")))
#  if (length(u_eType) == 2L) {
#    if (u_eType[1] == 2 && u_eType[2] == 5) u_eType = 2
#    else if (u_eType[1] == 3 && u_eType[2] == 6) u_eType = 3
#    else stop(paste("Multiple incompatible geometries:", 
#      paste(paste(WKB[as.integer(names(t_eType))], t_eType, sep=": "),
#        collapse="; ")))
#   }
    } else {
      if (!require_geomType %in% WKB[as.integer(names(t_eType))])
        stop(require_geomType, "not in", WKB[as.integer(names(t_eType))])
      u_eType <- match(require_geomType, WKB)
      keepGeoms <- WKB[eType] == require_geomType
      message("NOTE: keeping only ", sum(keepGeoms), " ", require_geomType,
        " of ", length(keepGeoms), " features\n")
    }

	if (length(gFeatures) != length(fids)) stop("Feature mismatch")

        if (any(isNULL)) {
            if (dropNULLGeometries) {
                warning(paste("Dropping null geometries:", paste(which(isNULL),
                    collapse=", ")))
                gFeatures <- gFeatures[!isNULL]
	        data <- data[!isNULL, , drop=FALSE]
                fids <- fids[!isNULL]
                gComments <- gComments[!isNULL]
            } else {
                warning(paste("Null geometries found:", paste(which(isNULL),
                    collapse=", ")))
                warning("dropNULLGeometries FALSE, returning only data for null-geometry features")
                return(data[isNULL, , drop=FALSE])
            }
        }

        if (!is.null(require_geomType)) {
                gFeatures <- gFeatures[keepGeoms]
	        data <- data[keepGeoms, , drop=FALSE]
                fids <- fids[keepGeoms]
                gComments <- gComments[keepGeoms]
        }

	if (u_eType == 1) { # points
		if (u_with_z == 0 || pointDropZ) {
                    if (swapAxisOrder) {
			coords <- do.call("rbind", lapply(gFeatures, 
				function(x) c(x[[1]][[2]], x[[1]][[1]])))
                    } else {
			coords <- do.call("rbind", lapply(gFeatures, 
				function(x) c(x[[1]][[1]], x[[1]][[2]])))
                    }
		} else {
                  if (elevated) {
                    if (swapAxisOrder) {
			coords <- do.call("rbind", lapply(gFeatures, 
			    function(x) c(x[[1]][[2]], x[[1]][[1]], 
                            ifelse(length(x[[1]]) == 2L, missing_3D, x[[1]][[3]]))))
                    } else {
			coords <- do.call("rbind", lapply(gFeatures, 
			    function(x) c(x[[1]][[1]], x[[1]][[2]],
                            ifelse(length(x[[1]]) == 2L, missing_3D, x[[1]][[3]]))))
                    } 
                  } else {
                    if (swapAxisOrder) {
			coords <- do.call("rbind", lapply(gFeatures, 
			    function(x) c(x[[1]][[2]], x[[1]][[1]],
                            ifelse(length(x[[1]]) == 2L, missing_3D, x[[1]][[3]]))))
                    } else {
			coords <- do.call("rbind", lapply(gFeatures, 
			    function(x) c(x[[1]][[1]], x[[1]][[2]], 
                            ifelse(length(x[[1]]) == 2L, missing_3D, x[[1]][[3]]))))
                    } 
                  } 
		}
#		data <- data.frame(dlist)
		row.names(data) <- NULL
		res <- SpatialPointsDataFrame(coords=coords, data=data,
			proj4string=oCRS)
	} else if (u_eType == 2) { # lines
		if (u_with_z != 0) warning("Z-dimension discarded")
		n <- length(gFeatures)
		lnList <- vector(mode="list", length=n)
		for (i in 1:n) {
			iG <- gFeatures[[i]]
			m <- length(iG)
			lnlist <- vector(mode="list", length=m)
			for (j in 1:m) {
				jG <- iG[[j]]
                                if (swapAxisOrder) {
				  lnlist[[j]] <- Line(cbind(jG[[2]], jG[[1]]))
                                } else {
				  lnlist[[j]] <- Line(cbind(jG[[1]], jG[[2]]))
                                }
			}
			lnList[[i]] <- Lines(lnlist, ID=as.character(fids[i]))
		}
		SL <- SpatialLines(lnList, proj4string=oCRS)
#		data <- data.frame(dlist, row.names=fids)
		res <- SpatialLinesDataFrame(SL, data)
	} else if (u_eType == 3) { # polygons
            if (u_with_z != 0) warning("Z-dimension discarded")
            if (useC) {
#                plList <- .Call("make_polygonslist", gFeatures,
#                    as.list(as.character(fids)), PACKAGE="rgdal")
		n <- length(gFeatures)
		plList <- vector(mode="list", length=n)
		for (i in 1:n) {
			iG <- gFeatures[[i]]
                        if (swapAxisOrder) {
                          iG <- lapply(iG, function(x) {
                            tmp <- x[[1]]; x[[1]] <- x[[2]]; x[[2]] <- tmp
                          })
                        }
                        if (addCommentsToPolygons) {
                            thisPL <- Polygons(.Call("make_Polygonlist",
                                iG, gComments[[i]], PACKAGE="rgdal"),
                                ID=as.character(fids[i]))
                            comment(thisPL) <- paste(gComments[[i]],
                                collapse=" ")
                        } else {
                            thisPL <- Polygons(.Call("make_Polygonlist",
                                iG, NULL, PACKAGE="rgdal"),
                                ID=as.character(fids[i]))
                        }
			plList[[i]] <- thisPL
                    }
            } else {
		n <- length(gFeatures)
		plList <- vector(mode="list", length=n)
		for (i in 1:n) {
			iG <- gFeatures[[i]]
                        if (swapAxisOrder) {
                          iG <- lapply(iG, function(x) {
                            tmp <- x[[1]]; x[[1]] <- x[[2]]; x[[2]] <- tmp
                          })
                        }
			m <- length(iG)
			pllist <- vector(mode="list", length=m)
			for (j in 1:m) {
				jG <- iG[[j]]
				cmat <- cbind(jG[[1]], jG[[2]])
				if (!identical(cmat[1,], cmat[nrow(cmat),])) {
				  cmat <- rbind(cmat, cmat[1,])
                                  warning(paste("Ring closed in Polygons",
				    i, "Polygon", j))
				}
				t0 <- try(pllist[[j]] <- Polygon(cmat),
                                    silent=TRUE)
                                if (inherits(t0, "try-error")) {
                                     print(cmat)
                                     print(t0)
                                     stop("i: ", i, ", j: ", j,
                                       ", Polygon error exit")
                                }
			}
                        thisPL <- Polygons(pllist, ID=as.character(fids[i]))
                        if (addCommentsToPolygons) {
                            comment(thisPL) <- paste(gComments[[i]],
                                collapse=" ")
                            if (!isTRUE(all.equal(as.logical(gComments[[i]]),
                              sapply(slot(thisPL, "Polygons"), slot, "hole"))))
                              warning("comment/hole mismatch, geometry:", i)
                        }
			plList[[i]] <- thisPL
                    }
		}
                rm(gFeatures)
                gc(verbose = FALSE)
		SP <- SpatialPolygons(plList, proj4string=oCRS)
                rm(plList)
                gc(verbose = FALSE)
#		data <- data.frame(dlist, row.names=fids)
		res <- SpatialPolygonsDataFrame(SP, data, match.ID=FALSE)
	} else stop(paste("Incompatible geometry:", u_eType))

	res
}

showWKT <- function(p4s, file=NULL, morphToESRI=FALSE, enforce_xy=NULL) {

	if (!is.character(p4s)) stop("invalid p4s object")
        stopifnot(length(p4s) == 1)
	if (!is.logical(morphToESRI)) stop("invalid morphToESRI object")
        stopifnot(length(morphToESRI) == 1)
        stopifnot(!is.na(morphToESRI))
        if (!is.null(enforce_xy)) {
            stopifnot(is.logical(enforce_xy))
            stopifnot(length(enforce_xy) == 1L)
            stopifnot(!is.na(enforce_xy))
        } else {
            enforce_xy <- get_enforce_xy()
        }
        attr(morphToESRI, "enforce_xy") <- enforce_xy
	res <- .Call("p4s_to_wkt", as.character(p4s), morphToESRI, 
		PACKAGE="rgdal")
	if (!is.null(file)) cat(res, "\n", sep="", file=file)
	res
}

showP4 <- function(wkt, morphFromESRI=FALSE, enforce_xy=NULL) {

	if (!is.character(wkt)) stop("invalid wkt object")
        stopifnot(length(wkt) == 1)
	if (!is.logical(morphFromESRI)) stop("invalid morphFromESRI object")
        stopifnot(length(morphFromESRI) == 1)
        stopifnot(!is.na(morphFromESRI))
        if (!is.null(enforce_xy)) {
            stopifnot(is.logical(enforce_xy))
            stopifnot(length(enforce_xy) == 1L)
            stopifnot(!is.na(enforce_xy))
        } else {
            enforce_xy <- get_enforce_xy()
        }
        attr(morphFromESRI, "enforce_xy") <- enforce_xy
	res <- .Call("wkt_to_p4s", as.character(wkt),
                morphFromESRI, PACKAGE="rgdal")
	res
}


showEPSG <- function(p4s, enforce_xy=NULL) {

	if (!is.character(p4s)) stop("invalid p4s object")
        stopifnot(length(p4s) == 1L)
        stopifnot(!is.na(p4s))
        if (!is.null(enforce_xy)) {
            stopifnot(is.logical(enforce_xy))
            stopifnot(length(enforce_xy) == 1L)
            stopifnot(!is.na(enforce_xy))
        } else {
            enforce_xy <- get_enforce_xy()
        }
        attr(p4s, "enforce_xy") <- enforce_xy

	res <- .Call("ogrAutoIdentifyEPSG", p4s, PACKAGE="rgdal")
	res
}

get_thin_PROJ6_warnings <- function() {
    get("thin_PROJ6_warnings", envir=.RGDAL_CACHE)
}

set_thin_PROJ6_warnings <- function(value) {
    stopifnot(is.logical(value))
    stopifnot(length(value) == 1L)
    stopifnot(!is.na(value))
    assign("thin_PROJ6_warnings", value, envir=.RGDAL_CACHE)
}

get_rgdal_show_exportToProj4_warnings <- function() {
  get("rgdal_show_exportToProj4_warnings", envir=.RGDAL_CACHE)
}

set_rgdal_show_exportToProj4_warnings <- function(value) {
    stopifnot(is.logical(value))
    stopifnot(length(value) == 1L)
    stopifnot(!is.na(value))
    assign("rgdal_show_exportToProj4_warnings", value, envir=.RGDAL_CACHE)
}

get_PROJ6_warnings_count <- function() {
    get("PROJ6_warnings_count", envir=.RGDAL_CACHE)
}

get_enforce_xy <- function() {
    get("enforce_xy", envir=.RGDAL_CACHE)
}

set_enforce_xy <- function(value) {
    stopifnot(is.logical(value))
    stopifnot(length(value) == 1L)
    stopifnot(!is.na(value))
    assign("enforce_xy", value, envir=.RGDAL_CACHE)
}


get_prefer_proj <- function() {
    get("prefer_proj", envir=.RGDAL_CACHE)
}

set_prefer_proj <- function(value) {
    stopifnot(is.logical(value))
    stopifnot(length(value) == 1L)
    stopifnot(!is.na(value))
    assign("prefer_proj", value, envir=.RGDAL_CACHE)
}


showSRID <- function(inSRID, format="WKT2", multiline="NO", enforce_xy=NULL, EPSG_to_init=TRUE, prefer_proj=NULL) {
    valid_WKT_formats <- c("SFSQL", "WKT1_SIMPLE", "WKT1", "WKT1_GDAL",
        "WKT1_ESRI", "WKT2_2015", "WKT2_2018", "WKT2") # add WKT2_2019 ??
    valid_formats <- c("PROJ", valid_WKT_formats)
    stopifnot(is.character(inSRID))
    stopifnot(length(inSRID) == 1L)
    stopifnot(nzchar(inSRID))
    stopifnot(is.character(format))
    stopifnot(length(format) == 1L)
    if (!(format %in% valid_formats)) stop("invalid format value")
    out_format <- as.integer(NA)
    if (format %in% valid_WKT_formats) out_format <- 1L
    if (format == "PROJ") out_format <- 2L
    if (!is.na(multiline) && is.logical(multiline)) {
      multiline <- ifelse(multiline, "YES", "NO")
    }
    stopifnot(!is.na(multiline))
    stopifnot(is.character(multiline))
    stopifnot(length(multiline) == 1L)
    if (!(multiline %in% c("YES", "NO"))) stop("invalid multiline value")
    in_format <- as.integer(NA)
    if (substring(inSRID, 1, 1) == " ") stop("string starts with space")
    if (substring(inSRID, 1, 1) == "+") in_format = 1L
    if (substring(inSRID, 1, 3) == "urn") in_format = 2L
    if (substring(inSRID, 1, 2) == "PR") in_format = 3L
    if (substring(inSRID, 1, 2) == "GE") in_format = 3L
    if (substring(inSRID, 1, 2) == "BA") in_format = 3L
    if (substring(inSRID, 1, 2) == "BO") in_format = 3L
    if (substring(inSRID, 1, 1) == "S") in_format = 3L
    if (substring(inSRID, 1, 2) == "CO") in_format = 3L
    if (substring(inSRID, 1, 3) == "ENG") in_format = 3L
    if (substring(inSRID, 1, 4) == "EPSG") in_format = 4L
    if (substring(inSRID, 1, 4) == "ESRI") in_format = 5L
    if (substring(inSRID, 1, 3) == "OGC") in_format = 5L
    epsg <- as.integer(NA)
    if (!is.null(prefer_proj)) {
      stopifnot(is.logical(prefer_proj))
      stopifnot(length(prefer_proj) == 1L)
      stopifnot(!is.na(prefer_proj))
    } else {
        prefer_proj <- get_prefer_proj()
    }
    if (!is.na(in_format) && in_format == 4L) {
        if (EPSG_to_init && !prefer_proj) {
            in_format = 1L
            inSRID <- paste0("+init=epsg:", substring(inSRID, 6, nchar(inSRID)))
        } else {
            epsg <- as.integer(substring(inSRID, 6, nchar(inSRID)))
        }
    }
    format <- paste0("FORMAT=", format)
    multiline <- paste0("MULTILINE=", multiline)
    if (!is.null(enforce_xy)) {
      stopifnot(is.logical(enforce_xy))
      stopifnot(length(enforce_xy) == 1L)
      stopifnot(!is.na(enforce_xy))
    } else {
        enforce_xy <- get_enforce_xy()
    }

    if (new_proj_and_gdal()) {
        if (!is.na(in_format)) {
            attr(in_format, "enforce_xy") <- enforce_xy
                if (prefer_proj) {
                  if (in_format == 1L && !grepl("\\+type\\=crs", inSRID))
                    inSRID <- paste0(inSRID, " +type=crs")
                  res <- try(.Call("P6_SRID_proj", as.character(inSRID),
                    as.character(format), as.character(multiline), 
                    in_format, as.integer(epsg),
                    as.integer(out_format), PACKAGE="rgdal"), silent=TRUE)
                } else {
                  res <- try(.Call("P6_SRID_show", as.character(inSRID),
                    as.character(format), as.character(multiline), 
                    in_format, as.integer(epsg),
                    as.integer(out_format), PACKAGE="rgdal"), silent=TRUE)
                }
                if (inherits(res, "try-error"))
                    stop(unclass(attr(res, "condition"))$message, "\n", inSRID) 
            no_towgs84 <- ((is.null(attr(res, "towgs84"))) || 
                (all(nchar(attr(res, "towgs84")) == 0)))
            if ((length(grep("towgs84|TOWGS84|Position Vector|Geocentric translations", c(res))) == 0L)
                && !no_towgs84) warning("TOWGS84 discarded")
            if ((!is.null(attr(res, "ellps"))) && 
                (nchar(attr(res, "ellps")) > 0L) && 
                (length(grep("ellps|ELLIPSOID", c(res))) == 0L)) {
                if (length(grep("datum|DATUM", c(res))) == 0L) {
                    msg <- paste0("Discarded ellps ", attr(res, "ellps"),
                        " in Proj4 definition: ", c(res))
                } else {
                    msg <- ""
                }
                if (get_rgdal_show_exportToProj4_warnings()) {
                  if (!get_thin_PROJ6_warnings()) {
                    if (nchar(msg) > 0) warning(msg)
                  } else {
                    if (nchar(msg) > 0 && get("PROJ6_warnings_count",
                        envir=.RGDAL_CACHE) == 0L) {
                        warning(paste0("PROJ/GDAL PROJ string degradation in workflow\n repeated warnings suppressed\n ", msg))
                      }
                      assign("PROJ6_warnings_count",
                            get("PROJ6_warnings_count",
                            envir=.RGDAL_CACHE) + 1L, envir=.RGDAL_CACHE)
                  }
                }
            }
#warning("Discarded ellps ", attr(res, "ellps"),
#                    " in Proj4 definition")
            if ((!is.null(attr(res, "datum"))) 
                && (nchar(attr(res, "datum")) > 0L)
# spatsoc tripped PROJ 8.0.0 ENSEMBLE 2021-02-22
                && (length(grep("datum|DATUM|ENSEMBLE", c(res))) == 0L)) {
                msg <- paste0("Discarded datum ", attr(res, "datum"),
                    " in Proj4 definition")
                if (!no_towgs84 && (length(grep("towgs84", c(res))) > 0L))
                    msg <- paste0(msg, ",\n but +towgs84= values preserved")
                if (get_P6_datum_hard_fail()) stop(msg)
                else {
                  if (get_rgdal_show_exportToProj4_warnings()) {
                    if (!get_thin_PROJ6_warnings()) {
                        warning(msg)
                    } else {
                        if (get("PROJ6_warnings_count",
                            envir=.RGDAL_CACHE) == 0L) {
                            warning(paste0("PROJ/GDAL PROJ string degradation in workflow\n repeated warnings suppressed\n ", msg))
                          }
                          assign("PROJ6_warnings_count",
                                get("PROJ6_warnings_count",
                                envir=.RGDAL_CACHE) + 1L, envir=.RGDAL_CACHE)
                    }
                  }
                }
            }
            res <- c(res)
        } else {
            stop("unknown input format")
        }
    } else {
        if (!is.na(in_format) && in_format == 1L) {
            res <- showWKT(inSRID)
        } else {
            stop("unknown input format")
        }
    }
    res
}
 

# thin wrappers 121001 Robert J. Hijmans

.gd_SetNoDataValue <- function(object, NAflag) {
	.Call("RGDAL_SetNoDataValue", object, as.double(NAflag), PACKAGE="rgdal")
}


.gd_SetGeoTransform <- function(object, geotrans) {
   .Call("RGDAL_SetGeoTransform", object, geotrans, PACKAGE="rgdal")
}


.gd_SetProject <- function(object, proj4string) {
# NOT UPDATED FOR PROJ >= 6
if (new_proj_and_gdal()) warning("NOT UPDATED FOR PROJ >= 6")
    .Call("RGDAL_SetProject", object, proj4string, PACKAGE="rgdal")
}

#RH: not tested yet.
.gd_SetProjectWkt <- function(object, crs, enforce_xy=NULL) {
    if (new_proj_and_gdal()) {
        iCRS <- slot(crs, "projargs")
        wkt2 <- comment(crs)
        if (!is.null(wkt2)) {
            if (!is.null(enforce_xy)) {
                stopifnot(is.logical(enforce_xy))
                stopifnot(length(enforce_xy) == 1L)
                stopifnot(!is.na(enforce_xy))
            } else {
                enforce_xy <- get_enforce_xy()
            }
            .Call("RGDAL_SetProject_WKT2", object, wkt2, enforce_xy, PACKAGE = "rgdal")
		} else {
			warning("NO WKT AVAILABLE FOR PROJ >= 6")
			.Call("RGDAL_SetProject", object, proj4string, PACKAGE="rgdal")
		}
	} else {
		.Call("RGDAL_SetProject", object, proj4string, PACKAGE="rgdal")	
	}
}


.gd_SetStatistics <- function(object, statistics) {
	.Call("RGDAL_SetStatistics", object, as.double(statistics), PACKAGE="rgdal")
}

.gd_SetRasterColorTable <- function(object, icT) {
    if (!is.matrix(icT)) {
        stopifnot(is.character(icT))
        icT <- t(col2rgb(icT))
    }
    stopifnot(is.matrix(icT))
    stopifnot(storage.mode(icT) == "integer")
    ricT <- nrow(icT)
    cicT <- ncol(icT)
    stopifnot(cicT == 3 || cicT == 4)
    .Call("RGDAL_SetRasterColorTable", object, icT, ricT, cicT,
        PACKAGE = "rgdal")
}

.gd_SetCategoryNames <- function(object, icN) {
    stopifnot(is.character(icN))
    .Call("RGDAL_SetCategoryNames", object, icN, PACKAGE = "rgdal")
}


GDALcall <- function(object, option, ...) {
	if (option == 'SetNoDataValue') {
		.gd_SetNoDataValue(object, ...)
	} else if (option == 'SetGeoTransform') {
		.gd_SetGeoTransform(object, ...)
	} else if (option == 'SetProjectWkt') {
		.gd_SetProjectWkt(object, ...)
	} else if (option == 'SetProject') {
		.gd_SetProject(object, ...)
	} else if (option == 'SetStatistics') {
		.gd_SetStatistics(object, ...)
	} else if (option == 'SetRasterColorTable') {
		.gd_SetRasterColorTable(object, ...)
	} else if (option == 'SetCategoryNames') {
		.gd_SetCategoryNames(object, ...)
	} else {
		stop('invalid option')
	}
}




.gd_transform <- function(projfrom, projto, n, x, y, z=NULL) {
# pkgdown work-around
# NOT UPDATED FOR PROJ >= 6
if (new_proj_and_gdal()) warning("NOT UPDATED FOR PROJ >= 6")
  if (is.na(get("has_proj_def.dat", envir=.RGDAL_CACHE))) {
    assign("has_proj_def.dat", .Call("PROJ4_proj_def_dat_Installed",
      PACKAGE="rgdal"), envir=.RGDAL_CACHE)
  }
  if ((!get("has_proj_def.dat", envir=.RGDAL_CACHE)) && (!PROJis6ormore())) {
      projfrom <- proj_def_bug_fix(projfrom)
      projto <- proj_def_bug_fix(projto)
  }
  if (is.null(z)) .Call("transform", projfrom, projto, n, x, y, NULL, PACKAGE="rgdal")
  else .Call("transform", projfrom, projto, n, x, y, z, PACKAGE="rgdal")
}

# exported version
rawTransform <- function(projfrom, projto, n, x, y, z=NULL, wkt=FALSE) {
	if (wkt) { 
		# the caller determines that projfrom and projto are wkt 
		# and that new_proj_and_gdal() returns TRUE
		# to avoid multiple warnings when the function is called repetitively
		return( .Call("transform_ng", projfrom, projto, NULL,
			n, x, y, z, NULL, PACKAGE="rgdal") )
	}
		
# pkgdown work-around
          if (new_proj_and_gdal()) {
              if (get_rgdal_show_exportToProj4_warnings()) {
                 if (!get_thin_PROJ6_warnings()) {
                    warning("Using PROJ not WKT2 strings")
                 } else {
                    if (get("PROJ6_warnings_count", envir=.RGDAL_CACHE) == 0L) {
                        warning(paste0("PROJ/GDAL PROJ string degradation in workflow\n repeated warnings suppressed\n ", " Using PROJ not WKT2 strings"))
                    }
                    assign("PROJ6_warnings_count",
                        get("PROJ6_warnings_count", envir=.RGDAL_CACHE) + 1L,
                        envir=.RGDAL_CACHE)
              }
          }

          if (length(grep("+init", projfrom)) > 0)
              projfrom <- slot(CRS(projfrom),  "projargs")
          if (length(grep("+init", projto)) > 0)
              projto <- slot(CRS(projto),  "projargs")
          if (is.null(z)) res <- .Call("transform_ng", # redundant if/else?
              paste0(projfrom, " +type=crs"), paste0(projto, " +type=crs"),
              NULL, n, x, y, NULL, NULL, PACKAGE="rgdal")
          else res <- .Call("transform_ng", paste0(projfrom, " +type=crs"),
              paste0(projto, " +type=crs"), NULL, n, x, y, z, NULL, PACKAGE="rgdal")
          return(res)
        }
        if (is.na(get("has_proj_def.dat", envir=.RGDAL_CACHE))) {
          assign("has_proj_def.dat", .Call("PROJ4_proj_def_dat_Installed",
          PACKAGE="rgdal"), envir=.RGDAL_CACHE)
        }
        if (!get("has_proj_def.dat", envir=.RGDAL_CACHE)) {
            projfrom <- proj_def_bug_fix(projfrom)
            projto <- proj_def_bug_fix(projto)
        }
	if (is.null(z)) .Call("transform", projfrom, projto, n, x, y, NULL, PACKAGE="rgdal")
	else .Call("transform", projfrom, projto, n, x, y, z, PACKAGE="rgdal")
}

suppressPackageStartupMessages(library(rgdal))
logo <- system.file("pictures/Rlogo.jpg", package="rgdal")[1]
x <- GDAL.open(logo)
try(getRasterData(x, band=4))
GDAL.close(x)
logoo <- paste(logo, "o", sep="")
try(GDAL.open(logoo))
try(GDAL.close(x))
try(new('GDALDriver', "GeoTIFF"))
fn <- system.file("pictures/erdas_spnad83.tif", package = "rgdal")[1]
x <- GDALinfo(fn)
try(erdas_spnad83 <- readGDAL(fn, offset=c(500, 1000), region.dim=c(400, 400), silent=TRUE))
try(erdas_spnad83 <- readGDAL(fn, region.dim=c(4000, 400), silent=TRUE))
x <- readGDAL(fn, silent=TRUE)
tf <- tempfile()
try(writeGDAL(x, tf, drivername="GTiff", type="Byte", options="INTERLEAVE=PIXIL"))
try(writeGDAL(x, tf, drivername="GeoTiff", type="Byte"))
dsn <- system.file("vectors", package = "rgdal")[1]
x <- try(ogrInfo(dsn=dsn, layer="citis"))
try(OGRSpatialRef(dsn=dsn, layer="scot_BMG"))
dsn <- system.file("vectors/test_trk2.gpx", package = "rgdal")[1]
try(readOGR(dsn=dsn, layer="trucks", verbose=FALSE))
cities <- readOGR(system.file("vectors", package = "rgdal")[1], "cities", verbose=FALSE)
td <- tempdir()
try(writeOGR(cities, td, "cities", driver="ESRI_Shapefile"))
x <- '{"type":"Point","coordinates":[-99.74,32.45]}"'
try(rgdal::readOGR(x, "OGRGeoJSON", verbose = FALSE))

writeOGR <- function(obj, dsn, layer, driver, dataset_options=NULL, layer_options=NULL, verbose=FALSE, check_exists=NULL, overwrite_layer=FALSE, delete_dsn=FALSE, morphToESRI=NULL, encoding=NULL, shp_edge_case_fix=FALSE, dumpSRS=FALSE) {
    if (missing(dsn)) stop("missing dsn")
    stopifnot(is.character(dsn))
    stopifnot(length(dsn) == 1L)
    if (file.exists(dsn)) dsn <- normalizePath(dsn)
    if (nchar(dsn) == 0) stop("empty dsn")
    stopifnot(is.logical(verbose))
    drvs <- ogrDrivers()
    mch <- match(driver, drvs$name)
    if (is.na(mch) || length(mch) > 1L)
        stop(paste("No such driver:", driver))
    if (!drvs$write[mch]) stop("Chosen driver cannot create files")
    if (driver == "KML" || driver == "GPX") {
      if (is.na(is.projected(obj))) warning(paste("Unknown coordinate", 
        "reference system: for KML driver should be geographical"))
      else if (is.projected(obj)) {
# added brackets 120330
         warning(paste("Projected coordinate",
        "reference system: for KML driver should be geographical"))
        proj4string(obj) <- CRS(as.character(NA))
      }
# fix for over-eager internal checking in the KML driver 090114
    }
    is_shpfile <- driver == "ESRI Shapefile"
    stopifnot(is.logical(is_shpfile))
    stopifnot(length(is_shpfile) == 1)
    stopifnot(!is.na(is_shpfile))
    stopifnot(inherits(obj, "Spatial"))
    if (gridded(obj)) {
        obj <- as(obj, "SpatialPointsDataFrame")
        if (verbose) warning("coercing gridded object to points")
    }
    if (inherits(obj, "SpatialPolygonsDataFrame")) {
        if (any(sapply(lapply(slot(obj, "polygons"), function(x) comment(x)),
            is.null))) {
            if (requireNamespace("rgeos", quietly=TRUE)) {
                obj <- rgeos::createSPComment(obj)
            } else {
                stop("Spatial Polygons objects should have valid ordering comments\nuse rgeos::createSPComment()")
            }
        }
    }
    if (!"data" %in% names(getSlots(class(obj))))
        stop("obj must be a SpatialPointsDataFrame, SpatialLinesDataFrame or\n    SpatialPolygonsDataFrame") 
    dfcls <- sapply(slot(obj, "data"), function(x) class(x)[1])
    dfcls <- gsub("POSIXct", "POSIXt", dfcls)
    dfcls <- gsub("POSIXlt", "POSIXt", dfcls)
    dfcls <- gsub("Date", "POSIXt", dfcls)
# fix for logical and better reporting Barry Rowlingson 091106
    known <- c("numeric", "character", "factor", "POSIXt", "integer", "logical")
    if (!all(dfcls %in% known)) 
        stop("Can't convert columns of class: ",
            paste(unique(dfcls[!dfcls %in% known]), collapse=","),
            "; column names: ", paste(names(obj@data)[!dfcls %in% known],
            collapse=","))
    dftof <- sapply(slot(obj, "data"), typeof)
    known <- c("double", "character", "integer", "logical", "list")
    if (!all(dftof %in% known))
        stop("Can't convert columns of type: ", 
            paste(unique(dftof[!dftof %in% known]),collapse=","), 
            "; column names: ", paste(names(obj@data)[!dftof %in% known], 
            collapse=","))
    if (any("list" %in% dftof)) {
         whch <- dfcls[which(dftof == "list")] != "POSIXt"
         if (any(whch)) 
             stop("Can't convert columns of type list:",
                 paste(names(obj@data)[whch], collapse=","))
    }

    if (is.null(check_exists)) {
        if (as.integer(getGDALVersionInfo("VERSION_NUM")) >= 1800L) {
            check_exists <- TRUE
        } else {
            check_exists <- FALSE
        }
    }

    odsn <- NULL
    if (check_exists) {
        already_exists <- FALSE
        ogrI <- .Call("ogrCheckExists", as.character(dsn),
            as.character(layer), PACKAGE = "rgdal")
        if (ogrI) {
            if (driver == attr(ogrI, "driver")) already_exists <- TRUE
        }
        if (already_exists) {
            if (overwrite_layer) {
# NASTY KLUDGE: GDAL 2 appears to handle shapefile deletion awkwardly
# and prefers the "single" *.shp definition of dsn
                if (strsplit(getGDALVersionInfo(), " ")[[1]][2] >= "2" && 
                    is_shpfile && file.info(dsn)$isdir) {
                    odsn <- dsn
                    dsn <- paste(dsn, "/", layer, ".shp", sep="")
                    if (verbose) warning(dsn, " substituted for ", odsn,
                        " for layer deletion")
                    odelete_dsn <- delete_dsn
                    delete_dsn <- TRUE
                    owarnZZ <- options("warn")$warn
                    options(warn=-1)
                }
                layer_del <- try(ogrDeleteLayer(dsn, layer, driver),
                    silent=TRUE)
                if (class(layer_del) == "try-error" && delete_dsn) {
                    ogrDeleteDataSource(dsn, driver)
                    if (verbose) warning("existing data source removed")
                }
                if (exists("odelete_dsn")) {
                    delete_dsn <- odelete_dsn
                    options(warn=owarnZZ)
                }
                if (verbose) warning("existing layer removed")
            } else {
                stop("layer exists, use a new layer name")
            }
        }
    }
    if (is.null(morphToESRI))
        morphToESRI <- ifelse(is_shpfile, TRUE, FALSE)
    stopifnot(is.logical(morphToESRI))
    stopifnot(length(morphToESRI) == 1)
    stopifnot(!is.na(morphToESRI))

    nf <- length(dfcls)
    ldata <- vector(mode="list", length=nf)
    ogr_ftype <- integer(nf)
    for (i in 1:nf) {
        if (dfcls[i] == "numeric" && dftof[i] == "double") {
            ldata[[i]] <- slot(obj, "data")[[i]]
            ogr_ftype[i] <- as.integer(2) #"OFTReal"
        } else if (dfcls[i] == "character" && dftof[i] == "character") {
            ldata[[i]] <- slot(obj, "data")[[i]]
            ogr_ftype[i] <- as.integer(4) #"OFTString"
        } else if (dfcls[i] == "factor" && dftof[i] == "integer") {
            ldata[[i]] <- as.character(slot(obj, "data")[[i]])
            ogr_ftype[i] <- as.integer(4) #"OFTString"
        } else if (dfcls[i] == "POSIXt" && dftof[i] == "integer") {
            ldata[[i]] <- as.character(format(slot(obj, "data")[[i]]))
            ogr_ftype[i] <- as.integer(4) #"OFTString"
        } else if (dfcls[i] == "POSIXt" && dftof[i] == "double") {
            ldata[[i]] <- as.character(format(slot(obj, "data")[[i]]))
            ogr_ftype[i] <- as.integer(4) #"OFTString"
        } else if (dfcls[i] == "POSIXt" && dftof[i] == "list") {
            ldata[[i]] <- as.character(format(slot(obj, "data")[[i]]))
            ogr_ftype[i] <- as.integer(4) #"OFTString"
        } else if (dfcls[i] == "integer" && dftof[i] == "integer") {
            ldata[[i]] <- slot(obj, "data")[[i]]
            ogr_ftype[i] <- as.integer(0) #"OFTInteger"
        } else if (dfcls[i] == "logical" && dftof[i] == "logical") {
# fix for logical and better reporting Barry Rowlingson 091106
            ldata[[i]] <- as.integer(slot(obj, "data")[[i]])
            ogr_ftype[i] <- as.integer(0) #"OFTInteger"
        } else stop(paste(dfcls[i], dftof[i], "unknown data type"))
        if (!is.null(encoding)) {
            if (ogr_ftype[i] == 4L) {
                ldata[[i]] <- iconv(ldata[[i]], from=encoding, to="UTF-8")
            }
        }
    }
    fld_names <- names(dfcls)
    if (!is.null(encoding)) {
        fld_names <- iconv(fld_names, from=encoding, to="UTF-8")
    }
    if (is_shpfile) {
        if (any(nchar(fld_names) > 10)) {
            fld_names <- abbreviate(fld_names, minlength=7)
            warning("Field names abbreviated for ESRI Shapefile driver")
            if (any(nchar(fld_names) > 10)) 
                fld_names <- abbreviate(fld_names, minlength=5)
        }
# fix for dots in DBF field names 121124
        if (length(wh. <- grep("\\.", fld_names) > 0)) {
            fld_names[wh.] <- gsub("\\.", "_", fld_names[wh.])
        }
    }
    if (length(fld_names) != length(unique(fld_names)))
       stop("Non-unique field names")
    nobj <- nrow(slot(obj, "data"))
# add FIDs 130502
    owarn <- options("warn")
    options("warn"=-1L)
    FIDs <- as.integer(row.names(obj))
    if (any(is.na(FIDs))) FIDs <- as.integer(0:(nobj-1))
    options("warn"=owarn$warn)
    attr(nf, "verbose") <- as.logical(verbose)
    driver <- as.character(driver)
    attr(driver, "is_shpfile") <- is_shpfile
    stopifnot(is.logical(shp_edge_case_fix))
    stopifnot(length(shp_edge_case_fix) == 1)
    stopifnot(!is.na(shp_edge_case_fix))
    attr(driver, "shp_edge_case_fix") <- shp_edge_case_fix && is_shpfile
    stopifnot(is.logical(dumpSRS))
    stopifnot(length(dumpSRS) == 1)
    stopifnot(!is.na(dumpSRS))
    attr(morphToESRI, "dumpSRS") <- dumpSRS
    
    pre <- list(obj, as.character(dsn), as.character(layer), 
        driver, as.integer(nobj), nf,
        as.character(fld_names), as.integer(ogr_ftype), ldata, 
        as.character(dataset_options), as.character(layer_options),
        morphToESRI, as.integer(FIDs))
    res <- .Call("OGR_write", pre, PACKAGE="rgdal")
    NAs <- !(fld_names == attr(res, "ofld_nms"))
    if (any(NAs) && is_shpfile) {
        hits <- which(NAs)
        warning("field name", ifelse(length(hits)==1, " ", "s "),
          paste(fld_names[hits], collapse=", "), " changed by driver to: ",
          paste(attr(res, "ofld_nms")[hits], collapse=", "))
    }
    if (verbose) {
        res <- list(object_type=res, ofld_nms=attr(res, "ofld_nms"),
            output_dsn=dsn, output_layer=layer,
            output_diver=driver, output_n=nobj, output_nfields=nf,
            output_fields=fld_names, output_fclasses=ogr_ftype, 
            dataset_options=dataset_options, layer_options=layer_options,
            morphToESRI=morphToESRI, FIDs=FIDs,
            shp_edge_case_fix=shp_edge_case_fix)
        return(res)
# add FIDs 130502
    } 
}

ogrDeleteLayer <- function(dsn, layer, driver) {
    if (missing(dsn)) stop("missing dsn")
    if (nchar(dsn) == 0) stop("empty name")
    if (missing(layer)) stop("missing layer")
    if (nchar(layer) == 0) stop("empty name")
    if (missing(driver)) stop("missing layer")
    if (nchar(driver) == 0) stop("empty name")
    res <- .Call("RGDAL_ogrDeleteLayer", as.character(dsn), as.character(layer),
        as.character(driver), PACKAGE = "rgdal")
    invisible(res)
}

ogrDeleteDataSource <- function(dsn, driver) {
    if (missing(dsn)) stop("missing dsn")
    if (nchar(dsn) == 0) stop("empty name")
    if (missing(driver)) stop("missing layer")
    if (nchar(driver) == 0) stop("empty name")
    res <- .Call("RGDAL_ogrDeleteDataSource", as.character(dsn), as.character(driver),
        PACKAGE = "rgdal")
    invisible(res)

}

suppressPackageStartupMessages(library(rgdal))
if (new_proj_and_gdal()) {
get_enforce_xy()
x0 <- showSRID("EPSG:4326", multiline=TRUE, EPSG_to_init=FALSE)
cat(x0, "\n")
set_enforce_xy(FALSE)
x1 <- showSRID("EPSG:4326", multiline=TRUE, EPSG_to_init=FALSE)
cat(x1, "\n")
set_enforce_xy(TRUE)
x2 <- showSRID("EPSG:4326", multiline=TRUE, EPSG_to_init=FALSE)
cat(x2, "\n")
get_enforce_xy()
y0 <- showSRID("EPSG:4326", multiline=TRUE, EPSG_to_init=FALSE, prefer_proj=TRUE)
cat(y0, "\n")
showSRID("EPSG:4326", format="PROJ", EPSG_to_init=FALSE, prefer_proj=TRUE)
set_enforce_xy(FALSE)
y1 <- showSRID("EPSG:4326", multiline=TRUE, EPSG_to_init=FALSE, prefer_proj=TRUE)
cat(y1, "\n")
showSRID("EPSG:4326", format="PROJ", EPSG_to_init=FALSE, prefer_proj=TRUE)
set_enforce_xy(TRUE)
y2 <- showSRID("EPSG:4326", multiline=TRUE, EPSG_to_init=FALSE, prefer_proj=TRUE)
cat(y2, "\n")
showSRID("EPSG:4326", format="PROJ", EPSG_to_init=FALSE, prefer_proj=TRUE)
}

suppressPackageStartupMessages(library(rgdal))
getPROJ4VersionInfo()
getGDALVersionInfo()
d <- system.file("vectors", package="rgdal")
#shps <- ogrListLayers(d)
(o <- OGRSpatialRef(d, "trin_inca_pl03"))
if (new_proj_and_gdal()) comment(o)
(o <- OGRSpatialRef(d, "kiritimati_primary_roads"))
if (new_proj_and_gdal()) comment(o)
(o <- OGRSpatialRef(d, "cities"))
if (new_proj_and_gdal()) comment(o)
(o <- OGRSpatialRef(d, "scot_BNG"))
if (new_proj_and_gdal()) comment(o)
(o <- OGRSpatialRef(file.path(d, "airports.gml"), "airports"))
if (new_proj_and_gdal()) comment(o)
(o <- OGRSpatialRef(file.path(d, "ps_cant_31.MIF"), "ps_cant_31"))
if (new_proj_and_gdal()) comment(o)
(o <- OGRSpatialRef(file.path(d, "Up.tab"), "Up"))
if (new_proj_and_gdal()) comment(o)
(o <- OGRSpatialRef(file.path(d, "test_trk2.gpx"), "tracks"))
if (new_proj_and_gdal()) comment(o)
(o <- OGRSpatialRef(file.path(d, "PacoursIKA2.TAB"), "PacoursIKA2"))
if (new_proj_and_gdal()) comment(o)
d <- system.file("pictures", package="rgdal")
#f <- c("SP27GTIF.TIF", "cea.tif", "erdas_spnad83.tif", "scaleoffset.vrt",
# "test_envi_class.envi")
ds <- GDAL.open(file.path(d, "SP27GTIF.TIF"))
(o <- getProjectionRef(ds, OVERRIDE_PROJ_DATUM_WITH_TOWGS84=TRUE))
if (new_proj_and_gdal()) comment(o)
(o <- getProjectionRef(ds, OVERRIDE_PROJ_DATUM_WITH_TOWGS84=FALSE))
if (new_proj_and_gdal()) comment(o)
GDAL.close(ds)
ds <- GDAL.open(file.path(d, "cea.tif"))
(o <- getProjectionRef(ds, OVERRIDE_PROJ_DATUM_WITH_TOWGS84=TRUE))
if (new_proj_and_gdal()) comment(o)
(o <- getProjectionRef(ds, OVERRIDE_PROJ_DATUM_WITH_TOWGS84=FALSE))
if (new_proj_and_gdal()) comment(o)
GDAL.close(ds)
ds <- GDAL.open(file.path(d, "erdas_spnad83.tif"))
(o <- getProjectionRef(ds, OVERRIDE_PROJ_DATUM_WITH_TOWGS84=TRUE))
if (new_proj_and_gdal()) comment(o)
(o <- getProjectionRef(ds, OVERRIDE_PROJ_DATUM_WITH_TOWGS84=FALSE))
if (new_proj_and_gdal()) comment(o)
GDAL.close(ds)
ds <- GDAL.open(file.path(d, "scaleoffset.vrt"))
(o <- getProjectionRef(ds, OVERRIDE_PROJ_DATUM_WITH_TOWGS84=TRUE))
if (new_proj_and_gdal()) comment(o)
(o <- getProjectionRef(ds, OVERRIDE_PROJ_DATUM_WITH_TOWGS84=FALSE))
if (new_proj_and_gdal()) comment(o)
GDAL.close(ds)
ds <- GDAL.open(file.path(d, "test_envi_class.envi"))
(o <- getProjectionRef(ds, OVERRIDE_PROJ_DATUM_WITH_TOWGS84=TRUE))
if (new_proj_and_gdal()) comment(o)
(o <- getProjectionRef(ds, OVERRIDE_PROJ_DATUM_WITH_TOWGS84=FALSE))
if (new_proj_and_gdal()) comment(o)
GDAL.close(ds)

# following upgrade request 070122

if(!exists("Sys.setenv", envir = baseenv()))
     Sys.setenv <- Sys.putenv
.RGDAL_CACHE <- new.env(FALSE, parent=globalenv())
assign(".rgdal_old.PROJ_LIB", "", envir=.RGDAL_CACHE)
assign(".rgdal_old.GDAL_DATA", "", envir=.RGDAL_CACHE)
assign(".rgdal_set.PROJ_LIB", "", envir=.RGDAL_CACHE)
assign(".rgdal_set.GDAL_DATA", "", envir=.RGDAL_CACHE)

#.First.lib <- function(lib, pkg) {
.onLoad <- function(lib, pkg) {
  load_stuff()
}

load_stuff <- function() {
  assign(".rgdal_old.PROJ_LIB", Sys.getenv("PROJ_LIB"), envir=.RGDAL_CACHE)
  assign(".rgdal_old.GDAL_DATA", Sys.getenv("GDAL_DATA"), envir=.RGDAL_CACHE)
  assign(".rgdal_old.NEEDED", FALSE, envir=.RGDAL_CACHE)
  if (file.exists(system.file("proj/nad.lst", package = "rgdal")[1])) {
    prj = system.file("proj", package = "rgdal")[1]
    if (PROJis6ormore() && 
      .Call("PROJ4VersionInfo", PACKAGE="rgdal")[[2]] < 700) {
      set_is <- set_proj_search_paths(prj)
    } else {
      Sys.setenv("PROJ_LIB"=prj)
    }
    assign(".rgdal_set.PROJ_LIB", Sys.getenv("PROJ_LIB"), envir=.RGDAL_CACHE)
    Sys.setenv("GDAL_DATA"=system.file("gdal", package = "rgdal")[1])
    assign(".rgdal_set.GDAL_DATA", Sys.getenv("GDAL_DATA"),
      envir=.RGDAL_CACHE)
    assign(".rgdal_old.NEEDED", TRUE, envir=.RGDAL_CACHE)
  } else if (.Platform$OS.type == "windows") {
    assign(".rgdal_OSGeo4W", Sys.getenv("OSGEO4W_ROOT"), envir=.RGDAL_CACHE)
  }
  assign("OVERRIDE_PROJ_DATUM_WITH_TOWGS84", TRUE, envir=.RGDAL_CACHE)
  assign("silent", TRUE, envir=.RGDAL_CACHE)
  assign("has_proj_def.dat", as.logical(NA), envir=.RGDAL_CACHE)
  assign("P6_datum_hard_fail", FALSE, envir=.RGDAL_CACHE)
  assign("transform_wkt_comment", new_proj_and_gdal(), envir=.RGDAL_CACHE)
  assign(".last_coordOp", "", envir=.RGDAL_CACHE)
  assign("thin_PROJ6_warnings", FALSE, envir=.RGDAL_CACHE)
  assign("PROJ6_warnings_count", 0L, envir=.RGDAL_CACHE)
  assign("enforce_xy", TRUE, envir=.RGDAL_CACHE)
  assign("prefer_proj", TRUE, envir=.RGDAL_CACHE)
  rgdal_show_exportToProj4_warnings <- options("rgdal_show_exportToProj4_warnings")
  if (!is.null(rgdal_show_exportToProj4_warnings)) {
    if (!(rgdal_show_exportToProj4_warnings %in% c("all", "thin", "none"))) {
# CURRENT DEFAULT: "all"
      rgdal_show_exportToProj4_warnings <- "all"
    }
  } else {
# CURRENT DEFAULT: "all"
    rgdal_show_exportToProj4_warnings <- "all"
  }
  if (rgdal_show_exportToProj4_warnings == "all") {
    assign("rgdal_show_exportToProj4_warnings", TRUE, envir=.RGDAL_CACHE)
  } else if (rgdal_show_exportToProj4_warnings == "thin") {
    assign("rgdal_show_exportToProj4_warnings", TRUE, envir=.RGDAL_CACHE)
    assign("thin_PROJ6_warnings", TRUE, envir=.RGDAL_CACHE)
  } else {
    assign("rgdal_show_exportToProj4_warnings", FALSE, envir=.RGDAL_CACHE)
  }
  local_RGDAL_Init() #.Call('RGDAL_Init', PACKAGE="rgdal")
}

local_RGDAL_Init <- function() .Call('RGDAL_Init', PACKAGE="rgdal")

.onAttach <- function(lib, pkg) {
  ver_ok <- getGDALCheckVersion()
  rver <- getGDALVersionInfo()

  gdl <- getGDAL_DATA_Path()
#  skip_writable <- Sys.getenv("PROJ_SKIP_READ_USER_WRITABLE_DIRECTORY")
#  if (nchar(skip_writable) == 0)
#    Sys.setenv("PROJ_SKIP_READ_USER_WRITABLE_DIRECTORY"="true")
  pl <- getPROJ4libPath()
  if (nchar(pl) == 0) {
    if (is.null(attr(pl, "search_path"))) pl <- "(autodetected)"
    else pl <- attr(pl, "search_path")
  }
  fn <- system.file("SVN_VERSION", package="rgdal")
  if (file.exists(fn)) {
    svn_version <- scan(fn, what=character(1), sep="\n", quiet=TRUE)
  } else {
    svn_version <- "(unknown)"
  }
  pdd <- .Call("PROJ4_proj_def_dat_Installed", PACKAGE="rgdal")
  if (is.null(pdd) && PROJis6ormore()) pdd <- TRUE
  if (is.na(get("has_proj_def.dat", envir=.RGDAL_CACHE))) {
    assign("has_proj_def.dat", pdd, envir=.RGDAL_CACHE)
  }

  Smess <- paste('Please note that rgdal will be retired by the end of 2023,\nplan transition to sf/stars/terra functions using GDAL and PROJ\nat your earliest convenience.\n\n', 'rgdal: version: ',
    utils::packageDescription("rgdal")$Version,
    ', (SVN revision ', svn_version, ')\n',
    'Geospatial Data Abstraction Library ',
    'extensions to R successfully loaded\n',
    'Loaded GDAL runtime: ', rver, ifelse(ver_ok, '\n',
    '\n   but rgdal build and GDAL runtime not in sync:\n   ... consider re-installing rgdal!!\n'),
    paste("Path to GDAL shared files: ", gdl[1], sep=""), "\n",
    ifelse(GDAL_iconv(), "",
        paste(" GDAL does not use iconv for recoding strings.\n")),
    paste('GDAL binary built with GEOS:', getGDALwithGEOS(), "\n"),
    'Loaded PROJ runtime: ', getPROJ4VersionInfo(), '\n',
    paste("Path to PROJ shared files: ", pl[1], sep=""), "\n",
    ifelse((get("has_proj_def.dat", envir=.RGDAL_CACHE)  || (PROJis6ormore())), "", "WARNING: no proj_defs.dat in PROJ.4 shared files\n"), sep="")
  CDN_enabled <- is_proj_CDN_enabled()
  if (!is.null(CDN_enabled)) {
    Smess <- paste(Smess, "PROJ CDN enabled: ", CDN_enabled, "\n", sep="")
    if (CDN_enabled) paste(Smess, "PROJ CDN directory: ",
        proj_CDN_user_writable_dir(), "\n", sep="")
  }
  splVersion <- version_sp_linkingTo()
  Smess <- paste(Smess, "Linking to sp version:", splVersion, "\n", sep="")
  spVcheck <- NULL
  if("sp" %in% .packages()) 
    spVcheck <- utils::packageVersion("sp") == splVersion
  if (!is.null(spVcheck) && !spVcheck) paste(Smess, 
    "sp version used to install rgdal and loaded sp version differ\n")
  if (PROJis6ormore()) Smess <- paste(Smess, "To mute warnings of possible GDAL/OSR exportToProj4() degradation,\nuse options(\"rgdal_show_exportToProj4_warnings\"=\"none\") before loading sp or rgdal.\n", sep="")
  if (nzchar(get(".rgdal_set.PROJ_LIB", envir=.RGDAL_CACHE))) {
    Smess <- paste(Smess, "Overwritten PROJ_LIB was ",
      get(".rgdal_set.PROJ_LIB", envir=.RGDAL_CACHE), "\n", sep="")
  }
  packageStartupMessage(Smess, appendLF = FALSE)
}

#.Last.lib <- function(lib, pkg) {
.onUnload <- function(libpath) {
  if (get(".rgdal_old.NEEDED", envir=.RGDAL_CACHE)) {
    Sys.setenv("PROJ_LIB"=get(".rgdal_old.PROJ_LIB", envir=.RGDAL_CACHE))
    Sys.setenv("GDAL_DATA"=get(".rgdal_old.GDAL_DATA", envir=.RGDAL_CACHE))
  }
  .Call('RGDAL_Exit', PACKAGE="rgdal")
}

rgdal_extSoftVersion <- function() {
  epsg_version <- EPSG_version()
  res <- c("GDAL"=strsplit(strsplit(getGDALVersionInfo(), ",")[[1]][1], " ")[[1]][2], "GDAL_with_GEOS"=as.character(getGDALwithGEOS()), "PROJ"=strsplit(strsplit(getPROJ4VersionInfo(), ",")[[1]][1], " ")[[1]][2], "sp"=version_sp_linkingTo())
  if (!is.null(epsg_version)) res <- c(res, "EPSG"=epsg_version)
  res
}

check_sp_version <- function() {
  splVersion <- version_sp_linkingTo()
  spVcheck <- NULL
  if("sp" %in% .packages()) 
    spVcheck <- utils::packageVersion("sp") == splVersion
  if (!is.null(spVcheck) && !spVcheck)
    warning("sp version used to install rgdal and loaded sp version differ")
}

SGDF2PCT <- function(x, ncolors=256, adjust.bands=TRUE) {
	if (!inherits(x, "SpatialGridDataFrame")) 
		stop("SpatialGridDataFrame required")
	if (length(names(slot(x, "data"))) != 3L)
		stop("Three data columns (red, green, blue) required")
	if (ncolors > 256) {
		warning("ncolors reset to maximum 256")
		ncolors <- 256
	}
	if (ncolors < 2) {
		warning("ncolors reset to minimum 2")
		ncolors <- 2
	}

	fullgrid(x) = TRUE
	d.dim <- dim(as.matrix(x[1]))
	d.drv <- new("GDALDriver", "GTiff")
	GTiff3B <- new("GDALTransientDataset", driver = d.drv, 
		rows = d.dim[2], cols = d.dim[1], bands = 3, type = "Byte", 
		handle = NULL)
	gp <- gridparameters(x)
	cellsize <- gp$cellsize
	offset <- gp$cellcentre.offset
	dims <- gp$cells.dim
	gt <- c(offset[1] - 0.5 * cellsize[1], cellsize[1], 0.0, 
		offset[2] + (dims[2] -0.5) * cellsize[2], 0.0, -cellsize[2])
	.Call("RGDAL_SetGeoTransform", GTiff3B, gt, PACKAGE = "rgdal")
	have_NAs <- FALSE
	for (i in 1:3) {
		band = as.matrix(x[i])
		if (any(is.na(band))) have_NAs <- TRUE
		if (!is.numeric(band)) stop("Numeric bands required")
# 101213 Michael Sumner
		if (adjust.bands || have_NAs) {
			bmax <- max(band, na.rm = TRUE)
			bmin <- min(band, na.rm = TRUE)
			if (bmax == bmin) {
                            if (ncolors < 256) bmax <- bmin + ncolors
                            else bmax <- bmin + 1
                        }
			band <- floor((band - bmin)/((bmax-bmin)/(255)))
		} else {
			if (!is.integer(band)) 
				stop("unadjusted band not integer")
			if (any(band[!is.na(band)] < 0) || 
				any(band[!is.na(band)] > 255)) 
				stop("unadjusted band out of range")
		}
		storage.mode(band) <- "integer"
		putRasterData(GTiff3B, band, i)
	}
# 101213 Michael Sumner
#	if (have_NAs) ncolors <- ncolors + 1
	dx <- RGB2PCT(GTiff3B, band=1:3, ncolors=ncolors, set.ctab=FALSE)
	GDAL.close(GTiff3B)
	output <- getRasterData(dx$dataset, band=1)
	idx <- as.vector(output)+1
	ct <- dx$pct
	res <- list(idx=idx, ct=ct[1:ncolors])
	GDAL.close(dx$dataset)
	res
}

vec2RGB <- function(vec, breaks, col) {
	if (!is.numeric(vec)) stop("vec must be numeric")
	if (!is.vector(vec)) stop("vec must be a vector")
	if (length(col) != (length(breaks)-1L)) 
		stop("length of col must be one less than length of breaks")
	idvec <- findInterval(vec, breaks, all.inside=TRUE)
	rgb_col <- col2rgb(col)
	res <- rgb_col[, idvec]
	storage.mode(res) <- "integer"
	t(res)
}

suppressPackageStartupMessages(library(rgdal))
load(system.file("etc/test_dfs.RData", package="rgdal"))
load(system.file("etc/obj_with_comments.RData", package="rgdal"))
Ps1_nc <- Ps1
comment(Ps1_nc) <- NULL
comment(slot(Ps1_nc, "polygons")[[1]]) <- NULL
drivers <- c("GeoJSON", "ESRI Shapefile")
drivers <- drivers[drivers %in% ogrDrivers()$name]
tfbase <- tempfile()
for (driver in drivers) {
tf <- paste0(tfbase, driver, "P", sep=".")
writeOGR(SpatialPolygonsDataFrame(P, data=df9), tf, "GeoJSON",
 driver=driver, verbose=TRUE)
#rP <- as(readOGR(tf, "GeoJSON", verbose=FALSE), "SpatialPolygons")
rP <- as(readOGR(tf, verbose=FALSE), "SpatialPolygons")
cat("P with driver:", driver, "\n")
unlink(paste(tf, "*", sep=""), recursive=driver == "ESRI Shapefile")
cat(" comment", isTRUE(all.equal(sapply(slot(rP, "polygons"), comment), sapply(slot(P, "polygons"), comment))), "\n")
cat(" coords", isTRUE(all.equal(lapply(slot(P, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "coords")), lapply(slot(rP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "coords")), check.attributes=FALSE)), "\n")
cat(" holes", isTRUE(all.equal(lapply(slot(P, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "hole")), lapply(slot(rP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "hole")), check.attributes=FALSE)), "\n")
tf <- paste0(tfbase, driver, "Ph", sep=".")
writeOGR(SpatialPolygonsDataFrame(Ph, data=df9), tf, "GeoJSON",
 driver=driver, verbose=TRUE)
#rP <- as(readOGR(tf, "GeoJSON", verbose=FALSE), "SpatialPolygons")
rP <- as(readOGR(tf, verbose=FALSE), "SpatialPolygons")
cat("Ph with driver:", driver, "\n")
unlink(paste(tf, "*", sep=""), recursive=driver == "ESRI Shapefile")
cat(" comment", isTRUE(all.equal(sapply(slot(rP, "polygons"), comment), sapply(slot(Ph, "polygons"), comment))), "\n")
cat(" coords", isTRUE(all.equal(lapply(slot(Ph, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "coords")), lapply(slot(rP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "coords")), check.attributes=FALSE)), "\n")
cat(" holes", isTRUE(all.equal(lapply(slot(Ph, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "hole")), lapply(slot(rP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "hole")), check.attributes=FALSE)), "\n")
tf <- paste0(tfbase, driver, "Phs", sep=".")
writeOGR(SpatialPolygonsDataFrame(Phs, data=df9), tf, "GeoJSON",
 driver=driver, verbose=TRUE)
#rP <- as(readOGR(tf, "GeoJSON", verbose=FALSE), "SpatialPolygons")
rP <- as(readOGR(tf, verbose=FALSE), "SpatialPolygons")
cat("Phs with driver:", driver, "\n")
unlink(paste(tf, "*", sep=""), recursive=driver == "ESRI Shapefile")
cat(" comment", isTRUE(all.equal(sapply(slot(rP, "polygons"), comment), sapply(slot(Phs, "polygons"), comment))), "\n")
cat(" coords", isTRUE(all.equal(lapply(slot(Phs, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "coords")), lapply(slot(rP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "coords")), check.attributes=FALSE)), "\n")
cat(" holes", isTRUE(all.equal(lapply(slot(Phs, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "hole")), lapply(slot(rP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "hole")), check.attributes=FALSE)), "\n")
tf <- paste0(tfbase, driver, "MP", sep=".")
writeOGR(SpatialPolygonsDataFrame(MP, data=df3), tf, "GeoJSON",
 driver=driver, verbose=TRUE)
#rP <- as(readOGR(tf, "GeoJSON", verbose=FALSE), "SpatialPolygons")
rP <- as(readOGR(tf, verbose=FALSE), "SpatialPolygons")
cat("MP with driver:", driver, "\n")
unlink(paste(tf, "*", sep=""), recursive=driver == "ESRI Shapefile")
cat(" comment", isTRUE(all.equal(sapply(slot(rP, "polygons"), comment), sapply(slot(MP, "polygons"), comment))), "\n")
cat(" coords", isTRUE(all.equal(lapply(slot(MP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "coords")), lapply(slot(rP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "coords")), check.attributes=FALSE)), "\n")
cat(" holes", isTRUE(all.equal(lapply(slot(MP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "hole")), lapply(slot(rP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "hole")), check.attributes=FALSE)), "\n")
tf <- paste0(tfbase, driver, "MPh", sep=".")
writeOGR(SpatialPolygonsDataFrame(MPh, data=df3), tf, "GeoJSON",
 driver=driver, verbose=TRUE)
#rP <- as(readOGR(tf, "GeoJSON", verbose=FALSE), "SpatialPolygons")
rP <- as(readOGR(tf, verbose=FALSE), "SpatialPolygons")
cat("MPh with driver:", driver, "\n")
unlink(paste(tf, "*", sep=""), recursive=driver == "ESRI Shapefile")
cat(" comment", isTRUE(all.equal(sapply(slot(rP, "polygons"), comment), sapply(slot(MPh, "polygons"), comment))), "\n")
cat(" coords", isTRUE(all.equal(lapply(slot(MPh, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "coords")), lapply(slot(rP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "coords")), check.attributes=FALSE)), "\n")
cat(" holes", isTRUE(all.equal(lapply(slot(MPh, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "hole")), lapply(slot(rP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "hole")), check.attributes=FALSE)), "\n")
load(system.file("etc/obj_without_comments.RData", package="rgdal"))
if (requireNamespace("rgeos", quietly=TRUE)) {
tf <- paste0(tfbase, driver, "aP", sep=".")
writeOGR(SpatialPolygonsDataFrame(aP, data=df9), tf, "GeoJSON",
 driver=driver, verbose=TRUE)
#rP <- as(readOGR(tf, "GeoJSON", verbose=FALSE), "SpatialPolygons")
rP <- as(readOGR(tf, verbose=FALSE), "SpatialPolygons")
cat("aP with driver:", driver, "\n")
unlink(paste(tf, "*", sep=""), recursive=driver == "ESRI Shapefile")
cat(" coords", isTRUE(all.equal(lapply(slot(aP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "coords")), lapply(slot(rP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "coords")), check.attributes=FALSE)), "\n")
cat(" holes", isTRUE(all.equal(lapply(slot(aP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "hole")), lapply(slot(rP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "hole")), check.attributes=FALSE)), "\n")
}
if (requireNamespace("rgeos", quietly=TRUE)) {
tf <- paste0(tfbase, driver, "aPh", sep=".")
writeOGR(SpatialPolygonsDataFrame(aPh, data=df9), tf, "GeoJSON",
 driver=driver, verbose=TRUE)
#rP <- as(readOGR(tf, "GeoJSON", verbose=FALSE), "SpatialPolygons")
rP <- as(readOGR(tf, verbose=FALSE), "SpatialPolygons")
cat("aPh with driver:", driver, "\n")
unlink(paste(tf, "*", sep=""), recursive=driver == "ESRI Shapefile")
cat(" coords", isTRUE(all.equal(lapply(slot(aPh, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "coords")), lapply(slot(rP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "coords")), check.attributes=FALSE)), "\n")
cat(" holes", isTRUE(all.equal(lapply(slot(aPh, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "hole")), lapply(slot(rP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "hole")), check.attributes=FALSE)), "\n")
}
if (requireNamespace("rgeos", quietly=TRUE)) {
tf <- paste0(tfbase, driver, "aMP", sep=".")
writeOGR(SpatialPolygonsDataFrame(aMP, data=df3), tf, "GeoJSON",
 driver=driver, verbose=TRUE)
#rP <- as(readOGR(tf, "GeoJSON", verbose=FALSE), "SpatialPolygons")
rP <- as(readOGR(tf, verbose=FALSE), "SpatialPolygons")
cat("aMP with driver:", driver, "\n")
unlink(paste(tf, "*", sep=""), recursive=driver == "ESRI Shapefile")
cat(" coords", isTRUE(all.equal(lapply(slot(aMP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "coords")), lapply(slot(rP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "coords")), check.attributes=FALSE)), "\n")
cat(" holes", isTRUE(all.equal(lapply(slot(aMP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "hole")), lapply(slot(rP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "hole")), check.attributes=FALSE)), "\n")
}
if (requireNamespace("rgeos", quietly=TRUE)) {
tf <- paste0(tfbase, driver, "aMPh", sep=".")
writeOGR(SpatialPolygonsDataFrame(aMPh, data=df3), tf, "GeoJSON",
 driver=driver, verbose=TRUE)
#rP <- as(readOGR(tf, "GeoJSON", verbose=FALSE), "SpatialPolygons")
rP <- as(readOGR(tf, verbose=FALSE), "SpatialPolygons")
cat("aMPh with driver:", driver, "\n")
unlink(paste(tf, "*", sep=""), recursive=driver == "ESRI Shapefile")
cat(" coords", isTRUE(all.equal(lapply(slot(aMPh, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "coords")), lapply(slot(rP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "coords")), check.attributes=FALSE)), "\n")
cat(" holes", isTRUE(all.equal(lapply(slot(aMPh, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "hole")), lapply(slot(rP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "hole")), check.attributes=FALSE)), "\n")
}
if (requireNamespace("rgeos", quietly=TRUE)) {
tf <- paste0(tfbase, driver, "Ps1", sep=".")
writeOGR(SpatialPolygonsDataFrame(Ps1, data=df1), tf, "GeoJSON",
 driver=driver, verbose=TRUE)
#rP <- as(readOGR(tf, "GeoJSON", verbose=FALSE), "SpatialPolygons")
rP <- as(readOGR(tf, verbose=FALSE), "SpatialPolygons")
cat("Ps1 with driver:", driver, "\n")
unlink(paste(tf, "*", sep=""), recursive=driver == "ESRI Shapefile")
cat(" coords", isTRUE(all.equal(lapply(slot(Ps1, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "coords")), lapply(slot(rP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "coords")), check.attributes=FALSE)), "\n")
cat(" holes", isTRUE(all.equal(lapply(slot(Ps1, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "hole")), lapply(slot(rP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "hole")), check.attributes=FALSE)), "\n")
}
if (requireNamespace("rgeos", quietly=TRUE)) {
tf <- paste0(tfbase, driver, "Ps1_nc", sep=".")
writeOGR(SpatialPolygonsDataFrame(Ps1_nc, data=df1), tf, "GeoJSON",
 driver=driver, verbose=TRUE)
#rP <- as(readOGR(tf, "GeoJSON", verbose=FALSE), "SpatialPolygons")
rP <- as(readOGR(tf, verbose=FALSE), "SpatialPolygons")
cat("Ps1_nc with driver:", driver, "\n")
unlink(paste(tf, "*", sep=""), recursive=driver == "ESRI Shapefile")
cat(" coords", isTRUE(all.equal(lapply(slot(Ps1, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "coords")), lapply(slot(rP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "coords")), check.attributes=FALSE)), "\n")
cat(" holes", isTRUE(all.equal(lapply(slot(Ps1, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "hole")), lapply(slot(rP, "polygons"), function(x) lapply(slot(x, "Polygons"), slot, "hole")), check.attributes=FALSE)), "\n")
}
}

GDALinfo <- function(fname, silent=FALSE, returnRAT=FALSE, returnCategoryNames=FALSE, returnStats=TRUE, returnColorTable=FALSE, OVERRIDE_PROJ_DATUM_WITH_TOWGS84=NULL, returnScaleOffset=TRUE, allowedDrivers=NULL, enforce_xy=NULL, options=NULL) {
	if (nchar(fname) == 0) stop("empty file name")
	x <- GDAL.open(fname, silent=silent,
              allowedDrivers=allowedDrivers, options=options)
	d <- dim(x)[1:2]
        dr <- getDriverName(getDriver(x))
        if (!is.null(enforce_xy)) {
            stopifnot(is.logical(enforce_xy))
            stopifnot(length(enforce_xy) == 1L)
            stopifnot(!is.na(enforce_xy))
        } else {
            enforce_xy <- get_enforce_xy()
        }
#	p4s <- .Call("RGDAL_GetProjectionRef", x, PACKAGE="rgdal")
	p4s <- getProjectionRef(x, OVERRIDE_PROJ_DATUM_WITH_TOWGS84=OVERRIDE_PROJ_DATUM_WITH_TOWGS84, enforce_xy=enforce_xy)
	if (!is.na(p4s) && nchar(p4s) == 0) p4s <- as.character(NA)
	gt <- .Call('RGDAL_GetGeoTransform', x, PACKAGE="rgdal")
        if (attr(gt, "CE_Failure") && !silent)
            warning("GeoTransform values not available")
	nbands <- .Call('RGDAL_GetRasterCount', x, PACKAGE="rgdal")
        mdata <- .Call('RGDAL_GetMetadata', x, NULL, PACKAGE="rgdal")
        subdsmdata <- .Call('RGDAL_GetMetadata', x, "SUBDATASETS",
            PACKAGE="rgdal")
        if (nbands < 1) {
#            warning("no bands in dataset")
            df <- NULL
        } else {
            band <- 1:nbands
            GDType <- character(nbands)
            hasNoDataValues <- logical(nbands)
            NoDataValues <- numeric(nbands)
            blockSize1 <- integer(nbands)
            blockSize2 <- integer(nbands)
            if (returnStats) {
              Bmin <- rep(as.numeric(NA), nbands)
              Bmax <- rep(as.numeric(NA), nbands)
              Bmn <- rep(as.numeric(NA), nbands)
              Bsd <- rep(as.numeric(NA), nbands)
            }
#            Pix <- character(nbands)
            if (returnRAT) RATlist <- vector(mode="list", length=nbands)
            if (returnCategoryNames)
                CATlist <- vector(mode="list", length=nbands)
            if (returnColorTable)
                colTabs <- vector(mode="list", length=nbands)
				
#RH 4feb2013
            if (returnScaleOffset) {
                scaleOffset <- matrix(0, ncol=2, nrow=nbands)
                colnames(scaleOffset) <- c('scale', 'offset')
            }
				
	
            for (i in seq(along = band)) {

                raster <- getRasterBand(x, band[i])
                GDType[i] <- .GDALDataTypes[(.Call("RGDAL_GetBandType",
                    raster, PACKAGE="rgdal"))+1]
                bs <- getRasterBlockSize(raster)
                blockSize1[i] <- bs[1]
                blockSize2[i] <- bs[2]
                if (returnStats) {
                  statsi <- .Call("RGDAL_GetBandStatistics", raster, silent,
                    PACKAGE="rgdal")
                  if (is.null(statsi)) {
                    Bmin[i] <- .Call("RGDAL_GetBandMinimum", raster,
                        PACKAGE="rgdal")
                    Bmax[i] <- .Call("RGDAL_GetBandMaximum", raster,
                        PACKAGE="rgdal")
                  } else {
                    Bmin[i] <- statsi[1]
                    Bmax[i] <- statsi[2]
                    Bmn[i] <- statsi[3]
                    Bsd[i] <- statsi[4]
                  }
                }
                if (returnRAT) {
                    RATi <- .Call("RGDAL_GetRAT", raster, PACKAGE="rgdal")
                    if (!is.null(RATi)) RATlist[[i]] <- RATi
                }
                if (returnCategoryNames) {
                    CATi <- .Call("RGDAL_GetCategoryNames", raster,
                        PACKAGE="rgdal")
                    if (!is.null(CATi)) CATlist[[i]] <- CATi
                }
                if (returnColorTable) {
                    colTabs[[i]] <- getBandColorTable(raster)
                }
				
#RH 4feb2013
                if (returnScaleOffset) {
                    scaleOffset[i,1] <- .Call('RGDAL_GetScale', raster,
                        PACKAGE="rgdal")
                    scaleOffset[i,2] <- .Call('RGDAL_GetOffset', raster,
                        PACKAGE="rgdal")
                }
				
                NDV <- .Call("RGDAL_GetBandNoDataValue", raster,
                    PACKAGE="rgdal")
                if (is.null(NDV)) {
                    hasNoDataValues[i] <- FALSE
                } else {
                    hasNoDataValues[i] <- TRUE
                    NoDataValues[i] <- NDV[1]
                }
#                Pix[i] <- .Call("RGDAL_GetBandMetadataItem",
#                    raster, "PIXELTYPE", "IMAGE_STRUCTURE", PACKAGE="rgdal")
            }
            df <- data.frame(GDType=GDType, hasNoDataValue=hasNoDataValues,
                NoDataValue=NoDataValues, blockSize1=blockSize1,
                blockSize2=blockSize2)
            if (returnStats) df <- cbind(df, data.frame(Bmin=Bmin,
                Bmax=Bmax, Bmean=Bmn, Bsd=Bsd))
        }
        
	GDAL.close(x)
#	res <- c(rows=d[1], columns=d[2], bands=nbands, ll.x=gt[1], ll.y=gt[4],
#		res.x=abs(gt[2]), res.y=abs(gt[6]), oblique.x=abs(gt[3]), 
#		oblique.y=abs(gt[5]))
### Modified: MDSumner 22 November 2008
        cellsize = abs(c(gt[2], gt[6]))
        ysign <- sign(gt[6])
        offset.y <- ifelse(ysign < 0, gt[4] + ysign * d[1] * abs(cellsize[2]),
            gt[4] +   abs(cellsize[2]))
        res <- c(rows = d[1], columns = d[2], bands = nbands, ll.x = gt[1],
            ll.y = offset.y, res.x = abs(gt[2]), res.y = abs(gt[6]),
            oblique.x = abs(gt[3]), oblique.y = abs(gt[5]))
#### end modification
	attr(res, "ysign") <- ysign
	attr(res, "driver") <- dr 
	attr(res, "projection") <- p4s 
	attr(res, "file") <- fname
        attr(res, "df") <- df
        attr(res, "sdf") <- returnStats
        attr(res, "mdata") <- mdata
        attr(res, "subdsmdata") <- subdsmdata
        if (returnRAT) attr(res, "RATlist") <- RATlist
        if (returnCategoryNames) attr(res, "CATlist") <- CATlist
        if (returnColorTable) attr(res, "ColorTables") <- colTabs
		
#RH 4feb2013    
        if (returnScaleOffset) attr(res, "ScaleOffset") <- scaleOffset
	class(res) <- "GDALobj"
	res
}

print.GDALobj <- function(x, ...) {
	cat("rows       ", x[1], "\n")
	cat("columns    ", x[2], "\n")
	cat("bands      ", x[3], "\n")
	cat("lower left origin.x       ", x[4], "\n")
	cat("lower left origin.y       ", x[5], "\n")
	cat("res.x      ", x[6], "\n")
	cat("res.y      ", x[7], "\n")
        cat("ysign      ", attr(x, "ysign"), "\n")
	cat("oblique.x  ", x[8], "\n")
	cat("oblique.y  ", x[9], "\n")
	cat("driver     ", attr(x, "driver"), "\n")
	cat("projection ", paste(strwrap(attr(x, "projection")),
		collapse="\n"), "\n")
	cat("file       ", attr(x, "file"), "\n")
        if (!is.null(attr(x, "df"))) {
            cat("apparent band summary:\n")
            print(attr(x, "df")[,1:5])
        }
        if (attr(x, "sdf")) {
            cat("apparent band statistics:\n")
            print(attr(x, "df")[,6:9])
        }
        if (!is.null(attr(x, "ScaleOffset"))) {
            somat <- attr(x, "ScaleOffset")
            rws <- which(somat[,1] != 1 | somat[,2] != 0)
            if (any(rws)) {
                cat("ScaleOffset:\n")
                rownames(somat) <- paste("band", 1:nrow(somat), sep="")
                print(somat[rws,])
            }
        }
        if (!is.null(attr(x, "mdata"))) {
            cat("Metadata:\n")
            cv <- attr(x, "mdata")
            for (i in 1:length(cv)) cat(cv[i], "\n")
        }
        if (!is.null(attr(x, "subdsmdata"))) {
            cat("Subdatasets:\n")
            cv <- attr(x, "subdsmdata")
            for (i in 1:length(cv)) cat(cv[i], "\n")
        }
        if (!is.null(attr(x, "RATlist"))) {
            RATs <- attr(x, "RATlist")
            nRAT <- length(RATs)
            if (nRAT == 1 ) cat("Raster attribute table:\n")
            else cat("Raster attribute tables (", nRAT, "):\n", sep="")
            for (i in 1:nRAT) {
                if (i > 1) cat("----------------------\n")
                RAT <- RATs[[i]]
                print(as.data.frame(RAT))
                cat(paste("  types:", paste(attr(RAT, "GFT_type"),
                    collapse=", ")), "\n")
                cat(paste("  usages:", paste(attr(RAT, "GFT_usage"),
                    collapse=", ")), "\n")
            }
        }
        if (!is.null(attr(x, "CATlist"))) {
            CATs <- attr(x, "CATlist")
            nCAT <- length(CATs)
            cat("Category names:\n")
            print(CATs)            
        }
        if (!is.null(attr(x, "ColorTables"))
            && length(attr(x, "ColorTables")) > 0)
            cat("Colour tables returned for bands:",
              paste(which(sapply(attr(x, "ColorTables"),
              function(x) !is.null(x))), collapse=" "), "\n")
	invisible(x)
}

asGDALROD_SGDF <- function(from) {
	x <- from
	d = dim(x)
	half.cell <- c(0.5,0.5)
	offset <- c(0,0)
	output.dim <- d[1:2]
#	p4s <- .Call("RGDAL_GetProjectionRef", x, PACKAGE="rgdal")
	p4s <- getProjectionRef(x, OVERRIDE_PROJ_DATUM_WITH_TOWGS84=NULL)
	if (!is.na(p4s) && nchar(p4s) == 0) p4s <- as.character(NA)
        if (new_proj_and_gdal()) wkt2 <- comment(p4s)
        oCRS <- CRS(c(p4s))
        if (new_proj_and_gdal()) comment(oCRS) <- wkt2
	gt = .Call('RGDAL_GetGeoTransform', x, PACKAGE="rgdal")
        if (attr(gt, "CE_Failure")) warning("GeoTransform values not available")
	if (any(gt[c(3,5)] != 0.0)) stop("Diagonal grid not permitted")
	data = getRasterData(x, list_out=TRUE)
	cellsize = abs(c(gt[2],gt[6]))
	ysign <- sign(gt[6])
	co.x <- gt[1] + (offset[2] + half.cell[2]) * cellsize[1]
	co.y <- ifelse(ysign < 0, gt[4] + (ysign*((output.dim[1] + 
		offset[1]) + (ysign*half.cell[1]))) * abs(cellsize[2]),
		gt[4] + (ysign*((offset[1]) + (ysign*half.cell[1]))) * 
		abs(cellsize[2]))
	cellcentre.offset <- c(x=co.x, y=co.y)
	grid = GridTopology(cellcentre.offset, cellsize, rev(output.dim))
#	if (length(d) == 2L)
#		df = list(band1 = as.vector(data))
#	else {
#		df <- vector(mode="list", length=d[3])
#		df[[1]] <- as.vector(data[,,1, drop = FALSE])
#		for (band in 2:d[3])
#			df[[band]] <- as.vector(data[,,band, drop = FALSE])
#		names(df) = paste("band", 1:d[3], sep="")
#	}
	return(SpatialGridDataFrame(grid = grid, 
		data = as.data.frame(data), proj4string=oCRS))
#		data = data.frame(df), proj4string=CRS(p4s)))
}

setAs("GDALReadOnlyDataset", "SpatialGridDataFrame", asGDALROD_SGDF)

asSGDF_GROD <- function(x, offset, region.dim, output.dim, p4s=NULL, ..., half.cell=c(0.5,0.5), OVERRIDE_PROJ_DATUM_WITH_TOWGS84=NULL, enforce_xy=NULL) {
	if (!extends(class(x), "GDALReadOnlyDataset"))
		stop("x must be or extend a GDALReadOnlyDataset")
	d = dim(x)
	if (missing(offset)) offset <- c(0,0)
	if (missing(region.dim)) region.dim <- dim(x)[1:2]
	odim_flag <- NULL
	if (!missing(output.dim)) odim_flag <- TRUE
	else {
		output.dim <- region.dim
		odim_flag <- FALSE
	}

# suggestion by Paul Hiemstra 070817
	if (is.null(p4s)) 
#	    p4s <- .Call("RGDAL_GetProjectionRef", x, PACKAGE="rgdal")
            if (!is.null(enforce_xy)) {
                stopifnot(is.logical(enforce_xy))
                stopifnot(length(enforce_xy) == 1L)
                stopifnot(!is.na(enforce_xy))
            } else {
                enforce_xy <- get_enforce_xy()
            }
	    p4s <- getProjectionRef(x, OVERRIDE_PROJ_DATUM_WITH_TOWGS84=OVERRIDE_PROJ_DATUM_WITH_TOWGS84, enforce_xy=enforce_xy)
	if (!is.na(p4s) && nchar(p4s) == 0) p4s <- as.character(NA)
        if (new_proj_and_gdal()) wkt2 <- comment(p4s)
        oCRS <- CRS(c(p4s))
        if (new_proj_and_gdal()) comment(oCRS) <- wkt2
	gt = .Call('RGDAL_GetGeoTransform', x, PACKAGE="rgdal")
        if (attr(gt, "CE_Failure")) warning("GeoTransform values not available")
	if (any(gt[c(3,5)] != 0.0)) stop("Diagonal grid not permitted")
	data = getRasterData(x, offset=offset, 
		region.dim=region.dim, output.dim=output.dim, ..., list_out=TRUE)
	if (!odim_flag) cellsize = abs(c(gt[2],gt[6]))
	else {
		icellsize = abs(c(gt[2],gt[6]))
		span <- icellsize * rev(d)
		cellsize <- span / rev(output.dim)
	}
	ysign <- sign(gt[6])
	co.x <- gt[1] + (offset[2] + half.cell[2]) * cellsize[1]
	co.y <- ifelse(ysign < 0, gt[4] + (ysign*((output.dim[1] + 
		offset[1]) + (ysign*half.cell[1]))) * abs(cellsize[2]),
		gt[4] + (ysign*((offset[1]) + (ysign*half.cell[1]))) * 
		abs(cellsize[2]))
	cellcentre.offset <- c(x=co.x, y=co.y)
	grid = GridTopology(cellcentre.offset, cellsize, rev(output.dim))
#	if (length(d) == 2L)
#		df = list(band1 = as.vector(data))
#	else {
#		df <- vector(mode="list", length=d[3])
#		df[[1]] <- as.vector(data[,,1, drop = FALSE])
#		for (band in 2:d[3])
#			df[[band]] <- as.vector(data[,,band, drop = FALSE])
#		names(df) = paste("band", 1:d[3], sep="")
#	}
#	df1 <- data.frame(df)
	df1 <- as.data.frame(data)
	data = SpatialGridDataFrame(grid = grid, 
		data = df1, proj4string=oCRS)
	return(data)
}

readGDAL = function(fname, offset, region.dim, output.dim, band, p4s=NULL, ..., half.cell=c(0.5,0.5), silent = FALSE, OVERRIDE_PROJ_DATUM_WITH_TOWGS84=NULL, allowedDrivers=NULL, enforce_xy=NULL, options=NULL) {
	if (nchar(fname) == 0) stop("empty file name")
	x = GDAL.open(fname, silent=silent,
              allowedDrivers=allowedDrivers, options=options)
	d = dim(x)
	if (missing(offset)) offset <- c(0,0)
	if (missing(region.dim)) region.dim <- dim(x)[1:2] # rows=nx, cols=ny
#	else d <- region.dim
	odim_flag <- NULL
        if (missing(band)) band <- NULL
        else {
                if (length(band) > 1L) d[3] <- length(band)
                else d <- d[1:2]
        }
# bug report Mike Sumner 070522
	if (!missing(output.dim)) odim_flag <- TRUE
	else {
		output.dim <- region.dim
		odim_flag <- FALSE
	}
	if (!silent) {
		cat(paste(fname, "has GDAL driver", getDriverName(getDriver(x)),"\n"))
		cat(paste("and has", d[1], "rows and", d[2], "columns\n"))
	}
# suggestion by Paul Hiemstra 070817
	if (is.null(p4s)) 
#	    p4s <- .Call("RGDAL_GetProjectionRef", x, PACKAGE="rgdal")
            if (!is.null(enforce_xy)) {
                stopifnot(is.logical(enforce_xy))
                stopifnot(length(enforce_xy) == 1L)
                stopifnot(!is.na(enforce_xy))
            } else {
                enforce_xy <- get_enforce_xy()
            }
	    p4s <- getProjectionRef(x, OVERRIDE_PROJ_DATUM_WITH_TOWGS84=OVERRIDE_PROJ_DATUM_WITH_TOWGS84, enforce_xy=enforce_xy)
	if (!is.na(p4s) && nchar(p4s) == 0) p4s <- as.character(NA)
        if (new_proj_and_gdal()) wkt2 <- comment(p4s)
        oCRS <- CRS(c(p4s))
        if (new_proj_and_gdal()) comment(oCRS) <- wkt2
	gt = .Call('RGDAL_GetGeoTransform', x, PACKAGE="rgdal")
        if (attr(gt, "CE_Failure")) warning("GeoTransform values not available")
	# [1] 178400     40      0 334000      0    -40
        opSilent <- get("silent", envir=.RGDAL_CACHE)
        assign("silent", silent, envir=.RGDAL_CACHE)
	if (any(gt[c(3,5)] != 0.0)) {
		data = getRasterTable(x, band=band, offset=offset, 
			region.dim=region.dim, ...)
		GDAL.close(x)
		coordinates(data) = c(1,2)
		proj4string(data) = CRS(p4s)
	} else {
#		cellsize = abs(c(gt[2],gt[6]))
		if (!odim_flag) cellsize = abs(c(gt[2],gt[6]))
		else {
			icellsize = abs(c(gt[2],gt[6]))
# bug report Jose M. Blanco Moreno 091004
			span <- icellsize * rev(region.dim)
# bug report Mike Sumner 070215
			cellsize <- span / rev(output.dim)
		}
		ysign <- sign(gt[6])
                if (ysign > 0) 
                    warning("Y axis resolution positive, examine data for flipping")
#		cells.dim = c(d[1], d[2]) # c(d[2],d[1])
# bug report Jose M. Blanco Moreno 091004
		co.x <- gt[1] + ((offset[2]/(cellsize[1]/abs(gt[2]))) + 
                    half.cell[2]) * cellsize[1]
		co.y <- ifelse(ysign < 0, gt[4] + (ysign*((output.dim[1] + 
# bug report Jose M. Blanco Moreno 091004
			(offset[1]/(cellsize[2]/abs(gt[6]))) + 
                        (ysign*half.cell[1])))) * abs(cellsize[2]),
			gt[4] + (ysign*((offset[1]) + (ysign*half.cell[1]))) * 
			abs(cellsize[2]))
		cellcentre.offset <- c(x=co.x, y=co.y)
#		cellcentre.offset = c(x = gt[1] + 0.5 * cellsize[1], 
#			y = gt[4] - (d[2] - 0.5) * abs(cellsize[2]))
		grid = GridTopology(cellcentre.offset, cellsize, 
			rev(output.dim))
#			rev(region.dim))
		data = getRasterData(x, band=band, offset=offset, 
			region.dim=region.dim, output.dim=output.dim, ..., list_out=TRUE)
		GDAL.close(x)
#		if (length(d) == 2L)
#			df = list(band1 = as.vector(data))
#		else {
#			df <- vector(mode="list", length=d[3])
#			df[[1]] <- as.vector(data[,,1, drop = FALSE])
#			for (band in 2:d[3])
#				df[[band]] <- as.vector(data[,,band, drop = FALSE])
#			df = as.data.frame(df)
#			names(df) = paste("band", 1:d[3], sep="")
#		}
		data = SpatialGridDataFrame(grid = grid, 
			data = as.data.frame(data), proj4string=oCRS)
	}
        assign("silent", opSilent, envir=.RGDAL_CACHE)
	return(data)
}

writeGDAL = function(dataset, fname, drivername = "GTiff", type = "Float32", 
		mvFlag = NA, options=NULL, copy_drivername = "GTiff",
                setStatistics=FALSE, colorTables=NULL, catNames=NULL, enforce_xy=NULL)
{
	if (nchar(fname) == 0) stop("empty file name")
        x <- gdalDrivers()
	copy_only <- as.character(x[!x$create & x$copy, 1])
        if (drivername %in% copy_only) {
	    tds.create <- create2GDAL(dataset=dataset,
                drivername=copy_drivername, type=type,
                mvFlag=mvFlag, fname=NULL, setStatistics=setStatistics,
                colorTables=colorTables, catNames=catNames, enforce_xy=enforce_xy)
            tds.copy <- copyDataset(tds.create, driver=drivername, fname=fname)
            GDAL.close(tds.create)
	    saveDataset(tds.copy, fname, options=options)
# RSB 120921
            GDAL.close(tds.copy)
        } else {
	    tds.out <- create2GDAL(dataset=dataset, drivername=drivername, 
		type=type, mvFlag=mvFlag, options=options, fname=fname,
                setStatistics=setStatistics,  colorTables=colorTables,
                catNames=catNames, enforce_xy=enforce_xy)
	    saveDataset(tds.out, fname, options=options)
# RSB 120921
            GDAL.close(tds.out)
        }
# RSB 081030 GDAL.close cleanup
#	tmp.obj <- saveDataset(tds.out, fname, options=options)
#        GDAL.close(tmp.obj)
	invisible(fname)
}

create2GDAL = function(dataset, drivername = "GTiff", type = "Float32", mvFlag = NA, options=NULL, fname=NULL, setStatistics=FALSE, colorTables=NULL, catNames=NULL, enforce_xy=NULL)
{
	stopifnot(gridded(dataset))
	fullgrid(dataset) = TRUE
	if (is.na(match(type, .GDALDataTypes)))
            stop(paste("Invalid type:", type, "not in:",
                paste(.GDALDataTypes, collapse="\n")))
# mvFlag issues Robert Hijmans 101109
        if (is.na(mvFlag)) {
            if (type %in% c('Byte', 'UInt16', 'Int16'))
                warning(paste("mvFlag=NA unsuitable for type", type))
        }
#	d.dim = dim(as.matrix(dataset[1])) RSB 081106
	gp = gridparameters(dataset)
	cellsize = gp$cellsize
	offset = gp$cellcentre.offset
	dims = gp$cells.dim
	d.drv = new("GDALDriver", drivername)
	nbands = length(names(slot(dataset, "data")))
        if (!is.null(options) && !is.character(options))
                stop("options not character")
	tds.out = new("GDALTransientDataset", driver = d.drv, 
		rows = dims[2], cols = dims[1],
        	bands = nbands, type = type, options = options, fname = fname,
		handle = NULL)
	gt = c(offset[1] - 0.5 * cellsize[1], cellsize[1], 0.0, 
		offset[2] + (dims[2] -0.5) * cellsize[2], 0.0, -cellsize[2])
	.Call("RGDAL_SetGeoTransform", tds.out, gt, PACKAGE = "rgdal")

        if (new_proj_and_gdal()) {
            iCRS <- slot(dataset, "proj4string")
            wkt2 <- comment(iCRS)
            if (!is.null(wkt2)) {
                if (!is.null(enforce_xy)) {
                    stopifnot(is.logical(enforce_xy))
                    stopifnot(length(enforce_xy) == 1L)
                    stopifnot(!is.na(enforce_xy))
                } else {
                    enforce_xy <- get_enforce_xy()
                }
                .Call("RGDAL_SetProject_WKT2", tds.out, wkt2, enforce_xy, PACKAGE = "rgdal")
            } else {
                if (getDriverName(getDriver(tds.out)) == "RST") 
                    stop("RST files must have a valid CRS")
            }
        } else {
	    p4s <- proj4string(dataset)
	    if (!is.na(p4s) && nchar(p4s) > 0) {
	        .Call("RGDAL_SetProject", tds.out, p4s, PACKAGE = "rgdal")
            } else {
                if (getDriverName(getDriver(tds.out)) == "RST") 
                    stop("RST files must have a valid CRS")
            }
        }
        if (!is.null(colorTables)) {
            stopifnot(is.list(colorTables))
            stopifnot(length(colorTables) == nbands)
            if (type != "Byte") {
#                colorTables <- NULL
                warning("colorTables valid for Byte type only in some drivers")
            }
        }
        if (!is.null(catNames)) {
            stopifnot(is.list(catNames))
            stopifnot(length(catNames) == nbands)
        }
	for (i in 1:nbands) {
		band = as.matrix(dataset[i])
		if (!is.numeric(band)) stop("Numeric bands required")
                if (setStatistics) {
                    statistics <- range(c(band), na.rm=TRUE)
                    statistics <- c(statistics, mean(c(band), na.rm=TRUE))
                    statistics <- c(statistics, sd(c(band), na.rm=TRUE))
                }
		if (!is.na(mvFlag))
			band[is.na(band)] = mvFlag
		putRasterData(tds.out, band, i)
                tds.out_b <- getRasterBand(dataset=tds.out, band=i)
		if (!is.na(mvFlag)) {
		    .Call("RGDAL_SetNoDataValue", tds.out_b, as.double(mvFlag),
		        PACKAGE = "rgdal")
		}
                if (setStatistics) {
                    .gd_SetStatistics(tds.out_b, as.double(statistics))
                }
                if (!is.null(colorTables)) {
                    icT <- colorTables[[i]]
                    if (!is.null(icT)) {
                        .gd_SetRasterColorTable(tds.out_b, icT)
                    }
                }
                if (!is.null(catNames)) {
                    icN <- catNames[[i]]
                    if (!is.null(icN)) {
                        .gd_SetCategoryNames(tds.out_b, icN)
                    }       
                }
	}
	tds.out
}


gdalDrivers <- function() getGDALDriverNames()

toSigned <- function(x, base) {
    if (any(x < 0)) stop("already signed")
    if (storage.mode(x) != "integer") stop("band not integer")
    b_2 <- (2^(base-1)-1)
    b <- 2^base
    x[x > b_2] <- x[x > b_2] - b
    as.integer(x)
}

toUnSigned <- function(x, base) {
    if (all(x >= 0)) stop("already unsigned")
    if (storage.mode(x) != "integer") stop("band not integer")
    b <- 2^base
    x[x < 0] <- x[x < 0] + b
    as.integer(x)
}

"GDALSpatialRef" <- function(fname, silent=FALSE, OVERRIDE_PROJ_DATUM_WITH_TOWGS84=NULL, allowedDrivers=NULL, enforce_xy=NULL, options=NULL) {
	if (nchar(fname) == 0) stop("empty file name")
	x <- GDAL.open(fname, silent=silent,
              allowedDrivers=allowedDrivers, options=options)
#        p4s <- .Call("RGDAL_GetProjectionRef", x, PACKAGE="rgdal")
        p4s <- getProjectionRef(x, OVERRIDE_PROJ_DATUM_WITH_TOWGS84=OVERRIDE_PROJ_DATUM_WITH_TOWGS84, enforce_xy=enforce_xy)
	GDAL.close(x)
        p4s
}

suppressPackageStartupMessages(library(rgdal))
set_thin_PROJ6_warnings(TRUE)
data(state)
xy <- cbind(state.center$x, state.center$y)
res <- project(xy, "+proj=lcc +lat_1=48 +lat_2=33 +lon_0=-100 +ellps=GRS80")
res1 <- project(res, "+proj=lcc +lat_1=48 +lat_2=33 +lon_0=-100 +ellps=GRS80",
 inv=TRUE)
stopifnot(isTRUE(all.equal(res1, xy)))
states <- data.frame(state.x77, state.center)
states <- states[states$x > -121,]
coordinates(states) <- c("x", "y")
proj4string(states) <- CRS("+proj=longlat +ellps=clrk66")
state.ll83 <- spTransform(states, CRS("+proj=longlat +ellps=GRS80"))
state.ll <- spTransform(state.ll83, CRS("+proj=longlat +ellps=clrk66"))
stopifnot(isTRUE(all.equal(coordinates(states), coordinates(state.ll))))
broke_proj <- FALSE
pv <- .Call("PROJ4VersionInfo", PACKAGE="rgdal")[[2]]
# https://github.com/OSGeo/PROJ/issues/1525
if (pv >= 600 && pv < 620) broke_proj <- TRUE
if (!broke_proj) {
(crds <- matrix(data=c(9.05, 48.52), ncol=2))
(a <- project(crds, paste("+proj=ob_tran +o_proj=longlat",
 "+o_lon_p=-162 +o_lat_p=39.25 +lon_0=180 +ellps=sphere +no_defs"),
 use_ob_tran=TRUE, verbose=TRUE))
stopifnot(isTRUE(all.equal(a, matrix(c(-5.917698, -1.87195), ncol=2), tolerance=.Machine$double.eps ^ 0.25)))
(a1 <- project(a, paste("+proj=ob_tran +o_proj=longlat",
 "+o_lon_p=-162 +o_lat_p=39.25 +lon_0=180 +ellps=sphere +no_defs"),
 inv=TRUE, use_ob_tran=TRUE, verbose=TRUE))
stopifnot(isTRUE(all.equal(a1, crds, tolerance=.Machine$double.eps ^ 0.25)))
spPoint <- SpatialPoints(coords=crds,
 proj4string=CRS("+proj=longlat +ellps=sphere +no_defs"))
a <- spTransform(spPoint, CRS(paste("+proj=ob_tran +o_proj=longlat",
 "+o_lon_p=-162 +o_lat_p=39.25 +lon_0=180 +ellps=sphere +no_defs")),
 use_ob_tran=TRUE)
stopifnot(isTRUE(all.equal(unname(coordinates(a)), matrix(c(-5.917698, -1.87195), ncol=2), tolerance=.Machine$double.eps ^ 0.25)))
a1 <- spTransform(a, CRS("+proj=longlat +ellps=sphere +no_defs"),
 use_ob_tran=TRUE)
stopifnot(isTRUE(all.equal(unname(coordinates(a1)), unname(coordinates(spPoint)), tolerance=.Machine$double.eps ^ 0.25)))
}
sp <- SpatialPoints(matrix(c(1, 1), nrow=1), proj4string=CRS("+init=epsg:4326"))
sp.tr <- spTransform(sp, CRS("+init=epsg:3857"))
stopifnot(isTRUE(all.equal(unname(coordinates(sp.tr)), matrix(c(111319.4908, 111325.1429), nrow=1))))

if(getRversion() < "3.3.0") {
  stop("Your version of R is too old. This package requires R-3.3.0 or newer on Windows.")
}

# Download gdal-3.* from rwinlib
VERSION <- commandArgs(TRUE)
#if(!file.exists(sprintf("../windows/gdal2-%s/include/gdal/gdal.h", VERSION))){
testfile <- sprintf("../windows/gdal3-%s/include/gdal-%s/gdal.h", VERSION, VERSION)
if(!file.exists(testfile)){
  if(getRversion() < "3.3.0") setInternet2()
#  download.file(sprintf("https://github.com/rwinlib/gdal2/archive/v%s.zip", VERSION), "lib.zip", quiet = TRUE)
  download.file(sprintf("https://github.com/rwinlib/gdal3/archive/v%s.zip", VERSION), "lib.zip", quiet = TRUE)
  dir.create("../windows", showWarnings = FALSE)
  unzip("lib.zip", exdir = "../windows")
  unlink("lib.zip")
}

EPSG_version <- function() {
    metadata <- NULL
    if (PROJis6ormore()) {
        if (requireNamespace("DBI", quietly=TRUE) && 
            requireNamespace("RSQLite", quietly=TRUE)) {
            shpr <- get_proj_search_paths()
            db <- DBI::dbConnect(RSQLite::SQLite(),
                dbname=file.path(shpr[length(shpr)], "proj.db"))
            md <- DBI::dbReadTable(db, "metadata")
            DBI::dbDisconnect(db)
            metadata <- md[md$key == "EPSG.VERSION", "value"]
        }
    }
    metadata
}

make_EPSG <- function(file) {
        if (.Call("PROJ4VersionInfo", PACKAGE = "rgdal")[[2]] >= 600L &&
            !missing(file)) {
            warning("New proj interface does not use file argument")
            file <- NULL
        }
        metadata <- NULL
        if (missing(file) || is.null(file)) {
            tf <- tempfile()
            n <- .Call("PROJcopyEPSG", tf, PACKAGE="rgdal")
            if (PROJis6ormore()) {
                if (n <= 0) stop("PROJ 6 database empty")
                EPSG <- read.csv(tf, header=TRUE, stringsAsFactors=FALSE)
                attr(EPSG, "metadata") <- EPSG_version()
                return(EPSG)
            }
            if (n > 0) file <- tf
            else stop("Error opening epsg file")
       } else {
# bug fix 20150914 Alain Benard
           if (file.exists(file)) {
               n <- length(readLines(file))
           } else {
               stop("Error opening", file)
           }
       }
       pre5 <- .Call("PROJ4VersionInfo", PACKAGE = "rgdal")[[2]] < 500L
       if (!pre5) n <- n - 1
#	epsg <- file(system.file("proj/epsg", package="rgdal"), open="r")
# report Michael Sumner 071017
	epsg <- file(file, open="r")
	note <- character(n)
	prj4 <- character(n)
	code <- integer(n)
        if (!pre5) metadata <- readLines(epsg, n=1)
	for (i in seq(along=code)) { # FIXME - broken in 5.0.0
		res <- strsplit(paste(readLines(epsg, n=2), collapse=""), 
			"[<>]")
		note[i] <- res[[1]][1]
		prj4[i] <- res[[1]][3]
		code[i] <- as.integer(res[[1]][2])
		if (is.na(code[i])) res <- readLines(epsg, n=1)
	}
	close(epsg)
        prj4 <- scan(con <- textConnection(prj4), what=character(0), sep="\n",
                 strip.white=TRUE, quiet=TRUE)
        close(con)
#	prj4 <- scan(textConnection(prj4), what=character(0), sep="\n", 
#		strip.white=TRUE, quiet=TRUE)
	EPSG <- data.frame(code=I(code), note=I(note), prj4=I(prj4))
	EPSG <- EPSG[-which(is.na(EPSG$note)),]
        attr(EPSG, "metadata") <- metadata
	EPSG
}

sub.GDROD = function(x, i, j, ... , drop = FALSE) {
	# x[rows, cols, bands, ...]
#	if (!require(sp))
#		stop("could not load package sp")
#	n.args = nargs()
	dots = list(...)
	if (!missing(drop))
		stop("don't supply drop: it needs to be FALSE anyway")
	d = dim(x)
	if (missing(i))
		rows = 1:d[1]
	else 
		rows = i
	if (missing(j))
		cols = 1:d[2]
	else
		cols = j
	if (length(dots) > 0L) {
		# fish for an unnamed argument, and rename it bands
		m  = match(names(dots),  c(""))
		if (any(!is.na(m))) 
			names(dots)[m[1]] = "band"
	}
	# process common arguments:
	gdal.args = dots
	if (is.null(gdal.args$half.cell)) half.cell <- c(0.5,0.5)
	else {
		half.cell <- gdal.args$half.cell
		gdal.args$half.cell <- NULL
	}

	gdal.args$dataset = x
	gdal.args$band = dots$band # NULL if not given
	if (is.null(gdal.args$offset))
#		gdal.args$offset = c(0,0)
		gdal.args$offset = c(min(rows) - 1, min(cols) - 1)
	if (is.null(gdal.args$region.dim))
		gdal.args$region.dim = c(max(rows) - gdal.args$offset[1], 
			max(cols) - gdal.args$offset[2])
	rows = rows - gdal.args$offset[1]
	cols = cols - gdal.args$offset[2]
	# further arguments to getRasterData:
	if (is.null(gdal.args$output.dim))
		gdal.args$output.dim = gdal.args$region.dim
	if (is.null(gdal.args$interleave))
		gdal.args$interleave = c(0, 0)
	if (is.null(gdal.args$as.is))
		gdal.args$as.is = TRUE
	
	# retrieve topology:
	gt = .Call('RGDAL_GetGeoTransform', x, PACKAGE="rgdal")
	# [1] 178400     40      0 334000      0    -40
        if (attr(gt, "CE_Failure")) warning("GeoTransform values not available")
#	p4s <- .Call("RGDAL_GetProjectionRef", x, PACKAGE="rgdal")
	p4s <- getProjectionRef(x, OVERRIDE_PROJ_DATUM_WITH_TOWGS84=NULL)
	if (!is.na(p4s) && nchar(p4s) == 0) p4s <- as.character(NA)
        if (new_proj_and_gdal()) wkt2 <- comment(p4s)
        oCRS <- CRS(c(p4s))
        if (new_proj_and_gdal()) comment(oCRS) <- wkt2

	# retrieve data:
	if (any(gt[c(3,5)] != 0)) {
		warning("grid has an orientation which is not suitable for sp grids; returning points...")
		data = do.call(getRasterTable, gdal.args)
		coordinates(data) = c(1,2)
	} else {
	        if (is.null(gdal.args$list_out))
		    gdal.args$list_out = FALSE
		data = do.call(getRasterData, gdal.args)
		# subset data:
		d = dim(data) # rows=nx, cols=ny
		if (length(d) == 3L) {
			if (!is.null(gdal.args$band))
				band = gdal.args$band
			else
				band = 1:d[3]
			data = data[cols,rows,band]
		} else
			data = data[cols,rows]
		cellsize = abs(c(gt[2] * (1 + gdal.args$interleave[2]),
				gt[6] * (1 + gdal.args$interleave[1])))
		d = dim(data) # rows=nx, cols=ny
		cells.dim = c(d[1], d[2]) # c(d[2],d[1])
		ysign <- sign(gt[6])
		co.x <- gt[1] + (gdal.args$offset[2] + half.cell[2]) *
			 cellsize[1]
		co.y <- ifelse(ysign < 0, gt[4] + (ysign*((
			gdal.args$region.dim[1] + 
			gdal.args$offset[1]) + (ysign*half.cell[1]))) * 
			abs(cellsize[2]), gt[4] + (ysign*(
			(gdal.args$offset[1]) + (ysign*half.cell[1]))) * 
			abs(cellsize[2]))
		cellcentre.offset <- c(x=co.x, y=co.y)
#		cellcentre.offset = c(x = gt[1] + (0.5 + 
#		gdal.args$offset[2])* cellsize[1], 
#			y = gt[4] - (d[2] - 0.5) * abs(cellsize[2]))
		grid = GridTopology(cellcentre.offset, cellsize, cells.dim)
		if (length(d) == 2L)
			df = data.frame(band1 = as.vector(data))
		else {
			df = as.vector(data[,,1])
			for (band in 2:d[3])
				df = cbind(df, as.vector(data[,,band]))
			df = as.data.frame(df)
			names(df) = paste("band", 1:d[3], sep="")
		}
		data = SpatialGridDataFrame(grid, df, proj4string=oCRS)
	}
	return(data)
}
setMethod("[", "GDALReadOnlyDataset", sub.GDROD)

list_coordOps <- function(src_crs, tgt_crs, area_of_interest=as.numeric(NA), strict_containment=FALSE, visualization_order=NULL) {
    stopifnot(is.character(src_crs))
    stopifnot(length(src_crs) == 1L)
    stopifnot(is.character(tgt_crs))
    stopifnot(length(tgt_crs) == 1L)
    stopifnot(is.numeric(area_of_interest))
    stopifnot((length(area_of_interest) == 1L && is.na(area_of_interest)) ||
        (length(area_of_interest) == 4L))
    stopifnot(is.logical(strict_containment))
    stopifnot(length(strict_containment) == 1L)
    if (!is.null(visualization_order)) {
        stopifnot(is.logical(visualization_order))
        stopifnot(length(visualization_order) == 1L)
        stopifnot(!is.na(visualization_order))
    } else {
        visualization_order <- get_enforce_xy()
    }
    res <- .Call("list_coordinate_ops", src_crs, tgt_crs, area_of_interest, 
        strict_containment, visualization_order, PACKAGE="rgdal")
    if (is.null(res)) stop("function not available without PROJ 6")
    names(res) <- c("description", "definition", "accuracy", "instantiable", "ballpark", "number_grids")
    is.na(res$accuracy) <- res$accuracy < 0
    grids <- res[[7]]
    res[[7]] <- NULL
    defs <- res$definition
    nzdefs <- nzchar(defs)
    res$definition[nzdefs] <- unname(sapply(defs[nzdefs], function(x) gsub(" ",
        " +", paste0("+", x))))
    res$definition[!nzdefs] <- as.character(NA)
    res <- as.data.frame(res, stringsAsFactors=FALSE)
    attr(res, "grids") <- grids
    attr(res, "src_crs") <- src_crs
    attr(res, "tgt_crs") <- tgt_crs
    attr(res, "area_of_interest") <- area_of_interest
    attr(res, "strict_containment") <- strict_containment
    attr(res, "visualization_order") <- visualization_order
    class(res) <- c("coordOps", class(res))
    res
}

prettify_wkt <- function(inSRID) {
    if (substring(inSRID, 1, 1) == " ") stop("string starts with space")
    if (substring(inSRID, 1, 1) == "+")
        res <- strwrap(inSRID, exdent=8, width=0.8*getOption("width"))
    if (substring(inSRID, 1, 3) == "urn") res <- inSRID
    if (substring(inSRID, 1, 1) == "P" || substring(inSRID, 1, 1) == "G" || substring(inSRID, 1, 1) == "B" || substring(inSRID, 1, 1) == "S") 
        res <- strwrap(gsub(",", ", ", inSRID), exdent=8,
            width=0.8*getOption("width"))
    if (substring(inSRID, 1, 4) == "EPSG") res <- inSRID
    if (substring(inSRID, 1, 4) == "ESRI") res <- inSRID
    if (substring(inSRID, 1, 3) == "OGC") res <- inSRID
    res
}

print.coordOps <- function(x, ...) {
    n <- nrow(x)
    cat("Candidate coordinate operations found: ", n, "\n")
    if (!all(is.na(attr(x, "area_of_interest"))))
        cat("Search specification:", paste("Area of interest: ", 
            paste(attr(x, "area_of_interest"),
            collapse=", ")), "\n")
    cat("Strict containment: ", attr(x, "strict_containment"), "\n")
    cat("Visualization order: ", attr(x, "visualization_order"), "\n")
    in_str <- prettify_wkt(attr(x, "src_crs"))
    if (length(in_str) == 1) cat("Source:", in_str, "\n")
    else {cat("Source: "); cat(in_str, sep="\n")}
    out_str <- prettify_wkt(attr(x, "tgt_crs"))
    if (length(out_str) == 1) cat("Target:", out_str, "\n")
    else cat("Target:", out_str, sep="\n")
    nos <- which(!x$instantiable)
    if (length(nos) > 0L) xx <- x[-nos,]
    else xx <- x
    xx <- xx[order(xx$accuracy),]
    if (is.na(xx$accuracy[1]))
      cat("Best instantiable operation has only ballpark accuracy", "\n")
    else cat("Best instantiable operation has accuracy:", xx$accuracy[1], "m\n")
    cat("Description: ")
    desc <- strwrap(xx$description[1], exdent=13, width=0.8*getOption("width"))
    if (length(desc) == 1L) cat(desc, "\n")
    else cat(desc, sep="\n")
    cat("Definition:  ")
    def <- strwrap(xx$definition[1], exdent=13, width=0.8*getOption("width"))
    if (length(def) == 1L) cat(def, "\n")
    else cat(def, sep="\n")
    if (length(nos) > 0L) {
        grds <- attr(x, "grids")
        for (i in seq(along=nos)) {
            grd <- grds[[nos[i]]]
            ii <- length(grd)
            if (ii > 0L) {
                cat("Operation", nos[i], "is lacking", ii,
                    ifelse(ii == 1L, "grid", "grids"),
                    "with accuracy", x$accuracy[nos[i]], "m\n")
                for (j in 1:ii) {
                    cat("Missing grid:", grd[[j]][[1]], "\n")
                    if (nzchar(grd[[j]][[2]])) cat("Name:", grd[[j]][[2]],
                        "\n")
                    if (nzchar(grd[[j]][[4]])) cat("URL:", grd[[j]][[4]],
                        "\n")
                }
            }
        }
    }
    
    invisible(x)    
}

best_instantiable_coordOp <- function(x) {
    stopifnot(inherits(x, "coordOps"))
    nos <- which(!x$instantiable)
    if (length(nos) > 0L) xx <- x[-nos,]
    else xx <- x
    xx <- xx[order(xx$accuracy),]
    if (is.na(xx$accuracy[1]))
      warning("Best instantiable operation has only ballpark accuracy")
    res <- xx$definition[1]
    attr(res, "description") <- xx$description[1]
    res
} 

%\VignetteIndexEntry{OGR shapefile encoding}
%\VignetteDepends{}
%\VignetteKeywords{spatial}
%\VignettePackage{rgdal}
\documentclass[a4paper,10pt]{article} 
\usepackage[utf8]{inputenc}
\usepackage[T1]{fontenc}
%\usepackage[dvips]{graphicx,color}
\usepackage{times}
\usepackage{hyperref}
\usepackage{natbib}
\usepackage[english]{babel}
\usepackage{xspace}

\usepackage{Sweave}
\usepackage{mathptm}
\usepackage{natbib}

\setkeys{Gin}{width=0.95\textwidth}
\newcommand{\strong}[1]{{\normalfont\fontseries{b}\selectfont #1}}
\let\pkg=\strong
\RequirePackage{alltt}
\newenvironment{example}{\begin{alltt}}{\end{alltt}}
\newenvironment{smallexample}{\begin{alltt}\small}{\end{alltt}}
\newcommand{\code}[1]{\texttt{\small #1}}
\def\RR{\textsf{R}\xspace}
\def\SP{\texttt{S-PLUS}\xspace}
\def\SS{\texttt{S}\xspace}

\title{OGR shapefile encoding} 
\author{Roger Bivand} 

\begin{document} 

\maketitle 

\section{Introduction}

Changes have taken place in the way that OGR, the part of GDAL that handles vector data, treats character strings, both the contents of string attribute fields, and the names of fields. A discussion document covers the principles recognised in work on encoding strings.\footnote{\url{http://trac.osgeo.org/gdal/wiki/rfc23_ogr_unicode}.} The document states that ``it is proposed to implement the CPLRecode() method using the iconv() and related functions when available.'' These mechanisms are similar to, but not identical with, those used in \RR itself. 

The implementation\footnote{\url{http://trac.osgeo.org/gdal/browser/trunk/gdal/port/cpl_recode.cpp}.} distinguishes between the use of iconv mechanisms in OGR, when GDAL is built with iconv support, and a fallback setting when iconv is not available. In both settings, UTF-8 encoding is intended to be use internally. Conversion from the layer in the data source to what we see inside \RR will differ depending on whether iconv support is available in GDAL, and on the encoding used in the \RR session.

This has had particular impact on the ESRI Shapefile driver, because this OGR format uses DBF files for storing attribute data. Impacts on other drivers are not known at present. A subsection has been created in the \RR wiki\footnote{\url{http://rwiki.sciviews.org/doku.php\#ogr_string_encodings}.} to gather user experiences.

Recoding support for the ESRI Shapefile driver was introduced in GDAL/OGR 1.9.0. Two mechanisms are described in the driver documentation.\footnote{\url{http://www.gdal.org/ogr/drv_shapefile.html}.} 

We can read thet ``the SHAPE\_ENCODING configuration option may be used to override the encoding interpretation of the shapefile with any encoding supported by CPLRecode or to "" to avoid any recoding.'' Reference is made here to option values set in a number of ways, possibly by an environment (shell) variable, but maintained within the running instance of GDAL.\footnote{\url{http://trac.osgeo.org/gdal/browser/trunk/gdal/port/cpl_conv.cpp}, functions CPLSetConfigOption() and CPLGetConfigOption().} It emerged during discussion of a ticket on this issue,\footnote{\url{http://trac.osgeo.org/gdal/ticket/4920}.} that the use of shell variables may not be portable, so functions in \pkg{rgdal} set and get the configure options through compiled code, in particular the SHAPE\_ENCODING configuration option.

The second mechanism does not seem very robust, and is based on storing and retrieving values set in the LDID byte of the DBF file header. The driver documentation says ``an attempt is made to read the LDID/codepage setting from the .dbf file and use it to translate string fields to UTF-8 on read, and back when writing. LDID "87 / 0x57" is treated as ISO8859\_1 which may not be appropriate.'' In the driver code, a listing is provided\footnote{\url{http://trac.osgeo.org/gdal/browser/trunk/gdal/ogr/ogrsf_frmts/shape/ogrshapelayer.cpp}, in function ConvertCodePage(), after line 180.}, referring to a Russian website\footnote{\url{http://www.autopark.ru/ASBProgrammerGuide/DBFSTRUC.HTM}, see Table 9.} as authority. Crucially, the most common case for ESRI Shapefiles written by ArcGIS under Windows is LDID 87, which this function treats as ISO8859\_1. In addition, users of this format may provide an extra file with the extension CPG, which may or may not be respected.

This mechanism is related to the rather incomplete description provided by ESRI in their knowledge base,\footnote{\url{http://support.esri.com/en/knowledgebase/techarticles/detail/21106}.} but which may contain hints that are helpful in exchanging shapefiles between \RR/\pkg{rgdal} via the OGR ESRI Shapefile driver. It is made plain that it is the user's responsibility to divine the appropriate encoding. 

For obvious reasons, it is very difficult to give a sensible representation in this vignette of the character strings involved, because the vignette itself contains encoded strings. I have encapsulated results on two platforms available to me, and have consulted with a user of OSX. Some results are given using \code{charToRaw} to provide a neutral framework permitting encodings to be compared across platforms. The key string used in this example may be found at this location: \url{http://goo.gl/maps/PNRgX} (Střítež nad Ludinou), using a shapefile provided by Lukáš Marek as a result of work provoked by Jeff Ranara reporting contorted Swedish strings during a course in Bergen.

\section{Does GDAL have iconv support?}

When the \pkg{rgdal} is loaded, a set of startup messages is displayed (unless supressed). They may include the line:

\begin{verbatim}
GDAL does not use iconv for recoding strings.
\end{verbatim}

\noindent
which is generated if:

<<echo=TRUE,eval=TRUE,results=hide>>=
library(rgdal)
@
<<echo=TRUE,eval=FALSE>>=
.Call("RGDAL_CPL_RECODE_ICONV", PACKAGE="rgdal")
@

\noindent
is \code{FALSE}. This test is not absolutely trustworthy when GDAL is dynamically linked to \pkg{rgdal}, because it reports the state of the GDAL configure variables set when GDAL and \pkg{rgdal} were built. It is trustworthy for the CRAN Windows and OSX binaries, because they are built with static linking to GDAL and its dependencies. In other cases, if the GDAL runtime binaries have been updated but the header files have not been, or \pkg{rgdal} has not been re-installed, the value returned may be misleading. 

Depending on the outcome of this test, and concentrating here on the ESRI Shapefile driver, we can fork between two cases. 

\subsection{LDID and codepage values}

First we will see how the LDID value may be retrieved. The \code{ogrInfo} reports the LDID for the ESRI Shapefile driver:

<<echo=TRUE,eval=TRUE>>=
dsn <- system.file("etc", package="rgdal")
layer <- "point"
oI <- ogrInfo(dsn, layer)
attr(oI, "LDID")
@

\noindent
In this case, Lukáš Marek also provided a CPG file, as this is said to be common practice to inform ArcGIS of the appropriate codepage for the DBF file:

<<echo=TRUE,eval=TRUE>>=
scan(paste(dsn, .Platform$file.sep, layer, ".cpg", sep=""),
 "character")
@

\noindent
This value does not match the LDID (which suits CP1252, the so-called ANSI codepage), but does match the Czech OEM codepage, for which the LDID should possibly be 31 (or 200). Most LDID values observed are 0 or 87. It is also possible that the value in the CPG file should be CP1250, rather than that used here.

To set the LDID value in writing a vector \code{Spatial*DataFrame} using the ESRI Shapefile driver, one may use the \code{layer\_options=} argument to \code{writeOGR}:

<<echo=TRUE,eval=FALSE>>=
writeOGR(mySDF, dsn, layer, driver="ESRI Shapefile",
 layer_options='ENCODING="LDID/31"')
@

\noindent
but note that the OGR format documentation says: ``The default value is "LDID/87". It is not clear what other values may be appropriate.''

\subsection{String representation in \RR}

The encoding in use in the \RR session may be seen by:

<<echo=TRUE,eval=TRUE>>=
Sys.getlocale("LC_CTYPE")
unlist(l10n_info())
@

\noindent
where \code{l10n\_info} may also return a \code{codepage} component (on Windows). Even when \RR is not running in a multibyte locale, as in the Windows GUI in some cases, it may be able to display marked multibyte charsets. OGR does not seem to mark multibyte charsets. \RR manuals document the ways in which internationalization, locales, and encoding are handled in general,\footnote{\url{http://cran.r-project.org/doc/manuals/r-release/R-admin.html\#Internationalization}.} and for data import and export.\footnote{\url{http://cran.r-project.org/doc/manuals/r-release/R-data.html\#Encodings}.} In addition, the OSX\footnote{\url{http://cran.r-project.org/bin/macosx/RMacOSX-FAQ.html\#Internationalization-of-the-R_002eapp}.} and Windows\footnote{\url{http://cran.r-project.org/bin/windows/base/rw-FAQ.html\#Languages-and-Internationalization}.} FAQ describe platform-specific questions. 

\section{GDAL with iconv}

Despite the uncertainty present in the LDID and codepage definitions, it is possible to assert control both when GDAL is built with iconv and without iconv. First the GDAL with iconv case. The data displayed here are pre-stored from this system:

<<echo=FALSE,eval=TRUE>>=
load(paste(dsn, .Platform$file.sep, "point_LinuxGDAL.RData", sep=""))
@
<<echo=TRUE,eval=FALSE>>=
sessionInfo()
@
<<echo=FALSE,eval=TRUE>>=
sI_1
@
\noindent
This system has \pkg{rgdal} installed from source, using GDAL also installed from source. Similar behaviour will be found on other similar systems, also including \pkg{rgdal} installed from source under OSX, built against the Kyngchaos GDAL framework. Here, we can check that iconv is available in GDAL:

<<echo=TRUE,eval=FALSE>>=
.Call("RGDAL_CPL_RECODE_ICONV", PACKAGE="rgdal")
@
<<echo=FALSE,eval=TRUE>>=
cpliconv_1
@

\noindent
We proceed by checking the CPL configure option:

<<echo=TRUE,eval=FALSE>>=
getCPLConfigOption("SHAPE_ENCODING")
@
<<echo=FALSE,eval=TRUE>>=
NULL
@

\noindent
It may be set using \code{setCPLConfigOption("SHAPE\_ENCODING", value)}, where \code{value} is the appropriate encoding, or \code{NULL} to unset \code{SHAPE\_ENCODING}. If \code{value} is the empty string, recoding is turned off in GDAL. Since we know that the encoding of the sample file is CP1250, we can import into \RR recoding to UTF-8 (the charset of the \RR session and the charset used internally by GDAL) in three ways (using \code{stringsAsFactors=FALSE} to make access to the string values easier):

<<echo=TRUE,eval=FALSE>>=
setCPLConfigOption("SHAPE_ENCODING", "CP1250")
pt1 <- readOGR(dsn, layer, stringsAsFactors=FALSE)
setCPLConfigOption("SHAPE_ENCODING", NULL)
charToRaw(pt1$NAZEV[1])
@
<<echo=FALSE,eval=TRUE>>=
pt2cr_1
@
<<echo=TRUE,eval=FALSE>>=
pt2 <- readOGR(dsn, layer, stringsAsFactors=FALSE, encoding="CP1250")
charToRaw(pt2$NAZEV[1])
@
<<echo=FALSE,eval=TRUE>>=
pt2cr_1
@
<<echo=TRUE,eval=FALSE>>=
setCPLConfigOption("SHAPE_ENCODING", "")
pt3 <- readOGR(dsn, layer, stringsAsFactors=FALSE)
setCPLConfigOption("SHAPE_ENCODING", NULL)
pt1i_1 <- iconv(pt3$NAZEV[1], from="CP1250", to="UTF-8")
charToRaw(pt1i_1)
@
<<echo=FALSE,eval=TRUE>>=
pt1icr_1
@

\noindent
The three methods are either to set the CPL configure option directly import the data, and then unset it; to use the \code{encoding=} argument to \code{readOGR} which sets and unsets the CPL configure option internally; or to turn off GDAL recoding, and use \code{iconv} inside \RR. The final approach would need extra care if the field names also need recoding, and the string data are stored in the \RR object in their original charset. This, however, may be desirable if the same object is to be exported back to, for example, ArcGIS. A fourth method is to turn off GDAL recoding and use iconv within \RR inside \code{readOGR} to recode from the given encoding to the session charset, here UTF-8:

<<echo=TRUE,eval=FALSE>>=
setCPLConfigOption("SHAPE_ENCODING", "")
pt4 <- readOGR(dsn, layer, stringsAsFactors=FALSE, use_iconv=TRUE,
 encoding="CP1250")
setCPLConfigOption("SHAPE_ENCODING", NULL)
charToRaw(pt4$NAZEV[1])
@
<<echo=FALSE,eval=TRUE>>=
pt1icr_1
@

\noindent
Note that \code{ogrInfo} also takes \code{encoding=} and \code{use\_iconv=} arguments. If we do not set the encoding, GDAL recoding is from CP1252, not CP1250, and is wrongly rendered in UTF-8:

<<echo=TRUE,eval=FALSE>>=
pt5 <- readOGR(dsn, layer, stringsAsFactors=FALSE)
charToRaw(pt5$NAZEV[1])
@
<<echo=FALSE,eval=TRUE>>=
ptcr_1
@
<<echo=TRUE,eval=FALSE>>=
all.equal(charToRaw(pt5$NAZEV[1]), charToRaw(pt4$NAZEV[1]))
@
<<echo=FALSE,eval=TRUE>>=
all.equal(ptcr_1, pt1icr_1)
@

\section{GDAL without iconv}

Next we turn to the GDAL without iconv case. The data displayed here are pre-stored from this system:

<<echo=FALSE,eval=TRUE>>=
load(paste(dsn, .Platform$file.sep, "point_WinCRAN.RData", sep=""))
@
<<echo=TRUE,eval=FALSE>>=
sessionInfo()
@
<<echo=FALSE,eval=TRUE>>=
sI
@
<<echo=TRUE,eval=FALSE>>=
unlist(l10n_info())
@
<<echo=FALSE,eval=TRUE>>=
unlist(l10n)
@
\noindent
This system has \pkg{rgdal} installed using the CRAN binary. Here, we can check that iconv is available in GDAL:

<<echo=TRUE,eval=FALSE>>=
.Call("RGDAL_CPL_RECODE_ICONV", PACKAGE="rgdal")
@
<<echo=FALSE,eval=TRUE>>=
cpliconv
@

\noindent
This means that if GDAL recoding is not turned off, default non-iconv stub recoding will be used. It will often not be sensible to permit this to happen, so users may protect the raw data by either of these two methods:

<<echo=TRUE,eval=FALSE>>=
setCPLConfigOption("SHAPE_ENCODING", "")
pt6 <- readOGR(dsn, layer, stringsAsFactors=FALSE)
setCPLConfigOption("SHAPE_ENCODING", NULL)
charToRaw(pt6$NAZEV[1])
@
<<echo=FALSE,eval=TRUE>>=
pt1cr
@
<<echo=TRUE,eval=FALSE>>=
pt7 <- readOGR(dsn, layer, stringsAsFactors=FALSE, encoding="")
charToRaw(pt7$NAZEV[1])
@
<<echo=FALSE,eval=TRUE>>=
pt1cr
@

The assumption here is that there may not be a suitable recoding matching the imported strings and the \RR console; users will need to experiment to see how \code{iconv} may be used within \RR.

\end{document}

---
title: "Why have CRS, projections and transformations changed?"
author: "Roger Bivand"
output:
  html_document:
    toc: true
    toc_float:
      collapsed: false
      smooth_scroll: false
    toc_depth: 2
bibliography: PROJ.bib
link-citations: yes
vignette: >
  %\VignetteIndexEntry{CRS, projections and transformations}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

# Introduction

Changes in the representation of coordinate reference systems (CRS), and of operations on coordinates, that have been occurring over decades must now be implemented in the way spatial objects are handled in R packages. Up to the 1990s, most spatial data simply used the coordinates given by the local mapping authority; for example, the  Meuse bank data set used a planar representation in metres, which turned out to be `EPSG:28992`, "Amersfoort / RD New". A major resource for finding out why CRS were specified as they were are [Clifford J. Mugnier's columns](https://www.asprs.org/asprs-publications/grids-and-datums) in *Photogrammetric Engineering & Remote Sensing*, references to which are available in **rgdal**; the Netherlands were covered in the February 2003 column:

```{r}
td <- tempfile()
dir.create(td)
Sys.setenv("PROJ_USER_WRITABLE_DIRECTORY"=td)
library(rgdal)
data("GridsDatums")
GridsDatums[grep("Netherlands", GridsDatums$country),]
```


While most national mapping agencies defined their own standard geographical and projected CRS, supranational bodies, such as military alliances and colonial administrations often imposed some regularity to facilitate operations across national boundaries. This also led to the creation of the European Petroleum Survey Group (EPSG), because maritime jurisdiction was not orderly, and mattered when countries sharing the same coastal shelf tried to assign conflicting exploration concessions. Experts from oil companies accumulated vast experience, which fed through to the International Standards Organization (ISO, especially TC 211) and the Open Geospatial Consortium (OGC).

Defining the CRS became necessary when integrating other data with a different CRS, and for displaying on a web map background. Many legacy file formats, such as the ESRI Shapefile format, did not mandate the inclusion of the CRS of positional data. Most open source software then used PROJ.4 strings as a flexible representation, but as internationally accepted standards have been reached, in particular ISO 19111, and improved over time by iteration, it is really necessary to change to a modern text representation, known as WKT2 (2019). Now it looks as though almost all corporations and mapping agencies accommodate this representation, and it has been adopted by **sp** through **rgdal**, **sf** and other packages.

```{r, echo=FALSE}
mvrun <- FALSE
#if (require(mapview, quietly=TRUE) && .Platform$OS.type == "unix" && require(curl, quietly=TRUE) && curl::has_internet()) mvrun <- TRUE
```

```{r, eval=FALSE}
demo(meuse, ask=FALSE, package="sp", echo=FALSE)
library(mapview)
x <- mapview(meuse, zcol="zinc")
mapshot(x, file="meuse.png")
```

```{r}
knitr::include_graphics(system.file("misc/meuse.png", package="rgdal"))
```


# Coordinate reference system representation

The **mapproj** package provided coordinate reference system and projection support for the **maps** package. From `mapproj/src/map.h`, line 20, we can see that the eccentricity of the Earth is defined as `0.08227185422`, corrresponding to the Clark 1866 ellipsoid [@iliffe]:

```{r, echo=FALSE}
odd_run <- FALSE
if (new_proj_and_gdal()) odd_run <- TRUE
```


```{r, eval=odd_run}
ellps <- projInfo("ellps")
(clrk66 <- ellps[ellps$name=="clrk66",])
```

```{r, eval=odd_run}
#eval(parse(text=clrk66$major))
#eval(parse(text=clrk66$ell))
a <- 6378206.4
b <- 6356583.8
print(sqrt((a^2-b^2)/a^2), digits=10)
```

With a very few exceptions, projections included in `mapproj::mapproject()` use the Clarke 1866 ellipsoid, with the remainder using a sphere with the Clarke 1866 major axis radius. The function returns coordinates for visualization in an unknown metric; no inverse projections are available.

Like many other free and open source software projects around 2000, R spatial development chose to use the best open source library and infrastructure then available, PROJ.4. Version 4.4 was published by Frank Warmerdam in 2000, based on Gerald Evenden's earlier work. This earlier work was a library for forward and inverse projection using a key-value string interface to describe the required projection [@evenden:90]. The key-value string is taken as `+key=value`, where `=value` could be omitted for some keys, and the definition of each projection is built up from space-separated key-value string, such as `+proj=utm +zone=25 +south` for Universal Transverse Mercator zone 25 in the southern hemisphere. 

## PROJ.4 2000-2018

Unlike **mapproj**, PROJ.4 had begun to introduce the `+ellps=` key in addition to projection parameters before PROJ.4.4, as some users would not treat Clark 1866 as their natural preference. The need for more care in geodetic specification became pressing from 2000, when civilian use of GPS became as accurate as military use; GPS was typically registered to a more modern geographical CRS than digitized printed maps had been, and most mapping agencies scrambled to update their products and services to "GPS coordinates". 

The `ellps=` key was followed by the `+datum=`, `+nadgrids=` and `+towgs84=` keys in successive releases to attempt to specify the geodetic model. The `+init=` key appeared to permit the look-up of sets of key-value strings in the packaged version of a given table with a known authority, typically `+init=epsg:<code>`, where `<code>` was the EPSG code of the coordinate reference system. Where `+towgs84=` was given, a three or seven parameter transformation to the WGS84 datum was provided as a comma-separated string, so that the coordinate reference system also included the inverse coordinate operation from projected coordinates to geographical coordinates defined by the WGS84 datum. This led to the need for a placeholder for geographical coordinates, set as `+proj=longlat` (or `lonlat`, or perhaps with reversed axis order `latlong` or `latlon`). Some of the issues involved were discussed in a [2010 blog post by Frank Warmerdam](http://fwarmerdam.blogspot.com/2010/03/in-last-few-weeks-i-believe-i-have-made.html).

The PROJ.4 framework functioned well for projection before it was expected to handle datum tranformation too. Within the remit of single mapping agencies, some adaptation could still provide help, say  using `+nadgrids=` in parts of North America (NAD, North American Datum), but where positional data from multiple agencies was being integrated, the framework was showing its age. For example, from about 2010 it was observed that the `+datum=` and `+towgs84=` keys sometimes provided contradictory values, leading functions in GDAL reading raster files to prefer `+towgs84=` values  to `+datum=` values. For users for whom accuracy of better than about 150m was irrelevant, using coordinate reference systems correctly defined in terms of the underlying geographical coordinates was less important, but this is still about five Landsat cells.

While the representation of coordinate reference systems (sometimes supplemented by coordinate operations to transform the underlying geographical coordinates to WGS84) as PROJ.4 strings continued to work adequately, changes were occurring. Many file formats chose to use WKT (well-known text) string representations, starting from the 2007 edition of the ISO standard [@iso19111]. This placed the file reading and writing functions offered by GDAL under stress, especially the `exportToProj4()` function, as an increasing number of specification components really did not map adequately from the WKT string representation to the PROJ.4 string representation. Another change was that pivoting through a chosen hub when transforming coordinates (in PROJ.4 WGS84) meant that accuracy was lost transforming from the source to the hub, and more was lost from the hub to the target. Why not transform from source to target in one step if possible?

## PR$\phi$J

Signalling changes, PROJ.4 changed its name to PR$\phi$J, and began burning through major version numbers. PROJ 5 (2018) introduced transformation pipelines [@knudsen+evers17; @evers+knudsen17], representing coordinate operations using a syntax similar to PROJ.4 strings, but showing the whole operation pipeline. PROJ 6 (2019) followed up by shifting from ad-hoc text files for holding coordinate reference system and coordinate operation metadata to an SQLite database. In an increasing number of cases, more accurate coordinate operations could be supported using open access transformation grids, and the grid files needed were now tabulated in the SQLite database. This database is distributed with PROJ, and kept in a directory on the PROJ search path, usually the only or final directory (`get_proj_search_paths()` returns a vector of current search paths). 
```{r, eval=odd_run}
(shpr <- get_proj_search_paths())
```

```{r, echo=FALSE, results='hide', eval=odd_run}
if (is.null(shpr)) odd_run <- FALSE
```

```{r, echo=FALSE, results='hide'}
run <- FALSE
if (require("RSQLite", quietly=TRUE)) run <- TRUE
```

```{r, eval=run && odd_run}
library(RSQLite)
db <- dbConnect(SQLite(), dbname=file.path(shpr[length(shpr)], "proj.db"))
dbListTables(db)
```

```{r, eval=run && odd_run}
(metadata <- dbReadTable(db, "metadata"))
```


PROJ 7 (2020) reconfigured the transformation grids, now using the Geodetic TIFF Grid (GTG) format, and created pathways for on-demand download (typically using a content download network (CDN) over CURL) of chunks of such grids to a local user-writable cache held in another SQLite database. After little change from the late 1990s to early 2018, PROJ.4 has incremented its major version number three times in three years, and by 2021 (PROJ 8), the pre-existing application programming interface will be history. In addition, GDAL 3 (2019) has tightened its links with PROJ >= 6, and `exportToProj4()` now says: 

> "Use of this function is discouraged. Its behavior in GDAL >= 3 / PROJ >= 6 is significantly different from earlier versions. In particular +datum will only encode WGS84, NAD27 and NAD83, and +towgs84/+nadgrids terms will be missing most of the time. PROJ strings to encode CRS should be considered as a a legacy solution. Using a AUTHORITY:CODE or WKT representation is the recommended way" (https://gdal.org/api/ogrspatialref.html).

For these reasons, **sf**  and **sp** are changing from PROJ.4 strings representing coordinate reference systems to WKT2:2019 strings, as described in [this r-spatial blog](https://r-spatial.org/r/2020/03/17/wkt.html). Most users who had been relying on file reading to set the coordinate reference systems of objects will not notice much difference, and legacy PROJ.4 strings can still be used to create new, authority-free definitions if need be. 

It will be useful to know that in general `"OGC:CRS84"` should be used instead of `"EPSG:4326"`, because the latter presupposes that Latitude is always ordered before Longitude. `"OGC:CRS84"` is the standard representation used by GeoJSON, with coordinates always ordered Longitude before Latitude. This is also the standard GIS and visualization order:


```{r, warning=TRUE, eval=odd_run}
cat(wkt(CRS(SRS_string="OGC:CRS84")), "\n")
```

Note that all `"GEOGCRS"` `"BBOX"` CRS bounding boxes are always specified latitude minimum, longitude minimum, latitude maximum, longitude maximum, even though the declared axis order is the opposite.

# Coordinate operations

The introduction of interactive mapping using **mapview** and **tmap** among other packages highlights the need to set the coordinate reference system (CRS) of objects correctly, so that zooming does not reveal embarrassing divergences. Using the location of the Broad Street pump disabled by Dr John Snow to stop a cholera epidemic in Soho, London, in 1854 [@brodyetal:00], we can start to see what steps are being taken. The point location of the pump is given in projected coordinates, defined in the British National Grid. The workflow used by `mapview::mapview()` is to transform first to WGS84 (OGC:CRS84) using `sf::st_transform()` if need be, before permitting **leaflet** to project to Web Mercator (EPSG:3857) internally.

```{r, warning=TRUE, eval=odd_run}
b_pump <- readOGR(system.file("vectors/b_pump.gpkg", package="rgdal"))
```
Reading the file loses the PROJ.4 `+datum=` key-value pair, but the WKT2:2019 string is complete:

```{r, warning=TRUE, eval=odd_run}
proj4string(b_pump)
```
```{r, eval=odd_run}
if (packageVersion("sp") > "1.4.1") {
  WKT <- wkt(b_pump)
} else {
  WKT <- comment(slot(b_pump, "proj4string"))
}
cat(WKT, "\n")
```

## Pipelines

PROJ.4 assumed that the from/source and to/target coordinate reference system definitions involved in coordinate operations each contained the specifications necessary to get from source to WGS84 and then on from WGS84 to target. PR$\phi$J drops this assumption, searching among many candidate coordinate operations for viable pipelines. The search is conducted using the tables given in the `proj.db` SQLite database, which is now backed by authorities, and regularly updated at each release to the current upstream state (see https://proj.org/operations/operations_computation.html for more details). The tables are searched to find lists of candidates.

```{r, eval=run && odd_run}
cov <- dbReadTable(db, "coordinate_operation_view")
cov[grep("OSGB", cov$name), c("table_name", "code", "name", "method_name", "accuracy")]
```

The same search can be conducted directly without using **RSQLite** to query the database tables, searching by source and target CRS, and in the near future also by area of interest. If we search using only the degraded PROJ.4 string, we only find a ballpark accuracy coordinate operation, yielding a pipeline with two steps, inverse projection to geographical coordinates in radians, and conversion from radians to degrees. Note that `rgdal::spTransform()` and its wrapper `sp::spTransform()` use PROJ for coordinate operations. Here we will use `"EPSG:4326"` briefly for exposition:

```{r, eval=odd_run}
list_coordOps(paste0(proj4string(b_pump), " +type=crs"), "EPSG:4326")
```

The description component "+ axis order change (2D)" refers to the `EPSG:4326` definition, which specifies Northings/Latitude as the first axis, and Eastings/Longitude as the second axis; in **sp**/**rgdal** workflows, it is assumed that GIS/visualization order with Eastings/Longitude as the first 2D axis and Northings/Latitude as the second axis is preferred. Because the input data are in GIS/visualization order already, the steps to swap axes to standards conformity and then back to GIS/visualization order cancel each other out.

We can see what is happening if we unset enforcing GIS/visualization order, with an `axisswap` coming into play:

```{r, eval=odd_run}
set_enforce_xy(FALSE)
list_coordOps(paste0(proj4string(b_pump), " +type=crs"), "EPSG:4326")
set_enforce_xy(TRUE)
```

While **rgdal** tries to impose GIS/visualization axis order on the `"EPSG:4326"` specification, one may end up with a WKT2 string with Latitude first, followed by Longitude without wishing to do so. Using `"OGC:CRS84"` is more secure, because it does not require such ad-hoc re-specification.

Setting the internal control option `set_transform_wkt_comment()` to `FALSE`, we use only the degraded PROJ.4 string when transforming. `spTransform()` undertakes the same search, chooses the best instantiable coordinate operation on its first pass, then uses that pipeline on all objects. The pipeline specification of the coordinate operation may be retrieved using `get_last_coordOp()`:

```{r, eval=odd_run}
set_transform_wkt_comment(FALSE)
isballpark <- spTransform(b_pump, CRS(SRS_string="OGC:CRS84"))
get_last_coordOp()
```

The coordinate returned is unfortunately in Ingestre Place, not Broad Street.

```{r, eval=odd_run}
print(coordinates(isballpark), digits=10)
```

Let us repeat the search using the WKT2 string; here we see that providing a well-specified CRS representation allows us to choose 2m accuracy for the coordinate operation. Further, we can also see that, had we had access to a named grid, we could have achieved 1m accuracy:


The Helmert transformation has parameters retrieved from the PROJ SQLite database (code 1314):


```{r, eval=run && odd_run}
helm <- dbReadTable(db, "helmert_transformation_table")
helm[helm$code == "1314", c("auth_name", "code", "name", "tx", "ty", "tz", "rx", "ry", "rz", "scale_difference")]
dbDisconnect(db)
```


Using the WKT2 CRS representation, we can achieve 2m accuracy (or as the table says in other fields: "Oil exploration. Accuracy better than 4m and generally better than 2m":

```{r, eval=odd_run}
set_transform_wkt_comment(TRUE)
is2m <- spTransform(b_pump, CRS(SRS_string="OGC:CRS84"))
get_last_coordOp()
```

The output point is close to the Broad Street pump:

```{r, eval=odd_run}
print(coordinates(is2m), digits=10)
```

It is over 100m West-North-West of the Ingestre Place position:

```{r, eval=odd_run}
c(spDists(isballpark, is2m)*1000)
```

```{r, echo=FALSE}
mrun <- FALSE
if (suppressPackageStartupMessages(require(maptools, quietly=TRUE))) mrun <- TRUE
```


```{r, eval=mrun && odd_run}
c(maptools::gzAzimuth(coordinates(isballpark), coordinates(is2m)))
```

This was about as good as one could get prior to PROJ 7 without downloading the missing grid file manually, and installing the downloaded file in a directory that would usually not be user-writable. The whole set of grids can be downloaded and installed manually for workgroups needing to be sure that the same grids are available to all users, as has been the case in the past as well.

The `rgdal::project()` uses the underlying geographical coordinate reference system, and does not transform, so using the degraded PROJ.4 string and WKT2 give the same output, and because the input is projected, we take the inverse:

```{r, eval=odd_run}
(a <- project(coordinates(b_pump), proj4string(b_pump), inv=TRUE, verbose=TRUE))
(b <- project(coordinates(b_pump), WKT, inv=TRUE))
```

The projected points only inverse project the projected coordinates using the specified projection

```{r, eval=odd_run}
all.equal(a, b)
c(spDists(coordinates(isballpark), a)*1000)
```

```{r, eval=odd_run}
list_coordOps(WKT, "OGC:CRS84")
```


### Areas of interest

The search for candidate coordinate operations may be expedited if the area of interest is provided. This is a vector of minumum longitude and latitude followed by maximum longitude and latitude, so we need to inverse project the bounding box of projected **sp** objects and coerce that matrix into the required order:

```{r, eval=run && odd_run}
if (is.projected(b_pump)) { 
  o <- project(t(unclass(bbox(b_pump))), wkt(b_pump), inv=TRUE)
} else {
  o <- t(unclass(bbox(b_pump)))
}
(aoi <- c(t(o + c(-0.1, +0.1))))
```
Without an area of interest, the coordinate operation look-up found 8 candidate operations, with a given area of interest, only 5 are found with the `strict_containment=` argument taking its default value of `FALSE`, so that the candidates found do not have to contain the area of interest strictly:

```{r, eval=run && odd_run}
nrow(list_coordOps(WKT, "OGC:CRS84", area_of_interest=aoi))
```
In this case, changing this argument to `TRUE` makes no further difference:

```{r, eval=run && odd_run}
nrow(list_coordOps(WKT, "OGC:CRS84", strict_containment=TRUE, area_of_interest=aoi))
```
Methods for projection and transformation use the area of interest implicit in the data object, controlled by an argument: `use_aoi`, default `TRUE`, for `+proj=ob_tran, `FALSE`.

## Transformation grid files

PROJ 7 introduced on-demand downloading of (chunks of) transformation grids from a content delivery network to a user-writable directory on the PROJ search path (usually the first path component). The status of the downloaded grids is stored in another SQLite database, `cache.db`. The PR$\phi$J user-writable CDN directory is set as soon as the internal search path is queried, and for most uses, the default value will allow all programs using PR$\phi$J such as R packages, QGIS, GRASS, etc., to access any downloaded grids. Grids are checked for staleness at regular intervals. This directory may be set to a non-default value with the `PROJ_USER_WRITABLE_DIRECTORY` environment variable *before* **rgdal** (and any other package using PR$\phi$J) is loaded and attached, from PR$\phi$J >= 7.1.0. The code used at the beginning of this vignette is repeated here for convenience:


```{r, eval=FALSE}
td <- tempfile()
dir.create(td)
Sys.setenv("PROJ_USER_WRITABLE_DIRECTORY"=td)
library(rgdal)
```

```{r, echo=FALSE, results='hide', eval=odd_run}
run <- run && (attr(getPROJ4VersionInfo(), "short") >= 710)
```

Let us check that **rgdal** is running with on-demand downloading disabled:

```{r, eval=run && odd_run}
is_proj_CDN_enabled()
```


We can enable on-demand download using a function that reports the value of the PR$\phi$J CDN user-writable directory; we see that with this setting, network download is enabled:

```{r, eval= run && odd_run}
enable_proj_CDN()
is_proj_CDN_enabled()
```

The reader will recall that the search path was stored in `shpr` above, the first element being the user-writable directory; at this point no SQLite database has been created:

```{r, eval= run && odd_run}
shpr[1]
try(file.size(file.path(shpr[1], "cache.db")))
```

When we then search for candidate coordinate operations, we see that the operation using an absent grid now sees that download is enabled, and proposes the 1m accuracy candidate, because the required grid can be downloaded (again using the area of interest):


```{r, eval= run && odd_run}
list_coordOps(WKT, "OGC:CRS84", area_of_interest=aoi)
```

On making the transformation, we may see that the coordinate operation takes longer than expected, because on first pass the grid is downloaded from the network:

```{r, eval= run && odd_run}
system.time(is1m <- spTransform(b_pump, CRS(SRS_string="OGC:CRS84")))
```

The coordinate operation used now specifies the grid in the pipeline:

```{r, eval= run && odd_run}
get_last_coordOp()
```

The coordinate values differ little from the 2m accuracy Helmert pipeline:

```{r, eval= run && odd_run}
print(coordinates(is1m), digits=10)
```

as we can see, the 1m accuracy point is 1.7m from the 2m accuracy point, just to the West:

```{r, eval= run && odd_run}
c(spDists(is2m, is1m)*1000)
```


```{r, eval= mrun && run && odd_run}
c(maptools::gzAzimuth(coordinates(is1m), coordinates(is2m)))
```

Now the SQLite grid cache database has been created and has grown in size:

```{r, eval=run && odd_run}
try(file.size(file.path(shpr[1], "cache.db")))
```

If we look in the SQLite database of downloaded grids, we see that the grid components that were downloaded. Here we have not yet used the area of interest to limit the number of chunks involved:

```{r, eval= run && odd_run}
library(RSQLite)
db <- dbConnect(SQLite(), dbname=file.path(shpr[1], "cache.db"))
(tbs <- dbListTables(db))
if (any("chunks" %in% tbs)) print(dbReadTable(db, "chunks"))
dbDisconnect(db)
```

Finally, we disable grid download to return to the status existing when **rgdal** was attached; the cached grids in this case, using an R temporary directory, will be discarded, but in usual workflows, grids are downloaded once, used often and updated seldom when the server versions change. 

```{r, eval= run && odd_run}
disable_proj_CDN()
is_proj_CDN_enabled()
```


The outcome positions are shown here; at zoom 18 we can see that the 1m accuracy green point matches the Open Street Map location of the pump very well:

```{r, eval=FALSE}
library(mapview)
x <- mapview(is2m, map.type="OpenStreetMap", legend=FALSE) + mapview(is1m, col.regions="green", legend=FALSE) + mapview(isballpark, col.regions="red", legend=FALSE)
mapshot(x, file="snow.png")
```

```{r}
knitr::include_graphics(system.file("misc/snow.png", package="rgdal"))
```

## Rebuilding CRS objects

Package authors of the almost 150 packages with **sp** objects with outdated `"CRS"` objects stored in `*.rda` or `*RData` files as data sets should update the `"proj4string"` slots of these objects and re-store. The `GGHB.IG` data set in the **CARBayesdata** package (version 2.1) can serve as an example.

```{r, eval=odd_run}
# library(CARBayesdata)
library(sp)
# data(GGHB.IG)
# orig <- slot(GGHB.IG, "proj4string")
(load(system.file("misc/GGHB.IG_CRS.rda", package="rgdal")))
orig
```
The original `"CRS"` object contains a `+datum=` key (deprecated in GDAL's `exportToProj4()` from GDAL 3) and a `+towgs84=` key, the handling of which has varied in different releases of PROJ >= 6 and GDAL 3. From **rgdal** 1.5-17 (development version), a PROJ function `proj_get_source_crs` is used by default, but may be avoided by setting the `sp::CRS()`  `get_source_if_boundcrs=` argument to FALSE. The function will return the source CRS of a `BOUNDCRS` object. In some versions of PROJ/GDAL, the inclusion of a `+towgs84=` key is taken as meaning that the user wishes to create an object with the given source CRS, but because a `+towgs84=` key is present, it is assumed that the target is `WGS 84` with the given transformation method. Proj4 strings with `+towgs84=` keys have no authority for the parameters given, and do not harmonise with the use of the EPSG database. 

To run the examples, we need the current development version of **sp** from `"rsbivand/sp"`:

```{r, eval=odd_run, warning=FALSE}
sp_version_ok <- length(grep("get_source_if_boundcrs", deparse(args(sp::CRS)))) > 0L
```

Running `sp::CRS()` takes the original Proj4 key-value pairs, and converts them into a WKT2 representation as well as returning an (unused) Proj4 string. This is kept in place for packages still expecting to see/use it, but all **rgdal** functions and methods use the WKT2 representation. Here, we do not try to handle the `BOUNDCRS` special case, something that leads to difficulties with some versions of PROJ later in workflows. The `BOUNDCRS` contains a `SOURCECRS` and a `TARGETCRS` (which also has swapped axes), while really all that was needed was the contents of the `SOURCECRS`:


```{r, eval=odd_run && sp_version_ok, warning=FALSE}
orig1 <- CRS(slot(orig, "projargs"), get_source_if_boundcrs=FALSE)
cat(wkt(orig1), "\n")
```

Using the default setting to remove the unintended `BOUNDCRS` representation, we can get what was (probably) intended (when the **sp** version is as was before this work-around was added, the following two chunks yield `BOUNDCRS` for some versions of PROJ/GDAL:

```{r, eval=odd_run, warning=FALSE}
orig1a <- CRS(slot(orig, "projargs"))
cat(wkt(orig1a), "\n")
```

The `sp::rebuild_CRS()` method for `"CRS"` objects (and thus that for `"Spatial"` objects) can be used to do the same as using `sp::CRS()` directly, with special treatment for a number of other objects inheriting from `"Spatial"`:

```{r, eval=odd_run, warning=FALSE}
orig1b <- rebuild_CRS(orig)
cat(wkt(orig1b), "\n")
```

In practice, `sp::rebuild_CRS()` methods may be used, but users should check the output for coherence, and the PROJ function for extracting `SOURCECRS` from `BOUNDCRS` avoided if it has unfortunate consequences.

### References

---
title: "Why have CRS, projections and transformations changed?"
author: "Roger Bivand"
output:
  html_document:
    toc: true
    toc_float:
      collapsed: false
      smooth_scroll: false
    toc_depth: 2
bibliography: PROJ.bib
link-citations: yes
vignette: >
  %\VignetteIndexEntry{CRS, projections and transformations}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

# Introduction

Changes in the representation of coordinate reference systems (CRS), and of operations on coordinates, that have been occurring over decades must now be implemented in the way spatial objects are handled in R packages. Up to the 1990s, most spatial data simply used the coordinates given by the local mapping authority; for example, the  Meuse bank data set used a planar representation in metres, which turned out to be `EPSG:28992`, "Amersfoort / RD New". A major resource for finding out why CRS were specified as they were are [Clifford J. Mugnier's columns](https://www.asprs.org/asprs-publications/grids-and-datums) in *Photogrammetric Engineering & Remote Sensing*, references to which are available in **rgdal**; the Netherlands were covered in the February 2003 column:

```{r}
td <- tempfile()
dir.create(td)
Sys.setenv("PROJ_USER_WRITABLE_DIRECTORY"=td)
library(rgdal)
data("GridsDatums")
GridsDatums[grep("Netherlands", GridsDatums$country),]
```


While most national mapping agencies defined their own standard geographical and projected CRS, supranational bodies, such as military alliances and colonial administrations often imposed some regularity to facilitate operations across national boundaries. This also led to the creation of the European Petroleum Survey Group (EPSG), because maritime jurisdiction was not orderly, and mattered when countries sharing the same coastal shelf tried to assign conflicting exploration concessions. Experts from oil companies accumulated vast experience, which fed through to the International Standards Organization (ISO, especially TC 211) and the Open Geospatial Consortium (OGC).

Defining the CRS became necessary when integrating other data with a different CRS, and for displaying on a web map background. Many legacy file formats, such as the ESRI Shapefile format, did not mandate the inclusion of the CRS of positional data. Most open source software then used PROJ.4 strings as a flexible representation, but as internationally accepted standards have been reached, in particular ISO 19111, and improved over time by iteration, it is really necessary to change to a modern text representation, known as WKT2 (2019). Now it looks as though almost all corporations and mapping agencies accommodate this representation, and it has been adopted by **sp** through **rgdal**, **sf** and other packages.

```{r, echo=FALSE}
mvrun <- FALSE
#if (require(mapview, quietly=TRUE) && .Platform$OS.type == "unix" && require(curl, quietly=TRUE) && curl::has_internet()) mvrun <- TRUE
```

```{r, eval=FALSE}
demo(meuse, ask=FALSE, package="sp", echo=FALSE)
library(mapview)
x <- mapview(meuse, zcol="zinc")
mapshot(x, file="meuse.png")
```

```{r}
knitr::include_graphics(system.file("misc/meuse.png", package="rgdal"))
```


# Coordinate reference system representation

The **mapproj** package provided coordinate reference system and projection support for the **maps** package. From `mapproj/src/map.h`, line 20, we can see that the eccentricity of the Earth is defined as `0.08227185422`, corrresponding to the Clark 1866 ellipsoid [@iliffe]:

```{r, echo=FALSE}
odd_run <- FALSE
if (new_proj_and_gdal()) odd_run <- TRUE
```


```{r, eval=odd_run}
ellps <- projInfo("ellps")
(clrk66 <- ellps[ellps$name=="clrk66",])
```

```{r, eval=odd_run}
#eval(parse(text=clrk66$major))
#eval(parse(text=clrk66$ell))
a <- 6378206.4
b <- 6356583.8
print(sqrt((a^2-b^2)/a^2), digits=10)
```

With a very few exceptions, projections included in `mapproj::mapproject()` use the Clarke 1866 ellipsoid, with the remainder using a sphere with the Clarke 1866 major axis radius. The function returns coordinates for visualization in an unknown metric; no inverse projections are available.

Like many other free and open source software projects around 2000, R spatial development chose to use the best open source library and infrastructure then available, PROJ.4. Version 4.4 was published by Frank Warmerdam in 2000, based on Gerald Evenden's earlier work. This earlier work was a library for forward and inverse projection using a key-value string interface to describe the required projection [@evenden:90]. The key-value string is taken as `+key=value`, where `=value` could be omitted for some keys, and the definition of each projection is built up from space-separated key-value string, such as `+proj=utm +zone=25 +south` for Universal Transverse Mercator zone 25 in the southern hemisphere. 

## PROJ.4 2000-2018

Unlike **mapproj**, PROJ.4 had begun to introduce the `+ellps=` key in addition to projection parameters before PROJ.4.4, as some users would not treat Clark 1866 as their natural preference. The need for more care in geodetic specification became pressing from 2000, when civilian use of GPS became as accurate as military use; GPS was typically registered to a more modern geographical CRS than digitized printed maps had been, and most mapping agencies scrambled to update their products and services to "GPS coordinates". 

The `ellps=` key was followed by the `+datum=`, `+nadgrids=` and `+towgs84=` keys in successive releases to attempt to specify the geodetic model. The `+init=` key appeared to permit the look-up of sets of key-value strings in the packaged version of a given table with a known authority, typically `+init=epsg:<code>`, where `<code>` was the EPSG code of the coordinate reference system. Where `+towgs84=` was given, a three or seven parameter transformation to the WGS84 datum was provided as a comma-separated string, so that the coordinate reference system also included the inverse coordinate operation from projected coordinates to geographical coordinates defined by the WGS84 datum. This led to the need for a placeholder for geographical coordinates, set as `+proj=longlat` (or `lonlat`, or perhaps with reversed axis order `latlong` or `latlon`). Some of the issues involved were discussed in a [2010 blog post by Frank Warmerdam](http://fwarmerdam.blogspot.com/2010/03/in-last-few-weeks-i-believe-i-have-made.html).

The PROJ.4 framework functioned well for projection before it was expected to handle datum tranformation too. Within the remit of single mapping agencies, some adaptation could still provide help, say  using `+nadgrids=` in parts of North America (NAD, North American Datum), but where positional data from multiple agencies was being integrated, the framework was showing its age. For example, from about 2010 it was observed that the `+datum=` and `+towgs84=` keys sometimes provided contradictory values, leading functions in GDAL reading raster files to prefer `+towgs84=` values  to `+datum=` values. For users for whom accuracy of better than about 150m was irrelevant, using coordinate reference systems correctly defined in terms of the underlying geographical coordinates was less important, but this is still about five Landsat cells.

While the representation of coordinate reference systems (sometimes supplemented by coordinate operations to transform the underlying geographical coordinates to WGS84) as PROJ.4 strings continued to work adequately, changes were occurring. Many file formats chose to use WKT (well-known text) string representations, starting from the 2007 edition of the ISO standard [@iso19111]. This placed the file reading and writing functions offered by GDAL under stress, especially the `exportToProj4()` function, as an increasing number of specification components really did not map adequately from the WKT string representation to the PROJ.4 string representation. Another change was that pivoting through a chosen hub when transforming coordinates (in PROJ.4 WGS84) meant that accuracy was lost transforming from the source to the hub, and more was lost from the hub to the target. Why not transform from source to target in one step if possible?

## PR$\phi$J

Signalling changes, PROJ.4 changed its name to PR$\phi$J, and began burning through major version numbers. PROJ 5 (2018) introduced transformation pipelines [@knudsen+evers17; @evers+knudsen17], representing coordinate operations using a syntax similar to PROJ.4 strings, but showing the whole operation pipeline. PROJ 6 (2019) followed up by shifting from ad-hoc text files for holding coordinate reference system and coordinate operation metadata to an SQLite database. In an increasing number of cases, more accurate coordinate operations could be supported using open access transformation grids, and the grid files needed were now tabulated in the SQLite database. This database is distributed with PROJ, and kept in a directory on the PROJ search path, usually the only or final directory (`get_proj_search_paths()` returns a vector of current search paths). 
```{r, eval=odd_run}
(shpr <- get_proj_search_paths())
```

```{r, echo=FALSE, results='hide', eval=odd_run}
if (is.null(shpr)) odd_run <- FALSE
```

```{r, echo=FALSE, results='hide'}
run <- FALSE
if (require("RSQLite", quietly=TRUE)) run <- TRUE
```

```{r, eval=run && odd_run}
library(RSQLite)
db <- dbConnect(SQLite(), dbname=file.path(shpr[length(shpr)], "proj.db"))
dbListTables(db)
```

```{r, eval=run && odd_run}
(metadata <- dbReadTable(db, "metadata"))
```


PROJ 7 (2020) reconfigured the transformation grids, now using the Geodetic TIFF Grid (GTG) format, and created pathways for on-demand download (typically using a content download network (CDN) over CURL) of chunks of such grids to a local user-writable cache held in another SQLite database. After little change from the late 1990s to early 2018, PROJ.4 has incremented its major version number three times in three years, and by 2021 (PROJ 8), the pre-existing application programming interface will be history. In addition, GDAL 3 (2019) has tightened its links with PROJ >= 6, and `exportToProj4()` now says: 

> "Use of this function is discouraged. Its behavior in GDAL >= 3 / PROJ >= 6 is significantly different from earlier versions. In particular +datum will only encode WGS84, NAD27 and NAD83, and +towgs84/+nadgrids terms will be missing most of the time. PROJ strings to encode CRS should be considered as a a legacy solution. Using a AUTHORITY:CODE or WKT representation is the recommended way" (https://gdal.org/api/ogrspatialref.html).

For these reasons, **sf**  and **sp** are changing from PROJ.4 strings representing coordinate reference systems to WKT2:2019 strings, as described in [this r-spatial blog](https://r-spatial.org/r/2020/03/17/wkt.html). Most users who had been relying on file reading to set the coordinate reference systems of objects will not notice much difference, and legacy PROJ.4 strings can still be used to create new, authority-free definitions if need be. 

It will be useful to know that in general `"OGC:CRS84"` should be used instead of `"EPSG:4326"`, because the latter presupposes that Latitude is always ordered before Longitude. `"OGC:CRS84"` is the standard representation used by GeoJSON, with coordinates always ordered Longitude before Latitude. This is also the standard GIS and visualization order:


```{r, warning=TRUE, eval=odd_run}
cat(wkt(CRS(SRS_string="OGC:CRS84")), "\n")
```

Note that all `"GEOGCRS"` `"BBOX"` CRS bounding boxes are always specified latitude minimum, longitude minimum, latitude maximum, longitude maximum, even though the declared axis order is the opposite.

# Coordinate operations

The introduction of interactive mapping using **mapview** and **tmap** among other packages highlights the need to set the coordinate reference system (CRS) of objects correctly, so that zooming does not reveal embarrassing divergences. Using the location of the Broad Street pump disabled by Dr John Snow to stop a cholera epidemic in Soho, London, in 1854 [@brodyetal:00], we can start to see what steps are being taken. The point location of the pump is given in projected coordinates, defined in the British National Grid. The workflow used by `mapview::mapview()` is to transform first to WGS84 (OGC:CRS84) using `sf::st_transform()` if need be, before permitting **leaflet** to project to Web Mercator (EPSG:3857) internally.

```{r, warning=TRUE, eval=odd_run}
b_pump <- readOGR(system.file("vectors/b_pump.gpkg", package="rgdal"))
```
Reading the file loses the PROJ.4 `+datum=` key-value pair, but the WKT2:2019 string is complete:

```{r, warning=TRUE, eval=odd_run}
proj4string(b_pump)
```
```{r, eval=odd_run}
if (packageVersion("sp") > "1.4.1") {
  WKT <- wkt(b_pump)
} else {
  WKT <- comment(slot(b_pump, "proj4string"))
}
cat(WKT, "\n")
```

## Pipelines

PROJ.4 assumed that the from/source and to/target coordinate reference system definitions involved in coordinate operations each contained the specifications necessary to get from source to WGS84 and then on from WGS84 to target. PR$\phi$J drops this assumption, searching among many candidate coordinate operations for viable pipelines. The search is conducted using the tables given in the `proj.db` SQLite database, which is now backed by authorities, and regularly updated at each release to the current upstream state (see https://proj.org/operations/operations_computation.html for more details). The tables are searched to find lists of candidates.

```{r, eval=run && odd_run}
cov <- dbReadTable(db, "coordinate_operation_view")
cov[grep("OSGB", cov$name), c("table_name", "code", "name", "method_name", "accuracy")]
```

The same search can be conducted directly without using **RSQLite** to query the database tables, searching by source and target CRS, and in the near future also by area of interest. If we search using only the degraded PROJ.4 string, we only find a ballpark accuracy coordinate operation, yielding a pipeline with two steps, inverse projection to geographical coordinates in radians, and conversion from radians to degrees. Note that `rgdal::spTransform()` and its wrapper `sp::spTransform()` use PROJ for coordinate operations. Here we will use `"EPSG:4326"` briefly for exposition:

```{r, eval=odd_run}
list_coordOps(paste0(proj4string(b_pump), " +type=crs"), "EPSG:4326")
```

The description component "+ axis order change (2D)" refers to the `EPSG:4326` definition, which specifies Northings/Latitude as the first axis, and Eastings/Longitude as the second axis; in **sp**/**rgdal** workflows, it is assumed that GIS/visualization order with Eastings/Longitude as the first 2D axis and Northings/Latitude as the second axis is preferred. Because the input data are in GIS/visualization order already, the steps to swap axes to standards conformity and then back to GIS/visualization order cancel each other out.

We can see what is happening if we unset enforcing GIS/visualization order, with an `axisswap` coming into play:

```{r, eval=odd_run}
set_enforce_xy(FALSE)
list_coordOps(paste0(proj4string(b_pump), " +type=crs"), "EPSG:4326")
set_enforce_xy(TRUE)
```

While **rgdal** tries to impose GIS/visualization axis order on the `"EPSG:4326"` specification, one may end up with a WKT2 string with Latitude first, followed by Longitude without wishing to do so. Using `"OGC:CRS84"` is more secure, because it does not require such ad-hoc re-specification.

Setting the internal control option `set_transform_wkt_comment()` to `FALSE`, we use only the degraded PROJ.4 string when transforming. `spTransform()` undertakes the same search, chooses the best instantiable coordinate operation on its first pass, then uses that pipeline on all objects. The pipeline specification of the coordinate operation may be retrieved using `get_last_coordOp()`:

```{r, eval=odd_run}
set_transform_wkt_comment(FALSE)
isballpark <- spTransform(b_pump, CRS(SRS_string="OGC:CRS84"))
get_last_coordOp()
```

The coordinate returned is unfortunately in Ingestre Place, not Broad Street.

```{r, eval=odd_run}
print(coordinates(isballpark), digits=10)
```

Let us repeat the search using the WKT2 string; here we see that providing a well-specified CRS representation allows us to choose 2m accuracy for the coordinate operation. Further, we can also see that, had we had access to a named grid, we could have achieved 1m accuracy:


The Helmert transformation has parameters retrieved from the PROJ SQLite database (code 1314):


```{r, eval=run && odd_run}
helm <- dbReadTable(db, "helmert_transformation_table")
helm[helm$code == "1314", c("auth_name", "code", "name", "tx", "ty", "tz", "rx", "ry", "rz", "scale_difference")]
dbDisconnect(db)
```


Using the WKT2 CRS representation, we can achieve 2m accuracy (or as the table says in other fields: "Oil exploration. Accuracy better than 4m and generally better than 2m":

```{r, eval=odd_run}
set_transform_wkt_comment(TRUE)
is2m <- spTransform(b_pump, CRS(SRS_string="OGC:CRS84"))
get_last_coordOp()
```

The output point is close to the Broad Street pump:

```{r, eval=odd_run}
print(coordinates(is2m), digits=10)
```

It is over 100m West-North-West of the Ingestre Place position:

```{r, eval=odd_run}
c(spDists(isballpark, is2m)*1000)
```

```{r, echo=FALSE}
mrun <- FALSE
if (suppressPackageStartupMessages(require(maptools, quietly=TRUE))) mrun <- TRUE
```


```{r, eval=mrun && odd_run}
c(maptools::gzAzimuth(coordinates(isballpark), coordinates(is2m)))
```

This was about as good as one could get prior to PROJ 7 without downloading the missing grid file manually, and installing the downloaded file in a directory that would usually not be user-writable. The whole set of grids can be downloaded and installed manually for workgroups needing to be sure that the same grids are available to all users, as has been the case in the past as well.

The `rgdal::project()` uses the underlying geographical coordinate reference system, and does not transform, so using the degraded PROJ.4 string and WKT2 give the same output, and because the input is projected, we take the inverse:

```{r, eval=odd_run}
(a <- project(coordinates(b_pump), proj4string(b_pump), inv=TRUE, verbose=TRUE))
(b <- project(coordinates(b_pump), WKT, inv=TRUE))
```

The projected points only inverse project the projected coordinates using the specified projection

```{r, eval=odd_run}
all.equal(a, b)
c(spDists(coordinates(isballpark), a)*1000)
```

```{r, eval=odd_run}
list_coordOps(WKT, "OGC:CRS84")
```


### Areas of interest

The search for candidate coordinate operations may be expedited if the area of interest is provided. This is a vector of minumum longitude and latitude followed by maximum longitude and latitude, so we need to inverse project the bounding box of projected **sp** objects and coerce that matrix into the required order:

```{r, eval=run && odd_run}
if (is.projected(b_pump)) { 
  o <- project(t(unclass(bbox(b_pump))), wkt(b_pump), inv=TRUE)
} else {
  o <- t(unclass(bbox(b_pump)))
}
(aoi <- c(t(o + c(-0.1, +0.1))))
```
Without an area of interest, the coordinate operation look-up found 8 candidate operations, with a given area of interest, only 5 are found with the `strict_containment=` argument taking its default value of `FALSE`, so that the candidates found do not have to contain the area of interest strictly:

```{r, eval=run && odd_run}
nrow(list_coordOps(WKT, "OGC:CRS84", area_of_interest=aoi))
```
In this case, changing this argument to `TRUE` makes no further difference:

```{r, eval=run && odd_run}
nrow(list_coordOps(WKT, "OGC:CRS84", strict_containment=TRUE, area_of_interest=aoi))
```
Methods for projection and transformation use the area of interest implicit in the data object, controlled by an argument: `use_aoi`, default `TRUE`, for `+proj=ob_tran, `FALSE`.

## Transformation grid files

PROJ 7 introduced on-demand downloading of (chunks of) transformation grids from a content delivery network to a user-writable directory on the PROJ search path (usually the first path component). The status of the downloaded grids is stored in another SQLite database, `cache.db`. The PR$\phi$J user-writable CDN directory is set as soon as the internal search path is queried, and for most uses, the default value will allow all programs using PR$\phi$J such as R packages, QGIS, GRASS, etc., to access any downloaded grids. Grids are checked for staleness at regular intervals. This directory may be set to a non-default value with the `PROJ_USER_WRITABLE_DIRECTORY` environment variable *before* **rgdal** (and any other package using PR$\phi$J) is loaded and attached, from PR$\phi$J >= 7.1.0. The code used at the beginning of this vignette is repeated here for convenience:


```{r, eval=FALSE}
td <- tempfile()
dir.create(td)
Sys.setenv("PROJ_USER_WRITABLE_DIRECTORY"=td)
library(rgdal)
```

```{r, echo=FALSE, results='hide', eval=odd_run}
run <- run && (attr(getPROJ4VersionInfo(), "short") >= 710)
```

Let us check that **rgdal** is running with on-demand downloading disabled:

```{r, eval=run && odd_run}
is_proj_CDN_enabled()
```


We can enable on-demand download using a function that reports the value of the PR$\phi$J CDN user-writable directory; we see that with this setting, network download is enabled:

```{r, eval= run && odd_run}
enable_proj_CDN()
is_proj_CDN_enabled()
```

The reader will recall that the search path was stored in `shpr` above, the first element being the user-writable directory; at this point no SQLite database has been created:

```{r, eval= run && odd_run}
shpr[1]
try(file.size(file.path(shpr[1], "cache.db")))
```

When we then search for candidate coordinate operations, we see that the operation using an absent grid now sees that download is enabled, and proposes the 1m accuracy candidate, because the required grid can be downloaded (again using the area of interest):


```{r, eval= run && odd_run}
list_coordOps(WKT, "OGC:CRS84", area_of_interest=aoi)
```

On making the transformation, we may see that the coordinate operation takes longer than expected, because on first pass the grid is downloaded from the network:

```{r, eval= run && odd_run}
system.time(is1m <- spTransform(b_pump, CRS(SRS_string="OGC:CRS84")))
```

The coordinate operation used now specifies the grid in the pipeline:

```{r, eval= run && odd_run}
get_last_coordOp()
```

The coordinate values differ little from the 2m accuracy Helmert pipeline:

```{r, eval= run && odd_run}
print(coordinates(is1m), digits=10)
```

as we can see, the 1m accuracy point is 1.7m from the 2m accuracy point, just to the West:

```{r, eval= run && odd_run}
c(spDists(is2m, is1m)*1000)
```


```{r, eval= mrun && run && odd_run}
c(maptools::gzAzimuth(coordinates(is1m), coordinates(is2m)))
```

Now the SQLite grid cache database has been created and has grown in size:

```{r, eval=run && odd_run}
try(file.size(file.path(shpr[1], "cache.db")))
```

If we look in the SQLite database of downloaded grids, we see that the grid components that were downloaded. Here we have not yet used the area of interest to limit the number of chunks involved:

```{r, eval= run && odd_run}
library(RSQLite)
db <- dbConnect(SQLite(), dbname=file.path(shpr[1], "cache.db"))
dbListTables(db)
dbReadTable(db, "chunks")
dbDisconnect(db)
```

Finally, we disable grid download to return to the status existing when **rgdal** was attached; the cached grids in this case, using an R temporary directory, will be discarded, but in usual workflows, grids are downloaded once, used often and updated seldom when the server versions change. 

```{r, eval= run && odd_run}
disable_proj_CDN()
is_proj_CDN_enabled()
```


The outcome positions are shown here; at zoom 18 we can see that the 1m accuracy green point matches the Open Street Map location of the pump very well:

```{r, eval=FALSE}
library(mapview)
x <- mapview(is2m, map.type="OpenStreetMap", legend=FALSE) + mapview(is1m, col.regions="green", legend=FALSE) + mapview(isballpark, col.regions="red", legend=FALSE)
mapshot(x, file="snow.png")
```

```{r}
knitr::include_graphics(system.file("misc/snow.png", package="rgdal"))
```

## Rebuilding CRS objects

Package authors of the almost 150 packages with **sp** objects with outdated `"CRS"` objects stored in `*.rda` or `*RData` files as data sets should update the `"proj4string"` slots of these objects and re-store. The `GGHB.IG` data set in the **CARBayesdata** package (version 2.1) can serve as an example.

```{r, eval=odd_run}
# library(CARBayesdata)
library(sp)
# data(GGHB.IG)
# orig <- slot(GGHB.IG, "proj4string")
(load(system.file("misc/GGHB.IG_CRS.rda", package="rgdal")))
orig
```
The original `"CRS"` object contains a `+datum=` key (deprecated in GDAL's `exportToProj4()` from GDAL 3) and a `+towgs84=` key, the handling of which has varied in different releases of PROJ >= 6 and GDAL 3. From **rgdal** 1.5-17 (development version), a PROJ function `proj_get_source_crs` is used by default, but may be avoided by setting the `sp::CRS()`  `get_source_if_boundcrs=` argument to FALSE. The function will return the source CRS of a `BOUNDCRS` object. In some versions of PROJ/GDAL, the inclusion of a `+towgs84=` key is taken as meaning that the user wishes to create an object with the given source CRS, but because a `+towgs84=` key is present, it is assumed that the target is `WGS 84` with the given transformation method. Proj4 strings with `+towgs84=` keys have no authority for the parameters given, and do not harmonise with the use of the EPSG database. 

To run the examples, we need the current development version of **sp** from `"rsbivand/sp"`:

```{r, eval=odd_run, warning=FALSE}
sp_version_ok <- length(grep("get_source_if_boundcrs", deparse(args(sp::CRS)))) > 0L
```

Running `sp::CRS()` takes the original Proj4 key-value pairs, and converts them into a WKT2 representation as well as returning an (unused) Proj4 string. This is kept in place for packages still expecting to see/use it, but all **rgdal** functions and methods use the WKT2 representation. Here, we do not try to handle the `BOUNDCRS` special case, something that leads to difficulties with some versions of PROJ later in workflows. The `BOUNDCRS` contains a `SOURCECRS` and a `TARGETCRS` (which also has swapped axes), while really all that was needed was the contents of the `SOURCECRS`:


```{r, eval=odd_run && sp_version_ok, warning=FALSE}
orig1 <- CRS(slot(orig, "projargs"), get_source_if_boundcrs=FALSE)
cat(wkt(orig1), "\n")
```

Using the default setting to remove the unintended `BOUNDCRS` representation, we can get what was (probably) intended (when the **sp** version is as was before this work-around was added, the following two chunks yield `BOUNDCRS` for some versions of PROJ/GDAL:

```{r, eval=odd_run, warning=FALSE}
orig1a <- CRS(slot(orig, "projargs"))
cat(wkt(orig1a), "\n")
```

The `sp::rebuild_CRS()` method for `"CRS"` objects (and thus that for `"Spatial"` objects) can be used to do the same as using `sp::CRS()` directly, with special treatment for a number of other objects inheriting from `"Spatial"`:

```{r, eval=odd_run, warning=FALSE}
orig1b <- rebuild_CRS(orig)
cat(wkt(orig1b), "\n")
```

In practice, `sp::rebuild_CRS()` methods may be used, but users should check the output for coherence, and the PROJ function for extracting `SOURCECRS` from `BOUNDCRS` avoided if it has unfortunate consequences.

### References

\name{CRS-class}
\docType{class}
\alias{CRS-class}
\alias{CRS}
\alias{CRSargs}
\alias{checkCRSArgs}
\alias{checkCRSArgs_ng}
\alias{print.CRS}
\alias{show,CRS-method}
\alias{RGDAL_checkCRSArgs}
\alias{compare_CRS}
\title{Class "CRS" of coordinate reference system arguments}
\description{Interface class to the PROJ.4 projection system. The class is defined as an empty stub accepting value NA in the sp package. If the rgdal package is available, then the class will permit spatial data to be associated with coordinate reference systems}
\section{Objects from the Class}{
Objects can be created by calls of the form \code{CRS("projargs")}, where "projargs" is a valid string of PROJ.4 arguments; the arguments must be entered exactly as in the PROJ.4 documentation, in particular there cannot be any white space in +<arg>=<value> strings, and successive such strings can only be separated by blanks. The initiation function calls the PROJ.4 library to verify the argument set against those known in the library, returning error messages where necessary. The complete argument set may be retrieved by examining the second list element returned by \code{validObject("CRS object")} to see which additional arguments the library will use (which assumptions it is making over and above submitted arguments). The function \code{CRSargs()} can be used to show the expanded argument list used by the PROJ.4 library.
}
\section{Slots}{
  \describe{
    \item{\code{projargs}:}{Object of class \code{"character"}: projection arguments; the arguments must be entered exactly as in the PROJ.4 documentation, in particular there cannot be any white space in +<arg>=<value> strings, and successive such strings can only be separated by blanks. }
  }
}

\section{Methods}{
  \describe{
    \item{show}{\code{signature(object = "CRS")}: print projection arguments in object }
  }
}

\usage{
checkCRSArgs(uprojargs)
checkCRSArgs_ng(uprojargs=NA_character_, SRS_string=NULL,
  get_source_if_boundcrs=TRUE)
compare_CRS(CRS1, CRS2)
}
\arguments{
  \item{uprojargs}{character string PROJ.4 projection arguments}
  \item{SRS_string}{default NULL, experimental in connection with adaptation to GDAL>=3/PROJ>=6; a valid WKT string or SRS definition such as \code{"OGC:CRS84"}}
  \item{get_source_if_boundcrs}{The presence of the \code{+towgs84=} key in a Proj4 string \code{projargs=} argument value may promote the output WKT2 CRS to BOUNDCRS for PROJ >= 6 and GDAL >= 3, which is a coordinate operation from the input datum to WGS84. This is often unfortunate, so a PROJ function is called through \pkg{rgdal} to retrieve the underlying source definition.}
  \item{CRS1, CRS2}{objects of class \code{"CRS"}}
}
\references{\url{https://proj.org/}}
\author{Roger Bivand \email{Roger.Bivand@nhh.no}}

\note{
Lists of projections may be seen by using the programs installed with the PROJ.4 library, in particular proj and cs2cs; with the latter, -lp lists projections, -le ellipsoids, -lu units, and -ld datum(s) known to the installed software (available in \pkg{rgdal} using \code{projInfo}). These are added to in successive releases, so tracking the website or compiling and installing the most recent revisions will give the greatest choice. Finding the very important datum transformation parameters to be given with the +towgs84 tag is a further challenge, and is essential when the datums used in data to be used together differ. Tracing projection arguments is easier now than before the mass ownership of GPS receivers raised the issue of matching coordinates from different argument sets (GPS output and paper map, for example). See \code{\link{GridsDatums}} and \code{\link{showEPSG}} for help in finding CRS definitions.

The 4.9.1 release of PROJ.4 omitted a small file of defaults, leading to reports of \dQuote{major axis or radius = 0 or not given} errors. From 0.9-3, rgdal checks for the presence of this file (proj_def.dat), and if not found, and under similar conditions to those used by PROJ.4, adds \dQuote{+ellps=WGS84} to the input string being checked by \code{checkCRSArgs} The \dQuote{+no_defs} tag ignores the file of defaults, and the default work-around implemented to get around this problem; strings including \dQuote{init} and \dQuote{datum} tags also trigger the avoidance of the work-around. Now messages are issued when a candidate CRS is checked; they may be suppressed using \code{suppressMessages}.}


\examples{
set_thin_PROJ6_warnings(TRUE)
CRSargs(CRS("+proj=longlat"))
try(CRS("+proj=longlat"))
CRSargs(CRS("+proj=longlat +datum=NAD27"))
CRSargs(CRS("+init=epsg:4267"))
CRSargs(CRS("+init=epsg:26978"))
CRSargs(CRS(paste("+proj=stere +lat_0=52.15616055555555", 
"+lon_0=5.38763888888889 +k=0.999908 +x_0=155000 +y_0=463000 +ellps=bessel", 
"+towgs84=565.237,50.0087,465.658,-0.406857,0.350733,-1.87035,4.0812", 
"+units=m")))
# see http://trac.osgeo.org/gdal/ticket/1987
CRSargs(CRS("+init=epsg:28992"))
crs <- CRS("+init=epsg:28992")
CRSargs(CRS(CRSargs(crs)))
library(sp)
data(meuse)
coordinates(meuse) <- c("x", "y")
proj4string(meuse) <- CRS("+init=epsg:28992")
CRSargs(CRS(proj4string(meuse)))
run <- new_proj_and_gdal()
if (run) {
c1 <- CRS(SRS_string="OGC:CRS84")
c2 <- CRS("+proj=longlat")
compare_CRS(c1, c2)
}
if (run) {
comment(c2) <- NULL
compare_CRS(c1, c2)
}
}
\keyword{classes}
\keyword{spatial}

\name{GridsDatums}
\alias{GridsDatums}
\docType{data}
\title{Grids and Datums PE&RS listing}
\description{A data.frame of years and months of Grids & Datums column publications by country and country code.}
\usage{data("GridsDatums")}
\format{
  A data frame with 241 observations on the following 4 variables.
  \describe{
    \item{\code{country}}{name of PE&RS column}
    \item{\code{month}}{issue month}
    \item{\code{year}}{publication year}
    \item{\code{ISO}}{ISO code for country}
  }
}
\details{The journal \emph{Photogrammetric Engineering & Remote Sensing}, run by the American Society for Photogrammetry and Remote Sensing (ASPRS), began publishing a more-or-less monthly column on the spatial reference systems used in different countries, including their datums. The column first appeared in September 1997, and continued until March 2016; subsequent columns are updated reprints of previous ones. Some also cover other topics, such as world and Martian spatial reference systems. They are written by Clifford J. Mugnier, Louisiana State University, Fellow Emeritus ASPRS. To access the columns, visit \url{https://www.asprs.org/asprs-publications/grids-and-datums}.}
\source{\url{https://www.asprs.org/asprs-publications/grids-and-datums}}
\examples{
data(GridsDatums)
GridsDatums[grep("Norway", GridsDatums$country),]
GridsDatums[grep("Google", GridsDatums$country),]
GridsDatums[grep("^Mars$", GridsDatums$country),]
}
\keyword{datasets}

\name{nor2k}
\alias{nor2k}
\docType{data}
\title{Norwegian peaks over 2000m}
\description{Norwegian peaks over 2000m, 3D SpatialPoints data.}
\usage{data(nor2k)}
\format{
  The format is:
Formal class 'SpatialPointsDataFrame' [package "sp"] with 5 slots
  ..@ data       :'data.frame':	300 obs. of  3 variables:
  .. ..$ Nr.    : int [1:300] 1 2 3 4 5 6 7 8 9 10 ...
  .. ..$ Navn   : chr [1:300] "Galdh?piggen" "Glittertinden" "Skagast?lstinden, Store (Storen)" "Styggedalstinden, Store, ?sttoppen" ...
  .. ..$ Kommune: chr [1:300] "Lom" "Lom" "Luster / Ardal" "Luster" ...
  ..@ coords.nrs : num(0) 
  ..@ coords     : num [1:300, 1:3] 463550 476550 439850 441450 441100 ...
  .. ..- attr(*, "dimnames")=List of 2
  .. .. ..$ : NULL
  .. .. ..$ : chr [1:3] "East" "North" "Height"
  ..@ bbox       : num [1:3, 1:2] 404700 6804200 2001 547250 6910050 ...
  .. ..- attr(*, "dimnames")=List of 2
  .. .. ..$ : chr [1:3] "East" "North" "Height"
  .. .. ..$ : chr [1:2] "min" "max"
  ..@ proj4string:Formal class 'CRS' [package "sp"] with 1 slots
  .. .. ..@ projargs: chr "+proj=utm +zone=32 +datum=WGS84 +ellps=WGS84 +towgs84=0,0,0"
}
\details{
Norwegian peaks over 2000m, coordinates in EUREF89/WGS84 UTM32N, names not fully updated, here converted to ASCII. 
}
\source{
\url{http://www.nfo2000m.no/}; \url{http://www.nfo2000m.no/Excel/2000m_data.xls}
}
\examples{
data(nor2k)
summary(nor2k)
## maybe str(nor2k) ; plot(nor2k) ...
}
\keyword{datasets}

\name{GDALMajorObject-class}
\docType{class}
\alias{GDALMajorObject-class}
\alias{getDescription}
\title{Class "GDALMajorObject" }

\description{"GDALMajorObject" is a virtual base class for all GDAL objects.}

\section{Objects from the Class}{
Objects can be created by calls of the form \code{new("GDALMajorObject", ...)}, but are only created for classes that extend this class.}

\section{Slots}{
  \describe{
    \item{\code{handle}:}{Object of class \code{"externalptr"}, used internally; not for public consumption}
  }
}

\section{Methods}{
No methods defined with class "GDALMajorObject" in the signature.
}
\usage{
getDescription(object)
}
\arguments{
  \item{object}{an object inheriting from "GDALMajorObject"}

}
\details{
  \describe{
    \item{\code{getDescription}:}{returns a descrption string associated
    with the object. No setter method is defined because GDAL dataset
    objects use the description to hold the filename attached to the
    dataset. It would not be good to change that mid-stream.}
  
  }
}

\references{\url{https://gdal.org/}}

\author{Timothy H. Keitt, modified by Roger Bivand}

\seealso{ \code{\link{GDALDriver-class}},
  \code{\link{GDALReadOnlyDataset-class}}, \code{\link{GDALDataset-class}} and
  \code{\link{GDALTransientDataset-class}} }

\examples{
driver <- new('GDALDriver', as.character(getGDALDriverNames()[1,1]))
driver
rm(driver)
logo <- system.file("pictures/logo.jpg", package="rgdal")[1]
x <- new("GDALReadOnlyDataset", logo)
x
getDescription(x)
dim(x)
GDAL.close(x)

}
\keyword{classes}

\name{project}
\alias{project}

\title{Projection of coordinate matrices}
\description{
  Interface to the PROJ.4 library of projection functions for geographical position data, no datum transformation possible. Use \code{spTransform()} for extended support.
}
\usage{
project(xy, proj, inv = FALSE, use_ob_tran=FALSE, legacy=TRUE,
 allowNAs_if_not_legacy=FALSE, coordOp = NULL, verbose = FALSE,
 use_aoi=TRUE)
}

\arguments{
  \item{xy}{ 2-column matrix of coordinates }
  \item{proj}{ character string of projection arguments; the arguments must be entered exactly as in the PROJ.4 documentation, in particular there cannot be any white space in +<arg>=<value> strings, and successive such strings can only be separated by blanks. }
  \item{inv}{ default FALSE, if TRUE inverse projection to geographical coordinates }
  \item{use_ob_tran}{default FALSE, if TRUE and \dQuote{+proj=ob_tran}, use General Oblique Transformation with internalised from/to projection reversal; the user oblique transforms forward rather than inverse.}
  \item{legacy}{default TRUE, if FALSE, use transform C functions (enforced internally for Windows 32-bit platforms)}
  \item{allowNAs_if_not_legacy}{used if legacy is FALSE, default FALSE; introduced to handle use of NAs as object separators in \pkg{oce}}
  \item{coordOp}{default NULL, for PROJ >= 6 used to pass through a pre-defined coordinate operation}
  \item{verbose}{default FALSE, for PROJ >=6 used to show the coordinate operation used}
  \item{use_aoi}{With PROJ >= 6, use the area of interest defined as the range of \code{xy} in limiting the search for candidate coordinate operations; set FALSE if \code{use_ob_tran} is TRUE}
}
\details{
  Full details of projection arguments available from website below, and examples in file "epsg" in the data directory installed with PROJ.4.

Note that from PROJ.4 4.9.3, the definition of UTM is changed from TMERC to ETMERC; see example.
}
\value{
  A two column matrix with projected coordinates.
}
\references{\url{https://proj.org/}}
\author{Barry Rowlingson, Roger Bivand \email{Roger.Bivand@nhh.no}}

\note{ The locations of Hawaii and Alaska in the data source are (putting it mildly) arbitrary, please avoid airlines using these positions.}

\seealso{ \code{\link[sp]{CRS-class}}, \code{\link{spTransform-methods}} }

\examples{
data(state)
res <- project(cbind(state.center$x, state.center$y),
 "+proj=lcc +lat_1=48 +lat_2=33 +lon_0=-100 +ellps=WGS84")
res1 <- project(res, "+proj=lcc +lat_1=48 +lat_2=33 +lon_0=-100 +ellps=WGS84",
 inv=TRUE)
summary(res1 - cbind(state.center$x, state.center$y))
plot(cbind(state.center$x, state.center$y), asp=1, type="n")
text(cbind(state.center$x, state.center$y), state.abb)
plot(res, asp=1, type="n")
text(res, state.abb)
broke_proj <- FALSE
pv <- .Call("PROJ4VersionInfo", PACKAGE="rgdal")[[2]]
# https://github.com/OSGeo/PROJ/issues/1525
if (pv >= 600 && pv < 620) broke_proj <- TRUE
if (!broke_proj) {
crds <- matrix(data=c(9.05, 48.52), ncol=2)
a <- project(crds, paste("+proj=ob_tran +o_proj=longlat",
 "+o_lon_p=-162 +o_lat_p=39.25 +lon_0=180 +ellps=sphere +no_defs"),
 use_ob_tran=TRUE)
a
#should be (-5.917698, -1.87195)
project(a, paste("+proj=ob_tran +o_proj=longlat",
 "+o_lon_p=-162 +o_lat_p=39.25 +lon_0=180 +ellps=sphere +no_defs"),
 inv=TRUE, use_ob_tran=TRUE)
#added after posting by Martin Ivanov
}
#
getPROJ4VersionInfo()
# Test for UTM == TMERC (<= 4.9.2) or UTM == ETMERC (> 4.9.2)
nhh <- matrix(c(5.304234, 60.422311), ncol=2)
nhh_utm_32N_P4 <- project(nhh, "+init=epsg:3044")
nhh_tmerc_P4 <- project(nhh, paste("+proj=tmerc +k=0.9996 +lon_0=9",
 "+x_0=500000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs"))
nhh_etmerc_P4 <- project(nhh, paste("+proj=etmerc +k=0.9996 +lon_0=9",
 "+x_0=500000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs"))
all.equal(nhh_utm_32N_P4, nhh_tmerc_P4, tolerance=1e-9, scale=1)
# UTM == TMERC: PROJ4 <=4.9.2
all.equal(nhh_utm_32N_P4, nhh_etmerc_P4, tolerance=1e-9, scale=1)
# UTM == ETMERC: PROJ4 > 4.9.2
unis <- matrix(c(15.653453, 78.222504), ncol=2)
unis_utm_33N_P4 <- project(unis, "+init=epsg:3045")
unis_tmerc_P4 <- project(unis, paste("+proj=tmerc +k=0.9996 +lon_0=15",
 "+x_0=500000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs"))
unis_etmerc_P4 <- project(unis, paste("+proj=etmerc +k=0.9996 +lon_0=15",
 "+x_0=500000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs"))
all.equal(unis_utm_33N_P4, unis_tmerc_P4, tolerance=1e-9, scale=1)
# UTM == TMERC: PROJ4 <=4.9.2
all.equal(unis_utm_33N_P4, unis_etmerc_P4, tolerance=1e-9, scale=1)
# UTM == ETMERC: PROJ4 > 4.9.2
#pv <- attr(getPROJ4VersionInfo(), "short")
#if (pv < 500) {  
# valgrind leakages in some cases for PROJ >= 5; many non-projection proj values added
# available projections and their inverses if provided
# For >=4.9.3 returns non-finite points rather than needing crash protection
projs <- as.character(projInfo()$name)
res <- logical(length(projs))
names(res) <- projs
msgs <- character(length(projs))
names(msgs) <- projs
owarn <- options("warn")$warn
options(warn=2L)
for (i in seq(along=res)) {
  iprs <- paste("+proj=", projs[i], sep="")
  xy <- try(project(cbind(0, 0), iprs, legacy=TRUE, use_aoi=FALSE), silent=TRUE)
  if (inherits(xy, "try-error")) {
    res[i] <- NA
    msgs[i] <- paste("fwd:", strsplit(xy, "\n")[[1]][2])
  } else if(any(abs(xy) > 1e+08)) {
    res[i] <- NA
    msgs[i] <- paste("fwd: huge value")
  } else {
    out <- try(project(xy, iprs, inv=TRUE, legacy=TRUE, use_aoi=FALSE), silent=TRUE)
    if (inherits(out, "try-error")) {
      res[i] <- NA
      msgs[i] <- paste("inv:", strsplit(out, "\n")[[1]][2])
    } else {
      res[i] <- isTRUE(all.equal(cbind(0,0), out))
    }
  }
}
options(warn=owarn)
df <- data.frame(res=unname(res), msgs=unname(msgs), row.names=names(res))
# projection and inverse projection failures
# fwd: missing parameters
# inv: mostly inverse not defined
df[is.na(df$res),]
# inverse not equal to input
# (see http://lists.maptools.org/pipermail/proj/2011-November/006015.html)
df[!is.na(df$res) & !df$res,]
# inverse equal to input
row.names(df[!is.na(df$res) & df$res,])
#}
# oce data representation with NAs
ll <- structure(c(12.1823368669203, 11.9149630062421, 12.3186076188739, 
12.6207597184845, 12.9955172054652, 12.6316117692658, 12.4680041846297, 
12.4366882666609, NA, NA, -5.78993051516384, -5.03798674888479, 
-4.60623015708619, -4.43802336997614, -4.78110320396188, -4.99127125409291, 
-5.24836150474498, -5.68430388755925, NA, NA), .Dim = c(10L, 
2L), .Dimnames = list(NULL, c("longitude", "latitude")))
try(xy0 <- project(ll, "+proj=moll", legacy=TRUE))
if (!PROJis6ormore()) { # legacy=TRUE PROJ >= 6
try(xy1 <- project(ll, "+proj=moll", legacy=FALSE, allowNAs_if_not_legacy=FALSE))
try(xy2 <- project(ll, "+proj=moll", legacy=FALSE, allowNAs_if_not_legacy=TRUE))
if (exists("xy0")) all.equal(xy0, xy2)
}
if (!exists("xy0")) xy0 <- structure(c(1217100.8468177, 1191302.229156,
1232143.28841193, 1262546.27733232, 1299648.82357849, 1263011.18154638,
1246343.17808186, 1242654.33986052, NA, NA, -715428.207551599,
-622613.577983058, -569301.605757784, -548528.530156422, -590895.949857199,
-616845.926397351, -648585.161643274, -702393.1160979, NA, NA), 
.Dim = c(10L, 2L), .Dimnames = list(NULL, c("longitude", "latitude")))
try(ll0 <- project(xy0, "+proj=moll", inv=TRUE, legacy=TRUE))
if (!PROJis6ormore()) { # legacy=TRUE PROJ >= 6
try(ll1 <- project(xy0, "+proj=moll", inv=TRUE, legacy=FALSE, allowNAs_if_not_legacy=FALSE))
try(ll2 <- project(xy0, "+proj=moll", inv=TRUE, legacy=FALSE, allowNAs_if_not_legacy=TRUE))
if (exists("ll0")) all.equal(ll0, ll2)
}
if (exists("ll0")) all.equal(ll0, ll)
}
\keyword{spatial}

\name{GDALTransientDataset-class}
\docType{class}
\alias{GDALTransientDataset-class}
\alias{initialize,GDALTransientDataset-method}
\title{Class "GDALTransientDataset"}
\description{\code{GDALTransientDataset} is identical to 
  \code{\link{GDALDataset-class}}
  except that transient datasets are not associated with any
  user-visible file. Transient datasets delete their associated file
  data when closed. See \code{\link{saveDataset}} and
  \code{\link{saveDatasetAs}}.}
\section{Objects from the Class}{
Objects can be created by calls of the form \code{new("GDALTransientDataset", driver, rows, cols, bands, type, options, fname, handle)}.
  \describe{
  \item{driver}{A "GDALDriver" object that determines the storage
    format}
  \item{rows}{Number of rows in the newly created dataset}
  \item{cols}{Number of columns in the newly created dataset}
  \item{bands}{Number of bands to create}
  \item{type}{A GDAL type name as listed in \code{.GDALDataTypes}}
  \item{options}{Driver specific options }
  \item{fname}{default NULL, used internally to pass through a file name with a required extension (RST driver has this problem)}
  \item{handle}{Used internally; not for public consumption}
  }
}
\section{Slots}{
  \describe{
    \item{\code{handle}:}{Object of class \code{"externalptr", from class "GDALDataset"}, used internally; not for public consumption}
  }
}
\section{Extends}{
Class \code{"GDALDataset"}, directly.
Class \code{"GDALReadOnlyDataset"}, by class "GDALDataset".
Class \code{"GDALMajorObject"}, by class "GDALDataset".
}
\section{Methods}{
  \describe{
    \item{closeDataset}{\code{signature(dataset = "GDALTransientDataset")}: ... }
    \item{initialize}{\code{signature(.Object = "GDALTransientDataset")}: ... }
  }
}

\author{Timothy H. Keitt, modified by Roger Bivand}

\seealso{See also \code{\link{GDALDriver-class}},
 \code{\link{GDALReadOnlyDataset-class}}}
\examples{
list.files(tempdir())
x <- new('GDALTransientDataset', driver=new('GDALDriver', "GTiff"), rows=100,
 cols=100, bands=3, type='Byte')
dim(x)
list.files(tempdir())
GDAL.close(x)
list.files(tempdir())
}
\keyword{classes}

\name{showWKT}
\alias{showWKT}
\alias{showEPSG}
\alias{showP4}
\alias{showSRID}
\alias{get_P6_datum_hard_fail}
\alias{set_P6_datum_hard_fail}
\alias{get_thin_PROJ6_warnings}
\alias{set_thin_PROJ6_warnings}
\alias{get_prefer_proj}
\alias{set_prefer_proj}
\alias{get_rgdal_show_exportToProj4_warnings}
\alias{set_rgdal_show_exportToProj4_warnings}
\alias{get_PROJ6_warnings_count}
\alias{OSRIsProjected}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{Show Well-Known Text spatial reference system metadata}
\description{
In modern workflows with PROJ >= 6 and GDAL >= 3, use only \code{showSRID()}
  DEPRECATED: Use GDAL/OGR spatial reference objects to convert a PROJ.4 representation to a Well-Known Text representation, and report an EPSG code if it can be determined by OGR SRS services.
}
\usage{
showWKT(p4s, file = NULL, morphToESRI = FALSE, enforce_xy = NULL)
showP4(wkt, morphFromESRI=FALSE, enforce_xy = NULL)
showEPSG(p4s, enforce_xy = NULL)
showSRID(inSRID, format="WKT2", multiline="NO", enforce_xy = NULL, EPSG_to_init=TRUE,
 prefer_proj=NULL)
get_P6_datum_hard_fail()
set_P6_datum_hard_fail(value)
get_thin_PROJ6_warnings()
set_thin_PROJ6_warnings(value)
get_prefer_proj()
set_prefer_proj(value)
get_rgdal_show_exportToProj4_warnings()
set_rgdal_show_exportToProj4_warnings(value)
get_PROJ6_warnings_count()
OSRIsProjected(obj)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{p4s}{A valid PROJ.4 string representing a spatial reference system}
  \item{file}{if not NULL, a file name to which the output Well-Known Text representation should be written}
  \item{morphToESRI}{default FALSE, morph the WKT string to the representation used by ESRI; set FALSE by default 2020-03-04 to avoid occasional Windows crashes seen when passing a very modern WKT string to an old version of PROJ: \url{https://github.com/edzer/sp/issues/75}}
  \item{enforce_xy}{(PROJ6+/GDAL3+) either use global setting (default NULL) or override policy for coordinate ordering easting/x as first axis, northing/y as second axis.}
  \item{wkt}{A valid WKT character string representing a spatial reference system}
  \item{morphFromESRI}{default TRUE, morph the WKT string from the representation used by ESRI}
  \item{inSRID}{Input coordinate reference string}
  \item{obj}{valid CRS object}
  \item{format}{Output format, default WKT2}
  \item{multiline}{Multiline output, either \code{"NO"} or \code{"YES"}}
  \item{EPSG_to_init}{default TRUE, workaround for PROJ 6.3.0 frailty leading to the dropping of +ellps= and +units=; DATUM seems to disappear in the internal definition}
  \item{prefer_proj}{default NULL, if TRUE, use PROJ compiled code directly, rather than FALSE using PROJ via GDAL SRS; if NULL, uses value shown by \code{get_prefer_proj()} set on startup to TRUE.}
%  \item{prefer_proj}{default FALSE, if TRUE, use the PROJ library directly rather than through GDAL}
  \item{value}{a logical value. For \code{set_P6_datum_hard_fail()}: by default, a deprecated/ignored input DATUM key/value pair on reading a file with PROJ6 will give a warning (default FALSE); if TRUE, an error is triggered, which may be trapped using \code{try}. For \code{set_thin_PROJ6_warnings()} default FALSE, can be set to TRUE to report only once and count the number of non-issues warnings, retrieved by \code{get_PROJ6_warnings_count()}. For \code{set_rgdal_show_exportToProj4_warnings()}, default in \pkg{rgdal} version 1.5.* TRUE, from 1.6 FALSE. The \code{options("rgdal_show_exportToProj4_warnings"="x")} may be used before loading \pkg{rgdal} to set the internal logical variables; if the option is set to \code{"all"}, all warnings reporting CRS degradation stemming from the GDAL OSR function \code{exportToProj4()} even if trivial are reported; if set to \code{"thin"}, all warnings are detected but thinned so that one report is given per function call; if set to "none", the degradations are detected but not reported.
}
}

\note{The \code{options("rgdal_show_exportToProj4_warnings"="x")} may be used before loading \pkg{rgdal} to set the internal logical variables; if the option is set to \code{"all"}, all warnings reporting CRS degradation stemming from the GDAL OSR function \code{exportToProj4()} even if trivial are reported; if set to \code{"thin"}, all warnings are detected but thinned so that one report is given per function call; if set to \code{"none"}, the degradations are detected but not reported.}

\value{
  For \code{showWKT}, a character string containing the WKT representation of the PROJ.4 string.
}
\references{\url{https://gdal.org/tutorials/osr_api_tut.html}}
\author{Roger Bivand}

\seealso{\code{\link[sp]{is.projected}}, \code{\link[sp]{CRS-class}}}
\examples{
set_thin_PROJ6_warnings(TRUE)
cities <- readOGR(system.file("vectors", package = "rgdal")[1], "cities")
readLines(system.file("vectors/cities.prj", package = "rgdal")[1])
showWKT(proj4string(cities))
showWKT("+init=epsg:28992")
showP4(showWKT("+init=epsg:28992"))
showEPSG("+proj=utm +zone=30")
showEPSG("+proj=longlat +ellps=WGS84")
exts <- rgdal_extSoftVersion()
run <- new_proj_and_gdal()
if (run) {
  cat(showSRID("EPSG:27700", multiline="YES"), "\n")
}
if (run) {
(prj <- showSRID("EPSG:27700", "PROJ"))
}
if (run) {
showSRID(paste0(prj, " +datum=OSGB36"), "WKT1")
}
if (run) {
showSRID(paste0(prj, " +towgs84=370.936,-108.938,435.682"), "WKT1")
}
if (run) {
showSRID(paste0(prj, " +nadgrids=OSTN15_NTv2_OSGBtoETRS.gsb"), "WKT1")
}
if (run) {
showSRID(paste0(prj, " +datum=OSGB36"), "WKT2")
}
if (run) {
showSRID(paste0(prj, " +towgs84=370.936,-108.938,435.682"), "WKT2")
}
if (run) {
showSRID(paste0(prj, " +nadgrids=OSTN15_NTv2_OSGBtoETRS.gsb"), "WKT2")
}
if (run) {
showSRID("ESRI:102761", "WKT2")
}
if (run) {
showSRID("OGC:CRS84", "WKT2")
}
if (run) {
showSRID("urn:ogc:def:crs:OGC:1.3:CRS84", "WKT2")
}
if (run) {
try(showSRID("", "WKT2"))
}

OSRIsProjected(CRS("+proj=longlat"))
OSRIsProjected(CRS("+proj=geocent"))
OSRIsProjected(CRS("+proj=geocent +units=km"))
}
\keyword{spatial}

\name{is_proj_CDN_enabled}
\alias{is_proj_CDN_enabled}
\alias{enable_proj_CDN}
\alias{disable_proj_CDN}
\alias{proj_CDN_user_writable_dir}
\alias{get_proj_search_paths}
\alias{set_proj_search_paths}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{PROJ search paths and content download network handling}
\description{
From PROJ 7 (and partly 7.1), it is becoming possible to use transformation grids downloaded on demand to improve coordinate operation accuracy from a content download network (CDN). These functions report on and control the use of the CDN.
}
\usage{
is_proj_CDN_enabled()
enable_proj_CDN()
disable_proj_CDN()
proj_CDN_user_writable_dir()
get_proj_search_paths()
set_proj_search_paths(paths)
}

\arguments{
  \item{paths}{a character vector of existing directories}
}

\details{
The PROJ user-writable CDN directory is set as soon as the internal search path is queried, and for most uses, the default value will allow all programs using PROJ such as R packages, QGIS, GRASS, etc., to access any downloaded grids. Grids are checked for staleness at regular intervals. This directory may be set to a non-default value with the PROJ_USER_WRITABLE_DIRECTORY environment variable before \pkg{rgdal} (and any other package using PROJ) is loaded and attached, from PROJ >= 7.1.0.
}
\value{
Logical values and/or character vector search paths, often NULL for earlier versions of PROJ. 
}
\references{\url{https://cdn.proj.org/}.}
\author{Roger Bivand}
\examples{
is_proj_CDN_enabled()
proj_CDN_user_writable_dir()
get_proj_search_paths()
}
\keyword{spatial}

\name{closeDataset-methods}
\docType{methods}
\title{closeDataset methods}
\alias{closeDataset-methods}
\alias{closeDataset}
\alias{closeDataset.default}
\alias{closeDataset,ANY-method}
\alias{closeDataset,GDALReadOnlyDataset-method}
\alias{closeDataset,GDALTransientDataset-method}
\description{Methods for closing GDAL datasets, used internally}
\usage{
	closeDataset(dataset)
	closeDataset.default(dataset)
}
\arguments{
\item{dataset}{GDAL dataset}
}
\section{Methods}{\describe{

\item{dataset = "ANY"}{default method, returns error}

\item{dataset = "GDALReadOnlyDataset"}{closes the "GDALReadOnlyDataset"}

\item{dataset = "GDALTransientDataset"}{closes the "GDALTransientDataset"}
}}
\keyword{methods}

\name{GDALReadOnlyDataset-class}
\docType{class}
\alias{GDALReadOnlyDataset-class}
\alias{getDriver}
%\alias{displayDataset}
\alias{GDAL.close}
\alias{GDAL.open}
\alias{getColorTable}
\alias{getGeoTransFunc}
\alias{initialize,GDALReadOnlyDataset-method}
\alias{dim,GDALReadOnlyDataset-method}
% \alias{[,GDALReadOnlyDataset-method}
% \alias{sub.GDROD}
\title{Class "GDALReadOnlyDataset"}
\description{\code{GDALReadOnlyDataset} is the base class for a GDAL Dataset
  classes. Only read operations are supported. Both \code{GDALDataset}
  and \code{GDALTransientDataset} inherit these read operations while
  providing additional write operations (see
  \code{\link{GDALDataset-class}}). \code{GDALReadOnlyDataset-class} inherits
  from \code{\link{GDALMajorObject-class}}.}

\section{Objects from the Class}{
Objects can be created by calls of the form \code{new("GDALReadOnlyDataset", filename, handle)}.
    ~~ describe objects here ~~ 
}
\section{Slots}{
  \describe{
    \item{\code{handle}:}{Object of class \code{"externalptr", from class "GDALMajorObject"} ~~ }
  }
}
\section{Extends}{
Class \code{"GDALMajorObject"}, directly.
}
\section{Methods}{
  \describe{
    \item{closeDataset}{\code{signature(dataset = "GDALReadOnlyDataset")}: ... }
    \item{dim}{\code{signature(x = "GDALReadOnlyDataset")}: ... }
    \item{initialize}{\code{signature(.Object = "GDALReadOnlyDataset")}: ... }
%     \item{"["}{\code{signature(.Object = "GDALReadOnlyDataset")}: requires package
% 	sp; selects rows and columns, and returns an object of class SpatialGridDataFrame
% 	if the grid is not rotated, or else of class SpatialPointsDataFrame. Any arguments
% 	passed to getRasterData (or in case of rotation getRasterTable) may be passed as
% 	named arguments; the first three unnamed arguments are row,col,band }
  }
}
\usage{
GDAL.close(dataset)
GDAL.open(filename, read.only = TRUE, silent=FALSE,
                 allowedDrivers = NULL, options=NULL)
getDriver(dataset)
%displayDataset(x, offset = c(0, 0), region.dim = dim(x),
%                           reduction = 1, band = NULL, col = NULL,
%                           max.dim = 500, ...)
getColorTable(dataset, band = 1)
getGeoTransFunc(dataset)
}
\arguments{
 \item{dataset}{An object inheriting from class 'GDALReadOnlyDataset'}
 \item{filename}{name of file to contain raster data object; will be normalized with \code{\link{normalizePath}} if it is a file}
% \item{x}{An object inheriting from class 'GDALReadOnlyDataset'}
% \item{offset}{Number of rows and columns from the origin (usually the
%    upper left corner) to begin reading from}
% \item{region.dim}{The number of rows and columns to read from the
%    dataset}
% \item{reduction}{scaling parameter for region.dim}
 \item{band}{The band number (1-based) to read from}
% \item{col}{a vector of colours or a colour function such as rainbow}
% \item{max.dim}{another scaling parameter}
% \item{...}{other plotting parameters passed through}
% \item{}{}
% \item{}{}
\item{read.only}{A logical flag indicating whether to open the file as a
  \code{GDALReadOnlyDataset} or as a writable \code{GDALDataset}}
\item{silent}{ logical; if TRUE, comment and non-fatal CPL driver errors suppressed }
\item{allowedDrivers}{a character vector of suggested driver short names may be provided starting from GDAL 2.0}
\item{options}{open options may be passed to raster drivers starting from GDAL 2.0; very few drivers support these options}
}
\details{
\code{GDAL.open} and \code{GDAL.close} are shorter versions of \code{new("GDALReadOnlyDataset", ...)} and \code{closeDataset()}. Because \code{GDAL.close} through \code{closeDataset()} uses the finalization mechanism to destroy the handles to the dataset and its driver, messages such as:

"Closing GDAL dataset handle 0x8ff7900...  destroyed ... done."

may appear when \code{GDAL.close} is run, or at some later stage. 
\code{getDriver} returns an object inheriting from class 'GDALDriver'.
%\code{displayDataset} and its use in \code{plot} for a pixmap object to produce a plot of the dataset, returning a pixmap object.
\code{getColorTable} returns the dataset colour table (currently does not support RGB imaging).
\code{getGeoTransFunc} returns a warping function.
  
}
\references{\url{https://gdal.org/}}

\author{Timothy H. Keitt, modified by Roger Bivand}

\seealso{See also \code{\link{GDALDriver-class}}, \code{\link{GDALDataset-class}}, \code{\link{GDALTransientDataset-class}}.}

\examples{
logo <- system.file("pictures/logo.jpg", package="rgdal")[1]
x <- new("GDALReadOnlyDataset", logo)
dim(x)
plot(density(getRasterTable(x)$band1))
#displayDataset(x)
#displayDataset(x, col=function(x){rev(cm.colors(x))})
#im <- displayDataset(x, col=function(x){rev(cm.colors(x))}, reset.par=FALSE)
#contour(1:attr(im, "size")[2], 1:attr(im, "size")[1],
# t(attr(im, "index"))[,attr(im, "size")[1]:1], nlevels = 1,
# levels = 100, col = 'black', add = TRUE)
GDAL.close(x)
logo <- system.file("pictures/Rlogo.jpg", package="rgdal")[1]
x <- new("GDALReadOnlyDataset", logo)
dim(x)
#displayDataset(x)
GDAL.close(x)
}
\keyword{classes}

\name{SGDF2PCT}
\alias{SGDF2PCT}
\alias{vec2RGB}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{Convert RGB three band to single band colour table}
\description{
  This function converts a three-band SpatialGridDataFrame into a single band of colour indices and a colour look-up table using \code{RGB2PCT}. \code{vec2RGB} uses given breaks and colours (like \code{image}) to make a three column matrix of red, green, and blue values for a numeric vector. 
}
\usage{
SGDF2PCT(x, ncolors = 256, adjust.bands=TRUE)
vec2RGB(vec, breaks, col)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{x}{a three-band SpatialGridDataFrame object}
  \item{ncolors}{a number of colours between 2 and 256}
  \item{adjust.bands}{default TRUE; if FALSE the three bands must lie each between 0 and 255, but will not be streched within those bounds}
  \item{vec}{a numeric vector}
  \item{breaks}{a set of breakpoints for the colours: must give one more breakpoint than colour}
  \item{col}{a list of colors}
}

\value{
The value returned is a list:
  \item{idx}{a vector of colour indices in the same spatial order as the input object}
  \item{ct}{a vector of RGB colours}
}
\references{\url{https://gdal.org/}}
\author{Roger Bivand}

\examples{
logo <- system.file("pictures/Rlogo.jpg", package="rgdal")[1]
SGlogo <- readGDAL(logo)
cols <- SGDF2PCT(SGlogo)
SGlogo$idx <- cols$idx
image(SGlogo, "idx", col=cols$ct)
SGlogo <- readGDAL(logo)
cols <- SGDF2PCT(SGlogo, ncolors=64)
SGlogo$idx <- cols$idx
image(SGlogo, "idx", col=cols$ct)
SGlogo <- readGDAL(logo)
cols <- SGDF2PCT(SGlogo, ncolors=8)
SGlogo$idx <- cols$idx
image(SGlogo, "idx", col=cols$ct)
data(meuse.grid)
coordinates(meuse.grid) <- c("x", "y")
gridded(meuse.grid) <- TRUE
fullgrid(meuse.grid) <- TRUE
summary(meuse.grid$dist)
opar <- par(no.readonly=TRUE)
par(mfrow=c(1,2), mar=c(1,1,1,1)+0.1)
image(meuse.grid, "dist", breaks=seq(0,1,1/10), col=bpy.colors(10))
RGB <- vec2RGB(meuse.grid$dist, breaks=seq(0,1,1/10), col=bpy.colors(10))
summary(RGB)
meuse.grid$red <- RGB[,1]
meuse.grid$green <- RGB[,2]
meuse.grid$blue <- RGB[,3]
cols <- SGDF2PCT(meuse.grid[c("red", "green", "blue")], ncolors=10,
 adjust.bands=FALSE)
is.na(cols$idx) <- is.na(meuse.grid$dist)
meuse.grid$idx <- cols$idx
image(meuse.grid, "idx", col=cols$ct)
par(opar)
# Note: only one wrongly classified pixel after NA handling/dropping
# The functions are not written to be reversible
sort(table(findInterval(meuse.grid$dist, seq(0,1,1/10), all.inside=TRUE)))
sort(table(cols$idx))
}
\keyword{spatial}

\name{SpatialGDAL-class}
\docType{class}
\alias{SpatialGDAL-class}
\alias{SpatialGDALWrite-class}
\alias{[,SpatialGDAL-method}
\alias{summary,SpatialGDAL-method}
\alias{open.SpatialGDAL}
\alias{copy.SpatialGDAL}
\alias{close.SpatialGDAL}
\alias{print.summary.SpatialGDAL}
\alias{coerce,SpatialGDAL,SpatialGridDataFrame-method}
\alias{coerce,SpatialGDAL,SpatialPixelsDataFrame-method}
\alias{$,SpatialGDAL-method}
\alias{[<-,SpatialGDALWrite-method}
\alias{[[,SpatialGDAL,ANY,missing-method}
\alias{$<-,SpatialGDAL-method}
\alias{[[<-,SpatialGDAL,ANY,missing-method}

\title{Class "SpatialGDAL" }
\description{  Class for spatial attributes that have spatial locations
on a (full) regular grid on file, not (yet) actually read. }
\section{Objects from the Class}{
Objects can be created by calls of the form \code{open.
SpatialGDAL(name), }, where \code{name} is the name of
the GDAL file.
}

\section{Slots}{
  \describe{
    \item{\code{points}:}{see \link[sp]{SpatialPoints}; points slot which is
	not actually filled with all coordinates (only with min/max) }
    \item{\code{grid}:}{see \link[sp]{GridTopology-class}; grid parameters }
    \item{\code{grid.index}:}{see \link[sp]{SpatialPixels-class}; this slot
	is of zero length for this class, as the grid is full }
    \item{\code{bbox}:}{Object of class \code{"matrix"}; bounding box }
    \item{\code{proj4string}:}{Object of class \code{"CRS"}; projection }
    \item{\code{data}:}{Object of class data.frame, containing attribute data }
  }
}
\section{Extends}{ Class \link[sp]{Spatial-class}, directly.  }
\section{Methods}{
  \describe{
    \item{[}{\code{signature(x = "SpatialGDAL", i, j, ...)}: 
		selects rows (i), columns (j), and bands (third argument); returns an object of 
		class \link[sp]{SpatialGridDataFrame-class}. Only the selection is actually read. }
	\item{[[}{\code{signature(i)}: reads band i and returns the values as a
	numeric vector}
  }
}
\usage{
\method{open}{SpatialGDAL}(con, \dots, silent = FALSE, allowedDrivers = NULL, options=NULL)
\method{close}{SpatialGDAL}(con, \dots)
copy.SpatialGDAL(dataset, fname, driver = getDriver(dataset@grod),
 strict = FALSE, options = NULL, silent = FALSE)
}
\arguments{
\item{con}{ file name of grid map for opening, SpatialGDAL object for closing }
\item{\dots}{ other arguments (currently ignored)}
\item{silent}{ logical; if TRUE, comment and non-fatal CPL driver errors suppressed }
\item{dataset}{ object of class SpatialGDAL}
\item{fname}{ file name of grid map }
\item{driver}{ GDAL driver name }
\item{strict}{TRUE if the copy must be strictly equivalent, or more normally FALSE indicating that the copy may adapt as needed for the output format}
\item{allowedDrivers}{a character vector of suggested driver short names may be provided starting from GDAL 2.0}
\item{options}{driver-specific options to be passed to the GDAL driver; only available for opening datasets from GDAL 2.0}
}
\note{Non-fatal CPL errors may be displayed for some drivers, currently for the AIG ArcInfo 9.3 binary raster driver using GDAL >= 1.6.2; the data has been read correctly, but the contents of the info directory did not meet the specifications used to reverse engineer the driver used in GDAL (see \url{https://trac.osgeo.org/gdal/ticket/3031})}

\author{ Edzer Pebesma, \email{edzer.pebesma@uni-muenster.de} }

\seealso{
  \code{\link[sp]{SpatialGridDataFrame-class}}, which is actually sub-classed.
}
\examples{
x <- open.SpatialGDAL(system.file("external/test.ag", package="sp")[1])
image(x[])
image(as(x, "SpatialGridDataFrame"))
summary(as(x, "SpatialGridDataFrame"))
spplot(as(x, "SpatialGridDataFrame"))
# select first 50 rows:
summary(x[1:50])
# select first 50 columns:
summary(x[,1:50])
# select band 1:
summary(x[,,1])
# select first 50 rows, first 50 columns, band 1:
summary(x[1:50,1:50,1])
# get values of first band:
summary(x[[1]])
close(x)
}
\keyword{classes}

\name{GDALDriver-class}
\docType{class}
\alias{GDALDriver-class}
\alias{getGDALDriverNames}
\alias{gdalDrivers}
\alias{getDriverName}
\alias{getDriverLongName}
\alias{initialize,GDALDriver-method}
\alias{getGDALVersionInfo}
\alias{getGDAL_DATA_Path}
\alias{getGDALCheckVersion}
\alias{getGDALwithGEOS}
\alias{getCPLConfigOption}
\alias{setCPLConfigOption}
\alias{rgdal_extSoftVersion}
\alias{get_cached_orig_PROJ_LIB}
\alias{get_cached_set_PROJ_LIB}
\alias{get_cached_orig_GDAL_DATA}
\alias{get_cached_set_GDAL_DATA}
\title{Class "GDALDriver": GDAL Driver Object}

\description{\code{GDALDriver} objects encapsulate GDAL file format
  drivers. \code{GDALDriver} inherits from \code{\link{GDALMajorObject-class}}.}

\section{Objects from the Class}{
Objects can be created by calls of the form \code{new("GDALDriver", name, handle)}, where name: a string giving the name of a GDAL driver, handle: used internally; not for public consumption (default = NULL).
}

\section{Slots}{
  \describe{
    \item{\code{handle}:}{Object of class \code{"externalptr", from class "GDALMajorObject"}, used internally; not for public consumption}
  }
}
\section{Extends}{
Class \code{"GDALMajorObject"}, directly.
}
\section{Methods}{
  \describe{
    \item{initialize}{\code{signature(.Object = "GDALDriver")}: drivername: a string giving the name of a GDAL driver, handle: used internally; not for public consumption (default = NULL) }
  }
}
\usage{
getGDALDriverNames()
gdalDrivers()
getDriverName(driver)
getDriverLongName(driver)
getGDALVersionInfo(str = "--version")
getGDALCheckVersion()
getGDALwithGEOS()
rgdal_extSoftVersion()
getCPLConfigOption(ConfigOption)
setCPLConfigOption(ConfigOption, value)
get_cached_orig_PROJ_LIB()
get_cached_orig_GDAL_DATA()
get_cached_set_PROJ_LIB()
get_cached_set_GDAL_DATA()
}
\arguments{
  \item{driver}{An object inheriting from class 'GDALDriver'}
  \item{str}{A string, may be one of \code{"--version"}, \code{"VERSION_NUM"}, \code{"RELEASE_DATE"}, \code{"RELEASE_NAME"}}
  \item{ConfigOption}{CPL configure option documented in \url{https://trac.osgeo.org/gdal/wiki/ConfigOptions} and elsewhere in GDAL source code}
  \item{value}{a string value to set a CPL option; NULL is used to unset the CPL option}
}

\details{
  \describe{
    \item{\code{getGDALDriverNames, gdalDrivers}:}{returns all driver names currently installed in GDAL, with their declared create and copy status (some drivers can create datasets, others can only copy from a prototype with a different driver.}
  
    \item{\code{getDriverName}:}{returns the GDAL driver name associated
      with the driver object.}

    \item{\code{getDriverLongName}:}{returns a longer driver name.}
    \item{\code{getGDALVersionInfo}:}{returns the version of the GDAL runtime shared object.}
    \item{\code{getGDALCheckVersion}:}{checks the version of the GDAL headers used when building the package (GDAL_VERSION_MAJOR, GDAL_VERSION_MINOR) - if the two versions differ, problems may arise (the C++ API/ABI may have changed), and rgdal should be re-installed}
    \item{\code{getGDALwithGEOS}:}{because drivers may behave differently if GDAL itself was built with GEOS support, the function uses a heuristic to check whether GDAL has access to the GEOS Union function or not}
    \item{\code{get_cached_orig_PROJ_LIB}, \code{get_cached_orig_GDAL_DATA}}{The values of environment variables PROJ_LIB and GDAL_DATA as read when this package was loaded}
    \item{\code{get_cached_set_PROJ_LIB}, \code{get_cached_set_GDAL_DATA}}{If not \code{""}, the values set when loading this package to point to metadata files included in CRAN binary packages}
}
}

\note{Loading the rgdal package changes the \code{GDAL\_DATA} environmental variable to the GDAL support files bundled with the package.}

\author{Timothy H. Keitt, modified by Roger Bivand}

\seealso{\code{\link{GDALMajorObject-class}}}

\examples{
gdalDrivers()
logo <- system.file("pictures/logo.jpg", package="rgdal")[1]
x <- new("GDALReadOnlyDataset", logo)
getDriver(x)
getDriverLongName(getDriver(x))
GDAL.close(x)
}
\keyword{classes}

\name{GDALcall}

\alias{GDALcall}
\alias{rawTransform}

\title{Wrapper functions to allow more direct calling of rgdal C code}

\description{
These functions allow more direct access to some of the rgdal C API. These are advanced methods intended for package developers only. 
}


\usage{
GDALcall(object, option, ...)
rawTransform(projfrom, projto, n, x, y, z=NULL, wkt=FALSE)
}


\arguments{
  \item{object}{GDALTransientDataset (option = 'SetGeoTransform', 'SetProject') or GDALRasterBand (the other options)}
  \item{option}{character. One of 'SetGeoTransform', 'SetProject', 'SetNoDataValue', 'SetStatistics', 'SetRasterColorTable' or 'SetCategoryNames') }
  \item{...}{additional arguments. The values to be set}
  \item{projfrom}{character. PROJ.4 coordinate reference system (CRS) description}
  \item{projto}{character. PROJ.4 CRS description}
  \item{n}{number of coordinates}
  \item{x}{x coordinates}
  \item{y}{y coordinates}
  \item{z}{z coordinates}
  \item{wkt}{default FALSE, if TRUE, the caller determines that projfrom and projto are wkt and that new_proj_and_gdal() returns TRUE to avoid multiple warnings when the function is called repetitively}
}

\value{
GDALcall does not return anything. 
rawTransform returns a matrix of transformed coordinates.
}

\author{Robert Hijmans}

\keyword{spatial}

\name{GDALReadOnlyDataset-methods}
\docType{methods}
\alias{GDALReadOnlyDataset-methods}
\alias{[,GDALReadOnlyDataset-method}
\alias{,GDALReadOnlyDataset-method}
\alias{coerce,GDALReadOnlyDataset,SpatialGridDataFrame-method}
\alias{sub.GDROD}
\title{subset methods for "GDALReadOnlyDataset"}
\description{subsets GDAL objects, returning a SpatialGridDataFrame object}
\section{Methods}{
  \describe{
     \item{"["}{\code{signature(.Object = "GDALReadOnlyDataset")}: requires package
 	sp; selects rows and columns, and returns an object of class SpatialGridDataFrame
 	if the grid is not rotated, or else of class SpatialPointsDataFrame. Any arguments
 	passed to getRasterData (or in case of rotation getRasterTable) may be passed as
 	named arguments; the first three unnamed arguments are row,col,band }
  }
}
\details{
The \code{[} method subsets a GDAL data set, returning a SpatialGridDataFrame object. Reading is
done on the GDAL side, and only the subset requested is ever read into memory.

Further named arguments to \code{[} are to either getRasterTable or getRasterData:
\describe{
 \item{as.is}{see \link{getRasterData}}
 \item{interleave}{see \link{getRasterData}}
 \item{output.dim}{see \link{getRasterData}}
}
the other arguments, \code{offset} and \code{region.dim} are
derived from row/column selection values.

An GDALReadOnlyDataset object can be coerced directly to a SpatialGridDataFrame
}

\author{Edzer Pebesma}

\seealso{See also \code{\link{readGDAL}} \code{\link{GDALDriver-class}},
\code{\link{GDALDataset-class}}, \code{\link{GDALTransientDataset-class}},
\code{\link[sp]{SpatialGridDataFrame-class}} .}

\examples{
library(grid)
logo <- system.file("pictures/logo.jpg", package="rgdal")[1]
x <- new("GDALReadOnlyDataset", logo)
dim(x)
x.sp = x[20:50, 20:50]
class(x.sp)
summary(x.sp)
spplot(x.sp)
GDAL.close(x)

logo <- system.file("pictures/Rlogo.jpg", package="rgdal")[1]
x.gdal <- new("GDALReadOnlyDataset", logo)
x = x.gdal[,,3]
dim(x)
summary(x)
spplot(x)
spplot(x.gdal[])
GDAL.close(x.gdal)

logo <- system.file("pictures/Rlogo.jpg", package="rgdal")[1]
x.gdal <- new("GDALReadOnlyDataset", logo)
x.as <- as(x.gdal, "SpatialGridDataFrame")
GDAL.close(x.gdal)
summary(x.as)
}
\keyword{classes}

\name{RGB2PCT}
\alias{RGB2PCT}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{Convert RGB three band to single band colour table}
\description{
  This function converts a three-band GDALReadOnlyDataset into a single band of colour indices as a GDALTransientDataset. 
}
\usage{
RGB2PCT(x, band, driver.name = 'MEM', ncolors = 256, set.ctab = TRUE)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{x}{a three-band GDALReadOnlyDataset object}
  \item{band}{a vector of numbers, recycled up to 3 in length}
  \item{driver.name}{default MEM}
  \item{ncolors}{a number of colours between 2 and 256}
  \item{set.ctab}{default TRUE, when the dithered dataset handle is returned, otherwise a list of the dataset and the PCT colour table}
}

\value{
The value returned is a either GDALTransientDataset or a list of a GDALTransientDataset and a colour table.
}
\references{\url{https://gdal.org/}}
\author{Tim Keitt}

\examples{
\dontrun{
logo <- system.file("pictures/Rlogo.jpg", package="rgdal")[1]
x <- GDAL.open(logo)
dim(x)
dx <- RGB2PCT(x, band=1:3)
displayDataset(dx, reset.par=FALSE)
dim(dx)
GDAL.close(x)
GDAL.close(dx)
}
}
\keyword{spatial}

\name{make_EPSG}
\alias{make_EPSG}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{Make a data frame of EPSG projection codes}
\description{
  Make a data frame of the European Petroleum Survey Group (EPSG) geodetic parameter dataset as distributed with PROJ.4 software (prior to PROJ 6.0.0, March 2019, only the CSV file, from March 2019 with PROJ >= 6 from the SQLite database). Because finding the correct projection specification is not easy, lists still known as EPSG lists are maintained, and more generally retrieved from databases. The data collated here are as distributed with PROJ.4.
}
\usage{
make_EPSG(file)
}

\arguments{
  \item{file}{file name of the file matching EPSG codes and PROJ.4 arguments, should usually be autodetected; not used for PROJ >= 6}
}
%- maybe also 'usage' for other objects documented here.

\value{
  returns a data frame with columns:
  \item{code}{integer column of EPSG code numbers}
  \item{note}{character column of notes as included in the file}
  \item{prj4}{character column of PROJ.4 arguments for the equivalent projection definitions}
  \item{prj_method}{extra character column from PROJ 6 showing the projection method}
  ...
}
\references{(unlinked because of certificate issues: https://epsg.org/home.html.}

\author{Roger Bivand}
\note{See also Clifford J. Mugnier's Grids \& Datums columns in Photogrammetric Engineering & Remote Sensing, \url{https://www.asprs.org/a/resources/grids/}, see also \code{\link{GridsDatums}}.}
\examples{
EPSG <- try(make_EPSG())
# from PROJ 6.0.0, EPSG data is no longer stored in a flat file
if (!inherits(EPSG, "try-error")) attr(EPSG, "metadata") 
# PROJ.4 version 5 and later include the EPSG version as an attribute
if (!inherits(EPSG, "try-error")) EPSG[grep("Oslo", EPSG$note), 1:2]
if (!inherits(EPSG, "try-error")) EPSG[1925:1927, 3]
if (!inherits(EPSG, "try-error")) EPSG[grep("Poland", EPSG$note), 1:2]
if (!inherits(EPSG, "try-error")) EPSG[grep("Amersfoort", EPSG$note), 1:2]
if (!inherits(EPSG, "try-error")) EPSG[grep("North Carolina", EPSG$note), 1:2]
if (!inherits(EPSG, "try-error")) EPSG[2202, 3]
}
\keyword{spatial}

\name{displayDataset}
\alias{displayDataset}
\title{Display a GDAL dataset}
\description{Display a GDAL dataset allowing for subscenes and decimation, allowing very large images to be browsed}
\usage{
displayDataset(x, offset=c(0, 0), region.dim=dim(x), reduction = 1, 
  band = 1, col = NULL, reset.par = TRUE, max.dim = 500, ...)
}
\arguments{
  \item{x}{a three-band GDALReadOnlyDataset object}
  \item{offset}{Number of rows and columns from the origin (usually the
    upper left corner) to begin reading from; presently ordered (y,x) - this may change}
  \item{region.dim}{The number of rows and columns to read from the
    dataset; presently ordered (y,x) - this may change}
  \item{reduction}{a vector of length 1 or 2 recycled to 2 for decimating the input data, 1 retains full resultion, higher values decimate}
  \item{band}{The band number (1-based) to read from}
  \item{col}{default NULL, attempt to use band colour table and default to grey scale if not available}
  \item{reset.par}{default TRUE - reset par() settings on completion}
  \item{max.dim}{default 500, forcing the image to a maximum dimension of the value}
  \item{...}{arguments passed to image.default()}
}
\value{
a list of the image data, the colour table, and the par() values on entry.
}
\references{\url{https://gdal.org/}}
\author{Tim Keitt}

\examples{
\dontrun{
logo <- system.file("pictures/Rlogo.jpg", package="rgdal")[1]
x <- GDAL.open(logo)
opar <- par(no.readonly=TRUE)
par(mfrow=c(2,2))
displayDataset(x, band=1, reset.par=FALSE)
displayDataset(x, band=2, reset.par=FALSE)
#displayDataset(x, band=3, reset.par=TRUE)
par(opar)
dx <- RGB2PCT(x, band=1:3)
displayDataset(dx, reset.par=FALSE)
GDAL.close(x)
GDAL.close(dx)
}
}
\keyword{spatial}

\name{spTransform-methods}
\docType{methods}
\title{ Methods for Function spTransform for map projection and datum transformation in package "rgdal" }
\alias{spTransform-methods}
\alias{spTransform}
%\alias{spTransform,ANY-method}
\alias{spTransform,SpatialPoints,CRS-method}
\alias{spTransform,SpatialPointsDataFrame,CRS-method}
\alias{spTransform,SpatialPixelsDataFrame,CRS-method}
\alias{spTransform,SpatialGridDataFrame,CRS-method}
%\alias{spTransform,Sline-method}
%\alias{spTransform,Slines-method}
\alias{spTransform,SpatialLines,CRS-method}
%\alias{spTransform,Sring-method}
%\alias{spTransform,Srings-method}
\alias{spTransform,SpatialPolygons,CRS-method}
\alias{spTransform.SpatialPoints}
\alias{spTransform.SpatialPointsDataFrame}
%\alias{spTransform.Sline}
%\alias{spTransform.Slines}
\alias{spTransform.SpatialLines}
\alias{spTransform.SpatialLinesDataFrame}
\alias{spTransform,SpatialLinesDataFrame,CRS-method}
%\alias{spTransform.Sring}
%\alias{spTransform.Srings}
\alias{spTransform.SpatialPolygons}
\alias{spTransform.SpatialPolygonsDataFrame}
\alias{spTransform,SpatialPolygonsDataFrame,CRS-method}
\alias{get_transform_wkt_comment}
\alias{set_transform_wkt_comment}
\alias{get_last_coordOp}
\alias{get_enforce_xy}
\alias{set_enforce_xy}

\section{Methods}{\describe{

\item{"ANY"}{default void method}

\item{"SpatialPoints", CRSobj = CRS}{returns transformed coordinates of an "SpatialPoints" object using the projection arguments in "CRSobj", of class CRS}

\item{"SpatialPointsDataFrame", CRSobj = CRS}{returns transformed coordinates of an "SpatialPointsDataFrame" object using the projection arguments in "CRSobj", of class CRS}

\item{"SpatialLines", CRSobj = CRS}{returns transformed coordinates of an "SpatialLines" object using the projection arguments in "CRSobj", of class CRS}

\item{"SpatialLinesDataFrame", CRSobj = CRS}{returns transformed coordinates of an "SpatialLinesDataFrame" object using the projection arguments in "CRSobj", of class CRS}

\item{"SpatialPolygons", CRSobj = CRS}{returns transformed coordinates of an "SpatialPolygons" object using the projection arguments in "CRSobj", of class CRS}

\item{"SpatialPolygonsDataFrame", CRSobj = CRS}{returns transformed coordinates of an "SpatialPolygonsDataFrame" object using the projection arguments in "CRSobj", of class CRS}
\item{"SpatialPixelsDataFrame", CRSobj = CRS}{Because regular grids will usually not be regular after projection/datum transformation, the input object is coerced to a SpatialPointsDataFrame, and the transformation carried out on that object. A warning: \dQuote{Grid warping not available, coercing to points} is given.}
\item{"SpatialGridDataFrame", CRSobj = CRS}{Because regular grids will usually not be regular after projection/datum transformation, the input object is coerced to a SpatialPointsDataFrame, and the transformation carried out on that object. A warning: \dQuote{Grid warping not available, coercing to points} is given.}

}}

\usage{
get_transform_wkt_comment()
set_transform_wkt_comment(value)
get_enforce_xy()
set_enforce_xy(value)
get_last_coordOp()
}

\arguments{
  \item{value}{A non-NA logical value}
}

\description{The \code{spTransform} methods provide transformation between datum(s) and conversion between projections (also known as projection and/or re-projection), from one unambiguously specified coordinate reference system (CRS) to another, prior to version 1.5 using Proj4 projection arguments. From version 1.5, Well-Known Text 2 (WKT2 2019) strings are used. For simple projection, when no Proj4 +datum tags are used, datum projection does not occur. When datum transformation is required, the datum should be defined with a valid value both in the CRS of the object to be transformed, and in the target CRS. In general datum is to be prefered to ellipsoid, because the datum always fixes the ellipsoid, but the ellipsoid never fixes the datum. 

In addition, before version 1.5 the +towgs84 tag should have been used where needed to make sure that datum transformation would take place. Parameters for +towgs84 were taken from the legacy bundled EPSG file if they are known unequivocally, but could be entered manually from known authorities. Not providing the appropriate +datum and +towgs84 tags led to coordinates being out by hundreds of metres. Unfortunately, there is no easy way to provide this information: the user has to know the correct metadata for the data being used, even if this can be hard to discover.

From version 1.5, \code{spTransform} uses the modern PROJ coordinate operation framework for transformations. This avoids pivoting through WGS84 if possible, and uses WKT2 (2019) strings for source and target CRS often constructed from the bundled EPSG SQLite database. The database is searched for feasible candidate coordinate operations, and the most accurate available is chosen. More details are available in a vignette: \code{vignette("CRS_projections_transformations")}.}

\note{The projection arguments had to be entered exactly as in the PROJ.4 documentation, in particular there cannot be any white space in +<arg>=<value> strings, and successive such strings can only be separated by blanks.
Note that warnings about different projections may be issued when the PROJ.4 library extends projection arguments; examine the warning to see if the differences are real.

Also note that re-projection and/or datum transformation will usually not work for regular grids. The term used for similar operations for regular grids is warping, which involved resampling to a regular grid in the target coordinate reference system.

The methods may take an optional argument \dQuote{use_ob_tran}, default FALSE, if TRUE and \dQuote{+proj=ob_tran}, use General Oblique Transformation with internalised from/to projection reversal (the user oblique transforms from longlat to oblique forward rather than inverse as suggested in PROJ.4 mailing list postings); these changes are intended to meet a need pointed out by Martin Ivanov (2012-08-15). A subsequent point raised by Martin Ivanov (2017-04-28) was that use of a projected CRS with \dQuote{+proj=ob_tran} led to errors, so mixing projected CRS and \dQuote{+proj=ob_tran} is blocked. Transform first \dQuote{+proj=ob_tran} to or from \dQuote{+proj=longlat}, and then on from geographical coordinates to those desired or the reverse - see example.

If a SpatialPoints object has three dimensions, the third will also be transformed, with the metric of the third dimension assumed to be meters if the vertical units metric is not given in the projection description with +vunits= or +vto_meter= (which is 1.0 by default) \url{https://proj.org/faq.html}.

Note that WGS84 is both an ellipse and a datum, and that since 1984 there have been changes in the relative positions of continents, leading to a number of modifications. This is discussed for example in \url{http://www.uvm.edu/giv/resources/WGS84_NAD83.pdf}; there are then multiple transformations between NAD83 and WGS84 depending on the WGS84 definition used. One would expect that \dQuote{+towgs84=} is a no-op for WGS84, but this only applies sometimes, and as there are now at least 30 years between now and 1984, things have shifted. It may be useful to note that \dQuote{+nadgrids=@null} can help, see these threads: \url{https://stat.ethz.ch/pipermail/r-sig-geo/2014-August/021611.html}, \url{http://lists.maptools.org/pipermail/proj/2014-August/006894.html}, with thanks to Hermann Peifer for assistance.

Note that from PROJ.4 4.9.3, the definition of UTM is changed from TMERC to ETMERC; see example.
}

\author{Roger Bivand \email{Roger.Bivand@nhh.no}}

\examples{
set_thin_PROJ6_warnings(TRUE)
# state
data(state)
states <- data.frame(state.x77, state.center)
states <- states[states$x > -121,]
coordinates(states) <- c("x", "y")
proj4string(states) <- CRS("+proj=longlat +ellps=clrk66")
summary(states)
state.ll83 <- spTransform(states, CRS("+proj=longlat +ellps=GRS80"))
summary(state.ll83)
state.merc <- spTransform(states, CRS=CRS("+proj=merc +ellps=GRS80"))
summary(state.merc)
state.merc <- spTransform(states,
 CRS=CRS("+proj=merc +ellps=GRS80 +units=us-mi"))
summary(state.merc)
# NAD
\dontrun{
if (PROJis6ormore() || (!PROJis6ormore() && projNAD())) {
states <- data.frame(state.x77, state.center)
states <- states[states$x > -121,]
coordinates(states) <- c("x", "y")
proj4string(states) <- CRS("+init=epsg:4267")
print(summary(states))
state.ll83 <- spTransform(states, CRS("+init=epsg:4269"))
print(summary(state.ll83))
state.kansasSlcc <- spTransform(states, CRS=CRS("+init=epsg:26978"))
print(summary(state.kansasSlcc))
SFpoint_NAD83 <- SpatialPoints(matrix(c(-103.869667, 44.461676), nrow=1),
 proj4string=CRS("+init=epsg:4269"))
SFpoint_NAD27 <- spTransform(SFpoint_NAD83, CRS("+init=epsg:4267"))
print(all.equal(coordinates(SFpoint_NAD83), coordinates(SFpoint_NAD27)))
print(coordinates(SFpoint_NAD27), digits=12)
print(coordinates(SFpoint_NAD83), digits=12)
}
}
data(meuse)
coordinates(meuse) <- c("x", "y")
proj4string(meuse) <- CRS("+init=epsg:28992")
# see http://trac.osgeo.org/gdal/ticket/1987
summary(meuse)
meuse.utm <- spTransform(meuse, CRS("+proj=utm +zone=32 +datum=WGS84"))
summary(meuse.utm)
cbind(coordinates(meuse), coordinates(meuse.utm))
\dontrun{
# Kiritimati
kiritimati_primary_roads <- readOGR(system.file("vectors",
 package = "rgdal")[1], "kiritimati_primary_roads")
kiritimati_primary_roads_ll <- spTransform(kiritimati_primary_roads,
 CRS("+proj=longlat +datum=WGS84"))
opar <- par(mfrow=c(1,2))
plot(kiritimati_primary_roads, axes=TRUE)
plot(kiritimati_primary_roads_ll, axes=TRUE, las=1)
par(opar)
# scot_BNG
scot_BNG <- readOGR(system.file("vectors", package = "rgdal")[1],
   "scot_BNG")
scot_LL <- spTransform(scot_BNG, CRS("+proj=longlat +datum=WGS84"))
plot(scot_LL, axes=TRUE)
grdtxt_LL <- gridat(scot_LL)
grd_LL <- gridlines(scot_LL, ndiscr=100)
summary(grd_LL)
target <- CRS(proj4string(scot_BNG))
grd_BNG <- spTransform(grd_LL, target)
grdtxt_BNG <- spTransform(grdtxt_LL, target)
opar <- par(mfrow=c(1,2))
plot(scot_BNG, axes=TRUE, las=1)
plot(grd_BNG, add=TRUE, lty=2)
text(coordinates(grdtxt_BNG),
   labels=parse(text=as.character(grdtxt_BNG$labels)))
par(opar)
}
# broke_proj
broke_proj <- FALSE
# https://github.com/OSGeo/PROJ/issues/1525
pv <- .Call("PROJ4VersionInfo", PACKAGE="rgdal")[[2]]
if (pv >= 600 && pv < 620) broke_proj <- TRUE
if (!broke_proj) { 
crds <- matrix(data=c(9.05, 48.52), ncol=2)
spPoint <- SpatialPoints(coords=crds,
 proj4string=CRS("+proj=longlat +ellps=sphere +no_defs"))
ob_tran_def <- paste("+proj=ob_tran +o_proj=longlat",
 "+o_lon_p=-162 +o_lat_p=39.25 +lon_0=180 +ellps=sphere +no_defs")
tg <- CRS(ob_tran_def)
# proj4string not propagated in GDAL 3.0.0
a <- spTransform(spPoint, tg, use_ob_tran=TRUE)
a
}
#should be (-5.917698, -1.87195)
if (!broke_proj) {
spTransform(a, CRS("+proj=longlat +ellps=sphere +no_defs"),
 use_ob_tran=TRUE)
}
if (!broke_proj) {
try(spTransform(a, CRS(paste("+proj=tmerc +lat_0=0 +lon_0=9 +k=1",
"+x_0=3500000 +y_0=0 +ellps=bessel +units=m +no_defs")),
 use_ob_tran=TRUE))
}
if (!broke_proj) {
spTransform(spPoint, CRS(paste("+proj=tmerc +lat_0=0 +lon_0=9 +k=1",
"+x_0=3500000 +y_0=0 +ellps=bessel +units=m +no_defs")))
}
if (!broke_proj) {
spTransform(spTransform(a, CRS("+proj=longlat +ellps=sphere +no_defs"),
 use_ob_tran=TRUE), CRS(paste("+proj=tmerc +lat_0=0 +lon_0=9 +k=1",
"+x_0=3500000 +y_0=0 +ellps=bessel +units=m +no_defs")))
}
crds1 <- matrix(data=c(7, 51, 8, 52, 9, 52, 10, 51, 7, 51), ncol=2,
byrow=TRUE, dimnames=list(NULL, c("lon", "lat")));
crds2 <- matrix(data=c(8, 48, 9, 49, 11, 49, 9, 48, 8, 48), ncol=2,
byrow=TRUE, dimnames=list(NULL, c("lon", "lat")));
crds3 <- matrix(data=c(6, 47, 6, 55, 15, 55, 15, 47, 6, 47), ncol=2,
byrow=TRUE, dimnames=list(NULL, c("lon", "lat")));
spLines <- SpatialLines(list(Lines(list(Line(crds1), Line(crds2),
Line(crds3)), ID="a")));
slot(spLines, "proj4string") <- CRS("+proj=longlat +ellps=sphere +no_defs");
bbox(spLines);
if (!broke_proj) {
spLines_tr <- spTransform(spLines, tg, use_ob_tran=TRUE);
bbox(spLines_tr)
}
if (!broke_proj) {
bbox(spTransform(spLines_tr, CRS("+proj=longlat +ellps=sphere"),
 use_ob_tran=TRUE))
}
if (!broke_proj) {
spPolygons <- SpatialPolygons(list(Polygons(list(Polygon(crds1),
Polygon(crds2), Polygon(crds3)), ID="a")));
slot(spPolygons, "proj4string") <- CRS("+proj=longlat +ellps=sphere +no_defs");
bbox(spPolygons);
}
if (!broke_proj) {
spPolygons_tr <- spTransform(spPolygons, tg, use_ob_tran=TRUE);
bbox(spPolygons_tr)
}
if (!broke_proj) {
bbox(spTransform(spPolygons_tr, CRS("+proj=longlat +ellps=sphere"),
 use_ob_tran=TRUE))
}
# added after posting by Martin Ivanov
\dontrun{
data(nor2k)
summary(nor2k)
nor2kNGO <- spTransform(nor2k, CRS("+init=epsg:4273"))
summary(nor2kNGO)
all.equal(coordinates(nor2k)[,3], coordinates(nor2kNGO)[,3])
# added after posting by Don MacQueen 
crds <- cbind(c(-121.524764291826, -121.523480804667), c(37.6600366036405, 37.6543604613483))
ref <- cbind(c(1703671.30566227, 1704020.20113366), c(424014.398045834, 421943.708664294))
crs.step1.cf <- CRS(paste("+proj=lcc +lat_1=38.43333333333333",
 "+lat_2=37.06666666666667 +lat_0=36.5 +lon_0=-120.5",
 "+x_0=2000000.0 +y_0=500000.0 +ellps=GRS80 +units=us-ft +no_defs",
 "+towgs84=-0.991,1.9072,0.5129,0.025789908,0.0096501,0.0116599,0.0"))
locs.step1.cf <- spTransform(SpatialPoints(crds,
 proj4string=CRS("+proj=longlat +datum=WGS84")), crs.step1.cf)
suppressWarnings(proj4string(locs.step1.cf) <- CRS(paste("+proj=lcc",
"+lat_1=38.43333333333333 +lat_2=37.06666666666667 +lat_0=36.5",
"+lon_0=-120.5 +x_0=2000000.0 +y_0=500000.0 +ellps=GRS80 +units=us-ft",
"+no_defs +nadgrids=@null")))
locs.step2.cfb <- spTransform(locs.step1.cf, CRS("+init=epsg:26743"))
coordinates(locs.step2.cfb) - ref
all.equal(unname(coordinates(locs.step2.cfb)), ref)
}
\dontrun{
# new_proj_and_gdal()
run <- new_proj_and_gdal()
if (run) {
# Test for UTM == TMERC (<= 4.9.2) or UTM == ETMERC (> 4.9.2)
nhh <- SpatialPointsDataFrame(matrix(c(5.304234, 60.422311), ncol=2),
 proj4string=CRS(SRS_string="OGC:CRS84"), data=data.frame(office="RSB"))
nhh_utm_32N_P4 <- spTransform(nhh, CRS("+init=epsg:3044"))
nhh_tmerc_P4 <- spTransform(nhh, CRS(paste("+proj=tmerc +k=0.9996",
 "+lon_0=9 +x_0=500000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs")))
nhh_etmerc_P4 <- spTransform(nhh, CRS(paste("+proj=etmerc +k=0.9996",
 "+lon_0=9 +x_0=500000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs")))
all.equal(coordinates(nhh_utm_32N_P4), coordinates(nhh_tmerc_P4),
 tolerance=1e-9, scale=1)
# UTM == TMERC: PROJ4 <=4.9.2
all.equal(coordinates(nhh_utm_32N_P4), coordinates(nhh_etmerc_P4),
 tolerance=1e-9, scale=1)
# UTM == ETMERC: PROJ4 > 4.9.2
unis <- SpatialPointsDataFrame(matrix(c(15.653453, 78.222504), ncol=2),
 proj4string=CRS(SRS_string="OGC:CRS84"), data=data.frame(office="UNIS"))
unis_utm_33N_P4 <- spTransform(unis, CRS("+init=epsg:3045"))
unis_tmerc_P4 <- spTransform(unis, CRS(paste("+proj=tmerc +k=0.9996 +lon_0=15",
 "+x_0=500000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs")))
unis_etmerc_P4 <- spTransform(unis, CRS(paste("+proj=etmerc +k=0.9996",
 "+lon_0=15 +x_0=500000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs")))
all.equal(coordinates(unis_utm_33N_P4), coordinates(unis_tmerc_P4),
 tolerance=1e-9, scale=1)
# UTM == TMERC: PROJ4 <=4.9.2
all.equal(coordinates(unis_utm_33N_P4), coordinates(unis_etmerc_P4),
 tolerance=1e-9, scale=1)
# UTM == ETMERC: PROJ4 > 4.9.2
}
}
}

\keyword{methods}
\keyword{spatial}

\name{readGDAL}
\alias{readGDAL}
\alias{asSGDF_GROD}
\alias{writeGDAL}
\alias{create2GDAL}
\alias{GDALinfo}
\alias{print.GDALobj}
\alias{GDALSpatialRef}
\title{Read/write between GDAL grid maps and Spatial objects}

\description{The functions read or write GDAL grid maps. They will set
the spatial reference system if available. \code{GDALinfo} reports the
size and other parameters of the dataset. \code{create2GDAL} creates
a GDAL data set from a SpatialGridDataFrame object, in particular to
be able to save to GDAL driver formats that only permit copying rather
than creation.}
\usage{ 
readGDAL(fname, offset, region.dim, output.dim, band, p4s=NULL, ...,
 half.cell=c(0.5, 0.5), silent = FALSE, OVERRIDE_PROJ_DATUM_WITH_TOWGS84=NULL,
 allowedDrivers = NULL, enforce_xy = NULL, options=NULL)
asSGDF_GROD(x, offset, region.dim, output.dim, p4s=NULL, ...,
 half.cell=c(0.5,0.5), OVERRIDE_PROJ_DATUM_WITH_TOWGS84=NULL, enforce_xy = NULL)
writeGDAL(dataset, fname, drivername = "GTiff", type = "Float32",
 mvFlag = NA, options=NULL, copy_drivername = "GTiff", setStatistics=FALSE,
 colorTables = NULL, catNames=NULL, enforce_xy = NULL)
create2GDAL(dataset, drivername = "GTiff", type = "Float32", mvFlag = NA,
 options=NULL, fname = NULL, setStatistics=FALSE, colorTables = NULL,
 catNames=NULL, enforce_xy = NULL)
GDALinfo(fname, silent=FALSE, returnRAT=FALSE, returnCategoryNames=FALSE,
 returnStats=TRUE, returnColorTable=FALSE,
 OVERRIDE_PROJ_DATUM_WITH_TOWGS84=NULL, returnScaleOffset=TRUE,
 allowedDrivers = NULL, enforce_xy = NULL, options=NULL)
GDALSpatialRef(fname, silent=FALSE, OVERRIDE_PROJ_DATUM_WITH_TOWGS84=NULL,
 allowedDrivers = NULL, enforce_xy = NULL, options=NULL)
}

\arguments{
\item{fname}{ file name of grid map; in \code{create2GDAL} provides a way 
to pass through a file name with driver-required extension for sensitive 
drivers }
\item{x}{A GDALReadOnlyDataset object}
\item{offset}{Number of rows and columns from the origin (usually the upper 
left corner) to begin reading from; presently ordered (y,x) - this may change}
\item{region.dim}{The number of rows and columns to read from the dataset; 
presently ordered (y,x) - this may change}
\item{output.dim}{The number of rows and columns to return in the created 
object using GDAL's method to take care of image decimation / replication; 
presently ordered (y,x) - this may change}
\item{band}{if missing, all bands are read}
\item{p4s}{PROJ4 string defining CRS, if default (NULL), the value is read 
from the GDAL data set} 
\item{half.cell}{Used to adjust the intra-cell offset from corner to centre, 
usually as default, but may be set to c=(0,0) if needed; presently ordered (y,x) - this may change}
\item{silent}{ logical; if TRUE, comment and non-fatal CPL driver errors 
suppressed }
\item{OVERRIDE_PROJ_DATUM_WITH_TOWGS84}{logical value, default NULL, which case the cached option set by \code{set_OVERRIDE_PROJ_DATUM_WITH_TOWGS84} is used. Ignored if the GDAL version is less than \dQuote{1.8.0} or if the CPLConfigOption variable is already set; see \code{\link{getProjectionRef}} for further details}
\item{allowedDrivers}{a character vector of suggested driver short names may be provided starting from GDAL 2.0}
\item{...}{ arguments passed to either \code{getRasterData}, or
\code{getRasterTable}, depending on rotation angles (see below);
see the rgdal documentation for the available options (subsetting
etc.) }
\item{dataset}{ object of class \link[sp]{SpatialGridDataFrame-class} or 
\link[sp]{SpatialPixelsDataFrame-class}}
\item{drivername, copy_drivername}{ GDAL driver name; if the chosen driver 
does not support dataset creation, an attempt is made to use the 
\code{copy_drivername} driver to create a dataset, and \code{copyDatset} 
to copy to the target driver }
\item{type}{ GDAL write data type, one of: \sQuote{Byte}, \sQuote{Int16}, 
\sQuote{Int32}, \sQuote{Float32}, \sQuote{Float64}; \sQuote{UInt16}, 
\sQuote{UInt32} are available but have not been tests }
\item{ mvFlag }{default NA, missing value flag for output file; the default 
value works for \sQuote{Int32}, \sQuote{Float32}, \sQuote{Float64}, but 
suitable in-range value that fits the data type should be used for other 
data types, for example 255 for \sQuote{Byte}, -32768 for \sQuote{Int16}, 
and so on; see Details below.}
\item{enforce_xy}{(PROJ6+/GDAL3+) either use global setting (default NULL) or override policy for coordinate ordering easting/x as first axis, northing/y as second axis.}
\item{options}{driver-specific options to be passed to the GDAL driver; only available for opening datasets from GDAL 2.0; 
see copying and creation details below}
\item{setStatistics}{default FALSE, if TRUE, attempt to set per-band 
statistics in the output file (driver-dependent)}
\item{colorTables}{default NULL, if not NULL, a list of length equal to the number of bands, with NULL components for bands with no color table, or either an integer matrix of red, green, blue and alpha values (0-255), or a character vector of colours. The number of colours permitted may vary with driver.}
\item{catNames}{default NULL, if not NULL, a list of length equal to the number of bands, with NULL components for bands with no category names, or a string vector of category names}
\item{returnRAT}{default FALSE, if TRUE, return a list with a Raster 
Attribute Table or NULL for each band}
\item{returnCategoryNames}{default FALSE, if TRUE, return a list with a character vector of CategoryNames or NULL for each band}
\item{returnStats}{default TRUE, return band-wise statistics if avaliable (from 0.7-20 set to NA if not available)}
\item{returnColorTable}{default FALSE; if TRUE return band-wise colour tables in a list attribute \dQuote{ColorTables}}
\item{returnScaleOffset}{default TRUE, return a matrix of bandwise scales and offsets}

%\item{clone}{if not NULL, a grid file to clone}
}
\details{In \code{writeGDAL}, if types other than \sQuote{Int32},
\sQuote{Float32}, \sQuote{Float64} are used, the \dQuote{mvFlag}
argument should be used to set a no data value other than the default
NA. Note that the flag only replaces NA values in the data being exported
with the value of the argument - it does not mark data values equal to
\dQuote{mvFlag} as missing. The value is stored in the file being written
in driver-specific ways, and may be used when the file is read. When the
default \dQuote{mvFlag=NA} is used, no NoDataValue is written to the file,
and the input data is written as is.

Also in \code{writeGDAL}, the \dQuote{options} argument may be
used to pass a character vector of one or more options to the
driver, for example \sQuote{options=\dQuote{INTERLEAVE=PIXEL}},
or \sQuote{options=c(\dQuote{INTERLEAVE=PIXEL},
\dQuote{COMPRESS=DEFLATE})}. Typical cases are given in the examples
below; it may also be necessary in some cases to escape quotation markes
if included in the string passed to the driver.}

\value{
\code{read.GDAL} returns the data in the file as a Spatial object. 

Usually, GDAL maps will be north-south oriented, in which case the rgdal
function \code{getRasterData} is used to read the data, and an object
of class \link[sp]{SpatialGridDataFrame-class} is returned.

Some map formats supported by GDAL are not north-south oriented grids. If
this is the case, \code{readGDAL} returns the data as a set of point
data, being of class \link[sp]{SpatialPointsDataFrame-class}. If the points
are on a 45 or 90 degree rotated grid, you can try to enforce gridding
later on by e.g. using \code{\link[sp]{gridded-methods}(x)=TRUE}.
}
\note{Non-fatal CPL errors may be displayed for some drivers, currently
for the AIG ArcInfo 9.3 binary raster driver using GDAL >= 1.6.2; the data
has been read correctly, but the contents of the info directory did not
meet the specifications used to reverse engineer the driver used in GDAL
(see \url{https://trac.osgeo.org/gdal/ticket/3031})}

\section{Warning}{Some raster files may have an erroneous positive
y-axis resolution step, leading to the data being flipped on the
y-axis. \code{readGDAL} will issue a warning: Y axis resolution
positive, examine data for flipping, when the step is positive, but
this need not mean that the data are flipped. Examine a display of the
data compared with your knowledge of the file to determine whether this
is the case (one known case is interpolation files created under Qgis
up to February 2010 at least). To retreive the correct orientation,
use \code{\link[sp]{flip}}.}

\seealso{ \code{\link[sp]{image}}, \code{\link[maptools]{asciigrid}} }

\author{Edzer Pebesma, Roger Bivand}

\examples{
set_thin_PROJ6_warnings(TRUE)
library(grid)
GDALinfo(system.file("external/test.ag", package="sp")[1])
x <- readGDAL(system.file("external/test.ag", package="sp")[1])
class(x)
image(x)
summary(x)
x@data[[1]][x@data[[1]] > 10000] <- NA
summary(x)
image(x)

x <- readGDAL(system.file("external/simple.ag", package="sp")[1])
class(x)
image(x)
summary(x)
x <- readGDAL(system.file("pictures/big_int_arc_file.asc", package="rgdal")[1])
summary(x)
cat("if the range is not 10000, 77590, your GDAL does not detect big\n")
cat("integers for this driver\n")
y = readGDAL(system.file("pictures/Rlogo.jpg", package = "rgdal")[1], band=1)
summary(y)
y = readGDAL(system.file("pictures/Rlogo.jpg", package = "rgdal")[1])
summary(y)
spplot(y, names.attr=c("red","green","blue"), 
	col.regions=grey(0:100/100),
	main="example of three-layer (RGB) raster image", as.table=TRUE)
data(meuse.grid)
gridded(meuse.grid) = ~x+y
proj4string(meuse.grid) = CRS("+init=epsg:28992")
fn <- tempfile()
writeGDAL(meuse.grid["dist"], fn)
GDALinfo(fn)
writeGDAL(meuse.grid["dist"], fn, setStatistics=TRUE)
GDALinfo(fn)
mg2 <- readGDAL(fn)
proj4string(mg2)
SP27GTIF <- readGDAL(system.file("pictures/SP27GTIF.TIF", 
package = "rgdal")[1], output.dim=c(100,100))
summary(SP27GTIF)
slot(SP27GTIF, "proj4string")
if (new_proj_and_gdal()) comment(slot(SP27GTIF, "proj4string"))
image(SP27GTIF, col=grey(1:99/100))
GDALinfo(system.file("pictures/cea.tif", package = "rgdal")[1])
(o <- GDALSpatialRef(system.file("pictures/cea.tif", package = "rgdal")[1]))
if (new_proj_and_gdal()) comment(o)
cea <- readGDAL(system.file("pictures/cea.tif", package = "rgdal")[1], 
output.dim=c(100,100))
summary(cea)
image(cea, col=grey(1:99/100))
slot(cea, "proj4string")
if (new_proj_and_gdal()) comment(slot(cea, "proj4string"))
fn <- system.file("pictures/erdas_spnad83.tif", package = "rgdal")[1]
erdas_spnad83 <- readGDAL(fn, offset=c(50, 100), region.dim=c(400, 400), 
output.dim=c(100,100))
summary(erdas_spnad83)
slot(erdas_spnad83, "proj4string")
if (new_proj_and_gdal()) comment(slot(erdas_spnad83, "proj4string"))
image(erdas_spnad83, col=grey(1:99/100))
erdas_spnad83a <- readGDAL(fn, offset=c(50, 100), region.dim=c(400, 400))
bbox(erdas_spnad83)
bbox(erdas_spnad83a)
gridparameters(erdas_spnad83)
gridparameters(erdas_spnad83a)
tf <- tempfile()
writeGDAL(erdas_spnad83, tf, drivername="GTiff", type="Byte", options=NULL)
erdas_spnad83_0 <- readGDAL(tf)
slot(erdas_spnad83_0, "proj4string")
if (new_proj_and_gdal()) comment(slot(erdas_spnad83_0, "proj4string"))
all.equal(erdas_spnad83, erdas_spnad83_0)
writeGDAL(erdas_spnad83, tf, drivername="GTiff", type="Byte", 
options="INTERLEAVE=PIXEL")
erdas_spnad83_1 <- readGDAL(tf)
slot(erdas_spnad83_1, "proj4string")
if (new_proj_and_gdal()) comment(slot(erdas_spnad83_1, "proj4string"))
all.equal(erdas_spnad83, erdas_spnad83_1)
writeGDAL(erdas_spnad83, tf, drivername="GTiff", type="Byte",
options=c("INTERLEAVE=PIXEL", "COMPRESS=DEFLATE"))
erdas_spnad83_2 <- readGDAL(tf)
slot(erdas_spnad83_2, "proj4string")
if (new_proj_and_gdal()) comment(slot(erdas_spnad83_2, "proj4string"))
all.equal(erdas_spnad83, erdas_spnad83_2)

x <- GDAL.open(system.file("pictures/erdas_spnad83.tif", package = "rgdal")[1])
erdas_spnad83 <- asSGDF_GROD(x, output.dim=c(100,100))
GDAL.close(x)
summary(erdas_spnad83)
image(erdas_spnad83, col=grey(1:99/100))

tf <- tempfile()
xx <- create2GDAL(erdas_spnad83, type="Byte")
xxx <- copyDataset(xx, driver="PNG")
saveDataset(xxx, tf)
GDAL.close(xx)
GDAL.close(xxx)
GDALinfo(tf)

tf2 <- tempfile()
writeGDAL(erdas_spnad83, tf2, drivername="PNG", type="Byte")
GDALinfo(tf2)

GT <- GridTopology(c(0.5, 0.5), c(1, 1), c(10, 10))
set.seed(1)
SGDF <- SpatialGridDataFrame(GT, data=data.frame(z=runif(100)))
opar <- par(mfrow=c(2,2), mar=c(1,1,4,1))
image(SGDF, "z", col=colorRampPalette(c("blue", "yellow"))(20))
title(main="input values")
pfunc <- colorRamp(c("blue","yellow"))
RGB <- pfunc(SGDF$z)
SGDF$red <- RGB[,1]
SGDF$green <- RGB[,2]
SGDF$blue <- RGB[,3]
image(SGDF, red="red", green="green", blue="blue")
title(main="input RGB")
tf <- tempfile()
writeGDAL(SGDF[c("red", "green", "blue")], tf, type="Byte", drivername="PNG")
t1 <- readGDAL(tf)
image(t1, red=1, green=2, blue=3)
title(main="output PNG RGB")
par(opar)

t0 <- meuse.grid["ffreq"]
fullgrid(t0) <- TRUE
t0$ffreq <- as.integer(t0$ffreq)-1
# convert factor to zero-base integer
CT <- c("red", "orange", "green", "transparent")
CT
cN <- c("annual", "2-5 years", "infrequent")
tf <- tempfile()
writeGDAL(t0, tf, type="Byte", colorTable=list(CT), catNames=list(cN),
 mvFlag=3L)
attr(GDALinfo(tf, returnStats=FALSE, returnCategoryNames=TRUE),
 "CATlist")[[1]]
\dontrun{
ds <- GDAL.open(tf)
displayDataset(ds, reset.par=FALSE)
t(col2rgb(getColorTable(ds)[1:4]))
GDAL.close(ds)
}
fn <- system.file("pictures/test_envi_class.envi", package = "rgdal")[1]
Gi <- GDALinfo(fn, returnColorTable=TRUE, returnCategoryNames=TRUE)
CT <- attr(Gi, "ColorTable")[[1]]
CT
attr(Gi, "CATlist")[[1]]
with <- readGDAL(fn)
with <- readGDAL(fn, silent=TRUE)
table(with$band1)
table(as.numeric(with$band1))
with1 <- readGDAL(fn, as.is=TRUE)
table(with1$band1)
spplot(with, col.regions=CT)
tf <- tempfile()
cN <- levels(with$band1)
with$band1 <- as.integer(with$band1)-1
writeGDAL(with, tf, drivername="ENVI", type="Int16", colorTable=list(CT),
 catNames=list(cN), mvFlag=11L)
cat(paste(readLines(paste(tf, "hdr", sep=".")), "\n", sep=""), "\n")
wGi <- GDALinfo(tf, returnColorTable=TRUE, returnCategoryNames=TRUE)
CTN <- attr(wGi, "ColorTable")[[1]]
CTN
attr(wGi, "CATlist")[[1]]
withN <- readGDAL(tf)
table(withN$band1)
withN1 <- readGDAL(tf, as.is=TRUE)
table(withN1$band1)
spplot(withN, col.regions=CTN)


# a file with scale and offset
fn <- system.file("pictures/scaleoffset.vrt", package = "rgdal")[1]
g <- GDALinfo(fn)
attr(g, 'ScaleOffset')
g

fl <- system.file("pictures/MR5905167_372.nc", package="rgdal")[1]
if (file.exists(fl)) {
  flstr <- paste0("NETCDF:\"", fl, "\":TEMP")
  if ("netCDF" \%in\% gdalDrivers()$name) GDALinfo(flstr)
}
}
\keyword{spatial}

\name{projInfo}
\alias{projInfo}
\alias{getPROJ4VersionInfo}
\alias{getPROJ4libPath}
\alias{projNAD}
\alias{GDALis3ormore}
\alias{PROJis6ormore}
\alias{new_proj_and_gdal}
\alias{GDAL_OSR_PROJ}
%\alias{checkCRSArgs}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{List PROJ.4 tag information}
\description{
  The \code{projInfo} function lists known values and descriptions for PROJ.4 tags for tag in \code{c("proj", "ellps", "datum", "units")}; \code{getPROJ4VersionInfo} returns the version of the underlying PROJ.4 release, \code{getPROJ4libPath} returns the value of the PROJ\_LIB environment variable, \code{projNAD} detects the presence of NAD datum conversion tables (looking for conus).
}
\usage{
projInfo(type = "proj")
getPROJ4VersionInfo()
getPROJ4libPath()
projNAD()
GDAL_OSR_PROJ()
GDALis3ormore()
PROJis6ormore()
new_proj_and_gdal()
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{type}{One of these tags: \code{c("proj", "ellps", "datum", "units")}}
}
\details{
 The output data frame lists the information given by the proj application with flags -lp, -le, -ld or -lu. From PROJ 6, \code{"datum"} is not available. From PROJ 7.1.0, \code{"units"} returns the conversion factor as numeric, not character.
}
\value{
  A data frame with a name and description column, and two extra columns for the "ellps" and "datum" tags.
}
\references{\url{https://proj.org/}}
\author{Roger Bivand \email{Roger.Bivand@nhh.no}}

\note{Loading the rgdal package may change the \code{PROJ\_LIB} environmental variable to the PROJ.4 support files if bundled with binary packages.}

\examples{
getPROJ4VersionInfo()
projInfo()
projInfo("ellps")
projInfo("units")
}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{spatial}

\name{list_coordOps}
\alias{list_coordOps}
\alias{best_instantiable_coordOp}
\alias{print.coordOps}

\title{List PROJ 6 coordinate operations}
\description{List PROJ 6 coordinate operations for a pair of source/target coordinate reference systems}
\usage{
list_coordOps(src_crs, tgt_crs, area_of_interest = as.numeric(NA), 
 strict_containment = FALSE, visualization_order = NULL)
best_instantiable_coordOp(x)
\method{print}{coordOps}(x, ...)
}

\arguments{
  \item{src_crs}{Source coordinate reference system string}
  \item{tgt_crs}{Target coordinate reference system string}
  \item{area_of_interest}{Numeric vector; either \code{NA}, or the xmin, ymin, xmax, ymax of the bounding box of the area of interest. This may be used to restrict the search for coordinate operations}
  \item{strict_containment}{default FALSE, permit partial matching of the area of interest; if TRUE strictly contain the area of interest. The area of interest is either as given, or as implied by the source/target coordinate reference systems (FIXME)}
  \item{visualization_order}{default NULL, taking the value of \code{get_enforce_xy()}; if TRUE always choose \code{x} or longitude for the first axis; if FALSE, follow the axis orders given by the coordinate reference systems when constructing the coordinate operation}
  \item{x}{an object of class \code{"coordOps"}}
  \item{...}{arguments possibly passed through, unused}
}
\details{(FIXME)}
\value{
A data frame with rows showing the coordinate operations found, and columns:
  \item{description}{String describing the operation}
  \item{definition}{PROJ pipeline for executing the operation}
  \item{accuracy}{Accuracy in meters, if negative, unknown}
  \item{instantiable}{Can this operation be carried out with available resources}
  \item{ballpark}{Does this operation only have ballpark accuracy}
  \item{number_grids}{The number of grids required for the operation}
The object has a \code{"grids"} attribute containing a nested list of grids for each coordinate operations found; if \code{number_grids == 0}, NULL, otherwise a list of grids. For each grid required, the short and long names of the grid are given, the package name if available in a PROJ grid package, and the download URL for that package. Three logical variables report whether the grid may be downloaded directly, whether it has an open license, and whether it is available. 
}
\references{
\url{https://proj.org/}
}
\author{Roger Bivand \email{Roger.Bivand@nhh.no}}

\note{
Fragile: work in progress
}
\examples{
run <- new_proj_and_gdal()
if (run) {
  discarded_datum <- showSRID("EPSG:27700", "PROJ")
  (x <- list_coordOps(paste0(discarded_datum, " +type=crs"), "OGC:CRS84"))
}
if (run) {
  best_instantiable_coordOp(x)
}
if (run) {
  restored_datum <- showSRID("EPSG:27700", "PROJ")
  list_coordOps(paste0(restored_datum, " +datum=OSGB36 +type=crs"), "OGC:CRS84")
}
if (run) {
  wkt_datum <- showSRID("EPSG:27700", "WKT2")
  (x <- list_coordOps(wkt_datum, "OGC:CRS84"))
}
if (run) {
  best_instantiable_coordOp(x)
}
if (run) {
  list_coordOps("EPSG:27700", "OGC:CRS84")
}
if (run) {
}
if (run) {
  discarded_datum <- showSRID("EPSG:22525", "PROJ")
  list_coordOps(paste0(discarded_datum, " +type=crs"), "EPSG:31985")
}
if (run) {
}
if (run) {
  wkt_datum <- showSRID("EPSG:22525", "WKT2")
  list_coordOps(wkt_datum, "EPSG:31985")
}
if (run) {
  (x <- list_coordOps("EPSG:22525", "EPSG:31985"))
}
if (run) {
  best_instantiable_coordOp(x)
}
}

\keyword{spatial}

\name{llgridlines}
\alias{llgridlines}
\title{Plot long-lat grid over projected data}
\description{ Plot long-lat grid over projected data }
\usage{
llgridlines(obj, easts, norths, ndiscr = 20, lty = 2, offset=0.5, side="WS",
llcrs = "+proj=longlat +datum=WGS84", plotLines = TRUE, plotLabels =
TRUE, ...)
}
\arguments{
  \item{obj}{object, deriving from \link[sp]{Spatial-class} having projection
  specified }
  \item{easts}{numeric; see \link[sp]{gridlines}}
  \item{norths}{numeric; see \link[sp]{gridlines}}
  \item{ndiscr}{numeric; see \link[sp]{gridlines}}
  \item{offset}{numeric; see \link[sp]{gridlines}}
  \item{side}{character, default \dQuote{WS}; see \link[sp]{gridlines}; available from \pkg{sp} 0.9-84}
  \item{lty}{line type to be used for grid lines}
\item{llcrs}{proj4string of longitude - latitude}
\item{plotLines}{logical; plot lines?}
\item{plotLabels}{logical; plot labels?}
\item{...}{ graphics arguments passed to plot function for lines
and text function for labels }
}

\value{
	none; side effect is that grid lines and lables are plotted
}
\seealso{\code{\link[sp]{is.projected}}, \code{\link[sp]{CRS-class}}}
\examples{
set_thin_PROJ6_warnings(TRUE)
data(meuse)
coordinates(meuse) = ~x+y
proj4string(meuse) <- CRS("+init=epsg:28992")
plot(meuse) 
llgridlines(meuse, lty=3)
plot(meuse) 
llgridlines(meuse, lty=3, side = "EN", offset = 0.2)
}
\keyword{spatial}

\name{GDALRasterBand-class}
\docType{class}
\alias{GDALRasterBand-class}
\alias{getRasterData}
\alias{getRasterTable}
\alias{getProjectionRef}
\alias{initialize,GDALRasterBand-method}
\alias{dim,GDALRasterBand-method}
\alias{getRasterBand}
\alias{getRasterBlockSize}
\alias{toSigned}
\alias{toUnSigned}
\alias{get_OVERRIDE_PROJ_DATUM_WITH_TOWGS84}
\alias{set_OVERRIDE_PROJ_DATUM_WITH_TOWGS84}
\title{Class "GDALRasterBand"}
\description{Returns a two-dimensional array with data from a raster band, used internally within functions}
\section{Objects from the Class}{
Objects can be created by calls of the form \code{new("GDALRasterBand", dataset, band)}.
 
}
\section{Slots}{
  \describe{
    \item{\code{handle}:}{Object of class \code{"externalptr", from class "GDALMajorObject"}, used internally; not for public consumption }
  }
}
\section{Extends}{
Class \code{"GDALMajorObject"}, directly.
}
\section{Methods}{
  \describe{
    \item{dim}{\code{signature(x = "GDALRasterBand")}: ... }
    \item{initialize}{\code{signature(.Object = "GDALRasterBand")}: ... }
  }
}
\usage{
getRasterData(dataset, band = NULL, offset = c(0, 0),
              region.dim = dim(dataset), output.dim = region.dim,
              interleave = c(0, 0), as.is = FALSE, list_out=FALSE) 

getRasterTable(dataset, band = NULL, offset = c(0, 0), 
  region.dim = dim(dataset))

getProjectionRef(dataset, OVERRIDE_PROJ_DATUM_WITH_TOWGS84 = NULL, enforce_xy = NULL)

getRasterBand(dataset, band = 1)

getRasterBlockSize(raster)

toSigned(x, base)

toUnSigned(x, base)

get_OVERRIDE_PROJ_DATUM_WITH_TOWGS84()
set_OVERRIDE_PROJ_DATUM_WITH_TOWGS84(value)

}
\arguments{
  \item{dataset}{An object inheriting from class 'GDALReadOnlyDataset'}
  \item{band}{The band number (1-based) to read from}
  \item{offset}{Number of rows and columns from the origin (usually the
    upper left corner) to begin reading from; presently ordered (y,x) - this may change}
  \item{region.dim}{The number of rows and columns to read from the
    dataset; presently ordered (y,x) - this may change}
  \item{output.dim}{Number of rows and columns in the output data; if
    smaller than \code{region.dim} the data will be subsampled}
  \item{interleave}{Element and row stride while reading data; rarely
    needed}
  \item{as.is}{If false, scale the data to its natural units; if the
    case of thematic data, return the data as factors}
  \item{list_out}{default FALSE, return array, if TRUE, return a list of vector bands}
  \item{raster}{An object of class GDALRasterBand}
  \item{x}{integer variable for conversion}
  \item{base}{If Byte input, 8, if Int16 or UInt16, 16}
  \item{OVERRIDE_PROJ_DATUM_WITH_TOWGS84}{logical value, default NULL, which case the cached option set by \code{set_OVERRIDE_PROJ_DATUM_WITH_TOWGS84} is used. Ignored if the GDAL version is less than \dQuote{1.8.0} or if the CPLConfigOption variable is already set}
  \item{enforce_xy}{(PROJ6+/GDAL3+) either use global setting (default NULL) or override policy for coordinate ordering easting/x as first axis, northing/y as second axis.}
  \item{value}{logical value to set OVERRIDE_PROJ_DATUM_WITH_TOWGS84}
}

\details{
  \describe{
    \item{\code{getRasterData}:}{retrieves data from the dataset as an array or list of bands; will try to convert relevant bands to factor if category names are available in the GDAL driver when returning a list.}
    
    \item{\code{getRasterTable}:}{retrieves data from the dataset as data frame.}
    
    \item{\code{getProjectionRef}:}{returns the geodetic projection in
      Well Known Text format.}

    \item{\code{getRasterBand}:}{returns a raster band}

    \item{\code{getRasterBlockSize}:}{returns the natural block size of
      the raster band. Use this for efficient tiled IO.}

    \item{\code{toSigned}:}{used to convert a band read as unsigned integer to signed integer}
    \item{\code{toUnSigned}:}{used to convert a band read as signed integer to unsigned integer}

  }
}

\note{The OVERRIDE_PROJ_DATUM_WITH_TOWGS84 argument is used to revert GDAL behaviour to pre-1.8.0 status; from 1.8.0, any input datum may be discarded if the input also includes a towgs84 tag in conversion to the PROJ.4 representation, see \url{https://trac.osgeo.org/gdal/ticket/4880} and \url{https://lists.osgeo.org/pipermail/gdal-dev/2012-November/034550.html}. The cached value of OVERRIDE_PROJ_DATUM_WITH_TOWGS84 will also be used in \code{open.SpatialGDAL}, \code{sub.GDROD}, and \code{asGDALROD_SGDF}, which do not have a suitable argument}


\author{Timothy H. Keitt, modified by Roger Bivand}

\seealso{See also \code{\link{GDALDriver-class}}, \code{\link{GDALDataset-class}}, \code{\link{GDALTransientDataset-class}}}

\examples{
logo <- system.file("pictures/logo.jpg", package="rgdal")[1]
x <- new("GDALReadOnlyDataset", logo)
plot(density(getRasterTable(x)$band1))
GDAL.close(x)
}
\keyword{classes}

\name{readOGR}
\alias{readOGR}
\alias{ogrInfo}
\alias{print.ogrinfo}
\alias{ogrFIDs}
\alias{ogrDrivers}
\alias{OGRSpatialRef}
\alias{ogrListLayers}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{Read OGR vector maps into Spatial objects}
\description{
  The function reads an OGR data source and layer into a suitable Spatial vector object. It can only handle layers with conformable geometry features (not mixtures of points, lines, or polygons in a single layer). It will set the spatial reference system if the layer has such metadata.

If reading a shapefile, the data source name (\code{dsn=} argument) is 
the folder (directory) where the shapefile is, and the layer is the 
name of the shapefile (without the .shp extension). For example to read 
\code{bounds.shp} from \code{C:/Maps}, do 
\code{map <- readOGR(dsn="C:/Maps", layer="bounds")}. The logic behind 
this is that typically one keeps all the shapefiles for a project in 
one folder (directory). 

As noted below, for other file type drivers, the \code{dsn=} argument is 
interpreted differently, and may be the file name itself, as for example 
with the \code{GPX} driver for reading GPS data as \code{layer="tracks"} 
lines or \code{layer="track_points"} points.

}
\usage{
readOGR(dsn, layer, verbose = TRUE, p4s=NULL, 
 stringsAsFactors=as.logical(NA), 
 drop_unsupported_fields=FALSE,
 pointDropZ=FALSE, dropNULLGeometries=TRUE,
 useC=TRUE, disambiguateFIDs=FALSE, addCommentsToPolygons=TRUE,
 encoding=NULL, use_iconv=FALSE, swapAxisOrder=FALSE, require_geomType = NULL,
 integer64="no.loss", GDAL1_integer64_policy=FALSE, morphFromESRI = NULL,
 dumpSRS = FALSE, enforce_xy = NULL, D3_if_2D3D_points=FALSE, missing_3D=0)
ogrInfo(dsn, layer, encoding=NULL,
 use_iconv=FALSE, swapAxisOrder=FALSE, require_geomType = NULL,
 morphFromESRI = NULL, dumpSRS = FALSE, enforce_xy = NULL,
 D3_if_2D3D_points=FALSE)
ogrFIDs(dsn, layer)
ogrDrivers()
OGRSpatialRef(dsn, layer, morphFromESRI=NULL, dumpSRS = FALSE, driver = NULL,
 enforce_xy = NULL)
ogrListLayers(dsn)
\method{print}{ogrinfo}(x, ...)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{dsn}{data source name (interpretation varies by driver --- for some drivers, dsn is a file name, but may also be a folder)}
  \item{layer}{layer name (varies by driver, may be a file name without extension). From rgdal 1.2.*, layer may be missing, in which case ogrListLayers examines the dsn, and fails if there are no layers, silently reads the only layer if only one layer is found, and reads the first layer if multiple layers are present, issuing a warning that layer should be given explicitly.}
  \item{verbose}{report progress}
  \item{p4s}{PROJ4 string defining CRS, if default NULL, the value is read from the OGR data set} 
  \item{stringsAsFactors}{logical: should character vectors be converted to factors? Default NA, which uses the deprecated \code{default.stringsAsFactors()} in R < 4.1.0 (see \code{link[base]{data.frame}}). Before R 4, strings were converted to factors by default, as argument value TRUE. See \url{https://developer.r-project.org/Blog/public/2020/02/16/stringsasfactors/index.html} for details of changes.}
  \item{drop\_unsupported\_fields}{default FALSE, if TRUE skip fields other than String, Integer, and Real; Date, Time and DateTime are converted to String}
  \item{pointDropZ}{default FALSE, if TRUE, discard third coordinates for point geometries; third coordinates are alway discarded for line and polygon geometries}
  \item{dropNULLGeometries}{default TRUE, drop both declared NULL geometries, and empty geometries with no coordinates; if FALSE, return a data frame with the attribute values of the NULL and empty geometries. From 1.3-6, setting FALSE also works when there are no geometries at all, returning a data.frame including all FIDs}
  \item{useC}{default TRUE, if FALSE use original interpreted code in a loop}
  \item{disambiguateFIDs}{default FALSE, if TRUE, and FID values are not unique, they will be set to unique values 1:N for N features; problem observed in GML files}
  \item{addCommentsToPolygons}{default TRUE, may be set FALSE for legacy behaviour; used to indicate which interior rings are holes in which exterior rings in conformance with OGC SFS specifications}
  \item{encoding}{default NULL, if set to a character string, and the driver is \dQuote{ESRI Shapefile}, and use_iconv is FALSE, it is passed to the CPL Option \dQuote{SHAPE_ENCODING} immediately before reading the DBF of a shapefile. If use_iconv is TRUE, and encoding is not NULL, it will be used to convert input strings from the given value to the native encoding for the system/platform.}
  \item{use_iconv}{default FALSE; if TRUE and encoding is not NULL, it will be used to convert input strings from the given value to the native encoding for the system/platform.}
  \item{swapAxisOrder}{default FALSE, if TRUE, treat y coordinate as Easting, x as Northing, that is the opposite to the assumed order; this may be needed if some OGR read drivers do not behave as expected}
  \item{require_geomType}{, default NULL, if one of: \code{c("wkbPoint", "wkbLineString", "wkbPolygon")}, then in input with multiple geometry types, the chosen type will be read}
  \item{integer64}{default \dQuote{no.loss} (from rgdal 1.2.*). From GDAL 2, fields to be read may also take Integer64 values. As R has no such storage mode, three options are offered, analogous with \code{\link{type.convert}} for numeric conversion: \dQuote{allow.loss} which clamps to 32-bit signed integer (default < rgdal 1.2), \dQuote{warn.loss} - as \dQuote{allow.loss} but warns when clamping occurs, and \dQuote{no.loss}, which reads as a character string using the formatting applied by default by GDAL (default >= rgdal 1.2). The use of 64-bit integers is usually a misunderstanding, as such data is almost always a long key ID.}
  \item{GDAL1_integer64_policy}{default FALSE, if TRUE, Integer64 fields are read as doubles}
  \item{morphFromESRI}{default NULL, morph from ESRI WKT1 dialect}
  \item{dumpSRS}{dump SRS to stdout from inside GDAL to debug conversion - developer use only}
  \item{enforce_xy}{(PROJ6+/GDAL3+) either use global setting (default NULL) or override policy for coordinate ordering easting/x as first axis, northing/y as second axis.}
  \item{D3_if_2D3D_points}{https://github.com/r-spatial/sf/issues/1683 case of mixed 2D/3D track points - set TRUE to 3D to pass}
  \item{missing_3D}{default 0, may be finite real numbers; https://github.com/r-spatial/sf/issues/1683}
  \item{driver}{default NULL, driver found using \code{ogrListLayers} from the data source; otherwise already known and passed through from a calling function}
  \item{x}{ogrinfo object}
  \item{...}{other arguments to print method}
}
\details{
  The drivers available will depend on the installation of GDAL/OGR, and can vary; the \code{ogrDrivers()} function shows which are available, and which may be written (but all are assumed to be readable). Note that stray files in data source directories (such as *.dbf) may lead to suprious errors that accompanying *.shp are missing.
}
\value{
  A Spatial object is returned suiting the vector data source, either a SpatialPointsDataFrame (using an AttributeList for its data slot directly), a SpatialLinesDataFrame, or a SpatialPolygonsDataFrame.
}
\references{\url{https://gdal.org/drivers/vector/index.html}, \url{https://resources.oreilly.com/examples/9780596008659}}
\author{Roger Bivand}

\note{The bases for this implementation are taken from functions in Barry Rowlingson's draft Rmap package, and from Radim Blazek's v.in.ogr program in GRASS.

Please note that the OGR drivers used may not handle missing data gracefully, and be prepared to have to correct for this manually. From rgdal 0.5-27, missing value handling has been improved, and OGR unset field values are set to NA in R, but drivers and external files may vary in their representations of missing values.

In addition, from 0.6-9 date and time fields are read as strings rather than being treated as unsupported; NULL geometries are identified and dropped. There are differences in the reporting of NULL geometries between \code{ogrInfo} and \code{readOGR} - in \code{ogrInfo}, only declared NULL geometries are reported, but in \code{readOGR}, any line or polygon geometries with no coordinates are assigned NULL geometry status as well. An attempt is made to close unclosed rings in polygon geometries. 

For reading GPX files, refer to the OGR GPX format documentation for the use of layer tags: \dQuote{waypoints}, \dQuote{tracks}, \dQuote{routes}, \dQuote{track_points} and \dQuote{route_points} - reading GPX files requires a build of GDAL/OGR with the expat XML library. 

From 0.6-10, attempts are made to detect deleted features still present in the layer, but not read. Apparently features deleted in Qgis are only marked as deleted, but are still in the layer. These are not NULL geometries, but still need to be handled. An attempt is made to check the FID values, and \code{ogrFIDs} now returns attributes permitting this oddity to be detected. Such deleted features were seen as NULL in 0.6-9, but are not such.

From 0.7-24, if the layer has no fields, a single field containing the FID values is placed in the data slot of the returned object.

From 0.7-24, attempts are begun to provide users with arguments to control reading from OGR/shapefile driver when the encoding is inappropriate (especially the setting of LDID in shapefile DBFs, and the SHAPE_ENCODING environment variable).

While there is no certainty, newer drivers such as KML, GML, SQLite and Geopackage (GPKG) may encode string fields as UTF-8. Users are advised to explore this on a case to case basis using \code{\link[base]{Encoding}} on string fields of input objects.

Because of the representation of DateTime data in OGR, decimal seconds in input data are rounded to integer seconds, see: \url{https://trac.osgeo.org/gdal/ticket/2680}.

Because some drivers support reading string, integer and real list fields, support has been introduced into \code{ogrInfo} from version 0.9-1 to report their presence and the maximum counts of list items. This may lead to the introduction of the \code{-splitlistfields} facility from the command line utility \code{ogrinfo}. In addition, \code{ogrInfo} reports that there are no features when counting FIDs in a while loop over features in \code{ogrFIDs} never enters the loop, despite the layer feature count reporting at least one feature.
}
\seealso{\code{\link[sp]{SpatialPointsDataFrame-class}}, \code{\link[sp]{SpatialLinesDataFrame-class}}, \code{\link[sp]{SpatialPolygonsDataFrame-class}}, \code{\link[maptools]{readShapePoly}}, \code{\link[base]{iconv}}}
\examples{
set_thin_PROJ6_warnings(TRUE)
ogrDrivers()
dsn <- system.file("vectors", package = "rgdal")[1]
ogrListLayers(dsn)
ogrInfo(dsn)
ogrInfo(dsn=dsn, layer="cities")
owd <- getwd()
setwd(dsn)
ogrInfo(dsn="cities.shp")
ogrInfo(dsn="cities.shp", layer="cities")
setwd(owd)
ow <- options("warn")$warn
options("warn"=1)
cities <- readOGR(dsn=dsn, layer="cities")
str(slot(cities, "data"))
if (new_proj_and_gdal()) comment(slot(cities, "proj4string"))
cities$POPULATION <- type.convert(as.character(cities$POPULATION),
  na.strings="-99", numerals="no.loss")
str(slot(cities, "data"))
cities <- readOGR(dsn=dsn, layer="cities", GDAL1_integer64_policy=TRUE)
str(slot(cities, "data"))
options("warn"=ow)
summary(cities)
table(Encoding(as.character(cities$NAME)))
ogrInfo(dsn=dsn, layer="kiritimati_primary_roads")
OGRSpatialRef(dsn=dsn, layer="kiritimati_primary_roads")
kiritimati_primary_roads <- readOGR(dsn=dsn, layer="kiritimati_primary_roads")
summary(kiritimati_primary_roads)
if (new_proj_and_gdal()) comment(slot(kiritimati_primary_roads, "proj4string"))
ogrInfo(dsn=dsn, layer="scot_BNG")
OGRSpatialRef(dsn=dsn, layer="scot_BNG")
scot_BNG <- readOGR(dsn=dsn, layer="scot_BNG")
summary(scot_BNG)
if (new_proj_and_gdal()) comment(slot(scot_BNG, "proj4string"))
if ("GML" \%in\% ogrDrivers()$name) {
  dsn <- system.file("vectors/airports.gml", package = "rgdal")[1]
  airports <- try(readOGR(dsn=dsn, layer="airports"))
  if (!inherits(airports, "try-error")) {
    summary(airports)
    if (new_proj_and_gdal()) comment(slot(airports, "proj4string"))
  }
}
dsn <- system.file("vectors/ps_cant_31.MIF", package = "rgdal")[1]
ogrInfo(dsn=dsn, layer="ps_cant_31")
ps_cant_31 <- readOGR(dsn=dsn, layer="ps_cant_31")
summary(ps_cant_31)
sapply(as(ps_cant_31, "data.frame"), class)
if (new_proj_and_gdal()) comment(slot(ps_cant_31, "proj4string"))
ps_cant_31 <- readOGR(dsn=dsn, layer="ps_cant_31", stringsAsFactors=FALSE)
summary(ps_cant_31)
sapply(as(ps_cant_31, "data.frame"), class)
dsn <- system.file("vectors/Up.tab", package = "rgdal")[1]
ogrInfo(dsn=dsn, layer="Up")
Up <- readOGR(dsn=dsn, layer="Up")
summary(Up)
if (new_proj_and_gdal()) comment(slot(Up, "proj4string"))
dsn <- system.file("vectors/test_trk2.gpx", package = "rgdal")[1]
test_trk2 <- try(readOGR(dsn=dsn, layer="tracks"))
if (!inherits(test_trk2, "try-error")) {
  summary(test_trk2)
  if (new_proj_and_gdal()) comment(slot(test_trk2, "proj4string"))
}
test_trk2pts <- try(readOGR(dsn=dsn, layer="track_points"))
if (!inherits(test_trk2pts, "try-error")) {
  summary(test_trk2pts)
  if (new_proj_and_gdal()) comment(slot(test_trk2pts, "proj4string"))
}
dsn <- system.file("vectors", package = "rgdal")[1]
ogrInfo(dsn=dsn, layer="trin_inca_pl03")
birds <- readOGR(dsn=dsn, layer="trin_inca_pl03")
summary(birds)
if (new_proj_and_gdal()) comment(slot(birds, "proj4string"))
dsn <- system.file("vectors/PacoursIKA2.TAB", package = "rgdal")[1]
try(ogrInfo(dsn, "PacoursIKA2"))
ogrInfo(dsn, "PacoursIKA2", require_geomType="wkbPoint")
plot(readOGR(dsn, "PacoursIKA2", require_geomType="wkbLineString"), col="red")
plot(readOGR(dsn, "PacoursIKA2", require_geomType="wkbPoint"), add=TRUE)
odir <- getwd()
setwd(system.file("vectors", package = "rgdal")[1])
ow <- options("warn")$warn
options("warn"=1)
ogrInfo("test64.vrt", "test64")
str(readOGR("test64.vrt", "test64", verbose=FALSE, integer64="allow.loss")$val)
str(readOGR("test64.vrt", "test64", verbose=FALSE, integer64="warn.loss")$val)
str(readOGR("test64.vrt", "test64", verbose=FALSE, integer64="no.loss")$val)
str(readOGR("test64.vrt", "test64", verbose=FALSE, stringsAsFactors=FALSE,
 integer64="no.loss")$val)
setwd(odir)
options("warn"=ow)
}
\keyword{spatial}

\name{writeOGR}
\alias{writeOGR}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{Write spatial vector data using OGR}
\description{
  The function is an interface with the OGR abstraction library for spatial vector data, allowing data to be written out using supported drivers. The drivers supported will depend on the local installation, and the capabilities of those drivers (many are read-only). The objects exported are SpatialPointsDataFrame, SpatialLinesDataFrame, or SpatialPolygonsDataFrame objects as defined in the sp package.
}
\usage{
writeOGR(obj, dsn, layer, driver, dataset_options = NULL,
 layer_options=NULL, verbose = FALSE, check_exists=NULL,
 overwrite_layer=FALSE, delete_dsn=FALSE, morphToESRI=NULL,
 encoding=NULL, shp_edge_case_fix=FALSE, dumpSRS = FALSE)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{obj}{a SpatialPointsDataFrame, SpatialLinesDataFrame, or a SpatialPolygonsDataFrame object.}
  \item{dsn}{data source name (interpretation varies by driver --- for some drivers, dsn is a file name, but may also be a folder)}
  \item{layer}{layer name (varies by driver, may be a file name without extension)}
  \item{driver}{a character string equal to one of the driver names returned by \code{\link{ogrDrivers}}}
  \item{dataset_options}{a character vector of options, which vary by driver, and should be treated as experimental}
  \item{layer_options}{a character vector of options, which vary by driver, and should be treated as experimental}
  \item{verbose}{if TRUE, returns a list of information about the attempted write operation}
  \item{check_exists}{default NULL, which tests for the GDAL version, and sets FALSE if < 1.8.0, or TRUE for >= 1.8.0}
  \item{overwrite_layer}{default FALSE, if TRUE and \code{check_exists=TRUE}, delete the existing layer of the same name from the data source before writing the new layer; this will delete data and must be used with extreme caution, its behaviour varies between drivers, and accommodates changes that may appear in GDAL 1.8}
  \item{delete_dsn}{default FALSE, may be set to TRUE if \code{overwrite_layer} reports that the data source cannot be updated; this will delete data and must be used with extreme caution, its behaviour varies between drivers, and accommodates changes that may appear in GDAL 1.8}
  \item{morphToESRI}{default NULL, in which case set TRUE if driver is \dQuote{ESRI Shapefile} or FALSE otherwise; may be used to override this default}
  \item{encoding}{default NULL, if set to a character string, it will be used to convert output strings from the given value to UTF-8 encoding.}
  \item{shp_edge_case_fix}{default FALSE, if TRUE, attempt to work around MULTIPOLYGON to POLYGON degradation in ESRI Shapefile output with two touching exterior rings in a single feature (not yet implemented).}
  \item{dumpSRS}{dump SRS to stdout from inside GDAL to debug conversion - developer use only}
}
\details{
  Working out which combination of dsn, layer, and driver (and option) values give the desired output takes time and care, and is constrained by the ability of drivers to write output; many are read-only. Use of the references given is highly advisable, with searches in the archives of other software using GDAL/OGR. Note that for the \dQuote{ESRI Shapefile} driver and GDAL >= 1.9, the \code{layer_options} value of \sQuote{ENCODING=\dQuote{LDID/CP1252}} or other values found on \url{http://www.autopark.ru/ASBProgrammerGuide/dbfstruc.htm} to set the encoding byte of the output DBF file (link refered to in ogr/ogrsf_frmts/shape/ogrshapelayer.cpp. The effect of setting the LDID may vary depending on whether GDAL is built with iconv or not, and on the setting of the CPL Option \dQuote{SHAPE_ENCODING}.

While there is no certainty, newer drivers such as KML, GML, SQLite and Geopackage (GPKG) may encode string fields as UTF-8. Users are advised to explore this on a case to case basis using \code{\link[base]{Encoding}} on string fields of objects to be output, converting where necessary with \code{\link[base]{iconv}} or assigning the appropriate value with \code{\link[base]{Encoding}}.

}

\value{
 if verbose=TRUE, a list of information about the attempted write operation
}

\note{Only a subset of possible data slot column classes may be written out; if the function returns an error that the data type of stated columns is unknown, examine the classes and check that they are one of \code{c("numeric", "character", "factor", "POSIXt", "integer", "logical")}, and if not convert to such classes. Classes \code{c("factor", "POSIXt")} are converted to character strings, and \code{c("logical")} to integer internally.

For writing with the KML and GPX drivers, note that the geometries should be in geographical coordinates with datum WGS84.
}
\section{Warning}{
The \code{overwrite_layer} and \code{delete_dsn} arguments are provided only for experienced script writers who need to be able to destroy data, for example during repetetive simulation runs. They should never be used by anyone who is not confident about deleting files.
}
\section{\code{writeOGR} Polygon bug in 1.1-1}{
In fixing a bug in the correct handling of SFS polygon geometries in version 1.1-1, a further bug was introduced affecting cases of wkbPolygon (not wkbMultiPolygon) output where SFS hole status in the output object was (correctly) defined in the comment to Polygons objects. The error only occurred when all the Polygons objects had one exterior ring, and zero or more interior rings. The error led to the coordinates of the rings cumulating, because the rings were not emptied before assigning the next ring. Version 1.1-2 corrects the error; thanks to  JamesWorrall for a complete bug report \url{https://stat.ethz.ch/pipermail/r-sig-geo/2015-December/023796.html}.
}
\references{\url{https://gdal.org/drivers/vector/index.html}, \url{https://resources.oreilly.com/examples/9780596008659}}
\author{Roger Bivand}

\seealso{\code{\link{readOGR}}}
\examples{
set_thin_PROJ6_warnings(TRUE)
cities <- readOGR(system.file("vectors", package = "rgdal")[1], "cities")
is.na(cities$POPULATION) <- cities$POPULATION == -99
summary(cities$POPULATION)
td <- file.path(tempdir(), "rgdal_examples"); dir.create(td)
# BDR 2016-12-15 (MapInfo driver fails writing to directory with ".")
if(nchar(Sys.getenv("OSGEO4W_ROOT")) > 0) {
    OLDPWD <- getwd()
    setwd(td)
    td <- "."
}
writeOGR(cities, td, "cities", driver="ESRI Shapefile")
try(writeOGR(cities, td, "cities", driver="ESRI Shapefile"))
writeOGR(cities, td, "cities", driver="ESRI Shapefile", overwrite_layer=TRUE)
cities2 <- readOGR(td, "cities")
summary(cities2$POPULATION)
if ("SQLite" \%in\% ogrDrivers()$name) {
  tf <- tempfile()
  try(writeOGR(cities, tf, "cities", driver="SQLite", layer_options="LAUNDER=NO"))
}
if ("GeoJSON" \%in\% ogrDrivers()$name) {
  js <- '{
    "type": "MultiPolygon",
    "coordinates": [[[[102.0, 2.0], [103.0, 2.0], [103.0, 3.0], [102.0, 3.0],
    [102.0, 2.0]]], [[[100.0, 0.0], [101.0, 0.0], [101.0, 1.0], [100.0, 1.0],
    [100.0, 0.0]]]]
  }'
  spdf <- readOGR(js, layer='OGRGeoJSON')
  in1_comms <- sapply(slot(spdf, "polygons"), comment)
  print(in1_comms)
  tf <- tempfile()
  writeOGR(spdf, tf, "GeoJSON", driver="GeoJSON")
  #spdf1 <- readOGR(tf, "GeoJSON")
  spdf1 <- readOGR(tf)
  in2_comms <- sapply(slot(spdf1, "polygons"), comment)
  print(in2_comms)
  print(isTRUE(all.equal(in1_comms, in2_comms)))
}
\dontrun{if ("GML" \%in\% ogrDrivers()$name) {
  airports <- try(readOGR(system.file("vectors/airports.gml",
    package = "rgdal")[1], "airports"))
  if (class(airports) != "try-error") {
    writeOGR(cities, paste(td, "cities.gml", sep="/"), "cities", driver="GML")
    cities3 <- readOGR(paste(td, "cities.gml", sep="/"), "cities")
  }
}}
# The GML driver does not support coordinate reference systems
if ("KML" \%in\% ogrDrivers()$name) {
  data(meuse)
  coordinates(meuse) <- c("x", "y")
  proj4string(meuse) <- CRS("+init=epsg:28992")
  meuse_ll <- spTransform(meuse, CRS("+proj=longlat +datum=WGS84"))
  writeOGR(meuse_ll["zinc"], paste(td, "meuse.kml", sep="/"), "zinc", "KML")
}
list.files(td)
roads <- readOGR(system.file("vectors", package = "rgdal")[1],
 "kiritimati_primary_roads")
summary(roads)
if (strsplit(getGDALVersionInfo(), " ")[[1]][2] < "2") {
# For GDAL >= 2, the TAB driver may need a BOUNDS layer option
  writeOGR(roads, td, "roads", driver="MapInfo File")
  roads2 <- readOGR(paste(td, "roads.tab", sep="/"), "roads")
  summary(roads2)
}
scot_BNG <- readOGR(system.file("vectors", package = "rgdal")[1], "scot_BNG")
summary(scot_BNG)
if (strsplit(getGDALVersionInfo(), " ")[[1]][2] < "2") {
# For GDAL >= 2, the TAB driver may need a BOUNDS layer option
  writeOGR(scot_BNG, td, "scot_BNG", driver="MapInfo File")
  list.files(td)
  scot_BNG2 <- readOGR(paste(td, "scot_BNG.tab", sep="/"), "scot_BNG",
    addCommentsToPolygons=FALSE)
  summary(scot_BNG2)
}
writeOGR(scot_BNG, td, "scot_BNG", driver="MapInfo File",
 dataset_options="FORMAT=MIF")
list.files(td)
scot_BNG3 <- readOGR(paste(td, "scot_BNG.mif", sep="/"), "scot_BNG")
summary(scot_BNG3)
if(nchar(Sys.getenv("OSGEO4W_ROOT")) > 0) {
    setwd(OLDPWD)
}
}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{spatial}

\name{GDALDataset-class}
\docType{class}
\alias{GDALDataset-class}
\alias{putRasterData}
\alias{saveDataset}
\alias{copyDataset}
\alias{deleteDataset}
\alias{saveDatasetAs}
\alias{initialize,GDALDataset-method}
\title{Class "GDALDataset"}
\description{\code{GDALDataset} extends \code{\link{GDALReadOnlyDataset-class}} with data update commands.
}
\section{Objects from the Class}{
Objects can be created by calls of the form \code{new("GDALDataset", filename, handle)}, where name: a string giving the name of a GDAL driver, handle: used internally; not for public consumption (default = NULL).
}
\section{Slots}{
  \describe{
    \item{\code{handle}:}{Object of class \code{"externalptr", from class "GDALReadOnlyDataset"}, used internally; not for public consumption}
  }
}
\section{Extends}{
Class \code{"GDALReadOnlyDataset"}, directly.
Class \code{"GDALMajorObject"}, by class "GDALReadOnlyDataset".
}
\section{Methods}{
  \describe{
    \item{initialize}{\code{signature(.Object = "GDALDataset")}: ... }
  }
}
\usage{
putRasterData(dataset, rasterData, band = 1, offset = c(0, 0))
saveDataset(dataset, filename, options=NULL, returnNewObj=FALSE)
copyDataset(dataset, driver, strict = FALSE, options = NULL, fname=NULL)
deleteDataset(dataset)
saveDatasetAs(dataset, filename, driver = NULL, options=NULL)
}
\arguments{
  \item{dataset}{An object inheriting from class 'GDALDataset'}
  \item{rasterData}{A data array with \code{length(dim(rasterData)) = 2}}
  \item{band}{The band number (1-based) to read from}
  \item{offset}{Number of rows and columns from the origin (usually the
    upper left corner) to begin reading from}
  \item{filename}{name of file to contain raster data object; will be normalized with \code{\link{normalizePath}}}
  \item{returnNewObj}{until and including 0.5-27, \code{saveDataset} returned an invisible copy of the new file handle, which was then only finalized when the garbage collector ran. The old behaviour can be retained by setting to FALSE, the default behaviour is to close the hande and not return it.}
  \item{driver}{GDAL driver name to use for saving raster data object}
  \item{strict}{TRUE if the copy must be strictly equivalent, or more normally FALSE indicating that the copy may adapt as needed for the output format}
  \item{options}{Driver specific options (currently passed to GDAL)}
  \item{fname}{default NULL, used internally to pass through a file name with a required extension (RST driver has this problem)}
}
\details{
  \describe{
    
    \item{\code{putRasterData}:}{writes data contained in
      \code{rasterData} to the dataset, begining at \code{offset} rows
      and columns from the origin (usually the upper left corner). Data
      type conversion is automatic.}
    \item{\code{saveDataset}:}{saves a raster data object in a file using the driver of the object}
    \item{\code{saveDatasetAs}:}{saves a raster data object in a file using the specified driver}
    \item{\code{copyDataset}:}{make a copy of raster data object in a file using the specified driver}
    \item{\code{deleteDataset}:}{delete the file from which the raster data object was read (should only delete files opened as GDALDataset objects}
  }
}

\author{Timothy H. Keitt, modified by Roger Bivand}

\seealso{\code{\link{GDALDriver-class}},
  \code{\link{GDALReadOnlyDataset-class}}, \code{\link{GDALTransientDataset-class}}}

\keyword{classes}
